open Odoc_info
open Parameter
open Value
open Type
open Exception
open Class
open Module
open Odoc_html
open Doc_outils



class gen () =
  object (self)
    inherit Odoc_latex.latex as latex
    (*inherit Odoc_to_text.to_text as to_text
    *)	method latex_of_example t = [Odoc_info.Bold [Odoc_info.Raw "Example :"]]@t		
	method generate module_list = latex#generate module_list    

    (** Return the header of the TeX document. *)
    method latex_header = 
	"\\documentclass[11pt,a4paper,oneside]{article} \n "^
	"\\usepackage[latin1]{inputenc} \n "^
	"\\usepackage[T1]{fontenc} \n "^
	"\\usepackage[francais]{babel} \n "^
	"\\usepackage{indentfirst} \n "^
	"\\usepackage{pslatex} \n "^
	"\\usepackage{ae} \n "^
	"\\usepackage{amsmath} \n "^
	"\\usepackage{indentfirst} \n "^
	"\\usepackage{pslatex} \n "^
	"\\usepackage{ae} \n "^
	"\\usepackage{amsfonts} \n "^
	"\\usepackage{makeidx} \n "^
	(*"\\usepackage{vmargin} \n "^
        "\\usepackage{fullpage} \n"^
       *) "\\usepackage{url} \n"^
        "\\usepackage{ocamldoc}\n"^
	"\\newif\\ifpdf \n "^
	"\\ifx\\pdfoutput\\undefined \n "^
	"\\pdffalse % we are not running PDFLaTeX \n "^
	"\\else \n "^
	"\\pdfoutput=1 % we are running PDFLaTeX \n "^
	"\\pdftrue \n "^
	"\\fi \n "^
	"\\ifpdf \n "^
	"\\usepackage[colorlinks,bookmarks]{hyperref} \n "^
	"\\hypersetup{ \n "^
	"  pdfauthor={CALVEZ Benoit}, \n "^
	"  pdftitle={Documentation MGS} \n "^
	"   } \n "^
	"\\usepackage{thumbpdf} \n "^
	"\\let\\docext=\\pdfext \n "^
	"\\else \n "^
	"\\let\\docext=\\dviext \n "^
	"\\usepackage[bookmarks]{hyperref} \n "^
	"\\fi \n "^
	"\\newcommand{\\dlink}[3]{% \n "^
	"\\ifpdfoutput \n "^
	"\\ifx\\pdfext#3 \n "^
	"\\href{#1/#2.#3}{\\texttt{texdoc #2}}% \n "^
	"      \\else \n "^
	"\\texttt{texdoc #2}% \n "^
	"\\fi \n "^
	"\\else \n "^
	"\\href{#1/#2.#3}{\\mbox{\\texttt{texdoc #2}}}% \n "^
	"\\fi} \n "^
	" \n "^
	"\\ifpdf \n "^
	"\\usepackage[pdftex]{graphicx} \n "^
	"\\else \n "^
	"\\usepackage{graphicx} \n "^
	"\\fi \n "^
	"\\ifpdf \n "^
	"\\DeclareGraphicsExtensions{.pdf, .jpg, .tif} \n "^
	"\\else \n "^
	"\\DeclareGraphicsExtensions{.eps, .jpg} \n "^
	"\\fi \n "^
        "\\makeindex \n "^
	" \n "^
	(
       match !Args.title with
         None -> ""
       | Some s -> "\\title{"^(self#escape s)^"}\n"
      )^
      "\\begin{document}\n"^
      (*(match !Args.title with None -> "" | Some _ -> "\\maketitle\n")^
      (if !Args.with_toc then "\\tableofcontents\n" else "")
*)
	"\\begin{titlepage} \n "^
	"\\begin{flushright} \n "^
	"%\\includegraphics[width=3cm]{LogoIIE.jpg} \n "^
	"\\end{flushright} \n "^
	" \n "^
	"\\vspace{1cm} \n "^
	"\\begin{center} \n "^
	" \n "^
	"\\fbox{ \n "^
	"\\begin{tabular}{c} \n "^
	"    {\\Huge {\\bf {Documentation}}} \\\\ \n "^
	"    \\\\ \n "^
	"    {\\huge M.G.S.}\\\\ \n "^
	"\\end{tabular} \n "^
	"} \n "^
	" \n "^
	"\\end{center} \n "^
	" \n "^
	"\\vspace{10cm} \n "^
	"\\begin{center} \n "^
	" \n "^
	"\\large{SPECIF} \\\\ \n "^
	" \n "^
	"\\vspace{1cm} \n "^
	"LAMI\\\\ \n "^
	"2003 \n "^
	"\\end{center} \n "^
	"\\end{titlepage} \n "^
	"\\newpage  \n "^
	"\\tableofcontents  \n "^
	"\\newpage \n "



	method latex_of_voir t = 
			
			match t with 
			(Odoc_info.Raw s)::q ->let lexbuf = Lexing.from_string s in
                        		let result = (*Doc_lien_yacc.depart  *)Doc_lien_lex.lexer lexbuf in
                        		let chaine =match result with
					Module (n,c) -> "\\textbf{See also } : Module "^(self#subst subst_strings n)^" section \\ref{"^n^"} page \\pageref{"^n^"} "^(self#subst subst_strings c)^"\n"
                        		|ModFonct (n,f,c)-> let ff = Doc_outils.code f in
                                        	"\\textbf{See also } : Function  "^(self#subst subst_strings f)^" dans le Module "^(self#subst subst_strings n)^" section \\ref{"^n^"."^f^"} page \\pageref{"^n^"."^f^"} "^(self#subst subst_strings c)^"\n"
                        		|Fonction (f,c) -> let ff = Doc_outils.code f in
                                        	"\\textbf{See also } : Function  "^(self#subst subst_strings f)^" "^(self#subst subst_strings c)^"\n"
					|URL (u,c) -> 
						"\\textbf{See also } : \\url{"^u^"} "^(self#subst subst_strings c)^"\n"
                        		in
					(Odoc_info.Latex chaine)::q
			|_ -> t;





	
    (** @return [text] value for a value. *)
    method text_of_value v =
      let s_name = Name.simple (transforme_nom v.val_name) in
      Format.fprintf Format.str_formatter "@[<hov 2> %s :@ "
        s_name;
      let s = suppression_type
        (self#normal_type (Name.father (transforme_nom v.val_name)) v.val_type)
      in
      [ CodePre s ] @
      [Latex ("\\index{"^(self#label s_name)^"@\\verb`"^(self#label ~no_:false s_name)^"`}\n")] @
      (self#text_of_info v.val_info)


    (** Renur laTeX c de for a value. *)
    method latex_of_value v =
	let nom =transforme_nom v.val_name in
      Odoc_info.reset_type_names () ;
      self#latex_of_text
        ((Latex (self#make_label (self#value_label nom))) ::
         (self#text_of_value v))



    method latex_of_Code s =
      let s2 = self#escape_code s in
      let  subs = [
      (">", "$>$");
      ("<", "$<$");
      ("\n","\\\\\n");
     ] in
      (*let s3 = (Str.global_replace (Str.regexp "\n") ("\\\\\n") s2) in*)
    let  s3 = self#subst subs s2 in
      "{\\tt{"^s3^"}}"


    (** Generate the LaTeX file from a module list, in the {!Args.out_file} file. *)
    method generate module_list =
      if !Args.separate_files then
        (
         let f m =
           try
             let chanout =
               open_out ((Filename.concat !Args.target_dir (Name.simple m.m_name))^".tex")
             in
             self#generate_for_module chanout m ;
             close_out chanout
           with
             Failure s
           | Sys_error s ->
               prerr_endline s ;
               incr Odoc_info.errors
         in
         List.iter f module_list
        );

      try
        let chanout = open_out (Filename.concat !Args.target_dir !Args.out_file) in
        let _ = if !Args.with_header then output_string chanout self#latex_header else () in
        List.iter
          (fun m -> if !Args.separate_files then
            output_string chanout ("\\input{"^((Name.simple m.m_name))^".tex}\n")
          else
            self#generate_for_module chanout m
          )
          module_list ;
        let _ = if !Args.with_trailer then output_string chanout "\\printindex\n\\end{document}" else () in
        close_out chanout
      with
        Failure s
      | Sys_error s ->
          prerr_endline s ;
          incr Odoc_info.errors


    initializer
      tag_functions <- ("example", self#latex_of_example) :: tag_functions; 
      tag_functions <- ("voir", self#latex_of_voir) :: tag_functions; 
  end

let my_generator = new gen () 
let _ = Odoc_args.set_doc_generator (Some (my_generator :> Odoc_args.doc_generator))

