(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: def_achain.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG, AS
 * Modified: $Date: 2004-09-27 16:15:56 $
 * Version: $Revision: 1.2 $
 * -------------------------------------------------------------------
 *
 * Status:
 *
 *     Une chane correspond  la valuation des cellules abstraites d'un complexe abstrait.
 *     Aussi, on trouvera dans ces fichiers (def_, core_, etc.) la dfinition et
 *     la gestion des complexes d'une part et des chanes d'autre part.
 *
 *     Un complexe abstrait n'existe qu' travers ses cellules. Une cellule est identifie par un entier
 *     unique (issu d'un compteur qu on incrmente lors de chaque cration. Un dictionnaire permet
 *     galement de faire le lien entre une clef et la cellule associe. Une cellule est un
 *     enregistrement  5 champs :
 *         _ uid_acell (int)            : l'identifiant
 *         _ dim_acell (int)            : le dimension du simplexe
 *         _ faces_acell (acell list)   : la liste des faces de la cellule (les faces sont de dimension
 *                                        (dim_acell-1))
 *         _ cofaces_acell (acell list) : la liste des cofaces de la cellule (les cofaces sont de dimension
 *                                        (dim_acell+1))
 *         _ modifiable_acell (bool)    : un flag prcisant si on peut appliquer une fonction changeant la
 *                                        topologie sur la cellule
 *
 *
 *     Une chane abstraite est reprsente par une table d'association Hash entre les cellules et
 *     une valeurs. Pour des raison d'efficacit, la clef sera uniquement l'identifiant des cellules
 *     et la valeur est le couple (acell,valeur associe).
 *
 *=========================================================================
 *)


Global.__register_version "$RCSfile: def_achain.ml,v $" "$Revision: 1.2 $" "$Date: 2004-09-27 16:15:56 $";;

type acell = { 
  uid_acell                : int;               (* unique identifier de la acell                 *)
  dim_acell                : int;               (* dimension                                     *)
  mutable faces_acell      : acell list;        (* acell incidentes de dimension dim-1           *)
  mutable cofaces_acell    : acell list;        (* acell incidentes de dimension dim+1           *)
  mutable modifiable_acell : bool               (* peut-on modifier faces_acell ou cofaces_acell *)
};;

type 'a achain = (int, (acell*'a)) Hash.t;;
