(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: kernel_achain.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG, AS
 * Modified: $Date: 2007-10-23 15:10:42 $
 * Version: $Revision: 1.15 $
 * -------------------------------------------------------------------
 *
 * Status: Cf. AChain/def_achain.ml pour les infos sur les types achain et acell
 *          
 *
 *=========================================================================
 *)

open Global;;
open Valeur;;
open Environnement;;
open Dispatch;;

__register_version "$RCSfile: kernel_achain.ml,v $" "$Revision: 1.15 $" "$Date: 2007-10-23 15:10:42 $";;

print_string "(* loading kernel functions on ACHAIN into user-space ... ";
flush Pervasives.stdout;;


(*!*)


(***
  @module achain
  The abstract chain topological collection.

*)

(**
  @module achain
  @name 'new_acell
  @type int_seq_seq_acell

  Creates a new acell of a given dimension.
  [new_acell(d, fl, cfl)] creates a new acell [sp] of dimension [d]
  with faces [fl] and cofaces [cfl]. Faces and cofaces are given as
  sequences. The cofaces and faces of the acells in theses
  sequences are updated to reflect the creation of the new
  acell. However, existing achains that include a coface of [sp] are
  not updated.
*)
addenv_impure "'new_acell" (fun_3 __newAcell);;


(**
  @module achain
  @name 'new_vertex
  @type unit_acell

  Creates a new acell of dimension 0.
*)
addenv_impure "'new_vertex" (VAL_funct (fun _ -> VAL_acell(Core_achain.new_vertex())));;


(**
  @module achain
  @name 'new_edge
  @type acell_acell_acell

  Creates a new acell of dimension 1. The arguments are two acells of dimension 0.
*)
addenv_impure "'new_edge" (fun_2 __newEdge);;



(**
  @module achain
  @name 'new_face
  @type acell_acell_acell_acell

  Creates a new acell of dimension 2. The argument are three acells of dimension 1.
*)
addenv_impure "'new_face" (fun_3 __newFace);;

(**
  @module achain
  @name 'dimension
  @type acell_int
  @type achain_int

  Return the dimension of a acell or a achain.
*)
addenv_impure "'dimension" (fun_1 __dim);;

(**
  @module achain
  @name 'faces
  @type achain_acell_seq
  @type acell_seq

  Return the sequence of faces of a cell or the values in a chain.
  [faces(c)] returns the sequence of the faces of [c].
  [faces(ch,c)] returns the sequence of the values of the faces of [c] in the chain [ch].
*)
addenv_impure "'faces" (fun_1 __faces);;


(**
  @module achain
  @name 'cofaces
  @type achain_acell_seq
  @type acell_seq

  Return the sequence of cofaces of a cell or the values in a chain.
  [cofaces(c)] returns the sequence of the faces of [c].
  [cofaces(ch,c)] returns the sequence of the values of the cofaces of [c] in the chain [ch].
*)
addenv_impure "'cofaces" (fun_1 __cofaces);;

(**
  @module achain
  @name 'icells
  @type acell_int_seq
  @type achain_acell_int_seq

  Return the sequence of icells of a cell or the values in a chain.
  The icells of dimension [i] of a cell [c] are the cells of dimension [i] that are parts of
  [c] or that include [c].
  [icells(c,i)] returns the sequence of the icells of dimension [i] of [c].
  [icells(ch,c,i)] returns the sequence of the values of the icells of dimension [i] of [c]
  in the chain [ch].
*)
addenv_impure "'icells" (fun_2 __icells);;

(**
  @module achain
  @name 'ccells
  @type acell_int_seq
  @type achain_acell_int_seq

  Return the sequence of ccells of a cell or the values in a chain.
  The ccells through dimension [i] of a cell [c] are the cells with the same dimension as [c]
  that share an icell with [c].
  [ccells(c,i)] returns the sequence of the ccells through dimension [i] of [c].
  [ccells(ch,c,i)] returns the sequence of the values of the icells through dimension [i] of [c]
  in the chain [ch].
*)
addenv_impure "'ccells" (fun_2 __ccells);;

(**
  @module achain
  @name 'transitive_faces
  @type acell_seq
  @type achain_acell_seq
  Return the sequence of faces of a cell and the faces of the faces, recursively.

  @voir 'faces 'faces
*)
addenv_impure "'transitive_faces" (fun_1 __transitivefaces);;

(**
  @module achain
  @name 'transitive_cofaces
  @type acell_seq
  @type achain_acell_seq

  Return the sequence of cofaces of a cell and the cofaces of the cofaces, recursively.

  @voir 'cofaces 'cofaces
*)
addenv_impure "'transitive_cofaces" (fun_1 __transitivecofaces);;

(**
  @module achain
  @name 'common_faces
  @type acell_acell_seq
  @type achain_acell_acell_seq

  Return the sequence of the faces shared by two cells.

  @voir 'faces 'faces
*)
addenv_impure "'common_faces" (fun_2 __commonfaces);;

(**
  @module achain
  @name 'common_cofaces
  @type acell_acell_seq
  @type achain_acell_acell_seq

  Return the sequence of the cofaces shared by two cells.

  @voir 'cofaces 'cofaces
*)
addenv_impure "'common_cofaces" (fun_2 __commoncofaces);;

(**
  @module achain
  @name 'common_icells
  @type acell_acell_int_seq

  Return the sequence of the icells shared by two cells.

  @voir 'icells 'icells
*)
addenv_impure "'common_icells" (fun_3 __commonicells);;

(**
  @module achain
  @name 'common_ccells
  @type acell_acell_int_seq

  Return the sequence of the ccells shared by two cells.

  @voir 'ccells 'ccells
*)
addenv_impure "'common_ccells" (fun_3 __commonccells);;

(**
  @module achain
  @name 'common_transitive_faces
  @type acell_acell_seq
  @type achain_acell_acell_seq

  Return the sequence of the transitive faces shared by two cells.

  @voir 'faces 'faces
  @voir 'transitive_faces 'transitive_faces
  @voir 'common_faces 'common_faces
*)
addenv_impure "'common_transitive_faces" (fun_2 __commontransitivefaces);;

(**
  @module achain
  @name 'common_transitive_cofaces
  @type acell_acell_seq
  @type achain_acell_acell_seq

  Return the sequence of the transitive cofaces shared by two cells.

  @voir 'cofaces 'cofaces
  @voir 'transitive_cofaces 'transitive_cofaces
  @voir 'common_cofaces 'common_cofaces
*)
addenv_impure "'common_transitive_cofaces" (fun_2 __commontransitivecofaces);;

(**
  @module achain
  @name 'facespos
  @type achain_acell_seq

  Return the sequence of faces of a cell in a chain.
  [facespos(ch,c)] returns the sequence of the faces of [c] in the chain [ch].

  @voir 'faces 'faces
*)
addenv_impure "'facespos" (fun_1 __facespos);;


(**
  @module achain
  @name 'cofacespos
  @type achain_acell_seq

  Return the sequence of cofaces of a cell in a chain.
  [cofacespos(ch,c)] returns the sequence of the cofaces of [c] in the chain [ch].

  @voir 'cofaces 'cofaces
*)
addenv_impure "'cofacespos" (fun_1 __cofacespos);;

(**
  @module achain
  @name 'icellspos
  @type achain_acell_int_seq

  Return the sequence of icells of a cell in a chain.
  [icellspos(ch,c,i)] returns the sequence of the icells of dimension [i] of [c] in the chain [ch].

  @voir 'icells 'icells
*)
addenv_impure "'icellspos" (fun_2 __icellspos);;

(**
  @module achain
  @name 'ccellspos
  @type achain_acell_int_seq

  Return the sequence of ccells of a cell in a chain.
  [ccellspos(ch,c,i)] returns the sequence of the ccells through dimension [i] of [c] in the chain [ch].

  @voir 'ccells 'ccells
*)
addenv_impure "'ccellspos" (fun_2 __ccellspos);;

(**
  @module achain
  @name 'transitive_facespos
  @type achain_acell_seq

  Return the sequence of faces of a cell and the faces of the faces in a chain, recursively.

  @voir 'facespos 'facespos
*)
addenv_impure "'transitive_facespos" (fun_1 __transitivefacespos);;

(**
  @module achain
  @name 'transitive_cofacespos
  @type achain_acell_seq

  Return the sequence of cofaces of a cell and the cofaces of the cofaces in a chain, recursively.

  @voir 'cofacespos 'cofacespos
*)
addenv_impure "'transitive_cofacespos" (fun_1 __transitivecofacespos);;

(**
  @module achain
  @name 'common_facespos
  @type achain_acell_acell_seq

  Return the sequence of the faces shared by two cells in a chain.

  @voir 'facespos 'facespos
*)
addenv_impure "'common_facespos" (fun_2 __commonfacespos);;

(**
  @module achain
  @name 'common_cofacespos
  @type achain_acell_acell_seq

  Return the sequence of the cofaces shared by two cells in a chain.

  @voir 'cofacespos 'cofacespos
*)
addenv_impure "'common_cofacespos" (fun_2 __commoncofacespos);;

(**
  @module achain
  @name 'common_transitive_facespos
  @type achain_acell_acell_seq

  Return the sequence of the transitive faces shared by two cells in a chain.

  @voir 'facespos 'facespos
  @voir 'transitive_facespos 'transitive_facespos
  @voir 'common_facespos 'common_facespos
*)
addenv_impure "'common_transitive_facespos" (fun_2 __commontransitivefacespos);;

(**
  @module achain
  @name 'common_transitive_cofacespos
  @type achain_acell_acell_seq

  Return the sequence of the transitive cofaces shared by two cells in a chain.

  @voir 'cofacespos 'cofacespos
  @voir 'transitive_cofacespos 'transitive_cofacespos
  @voir 'common_cofacespos 'common_cofacespos
*)
addenv_impure "'common_transitive_cofacespos" (fun_2 __commontransitivecofacespos);;

(**
  @module collections
  @name 'achainify
  @type collection_achain

  Transform its argument into a achain.
  ['achainify(c)] returns the achain of the elements of the collection
  [c].
*)
addenv_impure "'achainify" (fun_1 __achainify);;

(**
  @module math
  @name 'acellof
  @type int_acell

  Conversion into an acell.
  If the argument is an [int], the result is the [acell] whose uid is
  equal to the [int].
*)
(addenv_impure "'acellof"      (fun_1 __acellof));;

(** 
  @module math
  @name 'acell_of
  @type int_acell
  
  Conversion into an acell.

  @alias math.'acellof
  @voir 'acellof 'acellof
*)
(addenv_impure "'acell_of"     (fun_1 __acellof));;


(**
  @module achain
  @name 'foldD
  @type funct_any_achain_int
  Fold a function over the elements of a particular dimension of a collection.

  [foldD] is an operator that acts over absract chains. 
  If [c1, c2, ..., cn] of dimension [d] are the elements of an achain  [c],
  then [foldD(f, zero, c, d)] computes [f(c1, f(c2, ... f(cn, zero)))]. If the
  collection has no element of dimension [d], then [zero] is returned.

  @voir 'fold fold
*)
addenv_impure "'foldD"         (fun_4 __foldD);;

(**
  @module achain
  @name 'iterD
  @type funct_achain_int_undef
  Iter a function over the elements of a particular dimension of a collection.

  [iterD] is an operator that acts over abstract chains. 
  The results is undef. It is used for the side-effect of the function
  applied to the collection's elements. 

  @voir 'iter iter
*)
addenv_impure "'iterD"         (fun_3 __iterD);;

(**
  @module achain
  @name 'mapD
  @type funct_achain_int_any
  Map a function over the elements of a particular dimension of a collection, others remain unchanged.

  [mapD] is an operator that acts over abstract chains. Map a function over
  the elements of a particular dimension of a collection, others remain unchanged.

  @voir 'map map
*)
addenv_impure "'mapD"         (fun_3 __mapD);;

(**
  @module achain
  @name 'mapD_indexed
  @type funct_achain_int_any
  Map a function over the elements and their positions of a particular dimension of
  a collection, others remain unchanged.

  @voir 'mapD mapD
  @voir 'map_indexed map_indexed
*)
addenv_impure "'mapD_indexed"         (fun_3 __mapDindexed);;

(**
  @module achain
  @name 'iterD_indexed
  @type funct_achain_int_undef
  Iter a function over the elements of a particular dimension of a collection with their position.

  [iter_indexed(f, c)] is an expression that iterates the function
  [f] over the elements of the collection [c]. The function [f] takes
  2 arguments: the first one is a position [p] and the second is the value
  at position [p] in the collection [c]. 

  @voir 'iterD iterD
  @voir 'iter_indexed iter_indexed
*)
addenv_impure "'iterD_indexed" (fun_3 __iterDindexed);;

(**
  @module achain
  @name 'foldD_indexed
  @type funct_any_achain_int_any
  Fold a function over the elements of a particular dimension of a collection with their position.

  @voir 'foldD foldD
  @voir 'fold_indexed fold_indexed
*)
addenv_impure "'foldD_indexed" (fun_4 __foldDindexed);;

(**
  @module achain
  @name 'forallD
  @type funct_achain_int_bool

  Universal predicate.
  ['forall(p, c, d)] checks if all elements of dimension [d] of the collection [c]
  satisfy the predicate [p]. If the collection is empty, [true] is returned.
*)
addenv_impure "'forallD" (fun_3 __forallD);;

(**
  @module achain
  @name 'existsD
  @type funct_achain_int_bool

  Existential predicate. 
  ['exists(p, c, d)] checks if at least one element of dimension [d] of
  the collection satisfies the predicate [p]. If the collection is empty,
  [false] is returned.
*)
addenv_impure "'existsD" (fun_3 __existsD);;


(** 
  @module achain
  @name 'filtration
  @type funct_achain_seq
  
  Computes the filtration of an abstract chain.

  [filtration(fct,ch)] returns the filtration of the abstract chain [ch]
  whose elements are partially ordered by the function [fct] and the
  face/coface relationship. A filtration is a total order on a set of
  achains. This total order is computed using two partial orders:
  {ul
  {- the first one partially compares two acells of different dimension.
  If [c1] is a transitive_face of [c2], [c1] is {i lower} than [c2], and
  reciprocally if [c1] is a transitive_coface of [c2], [c1] is
  {i greater} than [c2]. Otherwise, [c1] and [c2] can't be compared.}
  {- The second partial order is computed using [fct], that is a function
  of 3 arguments. [fct(d,v1,v2)] has to return an integer [i], that
  depends on [d] the dimension of the compared abstract cells, and the
  values [v1] and [v2] associated with these cells in [ch]. [i] is negative
  if [v1] is lower than [v2], positive when [v1] is greater than [v2], null
  if [v1 = v2]. In fact, [fct] is a total order between cells with same
  dimension}
  }
  As an example, consider the two following abstract chains [ch1] and [ch2]:

  [v1 := new_vertex() ;;]

  [v2 := new_vertex() ;;]

  [v3 := new_vertex() ;;]

  [e1 := new_edge(v1,v2) ;;]

  [e2 := new_edge(v2,v3) ;;]

  [ch1 := 1*v1 + 2*e1 + 3*v2 + 4*e2 + 5*v3 ;;]

  [ch2 := 1*v1 + 4*e1 + 3*v2 + 2*e2 + 5*v3 ;;]

  Let compute the filtration of [ch1] and [ch2] using the standard comparison
  function [compare] provided by mgs:

  [filtration((\d.(compare)),ch1) ;;]

  returns the sequence

  [(v1, v2, e1, v3, e2)]

  while

  [filtration((\d.(compare)),ch2) ;;]

  returns

  [(v1, v2, v3, e2, e1)]

*)
(addenv_impure "'filtration"     (fun_2 __filtration));;


(** 
  @module achain
  @name 'purge
  @type achain_achain
  
  Remove <undef> association in achain (used for optimizing transformation applications).
*)
(addenv_impure "'purge"     (fun_1 __purge));;

(** 
  @module achain
  @name 'boundary
  @type acell_achain

  returns the boundary of an acell as a chain of integers.
*)
(addenv_impure "'boundary"     (fun_1 __boundary));;
  
(** 
  @module achain
  @name 'coboundary
  @type acell_achain

  returns the coboundary of an acell as a chain of integers.
*)
(addenv_impure "'coboundary"     (fun_1 __coboundary));;
  



(**
  @module achain
  @name 'normalise
  @type funct_achain

  Build a new achain where the acell that have the same faces are
  quotiented.
  [normalise(fct, ch)] build a new achain where the acells are the
  acells of [ch] quotiented by the relation "having exactly the
  same faces". The values of the acell that have the same faces are
  combined in a random order using the binary function [fct] and
  becomes the value of the quotient acell.

  @voir 'quotient quotient
*)
(* FIXME addenv_impure "'normalise" (fun_2 __normalise);;*)


(**
  @module achain
  @name 'cartesian_prod
  @type funct_achain_achain_achain
  
  Return the cartesian product of the two achain arguments.
  The first argument is a function used to build the value that
  labels a  acell in the result, from the values of the
  originating acells in the achain arguments. 
*)
addenv_impure "'cartesian_prod" (fun_3 __cartesianProd);;

(**
  @module achain
  @name 'quotient
  @type funct_seq_achain

  Build a new achain where the acell have been quotiented.
  [quotient(fct, rel, ch)] build a new achain where the acells are the
  acells of [ch] quotiented by the relation {i rel}. This relation
  is represented by a sequence [rel] of couples. A couple is a
  sequence of two elements [(a, b)] specifying that [a] must be
  identified with [b]. The value of the acells that have been
  quotiented is computed using the fonction [fct] to combine (in a
  random order) the values of the initial acells.

  @voir 'normalise normalise
*)
(* FIXME addenv_impure "'quotient" (fun_3 __quotient);;*)


(**
  @module achain
  @name 'cofaces_in_achain
  @type achain_acell_seq

  Return the sequence of cofaces of a acell belonging to a given achain.
  If a acell belong to a achain, then all its face belongs also to the same 
  achain. However, the acell can have as coface, cofaces that belongs to other achains.
  The function ['cofaces] return all the cofaces of a acell. ['cofaces_in_achain(c, s)] 
  returns only the cofaces of [s] that belong also to the achain [c].

  @voir 'cofaces 'cofaces
  @voir 'transitive_cofaces_in_achain 'transitive_cofaces_in_achain
*)
(* FIXME addenv_impure "'cofaces_in_achain" (fun_2 __cofacesinachain);;*)



(**
  @module achain
  @name 'transitive_cofaces_in_achain
  @type achain_acell_seq

  Return the sequence of recursive cofaces of a acell belonging to a given achain.
  If a acell belong to a achain, then all its face belongs also to the same 
  achain. However, the acell can have as coface, cofaces that belongs to other achains.
  The function ['transitive_cofaces] return all the cofaces of a acell (and the cofaces
  of them, recursively). ['cofaces_in_achain(c, s)] returns only the recursive cofaces of [s] 
  that belong also to the achain [c].

  @voir 'cofaces_in_achain   'cofaces_in_achain
  @voir 'transitive_cofaces 'transitive_cofaces
*)
(* FIXME addenv_impure "'transitive_cofaces_in_achain" (fun_2 __transitivecofacesinachain);; *)


(**
  @module achain
  @name 'common_cofaces_in_achain
  @type achain_acell_acell_seq

  Return the sequence of the cofaces shared by two acells in a given achain.
*)
(* FIXME addenv_impure "'common_cofaces_in_achain" (fun_3 __commoncofacesinachain);;*)


(**
  @module achain
  @name 'common_transitive_cofaces_in_achain
  @type achain_acell_acell_seq

  Return the sequence of the transitive cofaces shared by two acells in a given achain.

  @voir 'transitive_cofaces transitive_cofaces
*)
(* FIXME addenv_impure "'common_transitive_cofaces_in_achain" (fun_3 __commontransitivecofacesinachain);;*)








(* --- fin de l'installation des fonctions --- *)

print_string "done *)";
print_newline();;

(* ======================================================= *)
