(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: topo_achain.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG, AS
 * Modified: $Date: 2007-09-12 13:36:06 $
 * Version: $Revision: 1.7 $
 * -------------------------------------------------------------------
 *
 * Status: voir le commentaire general dans topology.ml
 *
 *=========================================================================
 *)

open Global;;
open Valeur;;
open Terme;;
open Printer;;
open Printer_term;;
open Topology;;
open Def_achain;;
open Core_achain;;
open Printer_achain;;
open Type_achain;;
open Achain;;

__register_version "$RCSfile: topo_achain.ml,v $" "$Revision: 1.7 $" "$Date: 2007-09-12 13:36:06 $";;


exception FoundPos of acell ;;

let in_place_shuffle_array len a = 
  let nb  = ref (if (len <= 1)
		 then 0
		 else 
		   if (len = 2) 
		   then 1 
		   else len) (* Pour eviter que sur un tableau de longueur 2 on fasse 2 permutations inverses. *)
  in                         (* Le risque est moins grand sur un tableau de taille superieure.                 *)
    while (!nb > 0)
    do
      let src = Random.int len
      and dst = Random.int len
      in
	if (src = dst)
	then ()
	else (let tmp = a.(src)
	      in
		a.(src) <- a.(dst);
		a.(dst) <- tmp;
		decr nb)
    done;
    a
;;


class topo_achain ((c : 'collection), couleur) mgsval (vdim:int) (ndim:int) =
object(self)
  inherit [acell] generic_topo 
    string_of_acell                   (** afficheur de position    **)
    (c, couleur)                      (** collection a transformer **)
    mgsval                            (** l'argument complet       **)
    (if (Core_achain.size c != 0) then
       let n      = Global.crandom (Core_achain.size c) in
       let (k, _) = Core_achain.nth n c
       in k                             (** position de depart       **)
     else
       Core_achain.default_acell 
    )
    
  method dummy () = Printf.fprintf stderr "#### creation d'un topo_chain\n"
  val _avl  = Core_achain.get_acells_of_dim_in_achain vdim c
  val _avl_usage = (let tbl = Hash.create (Core_achain.size_d vdim c)
		    in
		      Core_achain.iter_d () vdim (fun k v -> Hash.add tbl  (acelluid k)  (k,PosFree) ) c;
		      tbl
		   )
  val mutable _batch = ([] : (Terme.path_type * 'position path * mgs_value) list)
  val mutable _last_index = Global.crandom (Core_achain.size_d vdim c)
  
  method state s = snd(Hash.find _avl_usage (acelluid s))
  method update_state f s = Hash.replace _avl_usage (acelluid s) (s, f(snd(Hash.find _avl_usage (acelluid s))))
  method next_usable _ =
    try
      Hash.iter_from
	_last_index 
	(fun _ (k,v) -> let dim = acelldim k in if (( dim = vdim)&&(v = PosFree)) then raise (FoundPos k))
	_avl_usage;
      raise NoMoreUsable
    with FoundPos p -> (assert
			  begin 
			    if !trace_opt then
			      (Printf.printf "---- topoCHAIN: next_usable: found %s\n" (string_of_acell p)); 
			    true
			  end;
			_last_index <- Hash.last_iter_index();
			p)
  method unvisit_all =
    Hash.iter (fun k (s,v) -> if (pos_visited v) then (Hash.replace _avl_usage k (s,pos_unvisit v))) _avl_usage
  method direction_from_to p1 p2 = VAL_undef "direction"
  method some_neighbors p = function
    | AllDirections    -> (let verteces = (Core_achain.ccells p ndim) in
			   let ans = List.filter
				       (function v -> (pos_usable (self#state v)))
				       verteces
			   in
			   let n = List.length ans in
			     (n, if (!match_random > 0) then
				   (in_place_shuffle_array n (Array.of_list ans))
			         else
				   (Array.of_list ans)
			     )
			  )

    | SomeDirections d -> (let err = "Error: cannot use direction in chain: "
				     ^(Array.fold_left (fun s x -> s^"\t"^(to_string (prsvalue x))^"\n") "" d)
			   in 
			     prerr_endline err;
			     raise (Erreur err))

  method get_value_from_pos p = snd(Hash.find _avl (acelluid p))
  method mgs_value_from_pos p = VAL_acell(p)

  method add_match pt p v = _batch <- (pt, p, v)::_batch


  method private ajoute_tous h (kind, chemin, rhs) = 
    let idx = ref 0
    and longueur_chemin = List.length chemin
    in
      (match rhs
       with
	   VAL_seq (seq, col) ->
	     (if (longueur_chemin <> (Array.length seq))
	      then raise (Erreur "in CHAIN(??), replacement must be of same arity... at this time !!")
	      else (List.iter                               (*: ('a -> 'b -> 'b) -> 'a list -> 'b -> 'b*)
		      (fun pos -> 
			 ((match seq.(!idx) with 
			      VAL_undef _ -> ()   (* C'est que l'on souhaite le supprimer *)
			    | x           -> Core_achain.replace h pos x);
			  incr idx
			 )
		      )
		      chemin
		   )
	     )
(*	 | VAL_undef _ ->
	     if (longueur_chemin = 1) then
	       ()
	     else
	       (raise (Erreur "in CHAIN(??), replacement must be of same arity... at this time !!"))

*)	 | VAL_achain(ch,coul) when (kind.a_flat = FlatPath) -> (
	       if (Pervasives.compare coul couleur = 0) then
		 Core_achain.iter
		   (fun s v -> Core_achain.update_achain h Achain.combine_mgsval v s)
		   ch
	       else
		 if (longueur_chemin = 1)
		 then(Core_achain.replace h (List.hd chemin) rhs)
		 else raise (Erreur "Topo_achain: ajoute_tous: replacement must be of same arity")
	     )
	 | VAL_achain(ch,coul) when (false) ->
	     if (Pervasives.compare coul couleur = 0) then
	       if (longueur_chemin = 1) then
		 let old_s = (List.hd chemin) in (
		   Core_achain.iter
		     (fun s v ->
			if ( (acelldim s > vdim) ) then
			  raise (Erreur "on le remplace par un truc plus petit")
			else
			  try (
			    match ((Core_achain.find c s), v) with
			      | (VAL_undef _, VAL_undef t) -> ()
			      | (x, VAL_undef t)           -> ()
			      | (VAL_undef t, x)           -> (Core_achain.replace h s x)
			      | (x1, x2)                   -> (Core_achain.replace h s x2)
			  ) with
			      Not_found -> (
				if ((acelldim s) = vdim) then (
				  (* C'est un peu ol-ol mais j'ai envie de mettre  jour les faces et cofaces *)
				  (* En fait, ce serait mieux de laisser  l'utilisateur le choix de mettre    *)
				  (* a jour les cofaces et les faces ou non... On peut par exemple penser  une *)
				  (* notation telle que :                                                       *)
				  (*   chain:{| valeur @ (acell)[cofaces=(cf1,cf2,...)] , ... |}              *)
				  (* La chaine inclue alors les cofaces en qst et les ajoute mais en leur       *)
				  (* associant <undef>. Ainsi lors de la fusion, si les cofaces sont dj       *)
				  (* prsentes (et leurs faces d'ailleur) leurs valeurs ne sont pas modifies   *)
				  (* Il faut alors s'assurer que les cellules plus grandes en dimension sont    *)
				  (* absente de la chaine avant de les rejeter... Voire mme : pkoi les         *)
				  (* rejetter...                                                                *)
				  let cofaces_old_s = Core_achain.cofaces old_s in
				    List.iter
				      (fun cfs ->
					 internal_add_face s cfs;
					 internal_add_coface cfs s
				      )
				      cofaces_old_s ;
				) ;
				(Core_achain.replace h s v)
			      )
		     )
		     ch)
	       else
		 raise (Erreur "Je sais pas encore trop koi faire l :)")
	     else
	       if (longueur_chemin = 1)
	       then(Core_achain.replace h (List.hd chemin) rhs)
	       else raise (Erreur "in CHAIN(??), replacement must be of same arity... at this time !!")

	 | x -> 
	     if (longueur_chemin = 1)
	     then (Core_achain.replace h (List.hd chemin) x)
	     else raise (Erreur "in CHAIN(??), replacement must be of same arity... at this time !!")
      )

  method build_result =
    let tbl       = Core_achain.create_achain (Hash.size _avl)
    in
      List.iter (self#ajoute_tous tbl) _batch;
      Hash.iter
	(fun k (s,v) -> 
	   if (pos_usable v) then
	     (Core_achain.update_achain tbl combine_mgsval (snd(Hash.find _avl k)) s)
	)
	_avl_usage;
      Core_achain.iter
	(fun s v -> if ((acelldim s) <> vdim) then Core_achain.update_achain tbl combine_mgsval v s)
	c;
      VAL_achain (tbl, couleur)

end
;;
