(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: topo_bag.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2006-04-16 16:04:03 $
 * Version: $Revision: 1.4 $
 * -------------------------------------------------------------------
 *
 * Status: voir le commentaire general dans topology.ml
 *          
 * Ici on developpe les instance de generic_topo qui correspondent 
 * aux collections Associatives/Commutative (et parfois idempotente)
 *
 *=========================================================================
 *)

open Global;;
open Topo_ac;;
open Valeur;;
open Printer;;

__register_version "$RCSfile: topo_bag.ml,v $" "$Revision: 1.4 $" "$Date: 2006-04-16 16:04:03 $";;


let opt_bag_matching = ref true ;;

class topo_bag (c, couleur) valeur card =
object (self)
  inherit [bag_collection] topo_ac (c, couleur, valeur) card

  method iterateur = Core_bag.iter_b
  method empty     = Core_bag.empty_b
  method add v c   = (match v with VAL_undef _ -> c | _ -> Core_bag.add_b v c)
  method union     = Core_bag.union_b
  method from_coll = (function VAL_bag (v, _) -> v
			|      x              -> raise (Erreur ("bag awaited, received : "^(to_string(prsvalue x)))))
  method to_coll x = VAL_bag x

  method occurence v = Core_bag.num_b v c

  method all_occurences () = (let tbl = Hashtbl.create 1
			      in
				Core_bag.iter_occurences_b (fun (v, n) -> Hashtbl.add tbl v n) c;
				tbl)

  method dispose_get_NvalueS_from_pos = true

  method get_NposS_from_pos position nombre valeur =
    let chemin       = ref []
    and pos_courante = ref position
    and pos_retour   = ref (-1)
    and ok           = ref false
    and restant      = ref (nombre - 1)
    and teste      p = Dispatch_sig.equal_mgs_value valeur (self#get_value_from_pos p)
    in
      if (teste !pos_courante)
      then 
	begin
	  chemin := !pos_courante :: !chemin;
	  incr pos_courante;
	  while ((!restant > 0) & (!pos_courante < card) & (teste !pos_courante))
	  do
	    if self#usable(!pos_courante)
	    then 
	      begin
		decr restant;
		chemin := !pos_courante :: !chemin;
	      end;
	    pos_retour := !pos_courante;
	    incr pos_courante
	  done;
	  
	  pos_courante := position - 1;
	  
	  while ((!restant > 0) & (!pos_courante > 0) & (teste !pos_courante))
	  do
	    if self#usable(!pos_courante)
	    then 
	      begin
		decr restant;
		chemin := !pos_courante :: !chemin;
	      end;
	    decr pos_courante
	  done;
	  
	  if (!restant = 0)
	  then Some (!chemin, !pos_retour)
	  else None
	end
      else None

end;;


open Topology;;
open Terme;;

type bag_state_property =
    { mutable nb_free    : int;
      mutable nb_visited : int;
      mutable nb_used1   : int;
      mutable nb_used2   : int;
      mutable value      : mgs_value;
    }
;;

class optimized_topo_bag (c, couleur) mgsval card =
object(self)
  inherit [int] generic_topo 
    string_of_int (*fun c -> (to_string(prsvalue c))*) (** afficheur de position    **)
    (c, couleur)                         (** collection a transformer **)
    mgsval                               (** la valeur mgs complete sur laquelle on travaille **)
    (Global.crandom card)                (** position de depart       **)


  val _used    = Array.init card (fun x -> { nb_free=0; nb_visited=0; nb_used1=0; nb_used2=0; value=VAL_undef "" })
  val _voisins = Array.make (max 0 card) 0
  val mutable _batch = ([] : (rule_property * 'position path * mgs_value) list)

  method iter_batch f = List.iter f _batch

  method from_coll = (function VAL_bag (v, _) -> v
			|      x              -> raise (Erreur ("bag awaited, received : "^(to_string(prsvalue x)))))

  method update_state _ _ = raise (Erreur "Topo_bag: update_state: function deprecated with this optimization")
  method state _          = raise (Erreur "Topo_bag: state: function deprecated with this optimization")

  (** Fonctions rimplmentes dans utilises update_state et state... **)
  method use     p =
    if (_used.(p).nb_visited = 0)
    then
      if (_used.(p).nb_free = 0)
      then raise (Erreur "Topo_bag: use: lack of usable element")
      else (_used.(p).nb_free    <- _used.(p).nb_free - 1 ; _used.(p).nb_used1 <- _used.(p).nb_used1 + 1)
    else (_used.(p).nb_visited <- _used.(p).nb_visited - 1 ; _used.(p).nb_used2 <- _used.(p).nb_used2 + 1)
(*; self#debug "use" *)
  method unuse   p = 
    if (_used.(p).nb_used2 = 0)
    then
      if (_used.(p).nb_used1 = 0)
      then raise (Erreur "Topo_bag: unuse: lack of used element")
      else (_used.(p).nb_used1 <- _used.(p).nb_used1 - 1 ; _used.(p).nb_free <- _used.(p).nb_free + 1)
    else (_used.(p).nb_used2 <- _used.(p).nb_used2 - 1 ; _used.(p).nb_visited <- _used.(p).nb_visited + 1)
(*; self#debug "unuse" *)
  method visit   p =
    if (_used.(p).nb_free = 0)
    then raise (Erreur "Topo_bag: use: lack of free element")
    else (_used.(p).nb_free <- _used.(p).nb_free - 1 ; _used.(p).nb_visited <- _used.(p).nb_visited + 1)
(*; self#debug "visit" *)
  method unvisit p =
    if (_used.(p).nb_visited <> 0)
    then (_used.(p).nb_visited <- _used.(p).nb_visited - 1 ; _used.(p).nb_free <- _used.(p).nb_free + 1)
(*; self#debug "unvisit" *)

  method used    p = (_used.(p).nb_used1 <> 0) || (_used.(p).nb_used2 <> 0)
  method visited p = (_used.(p).nb_visited <> 0)
  method usable  p = (_used.(p).nb_visited <> 0) || (_used.(p).nb_free <> 0)
  method move p    = assert(self#usable(p)); _cpos <- p

  method unvisit_all = 
    (*self#debug "avant unvisit_all" ;*)
    for i = 0 to (card-1) do
      if (_used.(i).nb_visited > 0) then (
	_used.(i).nb_free <- _used.(i).nb_free + _used.(i).nb_visited ;
	_used.(i).nb_visited <- 0
      )
    done
    (*; self#debug "aprs unvisit_all"*)
  method some_neighbors p = function
    | AllDirections    -> (let vois = ref 0 
			   in
			     Array.iteri (fun idx etat -> if (self#usable idx) then (_voisins.(!vois) <- idx; incr vois)) _used;
			     !vois, if (!match_random > 0) then
			              in_place_shuffle_array (!vois) (Array.copy _voisins)
			            else
			              (Array.copy _voisins)
			  )

    | SomeDirections d -> (let err = "Error: cannot use direction in set/bag: "
				     ^(Array.fold_left (fun s x -> s^"\t"^(to_string (prsvalue x))^"\n") "" d)
			   in 
			     prerr_endline err;
			     raise (Erreur err))

  method next_usable p =
    try
      for i = 0 to (card-1)
      do
	let position = (p+i) mod card
	in
	  if (_used.(position).nb_free <> 0) then raise (FoundIntPos position)
      done;
      raise NoMoreUsable
    with FoundIntPos p -> (assert begin if !trace_opt then
			     (Printf.printf "---- topoAC: next_usable: found %d\n" p); 
			     true
			   end;
			   p)

  method mgs_value_from_pos pos = (VAL_int pos)
  method get_value_from_pos p = _used.(p).value
  method direction_from_to p1 p2 = VAL_undef "direction"
  
  method add_match pt p v = _batch <- (pt, p, v)::_batch

  method add v c = (match v with VAL_undef _ -> c | _ -> Core_bag.add_b v c)
  method addn v n c = (match v with VAL_undef _ -> c | _ -> Core_bag.addn_b v n c)


  method ajoute_tous ac couleur kind = function
      | (VAL_seq (seq, col)) as valeur -> (if (kind.a_flat = FlatPath) 
					   then Array.fold_right (fun elem suite -> self#add elem suite) seq ac
					   else self#add valeur ac)

      | valeur  when ((is_collection valeur) &
		      (is_same_coll_kind valeur mgsval))  -> (if ((kind.a_flat = NoFlatPath) or (not (Type.is_subtype (collection_type valeur) couleur)))
							      then (self#add valeur ac)
							      else (Core_bag.union_b (self#from_coll(valeur)) ac))
      | x                              -> self#add x ac

  method build_result =
    let nouveau   = List.fold_right (fun (kind, chemin, rhs) nouveau -> self#ajoute_tous nouveau couleur kind rhs) _batch Core_bag.empty_b
    and untouched = (let ret = ref Core_bag.empty_b
		     in
		       Array.iteri (fun idx etat -> if (self#usable idx) then ret := self#addn (etat.value) (etat.nb_visited + etat.nb_free) !ret) _used;
		       !ret)
    in
      VAL_bag ((Core_bag.union_b nouveau untouched), couleur)

  method affiche_used = Array.iter (fun s -> Printf.printf "%s(%d,%d,%d,%d) "
				      (to_string(prsvalue s.value))
				      s.nb_free
				      s.nb_visited
				      s.nb_used1
				      s.nb_used2) _used ; Printf.printf "\n" ; flush stdout


  method debug f = Printf.printf "%s" f ; self#affiche_used

  method find v =
    try
      for i = 0 to (card-1)
      do
	let index = (self#cpos + i) mod card in
	let st = _used.(index)
	in
          if (self#usable(index) && (Dispatch_sig.equal_mgs_value v (st.value))) then
		raise (FoundIntPos index)
	  done ;
	  None
    with FoundIntPos index -> Some index

  method findN n v =
    try
      for i = 0 to (card-1)
      do
	let index = (self#cpos + i) mod card in
	let st = _used.(index)
	in
          if ((st.nb_free + st.nb_visited >= n) && (Dispatch_sig.equal_mgs_value v (st.value))) then
	    raise (FoundIntPos index)
	  done ;
	  None
    with FoundIntPos index ->
      let rec gen acc = function
	| 0 -> acc
	| n -> (gen (index::acc) (n-1))
      in
	Some (gen [] n, index)

  method dispose_get_NvalueS_from_pos = true

  method get_NposS_from_pos position nombre valeur =
      if (Dispatch_sig.equal_mgs_value valeur (self#get_value_from_pos position)) then
	if ((_used.(position).nb_free + _used.(position).nb_visited) >= nombre) then
	  let rec gen acc = function
	    | 0 -> acc
	    | n -> (gen ( position::acc) (n-1))
	  in
	    Some (gen [] nombre, position)
	else
	  None
      else
	None

  method occurence v = Core_bag.num_b v c

  method all_occurences () = 
    (let tbl = Hashtbl.create 23
     in
       Core_bag.iter_occurences_b (fun (v, n) -> Hashtbl.add tbl v n) c;
       tbl)

  initializer
  let idx = ref 0
  in
    Core_bag.iter_occurences_b
      (function (v,n) ->
	 _used.(!idx).value <- v ;
	 _used.(!idx).nb_free <- n ;
	 incr idx
      )
      c

end;;
