(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: gab.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2006-04-23 06:18:45 $
 * Version: $Revision: 1.4 $
 * -------------------------------------------------------------------
 *
 * Status:  
 *    Implantation au niveau "representation MGS" des fonctions sur les 
 *    gab. Voir commentaire dans Seq/seq.ml
 *
 *    Dans un rpertoire Xxx correspondant  une collection xxx, 
 *    le type xxx est defini dans def_xxx.ml et les operations sur 
 *    ce type sont dfinies dans core_xxx.ml. Le fichier kernel_xxx.ml
 *    contient la declaration dans la librairie MGS des fcts dfinies 
 *    dans xxx.ml
 *
 *    L'organisation de xxx.ml fichier suit l'ordre suivant:
 *
 *       1/ Fonctions de construction: ... -> seq (sauf extension)
 *             cons, join, append, shape, xxxify, iota, extension
 *
 *       2/ Prdicats, accs et observateur
 *             compare, val2bool, size, empty, count, 
 *             member, random, hd, tl, take
 *
 *       3/ Fold et autre itrations
 *             fold, map, iter, foldindexed, mapindexed, iterindexed,
 *             forall, exists, filter
 *
 *       4/ Neighbors, neighbor
 *
 *
 *       5/ Fonctions spcifiques
 *
 *
 *       6/ mise  jour du descripteur du type
 *              champs: make, xxxify et empty
 *
 * Ici :   xxx = gab
 *
 *=========================================================================
 *)


open Global;;
open Def_gab;;
open Valeur;;
open Core_gab;;
open Type;;
open Dispatch_sig;;



__register_version "$RCSfile: gab.ml,v $" "$Revision: 1.4 $" "$Date: 2006-04-23 06:18:45 $";;

(** alias **)
let gab_desc = Type_gab.gab_desc
;;



(** === Constructeurs ======================================================- **)


let _cons_any_gab (e, s) = VAL_gab(Core_gab.add_g e (fst s), snd s)

and _cons_undef_gab (_, s) = VAL_gab s

and _append_gab_gab (s1, s2) = VAL_gab(union_g (fst s1) (fst s2), commontype (snd s1) (snd s2))

and _join_gab_gab (s1, s2) = VAL_gab(union_g (fst s1) (fst s2), commontype (snd s1) (snd s2))

and _join_any_gab (e, s) = if true_coll e then Seq._join_any_any (e, VAL_gab(s)) else VAL_gab(add_g e (fst s), snd s)
			       
and _join_gab_any (s, e) = if true_coll e then Seq._join_any_any (VAL_gab(s), e) else VAL_gab(add_g e (fst s), snd s)

and _join_gab_undef (s, _) = VAL_gab s

and _join_undef_gab (_, s) = VAL_gab s

and _mult_int_gab (i,(g,t)) = VAL_gab(mult_g i g,t)

and _add_gab_gab (s1, s2) = VAL_gab(union_g (fst s1) (fst s2), commontype (snd s1) (snd s2))

and _add_any_gab (v, (g,t)) = VAL_gab(add_g v g, t)

and _add_gab_any ((g,t), v) = VAL_gab(add_g v g, t)

and _sub_gab_gab (s1, s2) = VAL_gab(diff_g (fst s1) (fst s2), commontype (snd s1) (snd s2))

and _sub_gab_any ((g,t), v) = VAL_gab(remove_g v g, t)

and _modulo_gab_int ((g, t), i) = VAL_gab(map_occ_g (fun n -> n mod i) g, t)
;;


let  _shape_gab (b, deco) = 
  let taille_g = Core_gab.cardinal_g b
  in VAL_funct 
       (function 
	  | VAL_seq (s, _) -> 
	      if (taille_g <> Array.length s)
	      then raise_error "shape (gab) : argument is not of the same size of the original"
	      else VAL_gab ((Array.fold_right (fun x y -> Core_gab.add_g x y) s Core_gab.empty_g), deco)
	  | _  -> raise_error "shape (gab) : a sequence as argument is required")
;;

let _sequify_gab s = VAL_seq(Array.of_list (elements_g (fst s)), Type_seq.seq_desc) ;;

let _gabify_gab s = VAL_gab(fst s, gab_desc)
and _gabify_seq s = VAL_gab(Array.fold_left (fun ss x -> add_g x ss) empty_g (fst s), gab_desc)
and _gabify_any x = 
  if not(true_coll x) then VAL_gab(singleton_g(x), gab_desc)
  else 
    Dispatch_sig.fold 
      (VAL_funct(fun x -> VAL_funct(function
				      | VAL_gab(y, t) -> VAL_gab((Core_gab.add_g x y), t)
				      | _ -> failwith "internal error: gabify_any")))
      gab_desc.empty 
      x
;;


let _iota_int_gab (n, (v, col)) = 
  if ( n <= 0)
  then raise (Erreur "iota_gab with a negative argument")
  else (let res = ref v
	in
	  for i = 0 to (n-1)
	  do
	    res := Core_gab.add_g (VAL_int i) !res
	  done;
	  VAL_gab (!res, col))
;;

let _inter_gab_gab ((s1,t1),(s2,t2)) =
  let t = Type.commontype t1 t2 in
    VAL_gab(Core_gab.inter_g s1 s2, t)
;;

let _diff_gab_gab ((s1,t1),(s2,t2)) =
  let t = Type.commontype t1 t2 in
    VAL_gab(Core_gab.diff_g s1 s2, t)
;;


(** Extension --------------------------------- **)


let rec extension_gab desc a = 
  VAL_gab(Array.fold_left add_elem Core_gab.empty_g a, desc)

and add_elem gab = function
  | VAL_seq(aa, _) -> Core_gab.addn_g aa.(1) (to_int aa.(0)) gab
  | _ -> raise (Erreur "Internal Error: extension: add_elem in extension_gab")
;;



(** === Prdicats, accs et observateurs =========================================== **)

(** _compareMgsVal_gab_gab est dfini dans core_gab *)

let _val2bool_gab (b, _) = VAL_bool(not(Core_gab.is_empty_g b))
;;


let _size_gab (b, _) = VAL_int(cardinal_g b)
and _empty_gab (b, _) = VAL_bool(b = Core_gab.empty_g)
;;



let _count_any_gab (a, (b, _)) = VAL_int (Core_gab.num_g a b)
;;

let _member_any_gab (a, (b, _)) = VAL_bool (Core_gab.mem_g a b) 



exception Ok of mgs_value;;

let _random_gab(s, _) = 
  let n = Core_gab.set_cardinal_g s in
    if n == 0 then VAL_undef "random: bad argument: empty gab" 
    else 
      let nn = ref(Random.int n) in
	try
	  Core_gab.iter_g (function x -> if !nn == 0 then raise (Ok x) else decr(nn)) s;
	  assert(false);
	  VAL_undef("internal error in _random_gab")
	with Ok x -> x
;;



let _hd_gab s = 
  let ss = fst s in if (0 = set_cardinal_g ss) then VAL_undef("hd on empty gab") else choose_g ss

and _tl_gab s = 
  let ss = fst s 
  in 
    if (0 = set_cardinal_g ss)
    then VAL_undef("tl on empty gab") 
    else VAL_gab(remove_g (choose_g ss) ss, snd s)
;;

let _take_gab_any ((g, _),s) = VAL_int(num_g s g) ;;

(** take_gab_int est pris en charge par le _take_any_int dfini dans Seq/seq.ml **)



(* === Fold et autre itrations ================================ *)

let _fold_funct_any_gab (fonc, zero, (gab, _)) = 
  Core_gab.fold_g  (fun v suite -> (app_2 fonc v suite)) gab zero

and _map_funct_gab (fonc, (gab, c)) = 
  VAL_gab ((Core_gab.fold_g (fun v suite -> Core_gab.add_g (fonc v) suite) gab Core_gab.empty_g), c)

and _iter_funct_gab (fonc, (gab, _)) = 
  (Core_gab.iter_g (fun x -> fonc x; ()) gab);  VAL_undef "iter gab"

and _mapindexed_funct_gab (fct, (gab, t)) = 
  VAL_gab ((Core_gab.fold_occurences_g (fun n v suite -> Core_gab.add_g (app_2 fct v (VAL_int n)) suite) gab Core_gab.empty_g), t)

and _iterindexed_funct_gab (fonc, (gab, _)) = 
  (Core_gab.iter_occurences_g (fun n x -> ignore(app_2 fonc x (VAL_int n))) gab);  VAL_undef "iter gab"

and _foldindexed_funct_any_gab (fct, zero, (gab, _)) = 
  (Core_gab.fold_occurences_g (fun n v suite -> app_3 fct v (VAL_int n) suite) gab zero)

and _forall_funct_gab (p, (b, _)) = VAL_bool(for_all_g (fun x -> Auxfct.bool_from_val(p x)) b)

and _exists_funct_gab (p, (b, _)) = VAL_bool(exists_g (fun x -> Auxfct.bool_from_val(p x)) b)

and _filter_funct_gab (p, (b, t)) = VAL_gab(filter_g (fun x -> Auxfct.bool_from_val(p x)) b, t)
;;





(* === Neighbors, neighbor ================================ *)

let _neighbors_gab_any ((t, _), v) = 
  let sz         = Core_gab.set_cardinal_g t
  and retourne v = VAL_seq (v, Type_seq.seq_desc)
  in
    if not (mem_g v t)
    then VAL_undef "neighbors (gab): undefined neighbor"
    else
      let ret = Array.make (sz-1) (VAL_undef "")
      and ptr = ref 0
      in
	Core_gab.iter_g (fun vv -> if (compare_mgs_value vv v <> 0) then (ret.(post_inc ptr) <- v)) t;
	retourne ret
;;





(* === Fonctions spcifiques ======================================================- *)


(* Flatten ----------------------------------------------------------- *)


let _flatten_gab (s, t) =
  let ss = Core_gab.fold_occurences_g
	     (fun n v suite -> match v with 
		| VAL_gab (a, _) -> Core_gab.union_g (Core_gab.mult_g n a) suite
		| x -> Core_gab.addn_g x n suite)
	     s 
	     Core_gab.empty_g
  in VAL_gab(ss, t)
;;

(*let _subset_gab_gab ((a,_),(b,_)) = VAL_bool(Core_gab.subset_g a b) ;;*)

(** Partition en 2 classes suivant un predicat ----------------------------------------------- *)


let   _partition_funct_gab (f, (s, t)) = 
  let (s1, s2) = Core_gab.partition_g (function x -> Auxfct.bool_from_val (f x)) s
  in VAL_gab(Core_gab.add_g (VAL_gab(s1, t)) (Core_gab.singleton_g (VAL_gab(s2, t))), 
	     t)
;;



(** Une fonction partition par quivalence sur les monoide ----------------------------------- *)

let rec _partitioneq_funct_gab (f, (s, t)) =
  let eql = ref [] 
  and fill acc (_, lx) = Core_gab.add_g (VAL_gab(make_gab_from_list !lx, t)) acc
  in 
    begin
      Core_gab.iter_g (function e -> update2 eql f e) s;
      VAL_gab(List.fold_left fill Core_gab.empty_g !eql, t)
    end
    
and make_gab_from_list l = 
  List.fold_left (fun acc x -> Core_gab.add_g x acc) Core_gab.empty_g l

and update2 eql f x = eql := Auxfct.update_equivalence f x !eql
;;



(* ================================================================ *)

(*** !!! Ne pas oublier de mettre  jour les champs
 ***     _ make 
 ***     _ xxxify
 ***     _ empty
 *** du descripteur de type 
 ***)

gab_desc.empty <- VAL_gab(Core_gab.empty_g, gab_desc);;
gab_desc.xxxify <- VAL_funct(gabify);;
gab_desc.make <- extension_gab gab_desc;;
