(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: printer_gab.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2006-04-23 05:22:18 $
 * Version: $Revision: 1.1 $
 * -------------------------------------------------------------------
 *
 * Les fonctions d'impressions sur les scalaires.
 *
 * Il faut implmenter par type Xxx:
 *
 *  -  _uprbval_'outBuffer_xxx : impression non-formatte de xxx
 *  -  _pfbval_'outBufferF_xxx : impression formatte de xxx
 *
 *
 *=========================================================================
 *)

open Buffer;;
open Global;;
open Valeur;;

open Dispatch_sig;;
open Environnement;;
open Printer;;
open Auxfct;;



__register_version "$RCSfile: printer_gab.ml,v $" "$Revision: 1.1 $" "$Date: 2006-04-23 05:22:18 $";;


(** Attention: les noms collisionnent dans Printf et Format.
 ** Aussi on n'ouvre pas ces modules mais on dfinit ici 2 alias
 **)
let uprintf = Printf.bprintf (** impression standard dans un buffer **)
and fprintf = Format.fprintf (** impression formatt dans un buffer **)
;;





(** uprval b x : est une fonction qui imprime une valeur mgs x dans son buffer b.
 ** Cette fonction est la fonction principale d'affichage. L'impression dans une 
 ** string ou bien sur la sortie standard ou bien dans un fichier derivent de 
 ** cette fonction
 **)
let rec _uprbval_'outBuffer_gab (b, (c, t)) =  
  plist
    (uprintf b "%s")
    (fun (n,v) -> uprintf b "%i.%a" n uprbval2 v)
    "("
    ", " 
    ("):"^t.name)
    (Core_gab.elements_occ_g c);
  valundef

(** _upobval1_'outBuffer_string_string_string_funct_XXX (b,pre,inter,post,f,x)
 ** prend un buffer non-format [b], un prefix [pre], un postfix
 ** [post], un separateur [inter], une fonction [f] et une valeur [x] de type XXX. 
 ** Elle ecrit de faon non-formate l'lment de type mgs [x] dans le buffer [b] 
 ** avec pour prfixe [pre], pour infixe [inter], pour suffixe [post].
 ** De plus, elle applique la fonction f sur chaque lment de [x].
 ** Puis, elle renvoie le type valundef
 **)

and plist_aux pps ppval sep x f = match x with
  | [] -> ()
  | a::[] -> ppval (f a); ()
  | a::l -> (ppval (f a)); (pps sep); plist_aux pps ppval sep l f

and _upobval1_'outBuffer_string_string_string_funct_gab (b,pre,inter,post,f,x) =
  Printf.bprintf b "%s " pre;
  plist_aux
    (Printf.bprintf b "%s")
    (Printf.bprintf b "%a" uprbval2)
    inter
    (Core_gab.elements_g (fst x))
    f
  ;
  Printf.bprintf b " %s" post;
  valundef
;;


(** _pfbval_'outBufferF_XXX (b, x) prend un buffer format [b] et une valeur [x] de type XXX. 
 ** Elle crit de faon formate [x] dans le buffer format [b], puis renvoie valundef
 **
 ** On le fait "de manire imprative" car la mise en place des ordre de formattage est "dlicate".
 ** On ne peut mettre l'ordre de formattage que directement dans la string de fprintf, pas dans
 ** un argument qui sera passer  un %s: on ne peut donc pas utiliser le Aux.plist
 **)

let rec _pfbval_'outBufferF_gab (b, (c, t)) =
  fprintf b "(@[<hov>";
  plist
    b
    (Core_gab.elements_occ_g c)
  ;
  fprintf b "@]):%s" t.name;     
  valundef

and plist b = function
  | []    -> ()
  | (n,a)::[] -> fprintf b "%i.%a" n prbval2 a; ()
  | (n,a)::l  ->  fprintf b "%i.%a,@ " n prbval2 a; plist b l

(** _pobval1_'outBufferF_int_string_string_string_funct_XXX (b,t,pre,inter,post,f,x)
 ** prend un buffer format [b], une taille d'indentation [t], une prefix [pre], un postfix
 ** [post], un separateur [inter], une fonction [f] et une valeur [x] de type XXX. 
 ** Elle ecrit de faon formate l'lment de type mgs [x] dans le buffer [b] 
 ** avec pour prfixe [pre], pour infixe [inter], pour suffixe [post], le tout
 ** avec l'indentation [t].
 ** De plus, elle applique la fonction f sur chaque lment de [x].
 ** Puis, elle renvoie le type valundef
 **)

and plist_aux2 pps ppval sep x f = match x with
  | [] -> ()
  | a::[] -> ppval (f a); ()
  | a::l -> (ppval (f a)); (pps sep); plist_aux2 pps ppval sep l f 

and _pobval1_'outBufferF_int_string_string_string_funct_gab (b,t,pre,inter,post,f,x) =
  Format.pp_open_box b t;
  Format.fprintf b "%s" pre;
  plist_aux2 (Format.fprintf b "%s@,") (Format.fprintf b "%a" pfbvalue2  ) inter (Core_gab.elements_g (fst x)) f;
  Format.fprintf b " %s@]@," post;
  valundef
;;

(* --- fin du fichier ----------------------------------- *)

