(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: core_gbf.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2007-06-26 01:48:18 $
 * Version: $Revision: 1.5 $
 * -------------------------------------------------------------------
 *
 * Status:  
 *     La representation d'un GBF.
 *
 *=========================================================================
 *)

Global.__register_version "$RCSfile: core_gbf.ml,v $" "$Revision: 1.5 $" "$Date: 2007-06-26 01:48:18 $";;

open Global;;
open Def_gbf;;
(*open Valeur;;*)
open Auxfct;;
(*open Printf;;*)
(*open Printer;;*)
open Dispatch_sig;;
(*open Terme;;*)


(* **************************************************************************************************** *)
(* **************************************************************************************************** *)

(* => nom du gbf
      liste des generateurs
      matrice des equations
      flag de warning
   <= index du gbf dans la table interne
*)
   

external smith_init             : string -> (string array) -> (int array) array -> bool -> int  = "smith_init"

external smith_compute_index    : int -> (int array) -> unit                                    = "smith_compute_index"
external smith_normalise        : int -> (int array) -> unit                                    = "smith_compute_index"

external normalise_back_and_mod : int -> int -> (int array) -> unit                             = "smith_normalise_back_and_mod"
external smith_back             : int -> (int array) -> unit                                    = "smith_back"


(* Fonction qui, a partir d'un vecteur normalise, retourne les coordonnees dans le groupe initial 
   => index gbf
      origine
*)

external normalise_to_initial   : int -> (int array) -> unit                                    = "smith_normalise_to_initial"


(* Predicat vrai si  : origine + chemin aboutit a destination
  => index gbf
     origine
     destination
     chemin
  <= booleen 
*)

external smith_neighborsq       : int -> (int array) -> (int array) -> (int array) -> bool = "smith_neighborsq"

(* Fonction qui a partir d'une origine et d'un chemin calcule les coordonnees de la destination telle que
   destination = origine + nb * chemin

   (fonctionne par effet de bord en ECRASANT destination)

   => index gbf
      nb
      origine
      chemin
      destination
   <= ()
*)


external smith_compute_neighbor : int -> int -> (int array) -> (int array) -> (int array) -> unit         = "smith_compute_neighbor"

(* Fonction retournant le nombre de generateurs <> 0 du gbf *)

external smith_dim : int -> int = "smith_dim"

(* Fonction permettant de recuperer (par effet de bord) l'index des deux premiers generateurs <>0 d'un GBF.

   => index gbf
   <= tab[0] = index du 1er generateur <> 0
      tab[1] = index du 2er generateur <> 0
*)

external smith_coeffs : int -> (int array) -> unit = "smith_coeffs" 

external smith_diagonale : int -> (int array) -> unit = "smith_diagonale" 

(* **************************************************************************************************** *)
(* **************************************************************************************************** *)


let gbf_tbl         = (Hash.create 7 : (string, gbf_def) Hash.t) (* Un nom de gbf -> ce qui le caracterise *)
and gbf_inv_tbl     = (Hash.create 7 : (string, string) Hash.t)  (* Un generateur -> nom de gbf            *)


let find_gbf_def n =
  try
    Hash.find gbf_tbl n
  with Not_found -> raise (Erreur ("Core_gbf: find_gbf_def: non existing Gbf "^n))

;;


let add_sub plus_moins (x, y) =
  let ((v1, ({gbf_id = id1} as gbf_id)), (v2, {gbf_id = id2})) = (x, y)
  in
    if (id1 <> id2)
    then raise (Erreur "Core_gbf: generic_add_sub: gbfs do not match")
    else (
      let ret = Array.make (Array.length v1) 0 in
	smith_compute_neighbor id1 plus_moins v1 v2 ret;
	(ret, gbf_id)
    )

and mult_entier_posgbf (x, y) =
  let (entier, (v2, ({gbf_id = id2} as gbf_id))) = (x, y) in
  let sz  = Array.length v2 in
  let ret = Array.make sz 0
  and tmp = Array.make sz 0
  in
    smith_compute_neighbor id2 entier tmp v2 ret;
    (ret, gbf_id)

and fold f (tbl:'a tgbf) zero = Hash.fold f tbl zero

and map f (tbl:'a tgbf) = 
  let ret     = Hash.copy tbl in
    Hash.iter (fun k v -> Hash.replace ret k (f k v)) ret;
    ret
;;


let rec gbf_fold ((m, o, e, g, z) as args) expr = 
  let ff = gbf_fold args
  in
    match expr
    with
	GBF_mult (i, s)       -> m i (ff s)
      | GBF_op   (oo, e1, e2) -> o (ff e1) oo  (ff e2)
      | GBF_equal (e1, e2)    -> e (ff e1) (ff e2)
      | GBF_gen   (s, n)      -> g s n
      | GBF_zero              -> z
;;

let rec string_of_gbf_sign = function
    Positif -> ("|", ">")
  | Negatif -> ("<", "|")

and string_of_gbf_op = function
    GBF_plus  -> "+"
  | GBF_moins -> "-"

and string_of_gbf_expr expr = 
  let m i s      = Printf.sprintf "%d%s" i s
  and o e1 oo e2 = Printf.sprintf "%s %s %s" e1 (string_of_gbf_op oo) e2
  and e e1 e2    = Printf.sprintf "%s = %s" e1 e2
  and g s n      = (let (g, d) = string_of_gbf_sign s in Printf.sprintf "%s%s%s" g n d)
  and z          = "0"
  in
    gbf_fold (m, o, e, g, z) expr

and string_of_genid l   = (Array.fold_right (fun x y -> (Printf.sprintf "<%s> " x)^y) l "")

;;


let rec string_of_gbf {gbf_name = n; gbf_type = t; generateurs = g; matrice = m; gbf_id = i} =
  let (op, clo) = string_of_gbf_type t
  and m_string  = string_of_matrix_gbf g m
  in
    Printf.sprintf "%s %s%s%s %s (SMITH : %d)" op (string_of_generateurs g) (if (m_string = "") then "" else " ; ") m_string clo i

and string_of_gbf_type = function
    Abelian    -> "<", ">"
  | NonAbelian -> "<|", "|>"

and string_of_matrix_gbf generateurs m = 
  let fn_fold (c, i) y = (
    let gen = generateurs.(i) in
      match c with
	|  0 -> ""
	|  1 -> Printf.sprintf "%s + " gen
	| -1 -> Printf.sprintf "-%s + " gen
	|  _ -> Printf.sprintf "%d%s + " c gen
  )^y in

  let ret =
    Array.map
      (fun mat-> (
	 match (
	   let r = Array.fold_right fn_fold (Array.mapi (fun idx v -> v,idx) mat) "" in
	     if (r = "")
	     then ""
	     else without_comma 2 r
	 ) with 
	   | "" -> ""
	   | x  -> x^" = 0")
      )
      m
  in
    clean_comma_string_of_string_array ret

and string_of_generateurs x = clean_comma_string_of_string_list (Array.to_list x)
;;


(*****************************************************************************************************
 Etant donne:
  _ une gbf_expr (possiblement en non forme normale)

  => retourne 
               _ le vecteur NORMALISE correspondant au chemin defini par la gbf_expr
                            *********

  Par exemple : 2|x> + 2|x> - |y> - 3|x> + 2<z|
                
                => VAL_posgf [| 1 ; -1 ; -2 |]
*****************************************************************************************************)

let rec gbf_pos_from_gbf_expr expr = 
  try
    let partie_de_generateurs = gens_of_gbf_expr expr  in
    let _, ({ generateurs = generateurs;
	      gbf_id      = gbf_int
	    } as def) = find_gbf_from_generator_names partie_de_generateurs
    in
      ((chemin_normalise_of_gbf_expr gbf_int generateurs expr), def)
  with Not_found -> raise (Erreur ("Core_gbf: gbf_pos_from_gbf_expr: generators related to any GBF definition: "^(string_of_gbf_expr expr)))


(*****************************************************************************************************
  Etant donne :
  _ une gbf_expr (possiblement en non forme normale)

  => retourne
               _ le vecteur des generateurs trouves dans l'expression

  Par exemple : 2|x> + 2|x> - |y> - 3|x> + 2<z|

              => [| "x" ; "z"; "y" |]

*****************************************************************************************************)

and gens_of_gbf_expr expr = Array.of_list (without_double (let m _ s     = s
							   and o e1 _ e2 = e1 @ e2
							   and e e1 e2   = e1 @ e2
							   and g _ n     = [n]
							   and z         = []
							   in
							     gbf_fold (m, o, e, g, z) expr))


(*****************************************************************************************************
 Etant donne:
  _ un vecteur de generateurs

  => retourne le couple (nom du gbf, gbf_def) si les generateurs correspondent bien a une definition de gbf.
     sinon raise Not_found
*****************************************************************************************************)
and find_gbf_from_generator_names l = 
  let nom = Hash.find gbf_inv_tbl l.(0)
  in
    nom, (find_gbf_def nom)


(*****************************************************************************************************
 Etant donne:
  _ un vecteur de generateurs (chaines de caracteres) comme ils sont definis dans un GBF
  _ un deplacement specifie par une expression de generateurs de gbfs

  => retourne le vecteur d'entier NORMALISE correspondant a un deplacement selon le chemin specifie
                                  *********

  Par exemple : [| "x"; "y"; "z" |]
                |x> - |y> + 2<z|
                => [| 1 ; -1 ; -2 |]
*****************************************************************************************************)

and chemin_normalise_of_gbf_expr gbf_int generateurs expr =
  let tbl = simplify_gbf_expr generateurs expr
  and nb  = Array.length generateurs in
  let r   = Array.make nb 0
  in
    for i = 0 to (nb-1)
    do
      (try
	 r.(i) <- Hash.find tbl generateurs.(i)
       with Not_found -> ())
    done;
    smith_normalise gbf_int r;
    r

and string_of_genint l      = (Array.fold_right (fun x y -> (Printf.sprintf "<%d> " x)^y) l "")


(******************************************************************************************************
 Etant donne la liste de hash-tables que l'on recoit :

      [ { ("x", 3) ; ("y", 2) } , { ("x", 25) ; ("y", 0) } , { ("x", 64) ; ("y", 5) }

   correspondant a la normalisation des equations 

   < x, y ; 3x - 2y  = 0,
            25x      = 0,
            64y + 5x = 0  >

   on retourne la matrice des facteurs :

         eq1  eq2  eq3
   x: |   3    25   5   |
   y: |  -2    0    64  |

***************************************************************************************************** *)
and simplify_to_matrix generateurs eq_tbl_array =
  let nb_gen       = Array.length generateurs
  and nb_equations = Array.length  eq_tbl_array in
  let ret          = Array.make_matrix nb_gen nb_equations 0
  in
    for i = 0 to (nb_gen - 1)
    do
      for j = 0 to (nb_equations - 1)
      do
	ret.(i).(j) <- (try
			  Hash.find eq_tbl_array.(j) generateurs.(i)
			with Not_found -> raise (Erreur "Core_gbf: simplify_to_matrix: unknown generators"))
      done
    done;
    ret

(*****************************************************************************************************
 Simplifie une equation entre generateurs, i.e., a l'issue de l'appel a la fonction, 
 chaque generateur apparait au plus une fois avec un coefficient. Etant donne l'equation :

   < x, y >   (3x - 2y  = x)

   on retourne la table de hash-code suivante :

   { ("x", 2) ; ("y", 2) }

  (l'algo est code sous la forme d'un systeme de re-ecriture :-) Le point fixe est
  atteint en une iteration
******************************************************************************************************)
and simplify_gbf_expr generateurs = 
  let rec oriente = (function 
			 GBF_equal (g1, g2) -> GBF_op(GBF_moins, g1, g2)
		       | x                  -> x)
		      
(* "pousse" le * devant un generateur *)
  and push_mult    = (function
			  GBF_op (op, g1, g2)    -> GBF_op(op, (push_mult g1), (push_mult g2))
			| (GBF_mult (i, g1) as e)-> (match g1
						     with
							 GBF_op(op, ng1, ng2) -> GBF_op(op, (push_mult (GBF_mult(i, ng1))), (push_mult (GBF_mult(i, ng2))))
						       | GBF_mult(ni, ng1)    -> push_mult (GBF_mult(ni*i, ng1))
						       | GBF_gen (_, _)       -> e
						       | GBF_zero             -> g1
						       | GBF_equal (_, _)     -> raise (Erreur "push_mult cannot be applied on an equality of gbf expressions"))
			| GBF_equal (_, _)       -> raise (Erreur "push_mult cannot be applied on an equality of gbf expressions")
			| x                      -> x)

(* retire tous les "-" ; ne doit s'appliquer que sur un terme ou le MULT est devant un generateur *)
  and remove_minus = (function
			  GBF_op (GBF_plus, g1, g2)  -> GBF_op(GBF_plus, (remove_minus g1), (remove_minus g2))
			| GBF_op (GBF_moins, g1, g2) -> GBF_op(GBF_plus, (remove_minus g1), (change_sign (remove_minus g2)))
			| GBF_equal (_, _)           -> raise (Erreur "remove_minus cannot be applied on an equality of gbf expressions")
			| x                          -> x)
(* change le signe d'une expression ; ne doit s'appliquer que sur un terme ou le MULT est devant un generateur *)
  and change_sign = (function
			 GBF_op (op, g1, g2)  -> GBF_op(op, (change_sign g1), (change_sign g2))
		       | GBF_mult (i, g1)     -> GBF_mult (-i, g1)
		       | GBF_equal (_, _)     -> raise (Erreur "change_sign cannot be applied on an equality of gbf expressions")
		       | GBF_gen (Positif, x) -> GBF_gen (Negatif, x)
		       | GBF_gen (Negatif, x) -> GBF_gen (Positif, x)
		       | GBF_zero             -> GBF_zero)

  and collect gen_table arg = (let tbl_locale = Hash.copy gen_table
			       in
				 (let rec update_gen valeur nom = (let v = (try
									      Hash.find tbl_locale nom
									    with Not_found -> raise (Erreur ("Generator %s is not defined in any GBF"^nom)))
								   in
								     Hash.replace tbl_locale nom (v+valeur);
								     ())
				  and coll_recur = function
				      GBF_mult (i, (GBF_gen (Positif, x))) -> update_gen (i) x
				    | GBF_mult (i, (GBF_gen (Negatif, x))) -> update_gen (-i) x
				    | GBF_op (GBF_plus, g1, g2)            -> coll_recur g1; coll_recur g2
				    | GBF_gen (Positif, x)                 -> update_gen 1 x
				    | GBF_gen (Negatif, x)                 -> update_gen (-1) x
				    | GBF_zero                             -> ()
				    | x                                    -> raise (Erreur ("This case can't happen in coll_recur "^(string_of_gbf_expr x)))

				  in
				    coll_recur arg;
				    tbl_locale)) in
    
  let table = (let r = Hash.create (Array.length generateurs)
	       in
		 Array.iter (function v -> Hash.add r v 0) generateurs;
		 r) 
  in
    (rond (collect table) (rond remove_minus (rond push_mult oriente)))


and size_col m = (if ((Array.length m) != 0) then (Array.length m.(0)) else 0)
and size_row m = Array.length m

and transpose m =
  let size_col = size_col m
  and size_row = size_row m in
  let ret    = Array.make_matrix size_col size_row m.(0).(0)
  in
    for i = 0 to size_col-1
      do
        for j = 0 to size_row-1
        do
          ret.(i).(j) <- m.(j).(i)
        done
    done;
    ret

and process kind nom (gens, equations) = 
  let generateurs = Array.of_list gens in
  let ret = (let (eq_norm_tbl_list, nombre) = (match kind
					       with
						   Abelian    -> let ret = List.map (simplify_gbf_expr generateurs)  equations in (ret, List.length ret)
						 | NonAbelian -> raise (Erreur "Core_gbf: process: not yet implemented non abelian groups")) in
	     let (smith, mat) = (let mat = (if (nombre = 0)
					    then (Array.make_matrix 0 0 0) 
					    else (transpose (simplify_to_matrix generateurs (Array.of_list eq_norm_tbl_list)))) in
				   ((smith_init nom generateurs mat true), mat))
	     in
	       { gbf_name            = nom;
		 gbf_type            = kind;
		 generateurs         = generateurs;
		 matrice             = mat;
		 gbf_id              = (if (smith = (-1)) then raise (Erreur "Core_gbf: process: error while computing Smith NF") else smith);
		 gbf_neighbors_array = (let nb_gens                = Array.length generateurs 
					in
					  Array.init (2*nb_gens) (function n -> (let ret = (Array.make nb_gens 0)
										 in
										   if (0 = n mod 2)
										   then ret.(n/2)   <- 1
										   else ret.((n-1) / 2) <- (-1);
										   smith_normalise smith ret;
										   ret)))

	       }
	    )
  in
    Hash.add gbf_tbl nom ret;                                               (* On ajoute dans la table  noms -> gbf         *)
    Array.iter (function gen -> Hash.add gbf_inv_tbl gen nom) generateurs;  (* Et dans la table inverse generateur -> noms *)
    Printf.printf "%s\n" (string_of_gbf ret)


(******************************************************************************************************
 Etant donne un genid parse, retourne le posgbf correspondant. Soit |a> on retourne
 le posgbf pour le GBF correspondant. 
 ******************************************************************************************************)

and posgbf_from_genid genid = 
  let generateur = (match genid
		    with
			GBF_gen (_, v) -> v
		      | _              -> raise (Erreur "Core_gbf: posgbf_from_genid: wrong genid type")) 
  in
    (try
       let (nom, ({generateurs =  gens; gbf_id = stub_id} as type_gbf)) = find_gbf_from_generator_names [|generateur|]
       in
	 ((chemin_normalise_of_gbf_expr stub_id gens genid), type_gbf)
     with Not_found -> raise (Erreur ("Core_gbf: posgbf_from_genid: generator "^generateur^" does not correspond to any known generator")))

;;


exception Not_all_equal;;  (* Pour s'assurer que dans un FOLLOWING tous les generateurs referent bien le meme GBF *)

let following extract_array arg direction_array =
  if (0 = Array.length direction_array)
  then raise (Erreur "Core_gbf: following_common_seq_seq: an empty sequence in following does not make sense")
  else
    begin
      let direction, ( {gbf_name    = nom;
			gbf_type    = gb_type;
			generateurs = gens_vect;
			matrice     = matrice;
			gbf_id      = smith_int} as ret) = (
	let ret = (
	  Array.fold_right 
		     (fun (coord, gbd) (coord_liste, gbf_def)  -> (
			if ((gbf_def = None) or (gbd = (of_some gbf_def)))
			then  (coord::coord_liste), (Some gbd)
			else raise Not_all_equal
		      ))
		     direction_array
		     ([], None)
	) in
	  (fst ret), (of_some (snd ret)) 
      ) in
      let hash_tbl    = Hash.create 17 in
      let nb_gens_gbf = Array.length gens_vect in
      let vect_of_gen = chemin_normalise_of_gbf_expr smith_int gens_vect in
      let calc_pos    = smith_compute_neighbor smith_int in

      let update_gbf_tbl ve valeur = (
	let copie = (Array.copy ve) in
	  Hash.replace hash_tbl copie valeur
      ) in

      let rec compute_gbf chemin_courant =
	(function
	   | (_, []) -> raise (Erreur "Core_gbf: following_common_seq_seq: cannot produce a GBF if no generator is specified")
	       
	   | (sth, [generateur]) -> (
	       let vecteur = extract_array sth in
	       let (val_sv, len_sv)   = vecteur, (Array.length vecteur)
	       and tmp = Array.make nb_gens_gbf 0 in
		 for i = 0 to (len_sv - 1) do
		   let courant = val_sv.(i) in
		     (*if (not (is_undef(courant)))
		     then*)
		     begin
		       calc_pos i chemin_courant generateur tmp;
		       update_gbf_tbl tmp courant
		     end
		 done
	     )
	       
	   | (sth, (h::t)) -> (
	       let vecteur = extract_array sth in
	       let (val_sv, len_sv)  = vecteur, (Array.length vecteur)
	       and tmp = Array.make nb_gens_gbf 0 in
		 for i = 0 to (len_sv - 1) do
		   compute_gbf (
		     calc_pos i chemin_courant h tmp;
		     tmp)
		     (val_sv.(i), t)
		 done
	     )
	) in
	compute_gbf (Array.make nb_gens_gbf 0) (arg, direction);
	hash_tbl, nom
    end

(*
and generic_gbf_print (b, x) = 
  let (tbl, deco_type) = x in
  let { gbf_name    = nom;
        gbf_type    = gb_type;
	generateurs = gens;
	gbf_id      = gbf_i
      } = (try
	     find_gbf_def_from_deco deco_type
	   with Not_found -> raise (Erreur (Printf.sprintf "Internal error, gbf %s not found in prbval_gbf" deco_type.name)))
  in
    if ((gb_type = Abelian) && ((2 = smith_dim gbf_i)) && (try
								     Hash.iter 
								       (fun _ v  -> (if ((is_int v) or (is_string v))
										     then ()
										     else raise Not_found))
								       tbl;
								     true
								   with Not_found -> false))
    then 
      begin
	let tmp_array_size         = Array.length gens in
	let tmp_array              = Array.make tmp_array_size 0 in
	let (idx_gen_0, idx_gen_1) = (let tmp = Array.make 2 0
				      in 
					smith_coeffs gbf_i tmp;
					tmp.(0), tmp.(1)) in
	let position_i_j i j      = (Array.fill tmp_array 0 tmp_array_size 0;
				     tmp_array.(idx_gen_0) <- i;
				     tmp_array.(idx_gen_1) <- j;
				     tmp_array) in
	let (max_x, max_y, min_x, min_y) = (ref Pervasives.min_int, ref Pervasives.min_int,
					    ref Pervasives.max_int, ref Pervasives.max_int)
	in
(*					   Printf.printf "idx_gen_0 : %d idx_gen_1 : %d \n" idx_gen_0 idx_gen_1; flush Pervasives.stdout; *)
	  Hash.iter (fun k _ -> (let x = k.(idx_gen_0)
					 and y = k.(idx_gen_1)
					 in
					   max_x := Pervasives.max !max_x x;
					   min_x := Pervasives.min !min_x x;
					   max_y := Pervasives.max !max_y y;
					   min_y := Pervasives.min !min_y y)) tbl;
(* Printf.printf "X(%d) Y(%d)\n\n" !max_x !max_y; *)
	  for i = !min_x to !max_x
	  do
	    for j = !min_y to !max_y
	    do
	      try
		(match (Hash.find tbl (position_i_j i j))
		 with
		     (VAL_int valeur) -> Printf.bprintf b "%3d  " valeur
		   | (VAL_string s)   -> Printf.bprintf b "%3s  " s
		   | _                -> raise (Erreur "this case cannot happen.00"))
	      with Not_found       -> Printf.bprintf b "---  "
	    done;
	    Printf.bprintf b "\n"
	  done;
	  mgsundef
      end
    else
      ((plist (bprintf b "%s")  (bprintf b "%a" Printer.unformatted_prbvalue2) "("   ", "    ("):"^(snd x).name)
	 (Hash.fold (fun k v suite -> v::suite) (fst x) []));
       mgsundef
      )
*)
;;

let some_neighbors gbf_def gbf_tbl pos array_voisin =
  (let { gbf_type            = gb_type;
	 generateurs         = gens;
	 matrice             = mat;
	 gbf_id              = gbf_idx;
	 gbf_neighbors_array = gna
       } = gbf_def in
   let array_voisin = (Array.map (function (c, {gbf_id = idx}) -> assert(idx=gbf_idx) ; c) array_voisin) in
   let voisins               = Array.make (Array.length gna) [||]
   and nb_gens               = Array.length gens in
   let origine               = Array.copy pos                                             (* Le point de depart    t              *)
   and dest                  = Array.make nb_gens 0 in                                    (* Le point origine+voisin              *)
   let compute_neighbor vois = smith_compute_neighbor gbf_idx 1 origine vois dest (* Le calcul de l'adresse destination   *) in
     
   let voisins_effectifs   = ref 0
			       
   and quantite = Array.length array_voisin
   in
     for i = 0 to (quantite - 1)
     do
       compute_neighbor
	 array_voisin.(i);
       if (Hash.mem gbf_tbl dest)                  (* Un element se trouve-t-il a cette position ? *)
       then (voisins.(!voisins_effectifs) <-  Array.copy dest;
	     incr voisins_effectifs)
     done;
       Array.sub voisins 0 !voisins_effectifs 
  )
;;

(*
let neighbor (tbl, d) (t', d') v =
  let gbf_def = find_gbf_def_from_deco d
  in
    if (gbf_def.gbf_name <> d'.gbf_name)
    then raise (Erreur "neighbor : in gbf, the path/neighbor is not from the Gbf of the argument")
    else (let tab = some_neighbors gbf_def tbl t' v in
	  let vals = Array.map (Hash.find tbl) tab
	  in
	    vals
	 )
*)
