(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: printer_gbf.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG, AS
 * Modified: $Date: 2007-06-26 01:48:18 $
 * Version: $Revision: 1.7 $
 * -------------------------------------------------------------------
 *
 * Status: Cf. AChain/def_achain.ml pour les infos sur les types achain et acell
 *          
 *
 *=========================================================================
 *)

open Global;;
open Auxfct;;
open Dispatch_sig;;
open Printf ;;
open Def_gbf ;;
open Core_gbf ;;
open Gbf ;;
open Printer;;
open Valeur;;



let rec _uprbval_'outBuffer_posgbf (b, x) =
  let r = Global.clean_comma_string_of_general_array string_of_int (fst x) (* Desole, mais je ne *)
  in                                                                       (* m'en sors pas avec *)
    bprintf b "%s:|%s>" (snd x).gbf_name r;                                (* ta fonction parray *)
    valundef

and _upobval1_'outBuffer_string_string_string_funct_posgbf (b,pre,inter,post,f,x) =
  let r = Global.clean_comma_string_of_general_array string_of_int (fst x)
  in
    bprintf b "%s:|%s>" (snd x).gbf_name r;
    valundef

let rec _pfbval_'outBufferF_posgbf (b, x) =
  let r = Global.clean_comma_string_of_general_array string_of_int (fst x)
  in
    Format.fprintf b "%s:|%s>" (snd x).gbf_name r;
    valundef

and _pobval1_'outBufferF_int_string_string_string_funct_posgbf (b,t,pre,inter,post,f,x) =
  let r = Global.clean_comma_string_of_general_array string_of_int (fst x)
  in
    Format.fprintf b "%s:|%s>" (snd x).gbf_name r;
    valundef

;;




(** --- Affichage formatt ------------------------------------------------------- **)


let  _pfbval_'outBufferF_gbf  (b, x) = 
  let (tbl, deco_type) = x in
  let { gbf_name    = nom;
        gbf_type    = gb_type;
	generateurs = gens;
	gbf_id      = gbf_i
      } = 
    try
      find_gbf_def_from_deco deco_type
    with Not_found -> 
      raise (Erreur (Printf.sprintf "Internal error, gbf %s not found in prbval_gbf" deco_type.name))
 in
  if ((gb_type = Abelian) 
      && ((2 = smith_dim gbf_i)) )
  then 
    begin
      let tmp_array_size         = Array.length gens in
      let tmp_array              = Array.make tmp_array_size 0 in
      let (idx_gen_0, idx_gen_1) = (let tmp = Array.make 2 0
				    in 
				      smith_coeffs gbf_i tmp;
				      tmp.(0), tmp.(1)) in
      let position_i_j i j      = (Array.fill tmp_array 0 tmp_array_size 0;
				     tmp_array.(idx_gen_0) <- i;
				     tmp_array.(idx_gen_1) <- j;
				     tmp_array) in
      let (max_x, max_y, min_x, min_y) = (ref Pervasives.min_int, ref Pervasives.min_int,
					  ref Pervasives.max_int, ref Pervasives.max_int)
      in
	Hash.iter (fun k _ -> (let x = k.(idx_gen_0)
			       and y = k.(idx_gen_1)
			       in
				 max_x := Pervasives.max !max_x x;
				 min_x := Pervasives.min !min_x x;
				 max_y := Pervasives.max !max_y y;
				 min_y := Pervasives.min !min_y y)) tbl;
	for i = !min_x to !max_x do
	  for j = !min_y to !max_y do
	    (try
	      match (Hash.find tbl (position_i_j i j)) with
		| (VAL_undef _)    -> Format.pp_print_string b "  ---"
 		| _ as v  -> Format.pp_print_string b (Printf.sprintf "%5s" (to_string (uprsval v )))
	    with Not_found       -> Format.pp_print_string b "  ---") ;
	    Format.fprintf b "@   "
	  done;
	  Format.fprintf b "@\n"
	done;
	mgsundef
    end
  else
    begin
      Format.fprintf b "@[<b 1>%s" "(";
      plist_aux 
	(Format.fprintf b "%s@,") 
	(Format.fprintf b "%a" prbvalue2) 
	", " 
	(Hash.fold (fun k v suite -> v::suite) (fst x) []);
      Format.fprintf b "%s@]@," ("):"^(snd x).name);
      mgsundef
    end

      
 and _pobval1_'outBufferF_int_string_string_string_funct_gbf (b,t,pre,inter,post,f,x) =      
   Format.pp_open_box b t;
   Format.fprintf b "%s" pre;
   plist_aux2
     (Format.fprintf b "%s@,") 
     (Format.fprintf b "%a" prbvalue2  ) 
     inter 
     (Hash.fold (fun k v suite -> v::suite) (fst x) [])
     f;
   Format.fprintf b " %s@]@," post;
   mgsundef

let  _uprbval_'outBuffer_gbf  (b, x) = 
  let (tbl, deco_type) = x in
  let { gbf_name    = nom;
        gbf_type    = gb_type;
	generateurs = gens;
	gbf_id      = gbf_i
      } = 
    try
      find_gbf_def_from_deco deco_type
    with Not_found -> 
      raise (Erreur (Printf.sprintf "Internal error, gbf %s not found in prbval_gbf" deco_type.name))
 in
  if ((gb_type = Abelian) 
      && ((2 = smith_dim gbf_i)) )
  then 
    begin
      let tmp_array_size         = Array.length gens in
      let tmp_array              = Array.make tmp_array_size 0 in
      let (idx_gen_0, idx_gen_1) = (let tmp = Array.make 2 0
				    in 
				      smith_coeffs gbf_i tmp;
				      tmp.(0), tmp.(1)) in
      let position_i_j i j      = (Array.fill tmp_array 0 tmp_array_size 0;
				     tmp_array.(idx_gen_0) <- i;
				     tmp_array.(idx_gen_1) <- j;
				     tmp_array) in
      let (max_x, max_y, min_x, min_y) = (ref Pervasives.min_int, ref Pervasives.min_int,
					  ref Pervasives.max_int, ref Pervasives.max_int)
      in
	Hash.iter (fun k _ -> (let x = k.(idx_gen_0)
			       and y = k.(idx_gen_1)
			       in
				 max_x := Pervasives.max !max_x x;
				 min_x := Pervasives.min !min_x x;
				 max_y := Pervasives.max !max_y y;
				 min_y := Pervasives.min !min_y y)) tbl;
	for i = !min_x to !max_x do
	  for j = !min_y to !max_y do
	    (try
	      match (Hash.find tbl (position_i_j i j)) with
		| (VAL_undef _)    -> bprintf b "  ---"
 		| _ as v           -> bprintf b "%s" (Printf.sprintf "%5s" (to_string (uprsval v )))
	    with Not_found       -> bprintf b "  ---") ;
	    bprintf b "   "
	  done;
	  bprintf b "\n" 
	done;
	mgsundef
    end
  else
    begin
      bprintf b "(";
      plist_aux 
	(bprintf b "%s,") 
	(bprintf b "%a" uprbvalue2) 
	", " 
	(Hash.fold (fun k v suite -> v::suite) (fst x) []);
      bprintf b "%s" ("):"^(snd x).name);
      mgsundef
    end

      
 and _upobval1_'outBuffer_string_string_string_funct_gbf (b,pre,inter,post,f,x) =      
   bprintf b "%s" pre;
   plist_aux2
     (bprintf b "%s") 
     (bprintf b "%a" uprbvalue2  ) 
     inter 
     (Hash.fold (fun k v suite -> v::suite) (fst x) [])
     f;
   bprintf b " %s" post;
   mgsundef

(*      
and _pzbval1_'outBufferF_int_string_string_string_funct_seq_gbf (b,t,pre,inter,post,f,l,x) =      
  Format.pp_open_box b t;
  Format.fprintf b "%s" pre;
  plist_aux2
    (Format.fprintf b "%s@,")
    (Format.fprintf b "%a" (fun b x->(fprintcollrecpar b x (VAL_seq l);()))  )
    inter
    ((Hash.fold (fun k v suite -> v::suite) (fst x) []))
    f;
  Format.fprintf b " %s@]@," post;
  mgsundef
*)
;;
