/*
 *
 *                              MGS
 *                              LaMI
 *                       Universite d'Evry Val d'Essonne
 *                   Olivier MICHEL, Jean-Louis GIAVITTO
 * File: $RCSfile: stub_smith.cc,v $ 
 * Path: %P%
 * Description:
 * Created:
 * Author: Olivier Michel & Jean-Louis Giavitto
 * Modified: $Date: 2004-12-14 14:58:24 $ at 
 * Version: $Revision: 1.1 $
 * Status:
 *
 */

#include <stdio.h>
#include <pair.h>
#include <list>
#include <vector>
#include <set>
#include <multiset.h>
#include <map>
#include <hash_map.h>


#undef __GNUG__

#define EXPORT_MGS_C_DEVELOPMENT

using namespace std;

#include "smith.C"


 

/********************************************************************************/
/********************************************************************************/
/********************************************************************************/

#define MAX_SMITH 1000

MGS::StringVect        *generateurs;
MGS::Equations         *equations;
MGS::CoefVect          *vecteur;
static int             smith_ptr = -1;
static MGS::SmithData* tab_smith[MAX_SMITH];



extern "C" 
{
  /* ======================= */
  /* INITIALISATION D'UN GBF */
  /* ======================= */

  void gbf_init_vect(void)
  {
    generateurs = new MGS::StringVect();
  }

  void gbf_add_generator(char *nom)
  {
    generateurs->push_back(nom);
  }


  /* Attention, n'appeler cette fonction qu'apres avoir definit
     tous les generateurs.
  */
  void gbf_init_equations(void)
  {
    equations = new MGS::Equations();
    vecteur   = new MGS::CoefVect(generateurs->size());
  }

  void gbf_add_coeff(int valeur)
  {
    static unsigned int current = 0;

    (*vecteur)[current++] = valeur;
    if (current == (generateurs->size()))
	{
	  current = 0;
	  equations->push_back(*vecteur);
	}
  }

  int gbf_make_smith(char *nom, bool warning)
  {
    if (smith_ptr < MAX_SMITH)
      {
	tab_smith[++smith_ptr] = 
	  new MGS::SmithData(nom, *generateurs, *equations, warning);
	
	if (tab_smith[smith_ptr]->L != NULL)
	  cerr << "## Matrice left == " << *(tab_smith[smith_ptr]->L) << endl;

	return(smith_ptr);
      }
    else return(-1);
  }

  /* ======================= */
  /* NORMALISATION           */
  /* ======================= */
  
  void gbf_normalise(int tab[], int len, int smith)
  {
    MGS::IntVect  vecteur(tab, tab+len);

    tab_smith[smith]->normalise(vecteur);

    for(int i=0 ; i < len; i++)
      tab[i] = vecteur[i];
  }

  /* ======================= */
  /* NORMALISATION INVERSE   */
  /* ======================= */
  
  void gbf_smith_normalise_to_initial(int tab[], int len, int smith)
  {
    MGS::IntVect  vecteur(tab, tab+len);

    tab_smith[smith]->back(vecteur);

    for(int i=0 ; i < len; i++)
      tab[i] = vecteur[i];
  }

  /* ======================= */
  /* QUOTIENT                */
  /* ======================= */
  
  void gbf_smith_normalise_back_and_mod(int ginit, int gquotient, int tab[], int len)
  {
    MGS::IntVect  vecteur(tab, tab+len);

    tab_smith[gquotient]->normalise(vecteur);
    tab_smith[gquotient]->back(vecteur);
    tab_smith[ginit]->modulo(vecteur);

    for(int i=0 ; i < len; i++)
      tab[i] = vecteur[i];
  }

 
  void gbf_smith_back(int ginit, int tab[], int len)
  {
    MGS::IntVect  vecteur(tab, tab+len);

    tab_smith[ginit]->modulo(vecteur);
    tab_smith[ginit]->back(vecteur);

    for(int i=0 ; i < len; i++)
       tab[i] = vecteur[i];
  }


  /* ======================= */
  /* VOISINS                 */
  /* ======================= */

  /* *************************************************************** */
  /* Origine, destination et chemin sont normalises.                 */
  /* *************************************************************** */

  int gbf_neighborsQ(int groupe, int origine[], int len, int destination[], int chemin[])
  {
    MGS::IntVect origine_v(origine, origine+len);
    MGS::IntVect destination_v(destination, destination+len);
    MGS::IntVect chemin_v(chemin, chemin+len);

    /*    tab_smith[groupe]->back(origine_v);
	  tab_smith[groupe]->modulo(origine_v);
    */

    tab_smith[groupe]->inc(origine_v, 1, chemin_v);
    for(int i = 0; i < len; i++)
      if (origine_v[i] != destination_v[i])
	return(0);
    return(1);
  }

  /* *************************************************************** */
  /* Origine et chemin sont normalises.                              */
  /* *************************************************************** */

  int gbf_compute_neighbors(int groupe, int origine[], int len, int nb, int chemin[], int destination[])
  {
    MGS::IntVect origine_v(origine, origine+len);
    MGS::IntVect chemin_v(chemin, chemin+len);

    /*    tab_smith[groupe]->back(origine_v);
	  tab_smith[groupe]->modulo(origine_v);
    */

    tab_smith[groupe]->inc(origine_v, nb, chemin_v);
    for(int i=0; i < len ; i++)
      destination[i] = origine_v[i];
    return(1);
  }

  /* *************************************************************** */
  /* Savoir si le nombre de veritables dimension est <= 2            */
  /* *************************************************************** */

  int gbf_smith_dim(int groupe)
  {
    /*    printf("gbf_smith_dim : %d\n",tab_smith[groupe]->dim()); */

    return(tab_smith[groupe]->dim());;
  }

  /* *************************************************************** */
  /* Recuperer les index des generateurs <>0 d'un gbf                */
  /* *************************************************************** */

  void gbf_smith_coeffs(int gbf, int tableau[])
  {

    /*    printf("Index du premier generateur %d et du second %d \n", tab_smith[gbf]->nth_coeff(1), tab_smith[gbf]->nth_coeff(2)); */
     tableau[0] = tab_smith[gbf]->nth_coeff(1);
     tableau[1] = tab_smith[gbf]->nth_coeff(2);
  }  


   int gbf_smith_diagonale(int gbf, int tableau[])
  {
     int dim =  tab_smith[gbf]->dim();
     for (int i = 0; i < dim; i++)
	{
	   int d = tab_smith[gbf]->image[i];
	   if (0 == d) return i;
	   else tableau[i] = d;
	}
     return (dim-1);
  }
}
