(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: core_del.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2005-03-07 17:24:52 $
 * Version: $Revision: 1.1 $
 * -------------------------------------------------------------------
 *
 * Status: 
 *=========================================================================
 *)

open Global;;
open Def_del;;
open Valeur;;
open Printer;;

__register_version "$RCSfile: core_del.ml,v $" "$Revision: 1.1 $" "$Date: 2005-03-07 17:24:52 $";;


(* ******************************************************************************** *)
(* ******************************************************************************** *)

(* => pour affichage de l'usage
   <= 
*)
external usage : unit -> unit = "delaunay_stub_usage"

(* => nombre de sommets
      dimension des sommets
      tableau des sommets ; si sommets (1, 2, 3) et (3, 4, 5) alors on attends a avoir [| [|1; 2; 3|]; [|3; 4; 5|] |]
      booleen vrai si l'on souhaite avoir des informations sur le graphe de Voronoi associe
   <= tableau de la liste des voisins tab[s] donne tous les voisins du sommet s
      *
      tableau des centres des cellules de Delaunay (i.e. les sommets du Graphe de Voronoi) avec (x, y, [liste des sommets du graphe de Delaunay adjacents au centes])
*)

(*                             nb     dim     sommets                           opts     Voronoi ?               *)
 external delaunay_neighbors : int -> int -> (float array) array -> string -> bool       ->

(* sommets voisins          vor_x   vor_y    voisins                    open?                                    *)

 (int list) array * ((float * float * (int list)) array) * (bool array)  = "delaunay_neighbors"






let del_tbl         = (Hash.create 7 : (string, (mgs_value del_def)) Hash.t) (* Un nom de delaunay -> dimension, fonction *)
;;

let process nom dim fonc voronoi = (if (Hash.mem del_tbl nom)
				    then raise (Erreur "a delaunay definition of that name already exists")
				    else Hash.add del_tbl nom (dim, (to_funct fonc), voronoi))
;;


let deco    (d, _, _) = d
and pos     (_, p, _) = p
and vois    (_, _, v) = v
and decopos (d, p, _) = d, p
and posdeco (d, p, _) = p, d
;;


let get_defs_from_delaunay nom = 
  (try
     Hash.find del_tbl nom
   with Not_found -> raise (Erreur (nom ^ " does not correspond to any delaunay definition")))
;;


(*****************************************************************************************************
  Etant donne :
  _ une mgs_value (VAL_int ou VAL_float)

  => retourne
               _ la valeur float associee (avec conversion int -> float si necessaire)

  Par exemple : (VAL_float 2.3) => 2.3
                (VAL_int 2)     => 2.0

  ATTENTION: existe deja. Cf. floatof dans kernel.ml. Et je crois que ca traite aussi les string. (JLG)
*****************************************************************************************************)

let float_of_mgs_value = function
  | VAL_int i   -> float_of_int i
  | VAL_float f -> f
  | _           -> raise (Erreur "illegal use of float_of_mgs_value")
;;

(*****************************************************************************************************
  Etant donne :
  _ un int        (pour un delaunay dans $R^n$, la valeur $n$)
  _ une mgs_value (VAL_seq, VAL_int ou VAL_float)

  => retourne
               _ le (float array) de dimension $n$ et contenant les elements du 2ieme argument

  Par exemple : 3 (VAL_seq (([|(VAL_int 2) ; (VAL_float 3.0) ; (VAL_int 1) |]), ...))
               => [| 2.0 ; 3.0 ; 1.0 |]
*****************************************************************************************************)

let float_array_of_val_seq dim = function
    VAL_seq (a, _)       -> (if (dim <> Array.length a)
			     then raise (Erreur (Printf.sprintf "all positions have to be of same dimension (%d)" dim))
			     else Array.init dim (fun idx -> float_of_mgs_value a.(idx)))
  | ((VAL_float _) as x) 
  | ((VAL_int _)   as x) -> (if (dim =1) then [| float_of_mgs_value x |]
			     else raise (Erreur (Printf.sprintf "all positions have to be of same dimension (%d)" dim)))
  | x                    -> raise (Erreur ("value not corresponding to a position in a delaunay "^(to_string (prsvalue x))))
;;



(*****************************************************************************************************
  Etant donne :
  _ f   : mgs_value "fonctionelle" (VAL_funct _) permettant de calculer la coordonnee dans $R^n$ de chaque sommet
  _ dim : int        (pour un delaunay dans $R^n$, la valeur $n$)
  _ v   : une mgs_value (VAL_int, VAL_float ou collection)

  => retourne
               _ le ((float array) array) de taille :
                                                      _ 1 si c'est une constante
                                                      _ size(v) si c'est une collection
                 dont la dimension de chaque element est dim

  Par exemple : 2 (VAL_set ((VAL_seq [|(VAL_int 2) ; (VAL_float 3.0) |]);
                            (VAL_seq [|(VAL_int 5) ; (VAL_float 4.0) |]);
                            (VAL_seq [|(VAL_int 8) ; (VAL_float 1.0) |]);
                            (VAL_seq [|(VAL_int 1) ; (VAL_float 2.0) |])))
      
               => [| [| 2.0; 3.0 |]; [| 5.0; 4.0 |]; [| 8.0; 1.0 |]; [| 1.0; 2.0 |] |]

  Remarque : les elements de la collection doivent etre des constantes ou des VAL_seq
*****************************************************************************************************)

let array_of_mgs_value fonc dim v = 
  let fonction element suite = ((float_array_of_val_seq dim (fonc element)), element)::suite  
  in
    match v
    with 
      | ((VAL_int _) as x)
      | ((VAL_bool _) as x)
      | ((VAL_string _) as x)
      | ((VAL_rec _) as x) 
      | ((VAL_posgbf _) as x)
(*       | ((VAL_splx _) as x) *)
      | ((VAL_gmap _) as x)
      | ((VAL_ncell _) as x)
      | ((VAL_float _) as x) -> [| (float_array_of_val_seq dim x), v |]
      |	VAL_seq (a, _)       -> Array.map (fun e      -> (float_array_of_val_seq dim (fonc e)), e) a
      | VAL_grf ((_, a), _)  -> Array.map (fun (d, _) -> (float_array_of_val_seq dim (fonc d)), d) a
      | VAL_set (a, _)       -> Array.of_list (Core_set.fold    fonction                     a  [])
      | VAL_bag (a, _)       -> Array.of_list (Core_bag.fold_b  fonction                     a  [])
      | VAL_gbf (a, _)       -> Array.of_list (Hash.fold (fun k v s -> fonction v s) a  [])
      | VAL_del ((a, _), _)  -> Array.map posdeco a
      | VAL_achain (a, _)    -> Array.of_list (Hash.fold (fun k (_,v) s -> fonction v s) a  [])

      | VAL_undef  _
      | VAL_funct _        
      | _ -> raise (Erreur "don't know that to do with such a value in a delaunay join")
;;


(*****************************************************************************************************
  Etant donne :
  _ un ('a array)

  => retourne
               _ true si le tableau contient au moins 3 elements

  Par exemple : [| 2.0 ; 3.0 ; 1.0 |]
               => true
*****************************************************************************************************)

(* Pour effectuer une triangulation de Delaunay, il faut au moins trois sommets *)
let check_size tab = (3 <= Array.length tab) ;;



let compute_delaunay_neighbors dim tab vor = 
(*  Printf.printf "***********A debugger************\n";
  flush Pervasives.stdout;
  [||], [| (2.0, 2.0, []) |], [| |]
*)
  delaunay_neighbors (Array.length tab) dim tab !Global.delaunay_opt vor



(*
let cpt = ref 0 ;;


let compute_delaunay_neighbors dim tab vor = 
  incr cpt;
  Printf.fprintf stderr "Appel a compute_delaunay %d \n" !cpt;
  flush stderr;
  delaunay_neighbors (Array.length tab) dim tab vor
*)

(*****************************************************************************************************
  Etant donne :
  _ dim : int
  _ v   : (mgs_value array)
  _ p   : (mgs_value array)
  _ c   : ('a color)

  => retourne
               _ VAL_del(__, c) tel que :
                                          _ les sommets $s_i$ sont ceux de v
                                          _ les $s_i$ sont positionnes en  p.(i)
                                          _ la liste des voisins de chaque $s_i$ etant calcule par le stub

  Par exemple : 2
                [| { deco = "S0", x = 4,  y = 6 } ; { deco = "S1", x = 7, y = 3 } ;
                   { deco = "S2", x = 7,  y = 6 } ; { deco = "S3", x = 7, y = 9 } ;
                   { deco = "S4", x = 10, y = 6 } |]
                [| [|4., 6.|] ; [|7., 3.|]; [|7., 6.|]; [|7., 9.|]; [|10., 6.|] |]
                .... (la couleur)

               => 
               [       0 := { deco = "S0", x = 4, y = 6 }@ [4., 6.] :> 1, 3, 2
                       1 := { deco = "S1", x = 7, y = 3 }@ [7., 3.] :> 4, 0, 2
                       2 := { deco = "S2", x = 7, y = 6 }@ [7., 6.] :> 1, 4, 0, 3
                       3 := { deco = "S3", x = 7, y = 9 }@ [7., 9.] :> 4, 0, 2
                       4 := { deco = "S4", x = 10, y = 6 }@ [10., 6.] :> 1, 3, 2
               ] : D2
*****************************************************************************************************)

let delaunay_from_all_sommets_positions dim all_sommets all_positions color voronoi =
  if (check_size all_sommets)
  then (let voisins, vor, open_d   = compute_delaunay_neighbors dim all_positions voronoi
	in
(*
	  if voronoi
	  then 
	    begin
	      Printf.printf "Il y a %d sommets dans le graphe de Voronoi :\n " (Array.length vor);
	      for i = 0 to (Array.length vor) - 1
	      do
		let x, y, l = vor.(i)
		in
		  Printf.printf "Centre (%f, %f) du graphe de voronoi a pour voisins %s du graphe de delaunay\n" x y (clean_comma_string_of_general_list string_of_int l)
	      done;
	    end;
*)
	  VAL_del ( ((Array.init (Array.length all_sommets) (fun i -> ((all_sommets.(i)), (all_positions.(i)), (voisins.(i))))), (vor, open_d)), color))
  else raise (Erreur "a delaunay graph requires at least 3 values to be constructed")


(* ************************************************************************************ *)
(* -- Teste si p est inferieur a q pour l'ordre polaire relatif a la demi droite Op1 -- *)
(* -- Ce code peut etre optimise                                                     -- *)
(* Adapte du source en Pascal de M.-P. Beal pour le calcul de l'enveloppe convexe       *)
(* de points dans R^2 selon l'algo dit de la "Marche de Graham" :                       *)
(* http://www-igm.univ-mlv.fr/~beal/Enseignement/Polytechnique/Td6/graham.html          *)
(* ------------------------------------------------------------------------------------ *)
(* On utilise cela pour trier les sommets de la cellule de Voronoi dans l'ordre horaire *)
(* ************************************************************************************ *)

type point          = {h : float; v : float}
and  voronoi_id     = int
and  voronoi_sommet = { id : voronoi_id; hv : point }
and  facette        = { c_hv : point; mutable adj : (voronoi_sommet list) ; mutable area : float}
;;


(* Calcule le determinant des vecteurs  0p et Oq *)
let rec determinant (o, p, q) = 
  let x = p.h -. o.h
  and y = p.v -. o.v
  and z = q.h -. o.h
  and t = q.v -. o.v
  in
    (x *. t) -. (y *. z)


(* Calcule le produit scalaire des vecteurs Op et Oq *)
and scalaire (o, p, q) =
  let x = p.h -. o.h
  and y = p.v -. o.v
  and z = q.h -. o.h
  and t = q.v -. o.v
  in
    (x *. z) +. (y *. t)


(* Si on met p q, on obtient les elements classes dans le sens anti-horaire *)
(* Avec      q p,                                              horaire.     *)
and inferieur_voronoi  o p1 q p = inferieur o p1 p.hv q.hv

and inferieur o p1 p q = 
  (* si O,p,q alignes du meme cote par rapport a 0, p avant q si Op superieur a Oq *)
     
  (* le cas p et q alignes : *)
  if (determinant(o, p, q) = 0.0) 
  then
    begin
      if (scalaire(o, p, q) > 0.0) 
      then
	begin
	  let d = ((p.h -. o.h) *. (p.h -. o.h) +. (p.v -. o.v) *. (p.v -. o.v)) in
	  let d = d -. ((q.h -. o.h) *. (q.h -. o.h) +. (q.v -. o.v) *. (q.v -. o.v))
	  in
	    (d > 0.0)
	end
      else
	((determinant(o, p1, p) > 0.0) or ((determinant(o, p1, p) = 0.0) && (scalaire(o, p1, p) > 0.0)))
    end
  else
    (* le cas p et q non alignes : *)
    begin
      let ret = ref false in

      let b = ((determinant(o, p1, p) > 0.0) && (determinant(o, p1, q) > 0.0)) 
      in
	if (b or ((determinant(o, p1, p) < 0.0) && (determinant(o, p1, q) < 0.0)))
	then 
	  begin
	    if (determinant(o, p, q) > 0.0) 
	    then ret := true
	    else 
	      if (determinant(o, p, q) < 0.0) 
	      then ret := false
	      else
   		(* si (determinant(o, p, q) = 0) , p avant q si op superieur a oq *)
		begin
		  let d = ((p.h -. o.h) *. (p.h -. o.h) +. (p.v -. o.v) *. (p.v -. o.v)) in
		  let d = d -. ((q.h -. o.h) *. (q.h -. o.h) +. (q.v -. o.v) *. (q.v -. o.v))
		  in
		    ret := (d > 0.0)
		end
	  end;
	  if ((determinant(o, p1, p) > 0.0) && (determinant(o, p1, q) < 0.0)) then ret := true;
	  if ((determinant(o, p1, p) < 0.0) && (determinant(o, p1, q) > 0.0)) then ret := false;

	  (* p sur la demi droite op1 du meme cote *)
	  if ((determinant(o, p1, p) = 0.0) && (scalaire(o, p1, p) > 0.0)) then ret := true;

	  (* p sur la  droite op1 , autre demi droite *)
	  if ((determinant(o, p1, p) = 0.0) && (scalaire(o, p1, p) < 0.0)) then ret := (determinant(o, p1, q) < 0.0);

	  (* q sur la demi droite op1 du meme cote *)
	  if ((determinant(o, p1, q) = 0.0) && (scalaire(o, p1, q) > 0.0)) then ret := false;

	  (* q sur la  droite op1 , autre demi droite *)
	  if ((determinant(o, p1, q) = 0.0) && (scalaire(o, p1, q) < 0.0)) then ret := (determinant(o, p1, p) > 0.0);

	  !ret
    end;
      

(* Soit un triangle defini par ses trois points A, B et C                         *)
(*                                                                                *)
(*         B             La surface est : sqrt(s(s-a)(s-b)(s-c)) ou s = (a+b+c)/2 *)
(*        /\             (s est le 1/2 perimetre)                                 *)
(*     c /  \ a                                                                   *)
(*      /    \                                                                    *)
(*     A------C                                                                   *)
(*        b                                                                       *)

(* Calcul de la surface d'un polygone etant donne un centre et      *)
(* les points du polygone classes dans le sens horaire. On calcule  *)
(* a chaque fois la surface du triangle c a1 a2 avec c le centre et *)
(* a1 et a2 deux sommets du polygone. On continue recursivement sur *)
(* les sommets prives de a1. On n'oublie pas de finir par an a1 qui *)
(* est le dernier triangle.                                         *)

and surface_simple a1 b1 c1 =
  let a = distance b1 c1
  and b = distance a1 c1
  and c = distance a1 b1 in
  let s = (a +. b +. c) /. 2.0
  in
    Pervasives.sqrt(s *. (s -. a) *. (s -. b) *. (s -. c))

and distance a b = Pervasives.sqrt(((b.hv.h -. a.hv.h) ** 2.0) +. ((b.hv.v -. a.hv.v) ** 2.0))

and surface c l = 
  (* prerr_endline "surface !"; *)
  let rec surf acc = function
      [] | [_]         -> raise (Erreur "severe internal error in computing the area of a Voronoi cell")
    | [h1; h2]         -> (acc +. (surface_simple c h1 h2)), h2               (* On retourne le dernier sommet du polygone *)
    | h1::(h2::t as f) -> surf (acc +. (surface_simple c h1 h2)) f in

  let fin, dernier = surf 0.0 l
  in
    assert(0 < List.length l);
    fin +. (surface_simple c (List.hd l) dernier)                             (* Pour calculer l'aire du dernier triangle ! *)

