(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: topo_del.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2005-03-07 17:24:52 $
 * Version: $Revision: 1.1 $
 * -------------------------------------------------------------------
 *
 * Status: voir le commentaire general dans topology.ml
 *          
 *
 *=========================================================================
 *)

open Global;;
open Valeur;;
open Printer;;
open Topology;;
open Topo_graph_generique;;
open Del;;

__register_version "$RCSfile: topo_del.ml,v $" "$Revision: 1.1 $" "$Date: 2005-03-07 17:24:52 $";;

module IntSet = Set.Make(struct type t = int let compare=compare end);;

exception FoundIntPos of int;;

class topo_del ((((tab as c) : 'collection), _), couleur) mgsval =
object(self)
  inherit [mgs_value Def_del.del_tab_element] topo_graphe_generique tab c couleur mgsval

  val mutable _dimension   = 0
  val mutable _fonction    = (fun x -> x) 
  val mutable _has_voronoi = false

  method get_voisins_from_tab    = Core_del.vois
  method get_decoration_from_tab = Core_del.deco


  method private un_match (kind, chemin, rhs) (fold_chemin, fold_valeur, fold_len_chemin, fold_nb_valeur) =
    let len_filtree = List.length chemin in
    let liste_valeurs, nb_valeurs = ( (* Printf.printf "On a filtre les elements aux positions %s\ncorrespondant a %s\npour donner %s\n"
				       (Gbf.clean_comma_string_of_general_list string_of_int chemin)
				       (Gbf.clean_comma_string_of_general_list (fun i -> "\n\t"^(to_string (prsvalue (Delaunay.deco tab.(i))))) chemin)
				       (to_string (prsvalue rhs)); *)

				     match rhs
				     with 
					 VAL_seq (t, _) -> (let r = (Array.to_list t)  (* Une sequence represente la liste des elements filtres *)
							    in
							      r, (Array.length t))

				       | VAL_undef _    -> [], 0                       (* Un <undef> represente la suppression d'un element     *)

				       | x              -> (if (len_filtree = 1) 
							    then [x], 1
							    else raise (Erreur "internal error on delaunay graphs with un_match")))
    in
      (List.fold_right (fun e s -> IntSet.add e s) chemin fold_chemin), (liste_valeurs@fold_valeur), (len_filtree+fold_len_chemin), (nb_valeurs+fold_nb_valeur)

  method build_result =
    let chemins, valeurs, len_chemin, nb_valeurs = List.fold_right self#un_match _batch (IntSet.empty, [], 0, 0) in (* Ensemble des chemins filtres et des valeurs calculees   *)
    let valeurs_tableau = Array.of_list valeurs in
    let taille_initiale = Array.length tab in
    let quantite_finale = (let nombre_non_filtre = taille_initiale - len_chemin   (* Si les chemins representent 12 valeurs, 5 n'ont pas ete filtree *)
			   in
			     nombre_non_filtre + nb_valeurs) in                   (* La taille finale est le nombre de non filtre + le nombre de valeurs produites *)     
    let tab_valeur      = Array.make quantite_finale (VAL_undef "") 
    and base            = ref 0
    and check           = ref 0
    in
      assert((IntSet.cardinal chemins) = len_chemin);
      assert((Array.length valeurs_tableau) = nb_valeurs);
      for i = 0 to (taille_initiale - 1)                                          (* On recupere deja tous ceux qui n'ont pas etes filtres *)
      do
	if (IntSet.mem i chemins)                                                 (* Si l'index est dans la liste des chemins, c'est qu'il a ete filtre -> ne rien faire *)
	then ()
	else (tab_valeur.(!base) <- Core_del.deco (tab.(i)); incr base)           (* Sinon, c'est qu'au contraire il doit etre conserve                                  *)
      done;

      for i = 0 to (nb_valeurs - 1)                                               (* On ajoute ensuite tous les elements qui ont etes produits                           *)
      do
	tab_valeur.(!base + i) <- valeurs_tableau.(i);
	assert(incr check; true)
      done;

      assert(!base + !check = quantite_finale);
            
      let all_positions = Array.map (fun e -> Core_del.float_array_of_val_seq _dimension (_fonction e))  tab_valeur in
      let voisins, vor, open_d  = Core_del.compute_delaunay_neighbors _dimension all_positions _has_voronoi
      in
	VAL_del ( ((Array.init quantite_finale (fun i -> ((tab_valeur.(i)), (all_positions.(i)), (voisins.(i))))), (vor, open_d)), couleur)

  initializer 
  let dim, fonc, vor = Core_del.get_defs_from_delaunay couleur.name
  in
    _dimension   <- dim;
    _fonction    <- fonc;
    _has_voronoi <- vor

end
;;
