(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: type_del.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2005-11-23 11:10:57 $
 * Version: $Revision: 1.2 $
 * -------------------------------------------------------------------
 *
 * Voir le fichier Common/type.ml
 * 
 *=========================================================================
 *)

open Global;;
open Valeur;;
open Type;;
open Dispatch_sig;;

__register_version "$RCSfile: type_del.ml,v $" "$Revision: 1.2 $" "$Date: 2005-11-23 11:10:57 $";;


(* --- Delaunay ---------------------------------------------------- *)


(*!*)(** 
       @module collections
       @name 'delaunay
       @type any_bool
       Delaunay type and predicate.

       A delaunay is a collection of elements munished with the
       topology induced by the delaunay graph of the element projected
       in an euclidean space. 

       Delaunay's are constructed types. The corresponding declaration
       begins with the keyword [delaunay]. This implies that the
       corresponding predicate is denoted only by its quoted name
       ['delaunay]. 
       
       @voir Delaunay the chapter delaunay in the documentation.
*)(*!*)

let del_desc = create_type
		 { name = "'delaunay";
		   concrete = false;
		   leibniz = true;
		   predicate = wrap_pred(is_leibniz);
		   empty = VAL_undef "dummy for empty field in type descriptor 'delaunay";
		   pospredicate = pospredicate_int;
		   make = make_dummy;
		   xxxify = fun_2 delaunayfy; 
		   derive_from = [leibniz_desc];
		   father_of = [];
		   uid = get_new_uid();
		 }
;;

del_desc.predicate <- wrap_pred(coll_pred del_desc);;
del_desc.empty <- VAL_del (([||], ([||], [||])), del_desc);;

let extension_del nom desc x      = raise (Erreur "extension_del not yet implemented")
;;

let  create_new_del nom i x b =
  let ret =
    try
      let t = Hash.find type_tbl nom
      in let err = "Error: type  "^nom^"  already defined."
      in prerr_endline err; prftype stderr t; raise (Erreur err)
    with Not_found -> 
      (if Hash.mem Environnement.var_tbl nom
       then prerr_endline ("Warning: type predicate  "^nom^"  hiddes a function with same name");
       let nt = create_type { name = nom;
		 	      concrete = true;
			      leibniz = true;
			      predicate = del_desc.predicate;
			      empty =  VAL_del (([||], ([||], [||])), del_desc);
			      pospredicate = pospredicate_int;
			      make = make_dummy;
			      xxxify = xxxify_dummy nom;
			      derive_from = [del_desc];
			      father_of = [];
			      uid = get_new_uid();
			    }
       in 
	 begin
	   nt.empty <- change_type nt nt.empty;
	   nt.predicate <- wrap_pred(coll_pred nt);
	   nt.make <- extension_del nom nt;
	   make_xxxify nt;
	   ignore(Environnement.addfct nom nt.predicate);
	   nt
	 end)
  in
    Core_del.process nom i x b;
    ret
