(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: topo_graph_generique.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2005-08-30 12:55:17 $
 * Version: $Revision: 1.4 $
 * -------------------------------------------------------------------
 *
 * Status: voir le commentaire general dans topology.ml
 *          
 * Ici on developpe les instance de generic_topo qui correspondent 
 * aux collections Associatives/Commutative (et parfois idempotente)
 *
 *=========================================================================
 *)

open Global;;
open Valeur;;
open Printer;;
open Topology;;
open Core_gbf;;
open Gbf;;

__register_version "$RCSfile: topo_graph_generique.ml,v $" "$Revision: 1.4 $" "$Date: 2005-08-30 12:55:17 $";;
(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: topo_graph_generique.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2005-08-30 12:55:17 $
 * Version: $Revision: 1.4 $
 * -------------------------------------------------------------------
 *
 * Status: voir le commentaire general dans topology.ml
 *          
 * Ici on developpe les instance de generic_topo qui correspondent 
 * aux collections Associatives/Commutative (et parfois idempotente)
 *
 *=========================================================================
 *)

open Global;;
open Valeur;;
open Printer;;
open Topology;;
open Core_graph_generique;;

__register_version "$RCSfile: topo_graph_generique.ml,v $" "$Revision: 1.4 $" "$Date: 2005-08-30 12:55:17 $";;

exception FoundIntPos of int;;


(* added by Julien. The same function is provided in Gbf/topo_gbf.ml : it should be common. *)
let in_place_shuffle_array len a = 
  let nb  = ref (if (len <= 1)
		 then 0
		 else 
		   if (len = 2) 
		   then 1 
		   else len) (* Pour eviter que sur un tableau de longueur 2 on fasse 2 permutations inverses. *)
  in                         (* Le risque est moins grand sur un tableau de taille superieure.                 *)
    while (!nb > 0)
    do
      let src = Random.int len
      and dst = Random.int len
      in
	if (src = dst)
	then decr nb
	else (let tmp = a.(src) 
	      in
		a.(src) <- a.(dst);
		a.(dst) <- tmp;
		decr nb)
    done;
    a
;;



class virtual ['graphe_element] topo_graphe_generique (tab : 'graphe_element array) (c : 'collection) couleur mgsval =
object(self)
  inherit [int] generic_topo 
    string_of_int                       (** afficheur de position    **)
    (c, couleur)                        (** collection a transformer **)
    mgsval                              (** l'argument complet       **)
    (Global.crandom (Array.length tab)) (** position de depart       **)


  method virtual get_voisins_from_tab    : ('graphe_element -> 'position list)
  method virtual get_decoration_from_tab : ('graphe_element -> mgs_value)

  val _taille_tab            = Array.length tab
  val _used                  = Array.make (Array.length tab) PosFree
  val mutable _batch         = ([] : (Terme.path_type * 'position path * mgs_value) list)

  method mgs_value_from_pos pos = (VAL_int pos)

  method state p             = _used.(p)
  method update_state fonc p = _used.(p) <- fonc(_used.(p))

  method unvisit_all = (for i = 0 to (_taille_tab - 1)
			do 
			  (let v = _used.(i)
			   in
			     if pos_visited(v)
			     then _used.(i) <- (pos_unvisit v))
			done)

  method next_usable p =
    try
      for i = 0 to (_taille_tab - 1)
      do
	let position = (p+i) mod _taille_tab
	in
	  if (_used.(position) = PosFree) then raise (FoundIntPos position)
      done;
      raise NoMoreUsable
    with FoundIntPos p -> (assert begin if !trace_opt then
			     (Printf.printf "---- topoGraphs: next_usable: found %d\n" p); 
			     true
			   end;
			   p)

  method add_match pt p v     = _batch <- (pt, p, v)::_batch
  method get_value_from_pos p = self#get_decoration_from_tab (tab.(p))

  method some_neighbors pos kind =
    let (nb, somms) = (
    match kind
    with
	SomeDirections _ -> raise (Erreur "directions not yet implemented in graphs")
      | AllDirections    -> (let (nb, liste) = List.fold_right 
						 (fun sommet (nb, liste) -> (if pos_usable(_used.(sommet))
									     then (nb+1), (sommet::liste)
									     else nb, liste))
						 (self#get_voisins_from_tab tab.(pos))
						 (0, [])
			     in
			       nb, (Array.of_list liste))
    )
    in
(*      Printf.printf "Etat de _used : %s\n" (Gbf.clean_comma_string_of_general_array 
					      (function 
						   PosFree    -> "free"
						 | PosVisited -> "visited"
						 | PosUsed_1  -> "used1"
						 | PosUsed_2  -> "used2") _used);
      Printf.printf "Sur la position %d, on retourne les voisins aux positions %s\n\n" pos (Gbf.clean_comma_string_of_general_array string_of_int somms);
*)

(* next line modified by Julien. Old line : (nb,somms) *)
      (nb,(Array.sub (if (!match_random > 0) then (in_place_shuffle_array nb somms) else somms) 0 nb))

						 
  method direction_from_to p1 p2 = VAL_undef "direction"
end
;;
