(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: core_prox.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2005-03-07 17:24:53 $
 * Version: $Revision: 1.1 $
 * -------------------------------------------------------------------
 *
 * Status: 
 *=========================================================================
 *)

open Global;;
open Def_prox;;
open Valeur;;
open Printer;;

__register_version "$RCSfile: core_prox.ml,v $" "$Revision: 1.1 $" "$Date: 2005-03-07 17:24:53 $";;




let prox_tbl         = (Hash.create 7 : (string, (Valeur.mgs_value prox_def)) Hash.t) (* Un nom de proximal -> fonction *)
;;

let process nom fonc = (if (Hash.mem prox_tbl nom)
			then raise (Erreur "a proximal definition of that name already exists")
			else Hash.add prox_tbl nom (to_funct fonc))
;;

let deco (x, _) = x
and vois (_, x) = x
;;


let get_defs_from_proximal nom = 
  (try
     Hash.find prox_tbl nom
   with Not_found -> raise (Erreur (nom ^ " does not correspond to any proximal definition")))
;;

let compute_proximal_neighbors f tab = 
  let taille  = Array.length tab in
  let voisins = Array.make taille []
  in
    for i = 0 to (taille-1)
    do
      for j = i+1 to (taille-1)
      do
	if (((to_funct (f tab.(i))) tab.(j)) = (VAL_bool true)) then (voisins.(i) <- j::voisins.(i);   (* update le sommet i     *)
								      voisins.(j) <- i::voisins.(j));  (* ainsi que son voisin ! *)
      done
    done;
    voisins
;;




let proximal_from_all_sommets all_sommets f color =
  let voisins    = compute_proximal_neighbors f all_sommets
  in
    VAL_prox ((Array.init (Array.length all_sommets) (fun i -> (all_sommets.(i)), (voisins.(i)))), color)
;;


let array_of_mgs_value v =
  let cons h t = h::t
  in
    match v
    with 
      | ((VAL_int _) as x)
      | ((VAL_bool _) as x)
      | ((VAL_string _) as x)
      | ((VAL_rec _) as x) 
      | ((VAL_posgbf _) as x)
(*       | ((VAL_splx _) as x)  *)
      | ((VAL_gmap _) as x)
      | ((VAL_ncell _) as x)
      | ((VAL_float _) as x) -> [| v |]
      |	VAL_seq (a, _)       -> a
      | VAL_grf ((_, a), _)  -> Array.map Core_graph.deco a
      | VAL_set (a, _)       -> Array.of_list (Core_set.fold    cons        a  [])
      | VAL_bag (a, _)       -> Array.of_list (Core_bag.fold_b  cons        a  [])
      | VAL_gbf (a, _)       -> Array.of_list (Hash.fold (fun k v s -> v::s) a  [])
      | VAL_del ((a, _), _)  -> Array.map Core_del.deco a
      | VAL_prox (a, _)      -> Array.map deco a
      | VAL_achain (a, _)    -> Array.of_list (Hash.fold (fun k (_,v) s -> v::s) a  [])

      | VAL_undef  _
      | VAL_funct _        
      | _ -> raise (Erreur "don't know that to do with such a value in a proximal join")
;;
