(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: type_prox.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2005-11-23 11:10:57 $
 * Version: $Revision: 1.2 $
 * -------------------------------------------------------------------
 *
 * Voir le fichier Common/type.ml
 * 
 *=========================================================================
 *)


open Global;;
open Valeur;;
open Type;;
open Dispatch_sig;;

__register_version "$RCSfile: type_prox.ml,v $" "$Revision: 1.2 $" "$Date: 2005-11-23 11:10:57 $";;

(* --- Proximaux ---------------------------------------------------- *)


(*!*)(** 
       @module collections
       @name 'proximaux
       @type any_bool
       Proximal type and predicate.

       A proximal is a collection of elements munished with the
       topology induced by a distance functions given by the user. 

       Proximaux are constructed types. The corresponding declaration
       begins with the keyword [proximal]. This implies that the
       corresponding predicate is denoted only by its quoted name
       ['proximal]. 
       
       @voir Proximal the chapter proximal in the documentation.
*)(*!*)

let prox_desc = create_type
		 { name = "'proximal";
		   concrete = false;
		   leibniz = true;
		   predicate = wrap_pred(is_leibniz);
		   empty = VAL_undef "dummy for empty field in type descriptor 'proximal";
		   pospredicate = pospredicate_int;
		   make = make_dummy;
		   xxxify = fun_2 proxify; 
		   derive_from = [leibniz_desc];
		   father_of = [];
		   uid = get_new_uid();
		 }
;;

prox_desc.predicate <- wrap_pred(coll_pred prox_desc);;
prox_desc.empty <- VAL_prox ([||], prox_desc);;

let extension_prox nom desc x      = raise (Erreur "extension_prox not yet implemented")
;;

let  create_new_prox nom x = 
  let ret =
    try
      let t = Hash.find type_tbl nom
      in let err = "Error: type  "^nom^"  already defined."
      in prerr_endline err; prftype stderr t; raise (Erreur err)
    with Not_found -> 
      (if Hash.mem Environnement.var_tbl nom
       then prerr_endline ("Warning: type predicate  "^nom^"  hiddes a function with same name");
       let nt = create_type { name = nom;
		 	      concrete = true;
			      leibniz = true;
			      predicate = prox_desc.predicate;
			      empty =  VAL_prox ([||], prox_desc);
			      pospredicate = pospredicate_int;
			      make = make_dummy;
			      xxxify = xxxify_dummy nom;
			      derive_from = [prox_desc];
			      father_of = [];
			      uid = get_new_uid();
			    }
       in 
	 begin
	   nt.empty <- change_type nt nt.empty;
	   nt.predicate <- wrap_pred(coll_pred nt);
	   nt.make <- extension_prox nom nt;
	   make_xxxify nt;
	   ignore(Environnement.addfct nom nt.predicate);
	   nt
	 end)
  in
    Core_prox.process nom x;
    ret
