(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: ptopology.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2005-09-30 14:07:59 $
 * Version: $Revision: 1.10 $
 * -------------------------------------------------------------------
 *
 * Status: 
 *          
 *=========================================================================
 *)

open Printf;;

open Global;;
open Valeur;;
open Terme;;
open Printer;;
open Printer_term;;


__register_version "$RCSfile: ptopology.ml,v $" "$Revision: 1.10 $" "$Date: 2005-09-30 14:07:59 $";;

type 'position pattern_variable_value =  ImmediateVal of mgs_value
					 | EltCollection of 'position
					 | PathCollection of 'position list
;;

type 'position path = 'position list * 'position list          (** un chemin  (elts filtrs, elts utiliss) **)
and  'position return_path = ('position path) option           (** chemin + env associe                     **)
and  'position env =  ('position pattern_variable_value) array (** liaison d'un nom a un chemin             **)
and  path_type = Terme.path_type                               (** type d'une regle                         **)
;;

type 'pattern compiled_rule = { rname : string option;
				motif : 'pattern;   (** defini apres l'usage, c'est pourquoi en parametre **)
				rtype : Terme.path_type;
				rhs   : mgs_value;
				nbvar : int }
;;

type 'position translated_rhs = ('position rewritten_elt) list
and 'position rewritten_elt =
  | Symb of string * (int * (mgs_value list) * (mgs_value list) ) * mgs_value
  | Filt of 'position * mgs_value
;;

 
(** Acces a la valeur d'une variable de pattern, a partir d'un 
 ** environnement sous forme de tableau: l'attribution d'un index 
 ** (de ce tableau) a une variable du pattern est faite par la methode  
 ** Patch.bind_pvar. La fonction get_val_of_rule_variable attend 
 ** un VAL_int qui identifie la variable. 
 **
 ** voir dans le fichier patch pour les indices des variables predefinies 
 **) 
 
let get_val_of_rule_variable =

  (** acces  la valeur d'une variable de filtre **)
  let get_value_of_pattern_variable ptopo = function 
    | 0 -> ptopo#collection 
    | i ->  
	(try  
           (match (ptopo#get_env i) with 
              | ImmediateVal v -> v
	      | EltCollection p -> ptopo#get_value_from_pos p
              | PathCollection l -> VAL_seq((Array.of_list(List.map (ptopo#get_value_from_pos) l)), Type_seq.seq_desc))
	 with _ ->  
           let err = "Error: internal: do not find pattern variable number "
		     ^(string_of_int i) 
                     ^" in current pattern environment (of size "
		     ^ (string_of_int ptopo#length_env)
		     ^")\n" 
           in print_endline err; flush stdout; raise (Erreur err)) 

  (** acces  la position d'une variable de filtre **)
  and get_pos_of_pattern_variable ptopo = function 
    | 0 -> raise (Erreur "Error: internal: pos of self")
    | i ->  
	(try  
           (match (ptopo#get_env i) with 
              | ImmediateVal v -> raise (Erreur "Error: internal: immediate of pos")
	      | EltCollection p -> ptopo#mgs_value_from_pos p
              | PathCollection l -> VAL_seq((Array.of_list(List.map (ptopo#mgs_value_from_pos) l)), Type_seq.seq_desc))
	 with _ ->  
           let err = "Error: internal: do not find pattern variable number "
		     ^(string_of_int i) 
                     ^" in current pattern environment (of size "
		     ^ (string_of_int ptopo#length_env)
		     ^")\n" 
           in print_endline err; flush stdout; raise (Erreur err)) 
	
  in (** dispatch entre les 2 types d'accs **)
    fun ptopo index -> let idx = to_int index in
      begin
	(* Printf.printf "On cherche la variable a l'index %d alors que la taille est %d\n" idx ptopo#length_env; *)
	if idx < 0 then get_pos_of_pattern_variable ptopo (0 - idx)
	else get_value_of_pattern_variable ptopo idx
      end
;; 

(* --- gestion de la partie droite ------------------------------------------*)

let rec complete_fcf_relation ht l =
  match l with
    | [] -> []
    | (Filt (x,y)) :: q -> (Filt (x,y)) :: (complete_fcf_relation ht q)
    | Symb (s,(d,lf,lcf),v) :: q -> (
	List.iter (fun f ->
		     let (face,coface) = try Hash.find ht f with Not_found -> ([],[]) in
		       Hash.replace ht f (face,(VAL_symbol s)::coface)
		  ) lf ;
	List.iter (fun cf ->
		     let (face,coface) = try Hash.find ht cf with Not_found -> ([],[]) in
		       Hash.replace ht cf ((VAL_symbol s)::face,coface)
		  ) lcf ;
	let (new_lf, new_lcf) =
	  let (fnd, cfnd) = try Hash.find ht (VAL_symbol s) with Not_found -> ([],[]) in
	    ( without_double (fnd@lf), without_double (cfnd@lcf) )
	in
	  (Symb (s,(d,new_lf,new_lcf),v))::(complete_fcf_relation ht q)
      )
;;
 

(* --- gestion du marquage ------------------------------------------*)

type 'position usage_position =
  | PosFree
  | PosUsed1 of ('position list)
  | PosFiltered1
  | PosVisited
  | PosUsed2 of ('position list)
  | PosFiltered2
;;


let rec delete_first_occ p = function
  | [] -> raise Not_found
  | e::q -> if (Pervasives.compare e p = 0) then q else (delete_first_occ p q)

and pos_visit = function
  | PosFree      -> PosVisited
  | PosUsed1 l   -> PosUsed2 l
  | PosFiltered1 -> PosFiltered2
  | _ -> raise (Erreur "Error: internal: cannot visit a visited or filtered position")

and pos_unvisit = function
  | PosVisited   -> PosFree
  | PosUsed2 l   -> PosUsed1 l
  | PosFiltered2 -> PosFiltered1
  | _ -> raise (Erreur "Error: internal: cannot unvisit a not visited position")

and pos_use p = function
  | PosFree      -> PosUsed1 [p]
  | PosVisited   -> PosUsed2 [p]
  | PosUsed1 l -> PosUsed1 (p::l)
  | PosUsed2 l -> PosUsed2 (p::l)
  | _ -> raise (Erreur "Error: internal: cannot use a filtered position")

and pos_unuse p = function
  | PosUsed1 l -> (
      try
	let new_l = delete_first_occ p l in
	  if (new_l = []) then PosFree else (PosUsed1 new_l)
      with
	  Not_found -> raise (Erreur "Error: internal: cannot unuse a not used by p position")
    )
  | PosUsed2 l -> (
      try
	let new_l = delete_first_occ p l in
	  if (new_l = []) then PosVisited else (PosUsed2 new_l)
      with
	  Not_found -> raise (Erreur "Error: internal: cannot unuse a not used by p position")
    )
  | PosFree      -> raise (Erreur "Error: internal: cannot unuse a not-used free position") 
  | PosVisited   -> raise (Erreur "Error: internal: cannot unuse a not-used free position") 
  | PosFiltered1 -> raise (Erreur "Error: internal: cannot unuse a not-used filtered position")
  | PosFiltered2 -> raise (Erreur "Error: internal: cannot unuse a not-used filtered position")

and pos_filter = function
  | PosFree     -> PosFiltered1
  | PosVisited  -> PosFiltered2
  | PosUsed1 _   -> raise (Erreur "Error: internal: cannot filter a used position")
  | PosUsed2 _   -> raise (Erreur "Error: internal: cannot filter a used position")
  | PosFiltered1 -> raise (Erreur "Error: internal: cannot filter an already filtered position")
  | PosFiltered2 -> raise (Erreur "Error: internal: cannot filter an already filtered position")

and pos_unfilter = function
  | PosFiltered1 -> PosFree
  | PosFiltered2 -> PosVisited
  | PosUsed1 _   -> raise (Erreur "Error: internal: cannot unfilter a used position")
  | PosUsed2 _   -> raise (Erreur "Error: internal: cannot unfilter a used position")
  | PosFree      -> raise (Erreur "Error: internal: cannot unfilter an not filtered position")
  | PosVisited   -> raise (Erreur "Error: internal: cannot unfilter an not filtered position")

and pos_visited = function
  | PosVisited   -> true
  | PosFiltered2 -> true
  | PosUsed2 _   -> true
  | _ -> false

and pos_visitable = function
  | PosFree      -> true
  | PosUsed1 _   -> true
  | PosFiltered1 -> true
  | _ -> false

and pos_used = function
  | PosUsed1 _ -> true
  | PosUsed2 _ -> true
  | _          -> false

and pos_usable = function
  | PosFiltered1 -> false
  | PosFiltered2 -> false
  | _ -> true

and pos_filtered = function
  | PosFiltered1 -> true
  | PosFiltered2 -> true
  | _            -> false

and pos_filterable = function
  | PosFree    -> true
  | PosVisited -> true
  | _ -> false

and string_of_pos = function
  | PosFree      -> "PosFree"
  | PosVisited   -> "PosVisited"
  | PosUsed1 _   -> "PosUsed1"
  | PosUsed2 _   -> "PosUsed2"
  | PosFiltered1 -> "PosFiltered1"
  | PosFiltered2 -> "PosFiltered2"


(** Interface vers une PTOPOLOGIE :
 ** c'est une classe abstraite et polymorphe en le type representant
 ** une position.
 **)


exception NoMoreUsable;;
exception NoMoreFilterable;;
exception NoMoreFound;;
exception FoundPath;;

class virtual ['position] generic_ptopo 
  (prpos : 'position -> string)
  (coll, couleur)
  (mgsval : mgs_value)
  (init_pos : 'position) =
object (self)

  method collection = mgsval

  method string_of_usage_position p = string_of_pos (self#state p)

  method prpos = prpos

  (** gestion du marquage **)

  method visit x =
    try
      self#update_state pos_visit x
    with
	Erreur str ->
	  raise (Erreur ( Printf.sprintf
			    "Cannot visit this position : %s (%s) with value %s\n\t => %s"
			    (prpos x)
			    (self#string_of_usage_position x)
			    (to_string(prsvalue (self#get_value_from_pos x)))
			    str
			)
		)
  method unvisit x =
    try
      self#update_state pos_unvisit x
    with
	Erreur str ->
	  raise (Erreur ( Printf.sprintf
			    "Cannot unvisit this position : %s (%s) with value %s\n\t => %s"
			    (prpos x)
			    (self#string_of_usage_position x)
			    (to_string(prsvalue (self#get_value_from_pos x)))
			    str
			)
		)

  method use p x =
    try
      self#update_state (pos_use p) x
    with
	Erreur str ->
	  raise (Erreur ( Printf.sprintf
			    "Cannot use this position : %s (%s) with value %s\n\t => %s"
			    (prpos x)
			    (self#string_of_usage_position x)
			    (to_string(prsvalue (self#get_value_from_pos x)))
			    str
			)
		)
  method unuse p x =
    try
      self#update_state (pos_unuse p) x
    with
	Erreur str ->
	  raise (Erreur ( Printf.sprintf
			    "Cannot unuse this position : %s (%s) with value %s\n\t => %s"
			    (prpos x)
			    (self#string_of_usage_position x)
			    (to_string(prsvalue (self#get_value_from_pos x)))
			    str
			)
		)

  method filter x =
    try
      self#update_state pos_filter x
    with
	Erreur str ->
	  raise (Erreur ( Printf.sprintf
			    "Cannot filter this position : %s (%s) with value %s\n\t => %s"
			    (prpos x)
			    (self#string_of_usage_position x)
			    (to_string(prsvalue (self#get_value_from_pos x)))
			    str
			)
		)
  method unfilter x =
    try
      self#update_state pos_unfilter x
    with
	Erreur str ->
	  raise (Erreur ( Printf.sprintf
			    "Cannot unfilter this position : %s (%s) with value %s\n\t => %s"
			    (prpos x)
			    (self#string_of_usage_position x)
			    (to_string(prsvalue (self#get_value_from_pos x)))
			    str
			)
		)

  method visited p           = pos_visited(self#state p)
  method visitable p         = pos_visitable(self#state p)
  method filtered p          = pos_filtered(self#state p)
  method filterable p        = pos_filterable(self#state p)
  method used p              = pos_used(self#state p)
  method usable p            = pos_usable(self#state p)

  method virtual state        : 'position -> 'position usage_position
  method virtual update_state : ('position usage_position -> 'position usage_position) -> 'position -> unit 
  method virtual next_usable     : 'position
  method virtual next_filterable : 'position
  method virtual common_position : 'position list -> 'position list -> 'position list -> 'position list -> 'position array
  method virtual use_transitive_faces_cofaces :'position -> 'position list
  method virtual dimension       : 'position -> int

  method get_val_of_rule_variable = VAL_funct(get_val_of_rule_variable self)
  method virtual get_value_from_pos : 'position -> mgs_value
  method virtual mgs_value_from_pos : 'position -> mgs_value
  method virtual pos_from_mgs_value : mgs_value -> 'position

  (** gestion de la reconstruction **)
  method virtual add_match   : path_type -> 'position path -> mgs_value -> unit
  method virtual build_result : mgs_value


  val mutable _env = ([||] : 'position env)
  method get_env i = _env.(i)
  method get_elt i = match (self#get_env i) with
      EltCollection v -> v
    | _               -> raise (Erreur "Ptopology: ppattern_dpt: instancie_incident_positions: waiting for an EltCollection")
  method length_env = Array.length _env
  method set_elt i v = _env.(i) <- EltCollection v
  method unset_elt i = _env.(i) <- ImmediateVal (VAL_undef "uninitialized pattern variable")
  method set_path i v = _env.(i) <- PathCollection v
  method set_immediate i v = _env.(i) <- ImmediateVal v
  method print_env =
    printf "       Environement:\n" ;
    Array.iteri
      (fun i v -> if (i<>0) then match v with
	   EltCollection v -> printf "\t%d -> %s\n" i (prpos v) ;
	 | ImmediateVal (VAL_undef _) -> printf "\t%d -> empty\n" i
	 | _ -> ()(*raise (Erreur "Ptopology: ppattern_dpt: print_env: waiting for an EltCollection")*)
      )
      _env;
    flush stdout 

  method resize_env max_size = 
    let mm = 1 + max_size  (** l'elt 0 est reserv  self **)
    in (if (mm > Array.length _env) 
	then _env <- Array.make mm (ImmediateVal (VAL_undef "uninitialized pattern variable")))


  val mutable _cpos = init_pos
  method cpos : 'position = _cpos
  method move p = assert(not(self#used p)); _cpos <- p
  method goto p = _cpos <- p
  method visit_current = self#visit(_cpos)
  method use_current = self#use (_cpos) (_cpos)
  method filter_current = self#filter(_cpos)
  method unvisit_current = self#unvisit(_cpos)
  method unuse_current = self#unuse (_cpos) (_cpos)
  method unfilter_current = self#unfilter(_cpos)
(*  method move_next_usable = try _cpos <- self#next_usable _cpos; true with NoMoreUsable -> false*)
  method init_pos = init_pos

  (** Itration sur les faces, les cofaces ou tout le monde **)

  method virtual iter_all     : ( 'position -> unit ) -> unit
  method virtual iter_faces   : ( 'position -> unit ) -> 'position -> unit
  method virtual iter_cofaces : ( 'position -> unit ) -> 'position -> unit
  method virtual isFaceOf     : 'position -> 'position -> bool
  method virtual isCoFaceOf   : 'position -> 'position -> bool

  (** gestion de la transformation **)
  method apply_one_rule (rule : ('position ppattern) compiled_rule) =
    let rule_name  = (rule.rname)
    and motif      = (rule.motif)
    and rule_type  = (rule.rtype)
    and exp        = (rule.rhs)
    and nbvar      = (rule.nbvar)
    and nb_applied = ref 0  (* Une regle a-t-elle ete appliquee *)
    in
      begin
	assert begin if !trace_opt then 
	  (printf "---- topo: apply_one_rule: %s motif: %s\n" 
	     (match rule_name with None -> "(anonymous)" | Some s -> s) motif#to_string);
	  flush stdout;
	  true
	end;
	self#resize_env nbvar;
	let empty = ref true in
	  (try
	     while true do
	       assert begin if !trace_opt then 
		 (printf "---- topo: apply_one_rule: match an occurence of the ppattern\n" ;
		  flush stdout) ;
		 true
	       end;
	       (match motif#filtre [] (self :> 'position generic_ptopo) with
		 | None ->
		     assert begin if !trace_opt then 
		       (printf "---- topo: apply_one_rule: failed motif %s\n" 
			  motif#to_string;
			flush stdout);
		       true
		     end;
		     raise NoMoreFound
		 | Some path ->
		     assert begin if !trace_opt then 
		       (printf
			  "---- topo: apply_one_rule: accept motif %s and neighbors are used\n" 
			  motif#to_string;
		        printf "----                       add to batch:\n" ;
			flush stdout);
		       true
		     end;
		     incr nb_applied;
		     let rhs = (pf1 exp) self#get_val_of_rule_variable in
		       assert begin if !trace_opt then
			 (Printer.prvalue rhs ; Printf.printf "\n" ;
			  flush stdout);
			 true
		       end;
		       self#add_match rule_type path rhs);
	     done
	   with
	       NoMoreFound      
	     | NoMoreFilterable 
	     | NoMoreUsable     -> ()
	  );
	  self#unvisit_all;
	!nb_applied
      end

  method patch rules =
    begin
      try
	List.iter (fun r -> let _ = self#apply_one_rule r in ()) rules;
	self#build_result
      with BreakTransformation res -> res
    end

  method trad_der t = match t with
    | VAL_seq(a,_) -> (
	match (Array.to_list a) with
	  | (VAL_int d)::(VAL_seq (f,_))::(VAL_seq (cf,_))::[] -> (d,(Array.to_list f),(Array.to_list cf))
	  | (VAL_int d)::f::(VAL_seq (cf,_))::[]               -> (d,[f],(Array.to_list cf))
	  | (VAL_int d)::(VAL_seq (f,_))::cf::[]               -> (d,(Array.to_list f),[cf])
	  | (VAL_int d)::f::cf::[]                             -> (d,[f],[cf])
	  | _ -> raise (Erreur "je crois qu il y a un pb")
      )
    | _ -> raise (Erreur "je crois qu il y a un pb")

  method trad_one p = match p with
    | VAL_seq(a,_) -> (
	match a.(0) with
	  | VAL_symbol s -> Symb (s, (self#trad_der a.(1)), a.(2))
	  | p            -> Filt ((self#pos_from_mgs_value p), a.(1))
      )
    | _ -> raise (Erreur "je crois qu il y a un pb")

  method trad_old a =
    Array.to_list (Array.map self#trad_one a)

  method trad a =
    let i = ref 0 and ret = ref [] in
    let l = Array.length a in
      while (!i <> l) do
	match a.(!i) with
	  | VAL_seq(a, t) when (t == Type_seq.pRhsSpec_desc) ->
	      ret := (self#trad a) @ !ret ; incr i
	  | VAL_symbol s ->
	      ret := Symb (s, (self#trad_der a.(!i+2)),
			     match a.(!i+1) with
				 VAL_symbol "PatchDefaultVal" -> VAL_undef "PatchDefaultVal"
			       | VAL_symbol "PatchUndefVal" -> VAL_undef "PatchUndefVal"
			       | v -> v) :: !ret ; i := !i + 3
	  | p ->
	      let p = (self#pos_from_mgs_value p) in
		ret := Filt (p, match a.(!i+1) with
				 VAL_symbol "PatchDefaultVal" -> self#get_value_from_pos p
			       | VAL_symbol "PatchUndefVal" -> VAL_undef "PatchUndefVal"
			       | v -> v) :: !ret ; i := !i + 3
      done ;
      complete_fcf_relation (Hash.create l) !ret
	

end(*

  method unuse_and_move path p  = (self#unuse_all path;
				   self#move p)

  method use_all   = List.iter self#use 
  method reuse_all = List.iter (fun x -> if self#usable(x) then self#use(x))
  method unuse_all = List.iter self#unuse
    
  val mutable _cpos = init_pos
  method cpos : 'position = _cpos
  method move p = assert(not(self#used p)); _cpos <- p
  method goto p = _cpos <- p
  method visit_current = self#visit(_cpos)
  method use_current = self#use(_cpos)
  method move_next_usable = try _cpos <- self#next_usable _cpos; true with NoMoreUsable -> false
  method init_pos = init_pos

    
	

  (** calcul de la direction qui va de p1 a p2.
   ** On suppose que le resultat est encore une position (?)  **)
  method virtual direction_from_to : 'position -> 'position -> mgs_value

  (** gestion du voisinage :  some_neighbors p dir
   ** les voisins definis suivant dir du point p.
   ** Ca retourne un entier n et le tableau des 0 a (n-1) voisins definis   **)
  method virtual some_neighbors : 'position -> direction_specification -> int * ('position array) 
    (** tous les voisins **)
  method neighbors p = self#some_neighbors p AllDirections


  method find : mgs_value -> 'position option = function _ ->
    raise (Erreur "Error: internal: cannot use method find on a non-set topology (to be changed !)")

end



(** patvar = la liste des index des variables qui nomment ce pattern
 ** is_elementary_path : le chemin est-il elementaire -longueur 1 mais different de P*- ou compose 
 **)


and virtual ['position] pattern (patvar : int list) (is_elementary_path : bool) =
object(self)

  method virtual to_string : string
  method print = printf "%s\n" (self#to_string); flush Pervasives.stdout
  method string_var = (List.fold_left (fun acc x -> acc^" "^(string_of_int x)) "^(" patvar)^")"
  method virtual to_full_string : string


  val need_matched_value = (patvar <> [])

  (** mthode qui met  jour *les* variables associes  ce pattern.
   ** On 'prcalcule' cette mthode en renvoyant des fct spcialises 
   ** pour 0 1, 2 et n variables  mettre  jour.
   **)
  method update_elem_variables =
    match patvar with
      | [] -> (fun _ _ -> ())
      | idx1::[] -> (fun topo v ->  topo#set_elt idx1 (List.hd v))
      | idx1::idx2::[] -> (fun topo v -> topo#set_elt idx1 (List.hd v); topo#set_elt idx2 (List.hd v))
      | _ ->  (fun topo v -> List.iter (fun idx -> topo#set_elt idx (List.hd v)) patvar)

  method update_path_variables =
    match patvar with
      | [] -> (fun _ _ -> ())
      | idx1::[] -> (fun topo v ->  topo#set_path idx1 v)
      | idx1::idx2::[] -> (fun topo v -> topo#set_path idx1 v; topo#set_path idx2 v)
      | _ ->  (fun topo v -> List.iter (fun idx -> topo#set_path idx v) patvar)

  method update_variables (topo : 'position generic_topo) v = 
    if is_elementary_path
    then self#update_elem_variables topo v
    else self#update_path_variables topo v

  (* lg minimal d'un chemin filtre *) 
  method virtual minimal_path_length : int

  (* la sous-classe end_pattern, cf. matching.ml sert de terminateur dans certaines 
   * expressions de patterns, et j'ai besoin de tester "1 coup en avance" si l'expression 
   * se termine 
   *)
  method is_end_pattern = false
 
  (* file : environnement -> topo -> path option 
   * le resultat : None = fail, Some p = OK with matched path pt and env is updated
   *)
  method virtual filtre : 'position generic_topo -> 'position return_path
			 
end


and virtual  ['position] linked_pattern 
  (patvar : int list) 
  (is_elementary_path : bool)
  (next   : 'position pattern) =
object(self)
  inherit ['position] pattern patvar is_elementary_path
end
;;




*)



and virtual ['position] ppattern =
object(self)

  method print = printf "%s\n" (self#to_full_string); flush Pervasives.stdout
  method string_of_cguard = function
    | CTest (exp)    -> (to_string(prsvalue exp))
    | CPred (id,exp) -> id^" = "^(to_string(prsvalue exp))
  method string_of_guard = function
    | Test (exp)    -> (to_string(Printer_term.prsterm exp))
    | Pred (id,exp) -> id^" = "^(to_string(Printer_term.prsterm exp))

  method virtual to_string : string
  method virtual to_full_string : string
  method virtual filtre : 'position list -> 'position generic_ptopo -> 'position return_path
  method virtual use_neighborhood : 'position generic_ptopo -> unit
end

and virtual ['position] ppattern_link
  (filtering : bool)
  (ppatid : int)
  (next   : 'position ppattern) =
object(self)
  inherit ['position] ppattern

  method use_neighborhood ptopo =
    raise (Erreur "PTopology: bug if i appear !!!!")
(*
    if filtering then
      let p = ptopo#get_elt ppatid in
	ignore (ptopo#use_transitive_faces_cofaces p)
    else
      next#use_neighborhood ptopo*)

  method eval_cguard (ptopo : ('position generic_ptopo)) = function
    | CTest (exp)    -> (
	let vguard = (pf1 exp) ptopo#get_val_of_rule_variable in
	assert begin if !trace_opt then 
	  (printf "         |    eval_cguard: guard = %a" Printer.prfvalue2 vguard);
	  true
	end;
	Auxfct.bool_from_val vguard
       )

    | CPred ("dim",exp) -> (
	let vdim = (pf1 exp) ptopo#get_val_of_rule_variable in
	  assert begin if !trace_opt then 
	    (printf "         |    eval_cguard: dim = %a\n" Printer.prfvalue2 vdim );
	    true
	  end;
	  match vdim with
	    | VAL_int i -> (ptopo#dimension ptopo#cpos) = i
	    | _         -> raise (Erreur "Ptopology: ppattern_link: eval_cguard: Waiting for an integer")
       )
	
    | CPred ("val",exp) -> (
	let vval = (pf1 exp) ptopo#get_val_of_rule_variable in
	  assert begin if !trace_opt then 
	    (printf "         |    eval_cguard: val = %a\n" Printer.prfvalue2 vval );
	    true
	  end;
	  Dispatch_sig.equal_mgs_value vval (ptopo#get_value_from_pos ptopo#cpos)
       )
	
    | CPred (id,exp) -> raise (Erreur "Ptopology: ppattern_link: eval_cguard: CPred NYI")

end

and ['position] select_ppattern
  (rank : int)
  (filtering : bool)
  (ppatid : int)
  (ppatname : string)
  (faces_id : int list)
  (faces_name : string list)
  (cofaces_id : int list)
  (cofaces_name : string list)
  (gl  : Terme.guards list)
  (cgl : Terme.cguards list)
  (next   : 'position ppattern) =
object(self)
  inherit ['position] ppattern_link filtering ppatid next


  val _faces   = List.combine faces_id faces_name
  val _cofaces = List.combine cofaces_id cofaces_name
  val _guards  = List.combine cgl gl

  method to_string = (if filtering then "" else "~")^ppatname^" ("^(string_of_int ppatid)^") - "^(next#to_string)
  method to_full_string = "rank "^(string_of_int rank)^": "^(if filtering then "" else "~")^ppatname
			  ^" ("^(string_of_int ppatid)^"):\n"
			  ^"\tfaces   : "^(List.fold_left
					     (fun acc (id,name) -> name^"("^(string_of_int id)^") "^acc)
					     ""
					     _faces
					  )^"\n"
			  ^"\tcofaces : "^(List.fold_left
					     (fun acc (id,name) -> name^"("^(string_of_int id)^") "^acc)
					     ""
					     _cofaces
					  )^"\n"
			  ^"\tguards  : "^(List.fold_left
					     (fun acc (cg,g) -> (self#string_of_cguard cg)^"("^(self#string_of_guard g)^") "^acc)
					     ""
					     _guards
					  )^"\n"
			  ^"-----------------------------------------------------------------------------\n"^(next#to_full_string)

  val tabul = String.make (2*rank) (' ')

  method filtre pul ptopo =
    let found_path = ref ([],[]) in

    let coucou_filter (lf,lcf) pos =
      assert begin if !trace_opt then 
	(printf "       %s\\_\n         %s|select: %s\n" tabul tabul (ptopo#prpos pos) );
	true
      end;
      if (ptopo#filterable pos) then (
	assert begin if !trace_opt then 
	  (printf "         %s|--> filterable; (what about the \"such as\" thing ???) \n" tabul );
	  true
	end ;
	ptopo#goto pos;
	ptopo#filter_current;
	ptopo#set_elt ppatid ptopo#cpos;

	if ( ( List.for_all (ptopo#isFaceOf pos) lf ) & ( List.for_all (ptopo#isCoFaceOf pos) lcf ) ) then (
	  assert begin if !trace_opt then 
	    (printf "         %s|--> \"such as\" thing checked; (what about guards ???) \n" tabul );
	    true
	  end ;
	  if (List.for_all (self#eval_cguard ptopo) cgl) then (
	    assert begin if !trace_opt then 
	      (printf "         %s|--> guards checked; (what about the next ppattern ???) \n" tabul );
	      true
	    end ;
	    match (next#filtre pul ptopo) with
		None   ->
		  assert begin if !trace_opt then 
		    (printf "       %s__|--> Next ppattern doesn't match...\n" tabul );
		    true
		  end ;
		  ptopo#unset_elt ppatid ; ptopo#unfilter pos
	      | Some (p_f,p_u) ->
		  assert begin if !trace_opt then 
		    (printf "       %s__|--> A patch is found\n" tabul );
		    true
		  end ;
		  let more_p_u = ptopo#use_transitive_faces_cofaces pos in
		    found_path := ((pos::p_f),more_p_u@p_u) ; raise FoundPath
	  ) else (
	    assert begin if !trace_opt then 
	      (printf "       %s__|---> guards aren't verified... \n" tabul );
	      true
	    end ;
	    ptopo#unset_elt ppatid ;
	    ptopo#unfilter pos;
	  )
	) else (
	  assert begin if !trace_opt then 
	    (printf "       %s__|---> \"such as\" thing not verified...\n" tabul );
	    true
	  end ;
	  ptopo#unset_elt ppatid ;
	  ptopo#unfilter pos;
	)
      ) else (
	assert begin if !trace_opt then 
	  (printf "       %s__|---> not filterable...\n" tabul );
	  true
	end
      )
    in
    let coucou_use (lf,lcf) pos =
      assert begin if !trace_opt then 
	(printf "       %s\\_\n         %s|select: %s\n" tabul tabul (ptopo#prpos pos) );
	true
      end;
      if ((ptopo#usable pos) & (not (List.mem pos pul))) then (
	assert begin if !trace_opt then 
	  (printf "         %s|--> usable; (what about the \"such as\" thing ???) \n" tabul );
	  true
	end ;
	ptopo#goto pos;
	ptopo#use_current;
	ptopo#set_elt ppatid ptopo#cpos;

	if ( ( List.for_all (ptopo#isFaceOf pos) lf ) & ( List.for_all (ptopo#isCoFaceOf pos) lcf ) ) then (
	  assert begin if !trace_opt then 
	    (printf "         %s|--> \"such as\" thing checked; (what about guards ???) \n" tabul );
	    true
	  end ;
	  if (List.for_all (self#eval_cguard ptopo) cgl) then (
	    assert begin if !trace_opt then 
	      (printf "         %s|--> guards checked; (what about the next ppattern ???) \n" tabul );
	      true
	    end ;
	    match (next#filtre (pos::pul) ptopo) with
		None   ->
		  assert begin if !trace_opt then 
		    (printf "       %s__|--> Next ppattern doesn't match... \n" tabul );
		    true
		  end ;
		  ptopo#unset_elt ppatid ; ptopo#unuse pos pos
	      | Some (p_f,p_u) ->
		  assert begin if !trace_opt then 
		    (printf "       %s__|--> A patch is found\n" tabul );
		    true
		  end ;
		  found_path := (p_f,(pos::p_u)) ; raise FoundPath
	  ) else (
	    assert begin if !trace_opt then 
	      (printf "       %s__|---> guards aren't verified...\n" tabul );
	      true
	    end ;
	    ptopo#unset_elt ppatid ;
	    ptopo#unuse pos pos;
	  )
	) else (
	  assert begin if !trace_opt then 
	    (printf "       %s__|---> \"such as\" thing not verified...\n" tabul );
	    true
	  end ;
	  ptopo#unset_elt ppatid ;
	  ptopo#unuse pos pos;
	)
      ) else (
	assert begin if !trace_opt then 
	  (printf "       %s__|---> not usable...\n" tabul );
	  true
	end
      )
    in
    let coucou = if (filtering) then coucou_filter else coucou_use in
      try
	(match (List.map ptopo#get_elt faces_id,List.map ptopo#get_elt cofaces_id) with
	   | [],[]      -> ptopo#iter_all (coucou ([],[]))
	   | f::lf,lcf  -> ptopo#iter_cofaces (coucou (lf,lcf)) f
	   | lf,cf::lcf -> ptopo#iter_faces (coucou (lf,lcf)) cf
	) ;
	None
      with FoundPath -> Some (!found_path)

end

and ['position] end_ppattern =
object(self)
  inherit ['position] ppattern

  method to_string = "<end ppattern>"
  method to_full_string = ""
  method filtre pul ptopo =
    assert begin if !trace_opt then (
      (printf "*******\n" );
      ptopo#print_env;
      (printf "*******\n" ));
      true
    end;
    Some ([],[])


  method use_neighborhood ptopo = ()

end





















(*
and ['position] indpt_ppattern
  (filtering : bool)
  (ppatid : int)
  (ppatname : string)
  (gl  : Terme.guards list)
  (cgl : Terme.cguards list)
  (next   : 'position ppattern) =
object(self)
  inherit ['position] ppattern_link filtering ppatid next

  method to_string = (if filtering then "" else "~")^ppatname^" ("^(string_of_int ppatid)^") - "^(next#to_string)
  method to_full_string = (if filtering then "" else "~")^ppatname^" ("^(string_of_int ppatid)^"):"
			  ^" indepdendant...\n"
			  ^"\tguards  : "^(List.fold_left
					     (fun acc (cg,g) -> (self#string_of_cguard cg)^"("^(self#string_of_guard g)^") "^acc)
					     ""
					     (List.combine cgl gl)
					  )^"\n"
			  ^"-----------------------------------------------------------------------------\n"^(next#to_full_string)

  method filtre pul ptopo =
    begin
      assert begin if !trace_opt then 
	(printf "    ~~ indpt_ppattern: %s\n" self#to_string ; flush stdout); 
	true 
      end;

      if (filtering) then (
	ptopo#goto ptopo#next_filterable;
	ptopo#visit_current;
	ptopo#filter_current;
	ptopo#set_elt ppatid ptopo#cpos;

	if (List.for_all (self#eval_cguard ptopo) cgl) then (
	  match (next#filtre pul ptopo) with
	      None   -> ptopo#unset_elt ppatid ; ptopo#unfilter (ptopo#cpos) ; None
	    | Some p -> Some ((ptopo#cpos)::p)
	) else (
	  ptopo#unset_elt ppatid ;
	  ptopo#unfilter (ptopo#cpos) ;
	  None
	)

      ) else (
	ptopo#goto ptopo#next_usable;
	ptopo#use_current ;
	ptopo#set_elt ppatid ptopo#cpos;

	if (List.for_all (self#eval_cguard ptopo) cgl) then (
	  match (next#filtre pul ptopo) with
	      None   -> ptopo#unset_elt ppatid ; ptopo#unuse (ptopo#cpos) (ptopo#cpos) ; ptopo#visit (ptopo#cpos) ; None
	    | Some p -> Some ((ptopo#cpos)::p)
	) else (
	  ptopo#unset_elt ppatid ;
	  ptopo#unuse (ptopo#cpos) (ptopo#cpos);
	  ptopo#visit (ptopo#cpos);
	  None
	)

      )

    end

end

and ['position] dpt_ppattern
  (filtering : bool)
  (ppatid : int)
  (ppatname : string)
  (faces_id : int list)
  (faces_name : string list)
  (tfaces_id : int list)
  (tfaces_name : string list)
  (cofaces_id : int list)
  (cofaces_name : string list)
  (tcofaces_id : int list)
  (tcofaces_name : string list)
  (gl  : Terme.guards list)
  (cgl : Terme.cguards list)
  (next   : 'position ppattern) =
object(self)
  inherit ['position] ppattern_link filtering ppatid next


  val _faces  = List.combine faces_id faces_name
  val _tfaces = List.combine tfaces_id tfaces_name
  val _cofaces  = List.combine cofaces_id cofaces_name
  val _tcofaces = List.combine tcofaces_id tcofaces_name

  method to_string = (if filtering then "" else "~")^ppatname^" ("^(string_of_int ppatid)^") - "^(next#to_string)
  method to_full_string = (if filtering then "" else "~")^ppatname^" ("^(string_of_int ppatid)^"):\n"
			  ^"\tfaces   : "^(List.fold_left
					     (fun acc (id,name) -> name^"("^(string_of_int id)^") "^acc)
					     ""
					     (List.combine faces_id faces_name)
					  )^"\n"
			  ^"\ttfaces  : "^(List.fold_left
					     (fun acc (id,name) -> name^"("^(string_of_int id)^") "^acc)
					     ""
					     (List.combine tfaces_id tfaces_name)
					  )^"\n"
			  ^"\tcofaces : "^(List.fold_left
					     (fun acc (id,name) -> name^"("^(string_of_int id)^") "^acc)
					     ""
					     (List.combine cofaces_id cofaces_name)
					  )^"\n"
			  ^"\ttcofaces: "^(List.fold_left
					     (fun acc (id,name) -> name^"("^(string_of_int id)^") "^acc)
					     ""
					     (List.combine tcofaces_id tcofaces_name)
					  )^"\n"
			  ^"\tguards  : "^(List.fold_left
					     (fun acc (cg,g) -> (self#string_of_cguard cg)^"("^(self#string_of_guard g)^") "^acc)
					     ""
					     (List.combine cgl gl)
					  )^"\n"
			  ^"-----------------------------------------------------------------------------\n"^(next#to_full_string)

  method instancie_incident_positions (ptopo : 'position generic_ptopo) =
    begin
      let instancied_faces = List.map ptopo#get_elt faces_id       in
      let instancied_cofaces = List.map ptopo#get_elt cofaces_id   in
      let instancied_tfaces = List.map ptopo#get_elt tfaces_id     in
      let instancied_tcofaces = List.map ptopo#get_elt tcofaces_id in
	ptopo#common_position instancied_faces instancied_cofaces instancied_tfaces instancied_tcofaces
    end

  method filtre pul ptopo =
    begin
      assert begin if !trace_opt then 
	(printf "    ~~ dpt_ppattern: %s\n" self#to_string ; flush stdout ; ptopo#print_env) ;
	true 
      end;

      let ipos = self#instancie_incident_positions ptopo in
      let lth  = Array.length ipos in
      let found_path = ref [] in
	assert begin if !trace_opt then 
	  (printf "       incident neighbors : %d\n       < " lth;
	   Array.iter (fun p -> printf "%s" ((ptopo#prpos p)^" ")) ipos;
	   printf ">\n"
	  );
	  true
	end;
	try (
	  for i = 0 to (lth - 1) do
	    assert begin if !trace_opt then 
	      (printf "       try with : %s\n" (ptopo#prpos (ipos.(i))) );
	      true
	    end;
	    if (filtering) then (
	      if (ptopo#filterable (ipos.(i))) then (
 		ptopo#goto (ipos.(i));
		ptopo#filter_current;
		ptopo#set_elt ppatid ptopo#cpos;
		
		if (List.for_all (self#eval_cguard ptopo) cgl) then (
		  match (next#filtre pul ptopo) with
		      None   -> ptopo#unset_elt ppatid ; ptopo#unfilter (ipos.(i))
		    | Some p -> found_path := p ; raise FoundPath

		) else (
		  ptopo#unset_elt ppatid ;
		  ptopo#unfilter (ipos.(i)) ;
		)
	      ) else ()
	    ) else (
	      if (ptopo#filterable (ipos.(i))) then (
		ptopo#goto (ipos.(i));
		ptopo#use_current ;
		ptopo#set_elt ppatid ptopo#cpos;
	    
		if (List.for_all (self#eval_cguard ptopo) cgl) then (
		  match (next#filtre pul ptopo) with
		      None   -> ptopo#unset_elt ppatid ; ptopo#unuse (ipos.(i)) (ipos.(i))
		    | Some p -> found_path := p ; raise FoundPath
		) else (
		  ptopo#unset_elt ppatid ;
		  ptopo#unuse (ipos.(i)) (ipos.(i));
		)
	      ) else ()
		
	    )

	  done ;
	  None
	) with FoundPath -> Some ((ptopo#cpos)::(!found_path))

    end

end

*)
;;
