(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: core_qmf.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2007-06-26 01:46:41 $
 * Version: $Revision: 1.11 $
 * -------------------------------------------------------------------
 *
 * Status: Cf. Qmf/def_qmf.ml pour les infos sur les types qmf et ncell
 *          
 *
 *=========================================================================
 *)


open Global;;
open Def_qmf;;
open Dispatch_sig;;

__register_version "$RCSfile: core_qmf.ml,v $" "$Revision: 1.11 $" "$Date: 2007-06-26 01:46:41 $";;


(*** functions from stubs ***)

(** fonctions de cration et de destruction **)

external create_gmap         : unit -> gmap                = "ml_create_gmap"
external delete_gmap         : gmap -> unit                = "ml_delete_gmap"
external copy_gmap           : gmap -> gmap                = "ml_copy_gmap"
external copy_notmarked_dart : gmap -> gmap -> int -> int  = "ml_copy_notmarked_dart"
external isFrozen            : gmap -> bool                = "ml_is_frozen"
external isNotFrozen         : gmap -> bool                = "ml_is_not_frozen"

(** manipulation des darts **)

external dart_null        : unit -> dart                            = "ml_dart_null"
external newDart          : gmap -> float -> float -> float -> dart = "ml_new_dart"
external copyDart         : gmap -> dart -> dart                    = "ml_copy_dart"
external alpha            : gmap -> dart -> int -> dart             = "ml_alpha"
external sew              : gmap -> dart -> dart -> int -> unit     = "ml_sew"
external isFree           : gmap -> dart -> int -> bool             = "ml_is_free"
external getNewMark       : gmap -> int                             = "ml_get_new_mark"
external freeMark         : gmap -> int -> unit                     = "ml_free_mark"
external markDart         : gmap -> dart -> int ->unit              = "ml_mark_dart"
external markNcell        : ncell -> int -> unit                    = "ml_mark_ncell"
external unmarkDart       : gmap -> dart -> int ->unit              = "ml_unmark_dart"
external unmarkNcell      : ncell -> int -> unit                    = "ml_unmark_ncell"
external isMarkedDart     : gmap -> dart -> int -> bool             = "ml_is_marked_dart"
external isMarkedNcell    : ncell -> int -> bool                    = "ml_is_marked_ncell"
external getNewDirectInfo : gmap -> int                             = "ml_get_new_directinfo"
external freeDirectInfo   : gmap -> int -> unit                     = "ml_free_directinfo"
external setDirectInfo    : gmap -> dart -> int -> dart -> unit     = "ml_set_directinfo"
external getDirectInfo    : gmap -> dart -> int -> dart             = "ml_get_directinfo"
external check_gmap       : gmap -> bool                            = "ml_check_gmap"
external compare_ncells   : ncell -> ncell -> int                   = "ml_compare_ncells"
external all_coverage     : gmap -> coverage                        = "ml_all_coverage"
external ncell_coverage   : ncell -> coverage                       = "ml_ncell_coverage"
external delete_coverage  : coverage -> unit                        = "ml_delete_coverage"
external reinit_coverage  : coverage -> unit                        = "ml_reinit_coverage"
external current_dart     : coverage -> dart                        = "ml_get_current_dart"
external cont_coverage    : coverage -> bool                        = "ml_cont_coverage"
external next_dart        : coverage -> unit                        = "ml_next_dart"

(** construction de cellules... **)
external new_ncell                 : gmap -> int -> ncell                  = "ml_new_ncell"
external new_ncell_polygon         : gmap -> int -> ncell                  = "ml_new_polygon"
external new_ncell_sphere          : gmap -> int -> int -> (ncell * ncell) = "ml_new_sphere"
external new_ncell_polyline        : gmap -> int -> ncell                  = "ml_new_polyline"
external new_ncell_square          : gmap -> int -> int -> int -> ncell    = "ml_new_square"


external insert_vertex             : ncell -> ncell array                                                    = "ml_insert_vertex"
external insert_edge               : ncell -> ncell -> ncell array                                           = "ml_insert_edge"
external insert_face               : ncell -> ncell array -> int -> ncell array                              = "ml_insert_face"
external create_edge               : ncell -> ncell -> ncell array                                           = "ml_create_edge"
external create_face               : (ncell * ncell) array -> int-> ncell array                              = "ml_create_face"
external create_volume             : (ncell array) -> int -> (ncell array) array -> int array -> int -> unit = "ml_create_volume"
external remove_ncell              : ncell -> ncell                                                          = "ml_remove_ncell"
external delete_ncell              : ncell -> ncell array                                                    = "ml_delete_ncell"
external glue2FacesBy4Verteces     : ((ncell*ncell)*(ncell*ncell)*(ncell*ncell)) -> unit                     = "ml_glue2FacesBy4Verteces"



external get_coord_from_vertex : ncell -> float array = "ml_get_coord_from_vertex"
external set_coord_of_vertex : ncell -> float -> float -> float -> unit = "ml_set_coord_of_vertex"

external load_moka : gmap -> string -> ncell array = "ml_load_moka"
external save_moka : gmap -> string -> unit        = "ml_save_moka"

external normal_form : ncell -> ncell = "ml_normal_form"

(* --- Accesseurs ----------------------------------------------------- *)

let rec get_dim_from_ncell  (_,_,d) = d

and get_dart_from_ncell (_,d,_) = d

and get_gmap_from_ncell (g,_,_) = g

and string_of_gmap g = string_of_int (Hashtbl.hash g)

and string_of_ncell (g,d,n) =
  "Ncell:("^(string_of_int (Hashtbl.hash g))
  ^","^(string_of_int (Hashtbl.hash d))
  ^","^(string_of_int n)
  ^",NF="^(string_of_int (Hashtbl.hash (let (_,d',_) = normal_form (g,d,n) in d') ) )
  ^",f="^(string_of_int (List.length (faces (g,d,n)) ))
  ^",cof="^(string_of_int (List.length (cofaces (g,d,n)) ))
  ^")"

(* --- Cration ------------------------------------------------------- *)

(* --- Manipulations ------------------------------------------------- *)

and new_vertex g = new_ncell g 0
and new_edge g = new_ncell g 1
and new_face g = new_ncell g 2
and new_volume g = new_ncell g 3

(* --- Requetes ------------------------------------------------------ *)

and faces ((_,_,d) as c) = match d with
    0 -> []
  | _ -> icells c (d-1)

and cofaces ((_,_,d) as c) = match d with
    3 -> []
  | _ -> icells c (d+1)

and icells ((g,d,n) as c) i =
  let mark = getNewMark g in
  let cov = ncell_coverage c in
  let ret = ref [] in
    while (cont_coverage cov) do
      let curr = current_dart cov in
	if (not (isMarkedDart g curr mark)) then (
	  ret := (g,curr,i)::(!ret) ;
	  markNcell (g,curr,i) mark
	);
	next_dart cov
    done;
    List.iter (fun c -> unmarkNcell c mark) (!ret);
    delete_coverage cov;
    freeMark g mark;
    !ret

and ccells ((g,d,n) as c) i =
  let rouge = getNewMark g in
  let bleue = getNewMark g in
  let tmp = ref [] in
  let ret = ref [] in
    markNcell c rouge;
    let covRed = ncell_coverage c in
      while (cont_coverage covRed) do
	let vncell = current_dart covRed in

	  if (not (isMarkedDart g vncell bleue)) then (
	    tmp := (g,vncell,i)::(!tmp);
	    markNcell (g,vncell,i) bleue;

	    let covBlue = ncell_coverage (g,vncell,i) in

	      while (cont_coverage covBlue) do
		let closeNcell = current_dart covBlue in

		  if (not (isMarkedDart g closeNcell rouge)) then (
		    ret := (g,closeNcell,n)::(!ret);
		    markNcell (g,closeNcell,n) rouge;
		  );

		  next_dart covBlue
	      done ;
	      
	      delete_coverage covBlue

	  );

	  next_dart covRed
      done ;

      delete_coverage covRed;

      List.iter (fun c -> unmarkNcell c bleue) (!tmp);
      List.iter (fun c -> unmarkNcell c rouge) (!ret);
      unmarkNcell c rouge;
      unmarkNcell c bleue;
      freeMark g rouge;
      freeMark g bleue;
      !ret

and transitive_faces ((g,d,n) as c) =
  let ret = ref [] in
    for i = n-1 downto 0 do
      ret := (!ret)@(icells c i)
    done ;
    !ret

and transitive_cofaces ((g,d,n) as c) =
  let ret = ref [] in
    for i = n+1 to 3 do
      ret := (!ret)@(icells c i)
    done ;
    !ret

and common_faces c1 c2 =
  let f1 = faces c1
  and f2 = faces c2 in
  let (g,_,_) = c1 in
  let m = getNewMark g in
    List.iter (fun c -> markNcell c m) f1;
    let ret = List.filter (fun c -> isMarkedNcell c m) f2 in
      List.iter (fun c -> unmarkNcell c m) f1;
      freeMark g m;
      ret

and common_cofaces c1 c2 =
  let f1 = cofaces c1
  and f2 = cofaces c2 in
  let (g,_,_) = c1 in
  let m = getNewMark g in
    List.iter (fun c -> markNcell c m) f1;
    let ret = List.filter (fun c -> isMarkedNcell c m) f2 in
      List.iter (fun c -> unmarkNcell c m) f1;
      freeMark g m;
      ret

and common_transitive_faces c1 c2 =
  let f1 = transitive_faces c1
  and f2 = transitive_faces c2 in
  let (g,_,_) = c1 in
  let m = getNewMark g in
    List.iter (fun c -> markNcell c m) f1;
    let ret = List.filter (fun c -> isMarkedNcell c m) f2 in
      List.iter (fun c -> unmarkNcell c m) f1;
      freeMark g m;
      ret

and common_transitive_cofaces c1 c2 =
  let f1 = transitive_cofaces c1
  and f2 = transitive_cofaces c2 in
  let (g,_,_) = c1 in
  let m = getNewMark g in
    List.iter (fun c -> markNcell c m) f1;
    let ret = List.filter (fun c -> isMarkedNcell c m) f2 in
      List.iter (fun c -> unmarkNcell c m) f1;
      freeMark g m;
      ret

;;




(**************************************************************************
 ** Fonctions de manipulation des qmfs
 **************************************************************************)

external qmfy        : gmap -> unit   = "ml_qmfy"
external norm        : gmap -> unit   = "ml_norm"

 (** valeur qu'on affecte par defaut aux simplexes qu'on cree 
  ** pour assurer la coherence de la structure de complexe.
  **)
let default_value = Valeur.mgsundef
;;



(* --- Manipulation de qmfs ---------------------------------------------------------- *)
exception Cut ;;

let map fct (((g,ha,d) as q):'a qmf) =
  let ha' = Array.copy ha in
    for i = 0 to ((Array.length ha')-1) do
      let ret = Hash.copy ha'.(i) in
	Hash.iter (fun cp v -> Hash.replace ret cp (fct cp v)) ret;
	ha'.(i) <- ret
    done ;
    (g,ha',d)

let iter fct (((g,ha,d) as q):'a qmf) =
  for i=0 to d do
    Hash.iter (fun c v -> fct c v) (ha.(i))
  done

let iter_d dim fct (((g,ha,d) as q):'a qmf) =
  if (dim <= d) then
    Hash.iter (fun c v -> fct c v) (ha.(dim))

let fold fct (((g,ha,d) as q):'a qmf) zero =
  let acc = ref zero in
    for i = 0 to d do
      acc := Hash.fold (fun c v suite -> fct c v suite) (ha.(i)) (!acc)
    done;
    !acc

let fold_d dim fct (((g,ha,d) as q):'a qmf) zero =
  if (dim <= d) then
    Hash.fold (fun c v suite -> fct c v suite) (ha.(dim)) zero
  else
    zero

let forall fct (((g,ha,d) as q):'a qmf) =
  let acc = ref true in
    for i = 0 to d do
      acc := (Hash.forall (fun c v -> fct c v) (ha.(i))) && (!acc)
    done;
    !acc

let forall_d dim fct (((g,ha,d) as q):'a qmf) =
  if (dim <= d) then
    (Hash.forall (fun c v -> fct c v) (ha.(dim))) && true
  else
    true

let exists fct (((g,ha,d) as q):'a qmf) =
  try
    for i = 0 to d do
      if (Hash.exists (fun c v -> fct c v) (ha.(i))) then raise Cut
    done;
    false
  with
    | Cut -> true

let exists_d dim fct (((g,ha,d) as q):'a qmf) =
  if (dim <= d) then
    Hash.exists (fun c v -> fct c v) (ha.(dim))
  else
    false

let size (((g,ha,d) as q):'a qmf) =
  let nb = ref 0 in
    for i = 0 to d do
      nb := !nb + Hash.size ha.(i)
    done;
    !nb

let size_d dim (((g,ha,d) as q):'a qmf) =
  if (dim <= d) then Hash.size ha.(dim) else 0

let find (((g,ha,d) as q):'a qmf) c =
  let dim = get_dim_from_ncell c in
    if (dim > d) then
      raise Not_found
    else
      Hash.find ha.(dim) (normal_form c)

let nth n (((g,ha,d) as q):'a qmf) =
  let i = ref 0 in
  let pos = ref n in
    try
      while ((!i)<=d) do
	let s = Hash.size ha.(!i) in
	  if ( (!pos) >= s) then
	    begin
	      pos := (!pos) - s ;
              i   := !i + 1
	    end
	  else
	    raise Cut
      done ;
      failwith "internal error on nth in core_qmf"
    with
	Cut -> Hash.nth (!pos) (ha.(!i))

let nth_d dim n (((g,ha,d) as q):'a qmf) =
  if (dim <= d) then Hash.nth n (ha.(dim)) else raise (Erreur "core_qmf: nth_d: dimension is too big...")

let remove (((g,ha,d) as q):'a qmf) c =
  if ( (get_dim_from_ncell c) <= d ) then
    Hash.remove (ha.(get_dim_from_ncell c)) (normal_form c)

let copy (((g,ha,d) as q):'a qmf) =
  let ha' = Array.init (Array.length ha) (fun i -> Hash.copy (ha.(i))) in
    (g,ha',d)

let hd q = nth 0 q

let last q = nth ((size q) - 1) q

let tl q =
  let ret = copy q in
  let k, v = hd ret
  in
    remove ret k;
    ret

let mem ((g,ha,d):'a qmf) c =
  let dim = get_dim_from_ncell c in
    if (dim > d) then
      false
    else
      Hash.mem ha.(dim) (normal_form c)

let compare valcmp (((g1,ha1,d1) as q1):'a qmf) (((g2,ha2,d2) as q2):'b qmf) =
  let gdiff = Pervasives.compare g1 g2 in
    if gdiff <> 0 then
      gdiff
    else
      let ddiff = Pervasives.compare d1 d2 in
	if ddiff <> 0 then
	  ddiff
	else
	  begin
	    let diff = ref 0 in
	      try
		for i = 0 to d1 do
		  diff := Hash.compare valcmp ha1.(i)  ha2.(i);
		  if ( !diff <> 0) then raise Cut
		done ;
		0
	      with Cut -> !diff
	  end
	
let to_list (((g,ha,d) as q):'a qmf) =
  fold (fun k v s -> (k, v)::s) q []
  

(** dimension **)
let get_dim_from_qmf ((g,ha,d):'a qmf) = d

let get_gmap_from_qmf ((g,ha,d):'a qmf) = g

and get_ncells_of_dim_in_qmf dim ((g,ha,d):'a qmf) =
  if (dim <= d) then 
    ha.(dim)
  else
    raise (Erreur "core_qmf: get_ncells_of_dim: dimension is too big...")

and set_ncells_of_dim_in_qmf dim ((g,ha,d):'a qmf) h = (** !!! Il fo que les cellules de h soit normalises !!! **)
  if (dim <= d) then 
    ha.(dim) <- h
  else
    raise (Erreur "core_qmf: set_ncells_of_dim: dimension is too big...")

and get_ncell ((g,ha,d):'a qmf) c =
  if ((get_dim_from_ncell c) > d) then
    raise Not_found
  else
    let _ = (Hash.find (ha.(get_dim_from_ncell c)) (normal_form c)) in (normal_form c) (* ???? *)

and get_value_from_ncell_in_qmf ((g,ha,d):'a qmf) ncell =
  if ((get_dim_from_ncell ncell) > d) then
    raise Not_found
  else
    (Hash.find (ha.(get_dim_from_ncell ncell)) (normal_form ncell))

and get_ncell_from_val_in_qmf ((g,ha,d):'a qmf) v =
  for i = 0 to d do
    Hash.fold (fun nc vv acc -> if v = vv then nc::acc else acc) (ha.(i)) []
  done


(** Renvoie un tableau dont l'lment i contient 
 ** la liste des ncells de dimension i d'une qmf 
 **)
let skelette ((g,ha,d):'a qmf) =
  Array.map (fun h -> Hash.to_list h) ha

let faces_qmf ((g,ha,d):'a qmf) nc =
  if ((d>=(get_dim_from_ncell nc))&
      (Hash.mem (ha.(get_dim_from_ncell nc)) (normal_form nc))) then (* vrifie que la ncell est dans la qmf *)
    let fl = faces nc
    in List.map (fun fp -> (normal_form fp),(Hash.find (ha.(get_dim_from_ncell fp)) (normal_form fp))) fl
  else
    let err = "Error: Core_qmf.faces_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)

let cofaces_qmf ((g,ha,d):'a qmf) nc =
  if ((d>=(get_dim_from_ncell nc))&
      (Hash.mem (ha.(get_dim_from_ncell nc)) (normal_form nc))) then (* vrifie que la ncell est dans la qmf *)
    let cl = cofaces nc
    in List.fold_left
	 (fun acc c ->
	    try ((normal_form c),Hash.find (ha.(get_dim_from_ncell c)) (normal_form c))::acc
	    with Not_found -> acc | Invalid_argument "Array.get" -> acc)
	 []
	 cl
  else
    let err = "Error: Core_qmf.faces_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)

let icells_qmf ((g,ha,d):'a qmf) nc i =
  if ((d>=(get_dim_from_ncell nc))&
      (Hash.mem (ha.(get_dim_from_ncell nc)) (normal_form nc))) then
    let l = icells nc i
    in List.fold_left
	 (fun acc c ->
	    try ((normal_form c),Hash.find (ha.(get_dim_from_ncell c)) (normal_form c))::acc
	    with Not_found -> acc | Invalid_argument "Array.get" -> acc)
	 []
	 l
  else
    let err = "Error: Core_qmf.icells_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)

let ccells_qmf ((g,ha,d):'a qmf) nc i =
  if ((d>=(get_dim_from_ncell nc))&
      (Hash.mem (ha.(get_dim_from_ncell nc)) (normal_form nc))) then
    let l = ccells nc i
    in List.fold_left
	 (fun acc c ->
	    try ((normal_form c),Hash.find (ha.(get_dim_from_ncell c)) (normal_form c))::acc
	    with Not_found -> acc | Invalid_argument "Array.get" -> acc)
	 []
	 l
  else
    let err = "Error: Core_qmf.ccells_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)

let transitive_faces_qmf ((g,ha,d):'a qmf) c =
  if ((d>=(get_dim_from_ncell c))&
      (Hash.mem (ha.(get_dim_from_ncell c)) (normal_form c))) then (* vrifie que la ncell est dans la qmf *)
    let fl = transitive_faces c
    in List.map (fun fp -> (normal_form fp),(Hash.find (ha.(get_dim_from_ncell fp)) (normal_form fp))) fl
  else
    let err = "Error: Core_qmf.transitive_faces_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)
 
let transitive_cofaces_qmf ((g,ha,d):'a qmf) c =
  if ((d>=(get_dim_from_ncell c))&
      (Hash.mem (ha.(get_dim_from_ncell c)) (normal_form c))) then (* vrifie que la ncell est dans la qmf *)
    let cl = transitive_cofaces c
    in List.fold_left
	 (fun acc c ->
	    try ((normal_form c),Hash.find (ha.(get_dim_from_ncell c)) (normal_form c))::acc
	    with Not_found -> acc | Invalid_argument "Array.get" -> acc)
	 []
	 cl
  else
    let err = "Error: Core_qmf.transitive_faces_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)

let common_faces_qmf ((g,ha,d):'a qmf) c1 c2 =
  if ((d>=(get_dim_from_ncell c1))&
      (Hash.mem (ha.(get_dim_from_ncell c1)) (normal_form c1))&&
      (d>=(get_dim_from_ncell c2))&
      (Hash.mem (ha.(get_dim_from_ncell c2)) (normal_form c2))) then
    let l = common_faces c1 c2
    in List.map (fun p -> (normal_form p),(Hash.find (ha.(get_dim_from_ncell p)) (normal_form p))) l
  else
    let err = "Error: Core_qmf.common_faces_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)

let common_cofaces_qmf ((g,ha,d):'a qmf) c1 c2 =
  if ((d>=(get_dim_from_ncell c1))&
      (Hash.mem (ha.(get_dim_from_ncell c1)) (normal_form c1))&&
      (d>=(get_dim_from_ncell c2))&
      (Hash.mem (ha.(get_dim_from_ncell c2)) (normal_form c2))) then
    let l = common_cofaces c1 c2
    in List.fold_left
	 (fun acc c ->
	    try ((normal_form c),Hash.find (ha.(get_dim_from_ncell c)) (normal_form c))::acc
	    with Not_found -> acc | Invalid_argument "Array.get" -> acc)
	 []
	 l
  else
    let err = "Error: Core_qmf.common_cofaces_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)

let common_transitive_faces_qmf ((g,ha,d):'a qmf) c1 c2 =
  if ((d>=(get_dim_from_ncell c1))&
      (Hash.mem (ha.(get_dim_from_ncell c1)) (normal_form c1))&&
      (d>=(get_dim_from_ncell c2))&
      (Hash.mem (ha.(get_dim_from_ncell c2)) (normal_form c2))) then
    let l = common_transitive_faces c1 c2
    in List.map (fun p -> normal_form p,(Hash.find (ha.(get_dim_from_ncell p)) (normal_form p))) l
  else
    let err = "Error: Core_qmf.common_transitive_faces_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)

let common_transitive_cofaces_qmf ((g,ha,d):'a qmf) c1 c2 =
  if ((d>=(get_dim_from_ncell c1))&
      (Hash.mem (ha.(get_dim_from_ncell c1)) (normal_form c1))&&
      (d>=(get_dim_from_ncell c2))&
      (Hash.mem (ha.(get_dim_from_ncell c2)) (normal_form c2))) then
    let l = common_transitive_cofaces c1 c2
    in List.fold_left
	 (fun acc c ->
	    try (normal_form c,Hash.find (ha.(get_dim_from_ncell c)) (normal_form c))::acc
	    with Not_found -> acc | Invalid_argument "Array.get" -> acc)
	 []
	 l
  else
    let err = "Error: Core_qmf.common_transitive_cofaces_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)

(* --- Construction de qmf ---------------------------------------------------------- *)

let new_qmf gmap = qmfy gmap ; (gmap,[|Hash.create 11;Hash.create 11;Hash.create 11;Hash.create 11|],3)

let create_qmf gmap dim = qmfy gmap ; (gmap,(Array.init (dim+1) (fun _ -> Hash.create 11)),dim)

let extends_to_dim ((g,ha,d):'a qmf) dim =
  if (dim <= d) then
    (g,ha,d)
  else
    if (dim > ((Array.length ha)-1)) then
      (g,Array.append ha (Array.init (dim-(Array.length ha)+1) (fun _ -> Hash.create 11)),dim)
    else
      (g,ha,dim)

(** --- une somme --------------- **)

(** on ajoute un couple (ncell,valeur)
 ** On force l'ajout de ncell (quitte  mettre  jour la valeur associe)
 ** et on ajoute les faces de ncell. Par contre, on ne force pas cet ajout
 ** car des valeurs ont peut tre dj t associes aux faces.
 **)

let add ((g,ha,d):'a qmf) c v =
  if (get_dim_from_ncell c > d) then
    raise (Erreur "Core_qmf.add: the dimension of the qmf is lower than the dimension of the ncell")
  else
    Hash.replace (ha.(get_dim_from_ncell c)) (normal_form c) v

let replace ((g,ha,d):'a qmf) c v =
  if (get_dim_from_ncell c > d) then
    raise (Erreur "Core_qmf.replace: the dimension of the qmf is lower than the dimension of the ncell")
  else
    Hash.replace (ha.(get_dim_from_ncell c)) (normal_form c) v

let replace_force ((g,ha,d):'a qmf) c v =
  if (get_dim_from_ncell c > d) then
    raise (Erreur "Core_qmf.replace_force: the dimension of the qmf is lower than the dimension of the ncell")
  else
    Hash.replace (ha.(get_dim_from_ncell c)) c v

let rec add_ncell (((g,ha,d) as q):'a qmf) c v =
  begin
    let ((g',ha',d') as q') = extends_to_dim q (get_dim_from_ncell c) in
    List.iter (fun f -> add_one_ncell q' (normal_form f) default_value) (transitive_faces c);
    add_force_one_ncell q' c v;
    q'
  end

and add_one_ncell (((g,ha,d) as q):'a qmf) c v = 
  if not(Hash.mem (ha.(get_dim_from_ncell c)) (normal_form c)) then (add_force_one_ncell q c v)

and add_force_one_ncell ((g,ha,d):'a qmf) c v = 
  Hash.replace (ha.(get_dim_from_ncell c)) (normal_form c) v
;;


(** rajoute une ncell ou bien combine la valeur si la ncell est deja present **)
let update_qmf (((g,ha,d) as q):'a qmf) combine v ((g',_,d') as c) =
  if (g = g') then
    let c' = normal_form c in
    try
      let v0 = if (d'>d) then (raise Not_found) else (Hash.find (ha.(d')) c')
      in Hash.replace (ha.(d')) c' (combine v0 v)
    with Not_found -> ignore(add_ncell q c' v)
  else
    raise (Erreur "Core_qmf.update_qmf: qmf and ncell are not in the same gmap")
;;
