(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: core_qmf.old.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2005-01-13 00:59:54 $
 * Version: $Revision: 1.1 $
 * -------------------------------------------------------------------
 *
 * Status: Cf. Qmf/def_qmf.ml pour les infos sur les types qmf et ncell
 *          
 *
 *=========================================================================
 *)


open Global;;
open Def_qmf;;
open Valeur;;
open Dispatch_sig;;

__register_version "$RCSfile: core_qmf.old.ml,v $" "$Revision: 1.1 $" "$Date: 2005-01-13 00:59:54 $";;


(*** fonctions de cration et de destruction ***)

external create_gmap     : unit -> gmap                  = "ml_create_gmap"
external delete_gmap     : gmap -> unit                  = "ml_delete_gmap"
external copy_gmap       : gmap -> gmap                  = "ml_copy_gmap"
external default_ncell   : gmap -> ncell                 = "ml_default_ncell"

(*** fonctions de comparaison ***)

let compare_ncell (c1,c2) = Pervasives.compare c1 c2 ;;
let compare_gmap (g1,g2) = Pervasives.compare g1 g2 ;;
let compare_dart (d1,d2) = Pervasives.compare d1 d2 ;;

(*** fonctions de manipulation des brins et des cellules ***)

external normal_form     : ncell -> ncell                 = "ml_normal_form"
external print_ncell     : ncell -> unit                  = "ml_print_ncell"
external string_of_gmap  : gmap  -> string                = "ml_string_gmap"
external string_of_ncell : ncell -> string                = "ml_string_ncell"
external get_coord_from_vertex     : ncell -> float array                                = "ml_get_coord_from_vertex"
external set_coord_of_vertex       : ncell -> float -> float -> float -> unit            = "ml_set_coord_of_vertex"

(** manipulation des darts **)
external alpha                     : gmap -> dart -> int -> dart             = "ml_alpha"
external isFree                    : gmap -> dart -> int -> bool             = "ml_isFree"
external isFreeNcell               : ncell -> int -> dart                    = "ml_isFreeNcell"
external sew                       : gmap -> dart -> dart -> int -> unit     = "ml_sew"
external newDart                   : gmap -> float -> float -> float -> dart = "ml_new_dart"
external copyDart                  : gmap -> dart -> dart                    = "ml_copy_dart"
external getNewMark                : gmap -> int                             = "ml_get_new_mark"
external freeMark                  : gmap -> int -> unit                     = "ml_free_mark"
external markDart                  : gmap -> dart -> int ->unit              = "ml_mark_dart"
external markNcell                 : gmap -> ncell -> int -> unit            = "ml_mark_ncell"
external unmarkDart                : gmap -> dart -> int ->unit              = "ml_unmark_dart"
external unmarkNcell               : gmap -> ncell -> int -> unit            = "ml_unmark_ncell"

(** construction de cellules... **)
external new_ncell                 : (gmap*int) -> ncell                   = "ml_new_ncell"
external new_vertex                : gmap -> ncell                         = "ml_new_vertex"
external new_edge                  : gmap -> ncell                         = "ml_new_edge"
external new_face                  : gmap -> ncell                         = "ml_new_face"
external new_volume                : gmap -> ncell                         = "ml_new_volume"
external new_ncell_polygon         : (gmap*int) -> ncell                   = "ml_new_polygon"
external new_ncell_sphere          : gmap -> int -> int -> (ncell * ncell) = "ml_new_sphere"
external new_ncell_polyline        : gmap -> int -> ncell                  = "ml_new_polyline"
external new_ncell_square          : gmap -> int -> int -> int -> ncell    = "ml_new_square"

external insert_vertex             : ncell -> ncell array                                                    = "ml_insert_vertex"
external insert_edge               : ncell -> ncell -> ncell array                                           = "ml_insert_edge"
external insert_face               : ncell -> ncell array -> int -> ncell array                              = "ml_insert_face"
external create_edge               : ncell -> ncell -> ncell array                                           = "ml_create_edge"
external create_face               : (ncell * ncell) array -> int-> ncell array                              = "ml_create_face"
external create_volume             : (ncell array) -> int -> (ncell array) array -> int array -> int -> unit = "ml_create_volume"
external remove_ncell              : ncell -> ncell                                                          = "ml_remove_ncell"
external delete_ncell              : ncell -> ncell array                                                    = "ml_delete_ncell"
external glue2FacesBy4Verteces     : ((ncell*ncell)*(ncell*ncell)*(ncell*ncell)) -> unit                     = "ml_glue2FacesBy4Verteces"

(** manipulation des cellules... **)

external faces                     : ncell -> ncell array        = "ml_faces"
external cofaces                   : ncell -> ncell array        = "ml_cofaces"
external icells                    : ncell -> int -> ncell array = "ml_incident"
external ccells                    : ncell -> int -> ncell array = "ml_close_ncells"

(** colle deux cellules **)
external load_ase                  : gmap -> string -> ncell array                       = "ml_load_ase"
external save_ase                  : gmap -> string -> unit                              = "ml_save_ase"
external load_moka                 : gmap -> string -> ncell array                       = "ml_load_moka"
external save_moka                 : gmap -> string -> unit                              = "ml_save_moka"

(***************************************************************************************)
(***************************************************************************************)
(***************************************************************************************)
(***************************************************************************************)
(***************************************************************************************)

let new_ncell_with_copy (g,n) = let g' = copy_gmap g in new_ncell (g',n) ;;
let new_vertex_with_copy g = new_ncell_with_copy (g,0) ;;
let new_edge_with_copy g = new_ncell_with_copy (g,1) ;;
let new_face_with_copy g = new_ncell_with_copy (g,2) ;;
let new_volume_with_copy g = new_ncell_with_copy (g,3) ;;
let new_ncell_polygon_with_copy (g,n) = let g' = copy_gmap g in new_ncell_polygon (g',n) ;;
let new_ncell_sphere_with_copy g m p = let g' = copy_gmap g in (new_ncell_sphere g' m p) ;;
let new_ncell_square_with_copy g ax ay dim = let g' = copy_gmap g in (new_ncell_square g' ax ay dim) ;;
let new_ncell_polyline_with_copy g n = let g' = copy_gmap g in (new_ncell_polyline g' n) ;;

(* --- Accesseurs ----------------------------------------------------- *)

let get_dim_from_ncell (_,_,d) = d ;;

let get_dart_from_acell (_,d,_) = d ;;

(* --- Cration ------------------------------------------------------- *)


(* --- Requetes ------------------------------------------------------ *)

let ncelldim (_,_,d) = d ;;

let ncelldart (_,d,_) = d ;;

let equalNcell c1 c2 = (compare_ncell (c1,c2)) = 0;;

let rec build_icells c = function
    0 -> [icells c 0]
  | i -> (icells c i)::(build_icells c (i-1))

and transitivefaces (g,d,n) =
  let ftab =
    if (n=0) then
      []
    else
      build_icells (g,d,n) (n-1)
  in
    Array.concat ftab
;;

let rec build_icocells c = function
    3 -> [icells c 3]
  | i -> (icells c i)::(build_icocells c (i+1))

and transitivecofaces (g,d,n) =
  let ftab =
    if (n=3) then
      []
    else
      build_icocells (g,d,n) (n+1)
  in
    Array.concat ftab
;;  

let commonfaces c1 c2 =
  let f1 = if ((ncelldim c1) = 0) then [] else Array.to_list(faces c1)
  and f2 = if ((ncelldim c2) = 0) then [] else Array.to_list(faces c2) in
  let cfaces = List.filter (fun  x -> List.mem x f2) f1 in
    Array.of_list cfaces

and commoncofaces c1 c2 = 
  (*Printf.printf "common coface start\n" ; flush Pervasives.stdout;*)
  let f1 = if ((ncelldim c1) = 3) then [] else Array.to_list(cofaces c1)
  and f2 = if ((ncelldim c2) = 3) then [] else Array.to_list(cofaces c2) in
  let ccofaces = List.filter (fun  x -> List.mem x f2) f1 in
    Array.of_list ccofaces

and commontransitivefaces c1 c2 =
  let f1 =
    if ((ncelldim c1) = 0) then
      []
    else 
      Array.to_list (Array.concat (build_icells c1 ((ncelldim c1) - 1)))
  and f2 =
    if ((ncelldim c2) = 0) then
      []
    else
      Array.to_list (Array.concat (build_icells c2 ((ncelldim c2) - 1))) in
  let cfaces = List.filter (fun  x -> List.mem x f2) f1 in
    Array.of_list cfaces

and commontransitivecofaces c1 c2 =
  let f1 =
    if ((ncelldim c1) = 3) then
      []
    else 
      Array.to_list (Array.concat (build_icocells c1 ((ncelldim c1) + 1)))
  and f2 =
    if ((ncelldim c2) = 3) then
      []
    else
      Array.to_list (Array.concat (build_icocells c2 ((ncelldim c2) + 1))) in
  let ccofaces = List.filter (fun  x -> List.mem x f2) f1 in
    Array.of_list ccofaces

;;


(**************************************************************************
 ** Fonctions de manipulation des qmfs
 **************************************************************************)


 (** valeur qu'on affecte par defaut aux simplexes qu'on cree 
  ** pour assurer la coherence de la structure de complexe.
  **)
let default_value = Valeur.mgsundef
;;


(* --- Manipulation de qmfs ---------------------------------------------------------- *)
exception Cut ;;

let map fct (((g,ha,d) as q):'a qmf) =
  let ha' = Array.copy ha in
    for i = 0 to ((Array.length ha')-1) do
      let ret = Hash.copy ha'.(i) in
	Hash.iter (fun cp v -> Hash.replace ret cp (fct cp v)) ret;
	ha'.(i) <- ret
    done ;
    (g,ha',d)

let iter fct (((g,ha,d) as q):'a qmf) =
  for i=0 to d do
    Hash.iter (fun c v -> fct c v) (ha.(i))
  done

let iter_d dim fct (((g,ha,d) as q):'a qmf) =
  if (dim <= d) then
    Hash.iter (fun c v -> fct c v) (ha.(dim))

let fold fct (((g,ha,d) as q):'a qmf) zero =
  let acc = ref zero in
    for i = 0 to d do
      acc := Hash.fold (fun c v suite -> fct c v suite) (ha.(i)) (!acc)
    done;
    !acc

let fold_d dim fct (((g,ha,d) as q):'a qmf) zero =
  if (dim <= d) then
    Hash.fold (fun c v suite -> fct c v suite) (ha.(dim)) zero
  else
    zero

let forall fct (((g,ha,d) as q):'a qmf) =
  let acc = ref true in
    for i = 0 to d do
      acc := (Hash.forall (fun c v -> fct c v) (ha.(i))) && (!acc)
    done;
    !acc

let forall_d dim fct (((g,ha,d) as q):'a qmf) =
  if (dim <= d) then
    (Hash.forall (fun c v -> fct c v) (ha.(dim))) && true
  else
    true

let exists fct (((g,ha,d) as q):'a qmf) =
  try
    for i = 0 to d do
      if (Hash.exists (fun c v -> fct c v) (ha.(i))) then raise Cut
    done;
    false
  with
    | Cut -> true

let exists_d dim fct (((g,ha,d) as q):'a qmf) =
  if (dim <= d) then
    Hash.exists (fun c v -> fct c v) (ha.(dim))
  else
    false

let size (((g,ha,d) as q):'a qmf) =
  let nb = ref 0 in
    for i = 0 to d do
      nb := !nb + Hash.size ha.(i)
    done;
    !nb

let size_d dim (((g,ha,d) as q):'a qmf) =
  if (dim <= d) then Hash.size ha.(dim) else 0

let find (((g,ha,d) as q):'a qmf) c =
  let dim = ncelldim c in
    if (dim > d) then
      raise Not_found
    else
      Hash.find ha.(dim) c

let nth n (((g,ha,d) as q):'a qmf) =
  let i = ref 0 in
  let pos = ref n in
    try
      while ((!i)<=d) do
	let s = Hash.size ha.(!i) in
	  if ( (!pos) >= s) then
	    begin
	      pos := (!pos) - s ;
              i   := !i + 1
	    end
	  else
	    raise Cut
      done ;
      failwith "internal error on nth in core_qmf"
    with
	Cut -> Hash.nth (!pos) (ha.(!i))

let nth_d dim n (((g,ha,d) as q):'a qmf) =
  if (dim <= d) then Hash.nth n (ha.(dim)) else raise (Erreur "core_qmf: nth_d: dimension is too big...")

let remove (((g,ha,d) as q):'a qmf) c =
  if ( (ncelldim c) <= d ) then
    Hash.remove (ha.(ncelldim c)) c

let copy (((g,ha,d) as q):'a qmf) =
  let ha' = Array.init (Array.length ha) (fun i -> Hash.copy (ha.(i))) in
    (g,ha',d)

let hd q = nth 0 q

let last q = nth ((size q) - 1) q

let tl q =
  let ret = copy q in
  let k, v = hd ret
  in
    remove ret k;
    ret

let mem ((g,ha,d):'a qmf) c =
  let dim = ncelldim c in
    if (dim > d) then
      false
    else
      Hash.mem ha.(dim) c

let compare valcmp (((g1,ha1,d1) as q1):'a qmf) (((g2,ha2,d2) as q2):'b qmf) =
  let gdiff = Pervasives.compare g1 g2 in
    if gdiff <> 0 then
      gdiff
    else
      let ddiff = Pervasives.compare d1 d2 in
	if ddiff <> 0 then
	  ddiff
	else
	  begin
	    let diff = ref 0 in
	      try
		for i = 0 to d1 do
		  diff := Hash.compare valcmp ha1.(i)  ha2.(i);
		  if ( !diff <> 0) then raise Cut
		done ;
		0
	      with Cut -> !diff
	  end
	
let to_list (((g,ha,d) as q):'a qmf) =
  fold (fun k v s -> (k, v)::s) q []
  
(** dimension **)
let get_dim_from_qmf ((g,ha,d):'a qmf) = d

let get_gmap_from_qmf ((g,ha,d):'a qmf) = g

and get_ncells_dim_in_qmf dim ((g,ha,d):'a qmf) =
  if (dim <= d) then 
    ha.(dim)
  else
    raise (Erreur "core_qmf: get_ncells_of_dim: dimension is too big...")

and set_ncells_dim dim ((g,ha,d):'a qmf) h =
  if (dim <= d) then 
    ha.(dim) <- h
  else
    raise (Erreur "core_qmf: set_ncells_of_dim: dimension is too big...")

and get_ncell ((g,ha,d):'a qmf) c =
  if ((ncelldim c) > d) then
    raise Not_found
  else
    let _ = (Hash.find (ha.(ncelldim c)) c) in c (* ???? *)

and get_value_from_ncell_in_qmf ((g,ha,d):'a qmf) ncell =
  if ((ncelldim ncell) > d) then
    raise Not_found
  else
    (Hash.find (ha.(ncelldim ncell)) ncell)

and get_ncell_from_val_in_qmf ((g,ha,d):'a qmf) v =
  for i = 0 to d do
    Hash.fold (fun nc vv acc -> if v = vv then nc::acc else acc) (ha.(i)) []
  done


(** Renvoie un tableau dont l'lment i contient 
 ** la liste des ncells de dimension i d'une qmf 
 **)
let skelette ((g,ha,d):'a qmf) =
  Array.map (fun h -> Hash.to_list h) ha

let faces_qmf ((g,ha,d):'a qmf) nc =
  if ((d>=(ncelldim nc))&(Hash.mem (ha.(ncelldim nc)) nc)) then (* vrifie que la ncell est dans la qmf *)
    let fl = faces nc
    in Array.map (fun fp -> fp,(Hash.find (ha.(ncelldim fp)) fp)) fl
  else
    let err = "Error: Core_qmf.faces_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)
 
let cofaces_qmf ((g,ha,d):'a qmf) nc =
  if ((d>=(ncelldim nc))&(Hash.mem (ha.(ncelldim nc)) nc)) then (* vrifie que la ncell est dans la qmf *)
    let cl = cofaces nc
    in Array.of_list (Array.fold_left
			(fun acc c -> try (c,Hash.find (ha.(ncelldim c)) c)::acc with Not_found -> acc | Invalid_argument "Array.get" -> acc)
			[]
			cl
		     )
  else
    let err = "Error: Core_qmf.faces_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)

let icells_qmf ((g,ha,d):'a qmf) nc i =
  if ((d>=(ncelldim nc))&&(Hash.mem (ha.(ncelldim nc)) nc)) then
    let l = icells nc i
    in Array.of_list (Array.fold_left
			(fun acc c -> try (c,Hash.find (ha.(ncelldim c)) c)::acc with Not_found -> acc | Invalid_argument "Array.get" -> acc)
			[]
			l
		     )
  else
    let err = "Error: Core_qmf.icells_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)

let ccells_qmf ((g,ha,d):'a qmf) nc i =
  if ((d>=(ncelldim nc))&&(Hash.mem (ha.(ncelldim nc)) nc)) then
    let l = ccells nc i
    in Array.of_list (Array.fold_left
			(fun acc c -> try (c,Hash.find (ha.(ncelldim c)) c)::acc with Not_found -> acc | Invalid_argument "Array.get" -> acc)
			[]
			l
		     )
  else
    let err = "Error: Core_qmf.ccells_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)

let transitive_faces_qmf ((g,ha,d):'a qmf) c =
  if ((d>=(ncelldim c))&&(Hash.mem (ha.(ncelldim c)) c)) then (* vrifie que la ncell est dans la qmf *)
    let fl = transitivefaces c
    in Array.map (fun fp -> fp,(Hash.find (ha.(ncelldim fp)) fp)) fl
  else
    let err = "Error: Core_qmf.transitive_faces_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)
 
let transitive_cofaces_qmf ((g,ha,d):'a qmf) c =
  if ((d>=(ncelldim c))&&(Hash.mem (ha.(ncelldim c)) c)) then (* vrifie que la ncell est dans la qmf *)
    let cl = transitivecofaces c
    in Array.of_list (Array.fold_left
			(fun acc c -> try (c,Hash.find (ha.(ncelldim c)) c)::acc with Not_found -> acc | Invalid_argument "Array.get" -> acc)
			[]
			cl
		     )
  else
    let err = "Error: Core_qmf.transitive_faces_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)

let common_faces_qmf ((g,ha,d):'a qmf) c1 c2 =
  if ((d>=(ncelldim c1))&&(Hash.mem (ha.(ncelldim c1)) c1)&&(d>=(ncelldim c2))&&(Hash.mem (ha.(ncelldim c2)) c2)) then
    let l = commonfaces c1 c2
    in Array.map (fun p -> p,(Hash.find (ha.(ncelldim p)) p)) l
  else
    let err = "Error: Core_qmf.common_faces_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)

let common_cofaces_qmf ((g,ha,d):'a qmf) c1 c2 =
  if ((d>=(ncelldim c1))&&(Hash.mem (ha.(ncelldim c1)) c1)&&(d>=(ncelldim c2))&&(Hash.mem (ha.(ncelldim c2)) c2)) then
    let l = commoncofaces c1 c2
    in Array.of_list (Array.fold_left
			(fun acc c -> try (c,Hash.find (ha.(ncelldim c)) c)::acc with Not_found -> acc | Invalid_argument "Array.get" -> acc)
			[]
			l
		     )
  else
    let err = "Error: Core_qmf.common_cofaces_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)

let common_transitive_faces_qmf ((g,ha,d):'a qmf) c1 c2 =
  if ((d>=(ncelldim c1))&&(Hash.mem (ha.(ncelldim c1)) c1)&&(d>=(ncelldim c2))&&(Hash.mem (ha.(ncelldim c2)) c2)) then
    let l = commontransitivefaces c1 c2
    in Array.map (fun p -> p,(Hash.find (ha.(ncelldim p)) p)) l
  else
    let err = "Error: Core_qmf.common_transitive_faces_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)

let common_transitive_cofaces_qmf ((g,ha,d):'a qmf) c1 c2 =
  if ((d>=(ncelldim c1))&&(Hash.mem (ha.(ncelldim c1)) c1)&&(d>=(ncelldim c2))&&(Hash.mem (ha.(ncelldim c2)) c2)) then
    let l = commontransitivecofaces c1 c2
    in Array.of_list (Array.fold_left
			(fun acc c -> try (c,Hash.find (ha.(ncelldim c)) c)::acc with Not_found -> acc | Invalid_argument "Array.get" -> acc)
			[]
			l
		     )
  else
    let err = "Error: Core_qmf.common_transitive_cofaces_qmf: ncell not in qmf"
    in prerr_endline err; raise(Erreur err)

(* --- Construction de qmf ---------------------------------------------------------- *)

let new_qmf gmap = (gmap,[||],-1)

let create_qmf gmap dim = (gmap,(Array.init (dim+1) (fun _ -> Hash.create 11)),dim)

let extends_to_dim ((g,ha,d):'a qmf) dim =
  if (dim <= d) then
    (g,ha,d)
  else
    if (dim > ((Array.length ha)-1)) then
      (g,Array.append ha (Array.init (dim-(Array.length ha)+1) (fun _ -> Hash.create 11)),dim)
    else
      (g,ha,dim)

(** --- une somme --------------- **)

(** on ajoute un couple (ncell,valeur)
 ** On force l'ajout de ncell (quitte  mettre  jour la valeur associe)
 ** et on ajoute les faces de ncell. Par contre, on ne force pas cet ajout
 ** car des valeurs ont peut tre dj t associes aux faces.
 **)

let add ((g,ha,d):'a qmf) c v =
  if (ncelldim c > d) then
    raise (Erreur "Core_qmf.add: the dimension of the qmf is lower than the dimension of the ncell")
  else
    Hash.replace (ha.(ncelldim c)) c v

let replace ((g,ha,d):'a qmf) c v =
  if (ncelldim c > d) then
    raise (Erreur "Core_qmf.replace: the dimension of the qmf is lower than the dimension of the ncell")
  else
    Hash.replace (ha.(ncelldim c)) c v

let rec add_ncell (((g,ha,d) as q):'a qmf) c v =
  begin
    let ((g',ha',d') as q') = extends_to_dim q (ncelldim c) in
    Array.iter (fun f -> add_one_ncell q' f default_value) (transitivefaces c);
    add_force_one_ncell q' c v;
    q'
  end

and add_one_ncell (((g,ha,d) as q):'a qmf) c v = 
  if not(Hash.mem (ha.(ncelldim c)) c) then (add_force_one_ncell q c v)

and add_force_one_ncell ((g,ha,d):'a qmf) c v = 
  Hash.replace (ha.(ncelldim c)) c v
;;


(** rajoute une ncell ou bien combine la valeur si la ncell est deja present **)
let update_qmf (((g,ha,d) as q):'a qmf) combine v ((g',_,d') as c) =
  if (g = g') then
    (*let c' = normal_form c in*)
    try
      let v0 = if (d'>d) then (raise Not_found) else (Hash.find (ha.(d')) c)
      in Hash.replace (ha.(d')) c (combine v0 v)
    with Not_found -> ignore(add_ncell q c v)
  else
    raise (Erreur "Core_qmf.update_qmf: qmf and ncell are not in the same gmap")
;;
