/*
 *
 *                              MGS
 *                              LaMI
 *                       Universite d'Evry Val d'Essonne
 *                   Olivier MICHEL, Jean-Louis GIAVITTO
 * File: $RCSfile: stub_gmap.cc,v $ 
 * Path: %P%
 * Description:
 * Created:
 * Author: Olivier Michel & Jean-Louis Giavitto & Julien Cohen & Antoine Spicher
 * Modified: $Date: 2005-12-16 15:11:28 $ at 
 * Version: $Revision: 1.7 $
 * Status:
 *
 */

#include "g-map-vertex.hh"
#include <list>

using namespace std;
using namespace GMap3d;

typedef struct _SGMapVertex
{
  CGMapVertex *gmap ;
  bool qmfied ;
  int DI_Norm[4] ;
} SGMapVertex;

extern "C" 
{
  /****************************************************************************************/
  /****************************************************************************************/
  /** Normalisation des cellules                                                         **/
  /****************************************************************************************/
  /****************************************************************************************/

  /* Retourne le brin d'adresse la plus petite dans une cellule */
  CDart *minDart(CGMapVertex* gmap, CDart *dart, int dim)
  {
    CDart* minDart = dart ;
    CCoverage *cov=gmap->getDynamicCoverage(dart,ORBIT_CELL[dim]);
    //cov->reinit() ;
    while (cov->cont()) {
      if ( (cov->operator*()) < minDart )
	minDart = (cov->operator*());
      cov->operator++();
    }
    delete cov;

    return (minDart) ;    
  }

  void normDart(SGMapVertex* sgmap, CDart* min, int dim, int mark)
  {
    CCoverage *cov=sgmap->gmap->getDynamicCoverage(min,ORBIT_CELL[dim]);
    //cov->reinit() ;
    while (cov->cont()) {
      sgmap->gmap->setDirectInfo(**cov,sgmap->DI_Norm[dim],(void*)min);
      sgmap->gmap->setMark(**cov,mark);
      cov->operator++();
    }
    delete cov;
  }

  void normOrbit(SGMapVertex* saGmap, CDart* d, TOrbit aOrbit)
  {
    CGMapVertex* aGmap = saGmap->gmap;
    int mark0 = aGmap->getNewMark();
    int mark1 = aGmap->getNewMark();
    int mark2 = aGmap->getNewMark();
    int mark3 = aGmap->getNewMark();

    CCoverage *cov=aGmap->getDynamicCoverage(d,aOrbit);
    //cov->reinit() ;
    while (cov->cont()) {
      if (!(aGmap->isMarked(**cov,mark0))) normDart(saGmap, (minDart(aGmap,**cov,0)), 0, mark0);
      if (!(aGmap->isMarked(**cov,mark1))) normDart(saGmap, (minDart(aGmap,**cov,1)), 1, mark1);
      if (!(aGmap->isMarked(**cov,mark2))) normDart(saGmap, (minDart(aGmap,**cov,2)), 2, mark2);
      if (!(aGmap->isMarked(**cov,mark3))) normDart(saGmap, (minDart(aGmap,**cov,3)), 3, mark3);
      cov->operator++();
    }
    delete cov;

    aGmap->unmarkAll(mark0); //FIXME se concentrer sur les brins marks
    aGmap->unmarkAll(mark1);
    aGmap->unmarkAll(mark2);
    aGmap->unmarkAll(mark3);
    aGmap->freeMark(mark0);
    aGmap->freeMark(mark1);
    aGmap->freeMark(mark2);
    aGmap->freeMark(mark3);
  }

  void normAll(SGMapVertex* saGmap)
  {
    CGMapVertex* aGmap = saGmap->gmap;
    int mark0 = aGmap->getNewMark();
    int mark1 = aGmap->getNewMark();
    int mark2 = aGmap->getNewMark();
    int mark3 = aGmap->getNewMark();

    CDynamicCoverageAll *cov = new CDynamicCoverageAll(aGmap);

    cov->reinit() ;
    while (cov->cont()) {
      if (!(aGmap->isMarked(**cov,mark0))) normDart(saGmap, (minDart(aGmap,**cov,0)), 0, mark0);
      if (!(aGmap->isMarked(**cov,mark1))) normDart(saGmap, (minDart(aGmap,**cov,1)), 1, mark1);
      if (!(aGmap->isMarked(**cov,mark2))) normDart(saGmap, (minDart(aGmap,**cov,2)), 2, mark2);
      if (!(aGmap->isMarked(**cov,mark3))) normDart(saGmap, (minDart(aGmap,**cov,3)), 3, mark3);
      cov->operator++();
    }
    delete cov;

    aGmap->unmarkAll(mark0);
    aGmap->unmarkAll(mark1);
    aGmap->unmarkAll(mark2);
    aGmap->unmarkAll(mark3);
    aGmap->freeMark(mark0);
    aGmap->freeMark(mark1);
    aGmap->freeMark(mark2);
    aGmap->freeMark(mark3);
  }

  CDart* normalForm(SGMapVertex *saGmap, CDart *dart, int dim)
  {
    return (saGmap->gmap->getDirectInfoAsDart(dart,saGmap->DI_Norm[dim]));
  }  

  void qmfy(SGMapVertex *saGmap)
  {
    saGmap->qmfied = true;
    normAll(saGmap);
  }

  /****************************************************************************************/
  /****************************************************************************************/
  /** Cration d'une gmap                                                                **/
  /****************************************************************************************/
  /****************************************************************************************/
  
  SGMapVertex *createGmap()
  {
    SGMapVertex *sg = (new SGMapVertex) ;
    sg->gmap = (new CGMapVertex) ;
    sg->DI_Norm[0] = sg->gmap->getNewDirectInfo() ;
    sg->DI_Norm[1] = sg->gmap->getNewDirectInfo() ;
    sg->DI_Norm[2] = sg->gmap->getNewDirectInfo() ;
    sg->DI_Norm[3] = sg->gmap->getNewDirectInfo() ;
    sg->qmfied = false;
    return sg ;
  }
  
  void deleteGmap(SGMapVertex * g)
  {
    if (g != NULL) {
      delete g->gmap;
      delete g ;
    }
  }
  
  void deleteDartsTable(CDart** darts)
  {
    delete darts ;
  }


  SGMapVertex *copyGmap(SGMapVertex *ssGmap)
  {

#define DUP(D) (sGmap->getDirectInfoAsDart(D, directInfoIndex))

    SGMapVertex *stGmap = createGmap() ;
    stGmap->qmfied = false;
    CGMapVertex *tGmap = stGmap->gmap ;
    CGMapVertex *sGmap = ssGmap->gmap ;

    int nbDuplicated = 0;
    int duplicated = sGmap->getNewMark();
    int tduplicated = tGmap->getNewMark();
    int directInfoIndex = sGmap->getNewDirectInfo();

    CDynamicCoverageAll it(sGmap);

    // Duplication des brins:
    for (; it.cont(); ++it) {
      CDart * newDart = tGmap->addMapDart();
      
      tGmap->setMark(newDart, tduplicated);

      (*it)->setDirectInfo(directInfoIndex, newDart);
      
      ++nbDuplicated;
    }


    // 0, 1, 2 et 3 coutures:
    for (it.reinit(); it.cont(); ++it) {
      CDart * d = (*it);
      
      if (!(sGmap->isFree0(d)) && tGmap->isFree0(DUP(d)))
	tGmap->linkAlpha0(DUP(d), DUP(sGmap->alpha0(d)));

      if (!(sGmap->isFree1(d)) && tGmap->isFree1(DUP(d)))
	tGmap->linkAlpha1(DUP(d), DUP(sGmap->alpha1(d)));

      if (!(sGmap->isFree2(d)) && tGmap->isFree2(DUP(d)))
	tGmap->linkAlpha2(DUP(d), DUP(sGmap->alpha2(d)));

      if (!(sGmap->isFree3(d)) && tGmap->isFree3(DUP(d)))
	tGmap->linkAlpha3(DUP(d), DUP(sGmap->alpha3(d)));

    }

    // Duplication des plongements:
    int treated = sGmap->getNewMark();
    int ttreated = tGmap->getNewMark();
   
    for (int i=ORBIT_0; i<=ORBIT_0123; ++i)
      if (sGmap->isOrbitUsed(TOrbit(i))) {
	// Duplication pour l'orbite i:
	for (it.reinit(); it.cont(); ++it) {
	  CDart * d = (*it);
	  
	  if (!sGmap->isMarked(d, treated)) {
	    CDart * owner = sGmap->getEmbeddingOwner(d, TOrbit(i));
	    CEmbedding * E = owner->getEmbedding(TOrbit(i));
	    CCoverage * cov = sGmap->getDynamicCoverage(d, TOrbit(i));
	    for (; cov->cont(); sGmap->setMark(**cov, treated), ++(*cov))
	      if (!tGmap->isMarked(DUP(**cov), ttreated)) {
		CEmbedding * dupE = E->copy();
		DUP(**cov)->addEmbedding(dupE);
		tGmap->markOrbit(DUP(**cov), TOrbit(i), ttreated);
		tGmap->unmarkOrbit(DUP(**cov), TOrbit(i), tduplicated);
	      }
	     delete cov;
	  }
	}
	
	// Dmarquage:
	if (it.type() == COVERAGE_ALL) {
	  sGmap->negateMaskMark(treated);
	  tGmap->negateMaskMark(ttreated);
	}
	else
	  for (it.reinit(); it.cont(); ++it)
	    {
	      if (sGmap->isMarked(*it, treated))
		sGmap->unmarkOrbit(*it, TOrbit(i), treated);
	      
	      if (tGmap->isMarked(DUP(*it), ttreated))
		tGmap->unmarkOrbit(DUP(*it), TOrbit(i), ttreated);
	    }
      }
    
    // Librations:
    sGmap->unmarkAll(treated) ;
    tGmap->unmarkAll(ttreated) ;
    sGmap->unmarkAll(duplicated) ;
    tGmap->unmarkAll(tduplicated) ;
    sGmap->freeMark(treated);
    tGmap->freeMark(ttreated);
    sGmap->freeDirectInfo(directInfoIndex);
    sGmap->freeMark(duplicated);
    tGmap->freeMark(tduplicated);
    
    return stGmap;
  }

  int copy_notmarked_dart(SGMapVertex *gsrc,SGMapVertex *gdst,int mark)
  {
    CGMapVertex *sGmap = gsrc->gmap ;
    CGMapVertex *tGmap = gdst->gmap ;

    int nbDuplicated = 0;
    int duplicated = sGmap->getNewMark();
    int tduplicated = tGmap->getNewMark();
    int directInfoIndex = sGmap->getNewDirectInfo();

    CDynamicCoverageAll it(sGmap);

    // Duplication des brins:
    for (; it.cont(); ++it) {
      if (!(sGmap->isMarked(*it, mark))) {
	CDart * newDart = tGmap->addMapDart();
	tGmap->updateVertex(newDart,CVertex(0,0,0));
      
	tGmap->setMark(newDart, tduplicated);

	(*it)->setDirectInfo(directInfoIndex, newDart);
      
	++nbDuplicated;
      }
    }


    // 0, 1, 2 et 3 coutures:
    for (it.reinit(); it.cont(); ++it) {
      CDart * d = (*it);
      
      if (!(sGmap->isMarked(*it, mark))) {

	if (!(sGmap->isFree0(d)) && tGmap->isFree0(DUP(d)))
	  if (!(sGmap->isMarked(sGmap->alpha0(d), mark)))
	    tGmap->linkAlpha0(DUP(d), DUP(sGmap->alpha0(d)));

	if (!(sGmap->isFree1(d)) && tGmap->isFree1(DUP(d)))
	  if (!(sGmap->isMarked(sGmap->alpha1(d), mark)))
	    tGmap->linkAlpha1(DUP(d), DUP(sGmap->alpha1(d)));

	if (!(sGmap->isFree2(d)) && tGmap->isFree2(DUP(d)))
	  if (!(sGmap->isMarked(sGmap->alpha2(d), mark)))
	    tGmap->linkAlpha2(DUP(d), DUP(sGmap->alpha2(d)));

	if (!(sGmap->isFree3(d)) && tGmap->isFree3(DUP(d)))
	  if (!(sGmap->isMarked(sGmap->alpha3(d), mark)))
	    tGmap->linkAlpha3(DUP(d), DUP(sGmap->alpha3(d)));
      }
    }

        // Librations:
    sGmap->unmarkAll(duplicated) ;
    tGmap->unmarkAll(tduplicated) ;
    sGmap->freeMark(duplicated);
    tGmap->freeMark(tduplicated);

    return directInfoIndex ;
  }



  int isFrozen(SGMapVertex* gmap)
  {
    if (gmap->qmfied) 
      return 1;
    else
      return 0;
  }

  /****************************************************************************************/
  /****************************************************************************************/
  /** Infos et manipulation des brins                                                    **/
  /****************************************************************************************/
  /****************************************************************************************/

  CDart* copyDart(SGMapVertex *saGmap, CDart *d)
  {
    CDart* c = saGmap->gmap->addMapDart();
    CVertex *v=saGmap->gmap->findVertex(d) ;
    if (v == NULL)
      return c;
    saGmap->gmap->updateVertex(c,CVertex(v->getX(),v->getY(),v->getZ()));
    return c;
  }

  CDart* newDart(SGMapVertex *saGmap, float x, float y, float z)
  {
    CDart* c = saGmap->gmap->addMapDart();
    saGmap->gmap->updateVertex(c,CVertex(x,y,z));
    return c;
  }
  
  int getNewMark(SGMapVertex *saGmap)
  {
    return (saGmap->gmap->getNewMark());
  }

  void freeMark(SGMapVertex *saGmap, int mark)
  {
    saGmap->gmap->freeMark(mark);
  }

  void markDart(SGMapVertex *saGmap, CDart *d, int mark)
  {
    saGmap->gmap->setMark(d,mark);
  }

  void markNcell(SGMapVertex *saGmap, CDart *d, int n, int mark)
  {
    saGmap->gmap->markOrbit(d,ORBIT_CELL[n],mark);
  }

  void unmarkDart(SGMapVertex *saGmap, CDart *d, int mark)
  {
    saGmap->gmap->unsetMark(d,mark);
  }

  void unmarkNcell(SGMapVertex *saGmap, CDart *d, int n, int mark)
  {
    saGmap->gmap->unmarkOrbit(d,ORBIT_CELL[n],mark);
  }

  int isMarkedDart(SGMapVertex *saGmap, CDart *d, int mark)
  {
    if (saGmap->gmap->isMarked(d,mark))
      return 1;
    else
      return 0;
  }

  int isMarkedNcell(SGMapVertex *saGmap, CDart *d, int dim, int mark)
  {
    if (saGmap->gmap->isWholeCellMarked(d,ORBIT_CELL[dim],mark))
      return 1;
    else
      return 0;
  }

  int getNewDirectInfo(SGMapVertex *saGmap)
  {
    return (saGmap->gmap->getNewDirectInfo());
  }

  void freeDirectInfo(SGMapVertex *saGmap, int di)
  {
    saGmap->gmap->freeDirectInfo(di);
  }

  void setDirectInfo(SGMapVertex *saGmap, CDart*d, int di, void* info)
  {
    saGmap->gmap->setDirectInfo(d,di,info);
  }

  void* getDirectInfo(SGMapVertex *saGmap, CDart*d, int di)
  {
    return (saGmap->gmap->getDirectInfo(d,di));
  }

  CDart* alpha(SGMapVertex *saGmap, CDart *dart, int i)
  {
    return (saGmap->gmap->alpha(dart,i));
  }

  void sew(SGMapVertex *saGmap, CDart *dart1, CDart *dart2, int i)
  {
    if ((saGmap->gmap->isFree(dart1,i)) && (saGmap->gmap->isFree(dart2,i))) {
      saGmap->gmap->dartSew(dart1,dart2,i);
    }
  }

  int isFree(SGMapVertex *saGmap, CDart *dart1, int i)
  {
    if (saGmap->gmap->isFree(dart1,i))
      return 1;
    else
      return 0;
  }

  int get_vertex_coord(SGMapVertex* sg, CDart* s, float* x, float* y, float* z)
  {
    CGMapVertex* g=sg->gmap;
    CVertex *v=g->findVertex(s) ;
    if (v == NULL)
      return -1;
    *x = v->getX() ;
    *y = v->getY() ;
    *z = v->getZ() ;
    return 0;
  }

  void set_vertex_coord(SGMapVertex* sg, CDart* c, double x, double y, double z)
  {
    sg->gmap->updateVertex(c,CVertex(x,y,z));
  }

  int checkGmap(SGMapVertex *saGmap)
  {
    if (saGmap->gmap->checkTopology())
      return 1;
    else
      return 0;
  }

  int compareNcells(SGMapVertex *saGmap, CDart* c1, CDart* c2, int dim)
  {

    CCoverage *cov = saGmap->gmap->getDynamicCoverage(c1,ORBIT_CELL[dim]) ;

    while(cov->cont()) {
      if ( (**cov) == c2 ) {
	delete cov;
	return 0;
      }
      ++(*cov);
    }

    delete cov;
    return (((int)c1)-((int)c2));  

  }

  /****************************************************************************************/
  /****************************************************************************************/
  /** Parcours des cellules                                                              **/
  /****************************************************************************************/
  /****************************************************************************************/

  CCoverage* getAllCoverage(SGMapVertex* sGmap)
  {
    return (new CDynamicCoverageAll(sGmap->gmap));
  }

  CCoverage* getNcellCoverage(SGMapVertex* sGmap, CDart* d, int n)
  {
    return (sGmap->gmap->getDynamicCoverage(d,ORBIT_CELL[n]));
  }

  void deleteCoverage(CCoverage* cov)
  {
    delete cov;
  }

  void reinitCoverage(CCoverage* cov)
  {
    cov->reinit();
  }

  CDart* getCurrentDart(CCoverage* cov)
  {
    return ((CDart*)**cov);
  }

  int cont(CCoverage* cov)
  {
    if (cov->cont())
      return 1;
    else
      return 0;
  }

  void next(CCoverage* cov)
  {
    ++(*cov);
  }


  /****************************************************************************************/
  /****************************************************************************************/
  /** Entres/sorties                                                                    **/
  /****************************************************************************************/
  /****************************************************************************************/

  int loadMoka(SGMapVertex *saGmap, const char *MokaFileName, CDart*** ret) //AsciiFormat/BinaryFormat
  {
#define NbMaxVolume 100
    CGMapVertex *aGmap = saGmap->gmap;
    int mark = aGmap->getNewMark() ;

    CStaticCoverageAll* cov = new CStaticCoverageAll(aGmap) ;
    while (cov->cont()) {
      CDart *d = **cov ; 
      aGmap->setMark(d,mark);
      cov->operator++() ;
    }
    delete cov;

    if (saGmap->gmap->load(MokaFileName,AsciiFormat)) {

      int i=0 ;
      *ret = new CDart*[NbMaxVolume] ;
      cov = new CStaticCoverageAll(aGmap) ;
      while ((i<NbMaxVolume) && (cov->cont())) {
	CDart *d = **cov ; 
	if (!(aGmap->isMarked(d,mark))) {
	  //normOrbit(saGmap,d,ORBIT_CELL[3]);
	  aGmap->markOrbit(d,ORBIT_CELL[3], mark);
	  (*ret)[i] = d;//normalForm(saGmap, d, 3);
	  i++;
	}
	cov->operator++() ;
      }
      delete cov;

      aGmap->unmarkAll(mark);
      aGmap->freeMark(mark);
      return i;
    }
    else {
      aGmap->unmarkAll(mark);
      aGmap->freeMark(mark);
      return 0;
    }

  }

  int saveMoka(SGMapVertex *saGmap, const char *MokaFileName) //AsciiFormat/BinaryFormat
  {
    if (saGmap->gmap->save(MokaFileName,AsciiFormat))
      return 1;
    else
      return 0;
  }


  /****************************************************************************************/
  /****************************************************************************************/
  /** Construction et manipulation de cellules                                           **/
  /****************************************************************************************/
  /****************************************************************************************/

  CDart* new_ncell(SGMapVertex *saGmap, int dim)
  {
    CGMapVertex* aGmap = saGmap->gmap ;
    CDart* d1=NULL; CDart* d2=NULL; CDart* d3=NULL; CDart* d4=NULL;
    CDart* d_1=NULL;CDart* d_2=NULL;CDart* d_3=NULL;CDart* d_4=NULL;
    CDart* tab[8] ;

    switch(dim)
      {
      case 0 :
	d1 = aGmap->addMapDart();
	set_vertex_coord(saGmap,d1,0,0,0);
	return d1 ;
	break ;
      case 1:
	tab[0] = aGmap->addMapDart();
	set_vertex_coord(saGmap,tab[0],0,0,0);
	tab[1] = aGmap->addMapDart();
	set_vertex_coord(saGmap,tab[1],0,0,0);
	aGmap->sew0(tab[0],tab[1]);
	return (tab[0]);
	break;
      case 2:
	d1 = aGmap->createTopoTriangle() ;
	set_vertex_coord(saGmap,d1,0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha01(d1),0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha10(d1),0,0,0);
	return (d1);
	break;
      case 3:
	d1 = aGmap->createTopoTriangle(); d_1 = aGmap->alpha1(d1) ;
	set_vertex_coord(saGmap,d1,0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha01(d1),0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha10(d1),0,0,0);
	d2 = aGmap->createTopoTriangle(); d_2 = aGmap->alpha1(d2) ;
	set_vertex_coord(saGmap,d2,0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha01(d2),0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha10(d2),0,0,0);
 	d3 = aGmap->createTopoTriangle(); d_3 = aGmap->alpha1(d3) ;
	set_vertex_coord(saGmap,d3,0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha01(d3),0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha10(d3),0,0,0);
 	d4 = aGmap->createTopoTriangle(); d_4 = aGmap->alpha1(d4) ;
	set_vertex_coord(saGmap,d4,0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha01(d4),0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha10(d4),0,0,0);

	aGmap->dartSew2(d1,d2);
       	d1=aGmap->alpha0(d1);d2=aGmap->alpha0(d2);
	aGmap->dartSew2(d1,d2);
 	d1=aGmap->alpha1(d1); d2=aGmap->alpha1(d2);
	aGmap->dartSew2(d1,d3); aGmap->dartSew2(d2,d_3);
	d1=aGmap->alpha0(d1);d2=aGmap->alpha0(d2);
	d3=aGmap->alpha0(d3);d_3=aGmap->alpha0(d_3);
	aGmap->dartSew2(d1,d3);aGmap->dartSew2(d2,d_3);

	aGmap->dartSew2(d_1,d4);aGmap->dartSew2(d_2,d_4);
	d_1=aGmap->alpha0(d_1);d_2=aGmap->alpha0(d_2);
	d4=aGmap->alpha0(d4);d_4=aGmap->alpha0(d_4);
	aGmap->dartSew2(d_1,d4);aGmap->dartSew2(d_2,d_4);

	d_3=aGmap->alpha1(d_3);d3=aGmap->alpha1(d3);
	d_4=aGmap->alpha1(d_4);d4=aGmap->alpha1(d4);
	aGmap->dartSew2(d_3,d_4);aGmap->dartSew2(d3,d4);
	d_3=aGmap->alpha0(d_3);d3=aGmap->alpha0(d3);
	d_4=aGmap->alpha0(d_4);d4=aGmap->alpha0(d4);
	aGmap->dartSew2(d_3,d_4);aGmap->dartSew2(d3,d4);

	return(aGmap->alpha1(d_1));

	break;
      }

    return NULL;
  }

  CDart* new_ncell_polygon(SGMapVertex *aGmap, int nbside)
  {
    return (aGmap->gmap->createRegularPolygon(nbside));
  }

  CDart* new_ncell_square(SGMapVertex *aGmap, int Ax,int Ay, int dim)
  {
    CDart* ASquareCorners[2][2] = {{NULL,NULL},{NULL,NULL}};
    aGmap->gmap->createSquareIMeshed(Ax,Ay,dim,ASquareCorners) ;
    return (ASquareCorners[0][0]);
  }

  void new_ncell_sphere(SGMapVertex *aGmap, int nbmeridians, int nbparallels, CDart **southPole, CDart **northPole)
  {
    aGmap->gmap->createSphere(nbmeridians, nbparallels, southPole, northPole) ;
  }

  CDart* new_ncell_polyline(SGMapVertex *aGmap, int nbverteces)
  {
    CDart *ret = NULL ;
    for (int i = 0;i<nbverteces;i++)
      ret = aGmap->gmap->prolongPolyline(ret,CVertex(0,0,0)) ;
    
    return ret;
  }

  void glue2FacesBy4Verteces(SGMapVertex *saGmap, CDart *c1, CDart *c2,
			     CDart *v1c1, CDart *v1c2,
			     CDart *v2c1, CDart *v2c2)
  {
    CGMapVertex* aGmap = saGmap->gmap ;
    CDart *new_v1c1 = NULL;
    CDart *new_v1c2 = NULL;
    CDart *new_v2c1 = NULL;
    CDart *new_v2c2 = NULL;
    int mv1 = aGmap->getNewMark() ;
    int mv2 = aGmap->getNewMark() ;
    aGmap->markOrbit(v1c1, ORBIT_VERTEX, mv1);
    aGmap->markOrbit(v1c2, ORBIT_VERTEX, mv1);
    aGmap->markOrbit(v2c1, ORBIT_VERTEX, mv2);
    aGmap->markOrbit(v2c2, ORBIT_VERTEX, mv2);

    CCoverage * cov = aGmap->getDynamicCoverage(c1, ORBIT_FACE);
    for (int i=0; cov->cont() && i<2; ++(*cov)) {
      CDart *d = **cov ;
      if (aGmap->isMarked(d,mv1) &&  aGmap->isMarked(aGmap->alpha0(d),mv2) ) {
	new_v1c1 = d ;
	new_v2c1 = aGmap->alpha0(d) ;
	i = 2;
      }
      else if (aGmap->isMarked(d,mv2) &&  aGmap->isMarked(aGmap->alpha0(d),mv1) ) {
	new_v2c1 = d ;
	new_v1c1 = aGmap->alpha0(d) ;
	i = 2;
      }
    }
    delete cov ;

    cov = aGmap->getDynamicCoverage(c2, ORBIT_FACE);
    for (int i=0; cov->cont() && i<2; ++(*cov)) {
      CDart *d = **cov ;
      if (aGmap->isMarked(d,mv1) &&  aGmap->isMarked(aGmap->alpha0(d),mv2) ) {
	new_v1c2 = d ;
	new_v2c2 = aGmap->alpha0(d) ;
	i = 2;
      }
      else if (aGmap->isMarked(d,mv2) &&  aGmap->isMarked(aGmap->alpha0(d),mv1) ) {
	new_v2c2 = d ;
	new_v1c2 = aGmap->alpha0(d) ;
	i = 2;
      }
    }
    delete cov ;

    if (!(aGmap->isFree2(new_v1c1))
	||!(aGmap->isFree2(new_v1c2))
	||!(aGmap->isFree2(new_v2c1))
	||!(aGmap->isFree2(new_v2c2)))
      {
	cerr << "Stub_gmap: glue2FacesBy4Verteces: fatal error: some darts are already sewn" << endl ;
	cerr.flush();
	exit(0);
      }
    aGmap->dartSew2(new_v1c1,new_v1c2);
    aGmap->dartSew2(new_v2c1,new_v2c2);
    aGmap->unmarkAll(mv1) ;
    aGmap->unmarkAll(mv2) ;
    aGmap->freeMark(mv1) ;
    aGmap->freeMark(mv2) ;

  }

  //on insere un sommet sur un arc
  CDart* insert_vertex(SGMapVertex *saGmap, CDart *e, CDart **re1, CDart **re2)
  {
    CGMapVertex *aGmap = saGmap->gmap ;
    CDart *e1, *e2, *v ;

    if (aGmap->isFree0(e)) {
      cerr << "Stub_gmap: insertVertex: warning: e is 0-free !!! strange... but why not" << endl ;
      cerr.flush();
      e2 = NULL;
    } else {
      e2 = aGmap->alpha0(e) ;
    }

    v  = aGmap->insertVertex(e) ;
    e1 = e ;

    *re1 = e1;
    *re2 = e2;
    return v;
  }

  //on insere un arc dans une face
  bool commonPCellOf2NCells(SGMapVertex *saGmap, int n, int p,
			    CDart *c1, CDart *c2,
			    CDart **cp1, CDart **cp2)
  {
    CGMapVertex* aGmap = saGmap->gmap ;

    /* marquage de toutes les cellules de dim p incidente  c1 */
    
    int mark = aGmap->getNewMark();
    CCoverage *cov=aGmap->getDynamicCoverage(c1,ORBIT_CELL[n]);
    cov->reinit() ;
    while (cov->cont()) {
      if (!(aGmap->isMarked(**cov,mark)))
	aGmap->markOrbit(**cov,ORBIT_CELL[p],mark);
      cov->operator++();
    }
    delete cov;

    /* On regarde autour de c2 s'il n y a pas un brin marqu... sinon, on choue */
    cov=aGmap->getDynamicCoverage(c2,ORBIT_CELL[n]);
    cov->reinit() ;
    while (cov->cont()) {
      if (aGmap->isMarked(**cov,mark)) {
	goto found1;
	break;
      }
      cov->operator++();
    }
    delete cov;
    aGmap->unmarkAll(mark);
    aGmap->freeMark(mark);
    return false;

  found1:
    *cp2 = (CDart *)(**cov);
    delete cov;
    
    /* Je marque la cellule de dim p  cote de c2 */
    int mark2 = aGmap->getNewMark();
    aGmap->markOrbit(*cp2,ORBIT_CELL[p], mark2);
    
    /* Je parcours c1 de dim n pour chopper un brin marqu par mark2 */
    /* J'en profite pour tout dmarqu */

    cov=aGmap->getDynamicCoverage(c1,ORBIT_CELL[n]);
    cov->reinit() ;
    while (cov->cont()) {
      if (aGmap->isMarked(**cov,mark)) {
	aGmap->unmarkOrbit(**cov,ORBIT_CELL[p],mark);
      }
      if (aGmap->isMarked(**cov,mark2)) {
	aGmap->unmarkOrbit(**cov,ORBIT_CELL[p],mark2);
	*cp1 = (CDart *)(**cov);
      }
      cov->operator++();
    }
    delete cov;
    aGmap->unmarkAll(mark); // normalement inutile
    aGmap->unmarkAll(mark2); // normalement inutile
    aGmap->freeMark(mark);
    aGmap->freeMark(mark2);

    return true;
  }

  CDart* insert_edge(SGMapVertex *saGmap, CDart *vv1, CDart *vv2, CDart **rf1, CDart **rf2)
  {
    CGMapVertex *aGmap = saGmap->gmap ;
    CDart *e, *f1, *f2;
    CDart *v1, *v2;

    if (!(commonPCellOf2NCells(saGmap,0,2,vv1,vv2,&v1,&v2))) {
      cerr << "Stub_gmap: insertEdge: fatal error: vv1 and vv2 must belong to the same face" << endl ;
      cerr.flush();
      exit(0);
    }

    /* On teste pour ne pas faire de face non oriente */
    bool flag=false;
    CDart* tmp = aGmap->alpha01(v1);
    while ( (tmp != v1) && !flag ) {
      if (tmp == v2) {
	flag = true;
      }
      tmp = aGmap->alpha01(tmp) ;
    }
    if (flag)
      v2 = aGmap->alpha1(v2) ;

    f1 = v1;
    f2 = (aGmap->isFree1(f1))?NULL:aGmap->alpha1(f1);

    e = aGmap->insertEdge(v1,v2);
    *rf1 = f1;
    *rf2 = f2;
    return e;
  }

  CDart* sharedDart(SGMapVertex *saGmap,CDart *vol,int p,CDart *ed,int n)
  {
    CDart *volp = vol ;
    CCoverage *cov= saGmap->gmap->getDynamicCoverage(ed,ORBIT_CELL[n]);
    cov->reinit() ;
    while (cov->cont()) {
      if ( compareNcells(saGmap,volp,**cov,p) == 0) {
	goto found2;
	break;
      }
      cov->operator++();
    }

    delete cov;
    return NULL;

  found2:
    CDart *shared = **cov;
    delete cov;
    return shared;
  }

  //on insere une face dans un volume
  CDart* insert_face(SGMapVertex *saGmap, CDart *vol, CDart **et, int nb_e, CDart **rv1, CDart **rv2)
  {
    CGMapVertex *aGmap = saGmap->gmap ;
    CDart *f = NULL, *v1, *v2;

    if (nb_e <= 1) {
      cerr << "Stub_gmap: insertFace: fatal error: nb_e must be greater than 1" << endl ;
      cerr.flush();
      exit(0);
    }

    for (int i=1;i<nb_e;i++)
      if (compareNcells(saGmap,vol,et[i],3) != 0) {
	CDart* tmp = sharedDart(saGmap,vol,3,et[i],1);
	if (tmp == NULL) {
	  cerr << "Stub_gmap: insertFace: fatal error: all vt must belong to the same volume" << endl ;
	  cerr.flush();
	  exit(0);
	}
	et[i] = tmp;
      }


    v1 = et[0];

    int aMark = aGmap->getNewMark() ;
    for (int i=0;i<nb_e;i++) {
      aGmap->setMark(et[i],aMark) ; //aGmap->markOrbit(et[i],ORBIT_CELL[1],aMark);
    }

    f = aGmap->insertFace(et[0],aMark,true);
    for (int i=0;i<nb_e;i++) {
      aGmap->unsetMark(et[i],aMark) ; //aGmap->unmarkOrbit(et[i],ORBIT_CELL[1],aMark);
    }
    v1 = f;
    v2 = aGmap->alpha3(v1);

    bool flag = true;
    for (CDynamicCoverage01 cov(aGmap,v2); flag & cov.cont(); ++cov)
      if ( !(aGmap->isFree2(*cov)) )
	flag = false;

    if (flag) {
      aGmap->unsew3(v2);
      aGmap->markOrbit(v2,ORBIT_CELL[2],aMark);
      aGmap->deleteMarkedDarts(aMark);
      v2 = NULL;
    }
    aGmap->freeMark(aMark);

    *rv1 = v1;
    *rv2 = v2;
    return f;
  }

  CDart *create_edge(SGMapVertex *saGmap, CDart *v1, CDart *v2)
  {
    CGMapVertex *aGmap = saGmap->gmap;
    CDart *vv1, *vv2;
    if ((!aGmap->isFree1(v1)) || (!aGmap->isFree2(v1)) || (!aGmap->isFree3(v1)) ||
	(!aGmap->isFree1(v1)) || (!aGmap->isFree2(v2)) || (!aGmap->isFree3(v2))
	) {
      cerr << "Stub_gmap: create_edge: fatal error: the both darts must be alpha 1,2 and 3 free"
	   << endl ;
      cerr.flush();
      exit(0);
    }

    if ((aGmap->isFree0(v1)) && (aGmap->isFree0(v2))) {
      vv1 = v1;
      vv2 = v2;
      aGmap->sew0(vv1,vv2);
    } else if ((!(aGmap->isFree0(v1))) && (!(aGmap->isFree0(v2)))) {
      vv1 = aGmap->addMapDart();
      vv2 = aGmap->addMapDart();
      aGmap->sew1(v1,vv1);
      aGmap->sew1(v2,vv2);
      aGmap->sew0(vv1,vv2);
    } else if ((!(aGmap->isFree0(v1))) && (aGmap->isFree0(v2))) {
      vv2 = aGmap->addMapDart();
      vv1 = v1;
      aGmap->sew0(v2,vv2);
      aGmap->sew1(vv1,vv2);
    } else {
      vv1 = aGmap->addMapDart();
      vv2 = v2;
      aGmap->sew0(v1,vv1);
      aGmap->sew1(vv2,vv1);
    }

    return vv1;
  }

  /* On donne un tableau de couple : On cherche  fusionner v_tab[i][0] avec v_tab[i][1]*/
  void can_create_face(SGMapVertex *saGmap, CDart ***v_tab, int i, int j)
  {
    CGMapVertex *aGmap = saGmap->gmap ;
    CDart *tmp1, *tmp2;
    if (!(commonPCellOf2NCells(saGmap,0,1,v_tab[i][1],v_tab[j][0],&tmp1,&tmp2))) {
      cerr << "Stub_gmap: create_face: fatal error: some edge doesn't exist" << endl ;
      cerr.flush();
      exit(0);
    }

    if ((!aGmap->isFree2(tmp1)) || (!aGmap->isFree3(tmp1)) ||
	(!aGmap->isFree2(tmp2)) || (!aGmap->isFree3(tmp2))
	) {
      cerr << "Stub_gmap: create_face: fatal error: the both darts must be alpha 2 and 3 free"
	   << endl ;
      cerr.flush();
      exit(0);
    }
    v_tab[i][1] = tmp1 ;
    v_tab[j][0] = tmp2 ;
  }

  void double_edge(SGMapVertex *saGmap, CDart* d1, CDart* d2, CDart** t1, CDart** t2)
  {
    CGMapVertex *aGmap = saGmap->gmap;
    if ((aGmap->isFree1(d1)) && (aGmap->isFree1(d2))) {
      *t1 = d1;
      *t2 = d2;
      return;
    }

    /* Il faut ddoubler */
    CDart *n1 = aGmap->addMapDart();
    CDart *n2 = aGmap->addMapDart();

    aGmap->dartSew2(n1,d1);
    aGmap->dartSew2(n2,d2);
    aGmap->dartSew0(n1,n2);

    *t1 = n1;
    *t2 = n2;
    return;   
  }

  CDart *create_face(SGMapVertex *saGmap, CDart ***v_tab, int v_nb)
  {
    CGMapVertex *aGmap = saGmap->gmap;
    CDart *t1, *t2;

    /* Trouver les bons brins ; on en profite pour les vrif...*/
    if (v_nb <=1) {
      cerr << "Stub_gmap: create_face: fatal error: v_nb must be greater than 1" << endl ;
      cerr.flush();
      exit(0);
    }
    for(int i=0;i<v_nb-1;i++) {
      can_create_face(saGmap,v_tab,i,i+1);
    }
    can_create_face(saGmap,v_tab,v_nb-1,0);

    /* On cherche  ddoubler les arcs appartenant dj  une face (alpha1 cousu) */
    for(int i=0;i<v_nb-1;i++) {
      double_edge(saGmap,v_tab[i][1],v_tab[i+1][0],&t1,&t2);
      v_tab[i][1] = t1;
      v_tab[i+1][0] = t2;
    }
    double_edge(saGmap,v_tab[v_nb-1][1],v_tab[0][0],&t1,&t2);
    v_tab[v_nb-1][1]=t1;
    v_tab[0][0]=t2;

    /* Reste  tourner une dernire fois pour faire les alpha1 !!!! */

    for(int i=0;i<v_nb;i++) {
      aGmap->dartSew1(v_tab[i][0],v_tab[i][1]);
    }

    return (v_tab[0][0]);

  }

  CDart *double_face(SGMapVertex *saGmap, CDart* f)
  {
    CGMapVertex *aGmap = saGmap->gmap;
    CCoverage* cov = NULL;
    int dInfo;

    dInfo = aGmap->getNewDirectInfo();

    cout << "passe1" << endl ; cout.flush() ;
    cov = aGmap->getStaticCoverage(f, ORBIT_FACE);
    cov->reinit();
    while(cov->cont()) {
      CDart* n = aGmap->addMapDart();
      aGmap->setDirectInfo(**cov,dInfo,(void*)n);
      ++*cov;
    }

    cout << "passe2" << endl ; cout.flush() ;
    cov->reinit();
    while(cov->cont()) {
      cout << "passe2.1" << endl ; cout.flush() ;
      CDart* d0 = (CDart*)aGmap->getDirectInfo(aGmap->alpha0(**cov),dInfo);
      cout << "passe2.2" << endl ; cout.flush() ;
      CDart* d1 = (CDart*)aGmap->getDirectInfo(aGmap->alpha1(**cov),dInfo);
//       cout << "passe2.3" << endl ; cout.flush() ;
//       CDart* d2 = (CDart*)aGmap->getDirectInfo(aGmap->alpha2(**cov),dInfo);
      cout << "passe2.4" << endl ; cout.flush() ;
      CDart* d3 = (CDart*)aGmap->getDirectInfo(aGmap->alpha3(**cov),dInfo);

      cout << "passe2.5" << endl ; cout.flush() ;
      CDart* n = (CDart*)aGmap->getDirectInfo(**cov,dInfo);

      cout << "passe2.6" << endl ; cout.flush() ;
      if ((n != d0)&&(aGmap->isFree0(n))) aGmap->dartSew0(n,d0);
      cout << "passe2.7" << endl ; cout.flush() ;
      if ((n != d1)&&(aGmap->isFree1(n))) aGmap->dartSew1(n,d1);
      cout << "passe2.8" << endl ; cout.flush() ;
//       On fait pas puisque ont est dans une face... Il faut pas change de face !!!!
//       if ((n != d2)&&(aGmap->isFree2(n))) aGmap->dartSew2(n,d2);
//       cout << "passe2.9" << endl ; cout.flush() ;
      if ((n != d3)&&(aGmap->isFree3(n))) aGmap->dartSew3(n,d3);
      cout << "passe2.10" << endl ; cout.flush() ;
      
      ++*cov;
    }
    
    cout << "passe3" << endl ; cout.flush() ;
    cov->reinit();
    while(cov->cont()) {
      aGmap->dartSew3(**cov,(CDart*)aGmap->getDirectInfo(**cov,dInfo));
      ++*cov;
    }

    delete cov;
    CDart* df = (CDart*)aGmap->getDirectInfo(f,dInfo);
    aGmap->freeDirectInfo(dInfo);
    return df;
  }

  CDart *create_volume(SGMapVertex *saGmap, CDart **f_tab, int f_nb, CDart ***v_tab, int *l_tab, int v_nb)
  {
    CGMapVertex *aGmap = saGmap->gmap;
    int nb_faces = 0, i, j;
    CCoverage* cov;

    cout << v_nb << endl ; cout.flush() ;
    /* On marque les cellules qui vont tre utilises */

    /* Vrification : chaque face doit tre 3-libre - on marque les cellules  doubler */
    if (v_nb <=1) {
      cerr << "Stub_gmap: volume_face: fatal error: v_nb must be greater than 1" << endl ;
      cerr.flush();
      exit(0);
    }

    int face_mark = aGmap->getNewMark();
    int double_mark = aGmap->getNewMark();
    for(i=0;i<v_nb;i++) {
      for(j=0;j<l_tab[i];j++) {
	cout << i << " " << j << endl ; cout.flush() ;
	CDart* f = v_tab[i][j] ;
	if (!(aGmap->isMarked(f,face_mark))) {
	  bool flag = false ;
	  aGmap->markOrbit(f,ORBIT_FACE,face_mark);
	  nb_faces++;
	  cov = aGmap->getDynamicCoverage(f,ORBIT_FACE) ;
	  while(cov->cont()) {
	    if (!(aGmap->isFree3(**cov))) {
	      delete cov;
	      cerr << "Stub_gmap: volume_face: fatal error: all faces must be 3-free" << endl ;
	      cerr.flush();
	      exit(0);
	    }
	    if (!(aGmap->isFree2(**cov)))
	      flag = true;
	    ++*cov;
	  }
	  delete cov;
	  if (flag)
	    aGmap->markOrbit(f,ORBIT_FACE,double_mark);
	}
      }
    }
    cout << nb_faces << endl ; cout.flush() ;

    /* On ddouble les faces marques par double_mark et on en profite pour dtruire cette marque */

    for(i=0;i<v_nb;i++) {
      for(j=0;j<l_tab[i];j++) {
	cout << i << " - " << j << endl ; cout.flush() ;
	CDart* f = v_tab[i][j] ;
	if (aGmap->isMarked(f,face_mark)) {
	  cout << "Je suis marque par face..." ; cout.flush() ;
	  if (aGmap->isMarked(f,double_mark)) {
	    cout << " et par double" << endl ; cout.flush() ;
	    v_tab[i][j] = double_face(saGmap,f) ;
	  } else {
	    cout << " mais pas par double" << endl ; cout.flush() ;
	  }
	  aGmap->unmarkOrbit(f,ORBIT_FACE,face_mark);
	} else {
	  cout << "Je ne suis pas marque par face..." ; cout.flush() ;
	  if (aGmap->isMarked(f,double_mark)) {
	    cout << " mais par double oui" << endl ; cout.flush() ;
	    v_tab[i][j] = aGmap->alpha3(f) ;
	  } else {
	    cout << " ni par double" << endl ; cout.flush() ;
	  }
	}
	
      }
    }

    return NULL;
  }



  /* Enlever une ncell en fusionnant les exactement 2 (n+1)cells qui la partage */
  CDart* basic_remove(SGMapVertex* saGmap, CDart* c, int dim)
  {
    CGMapVertex* aGmap = saGmap->gmap;

    /* On vrifie que la cellule a exactement 2 cofaces */
    int nb_cofaces ;
    int dim_coface = dim + 1;
    switch (dim) {
    case 0:
      {
	aGmap->countCells(c,ORBIT_CELL[dim],NULL,&nb_cofaces,NULL,NULL,NULL);
	break;
      }
    case 1:
      {
	aGmap->countCells(c,ORBIT_CELL[dim],NULL,NULL,&nb_cofaces,NULL,NULL);
	break;
      }
    case 2:
      {
	aGmap->countCells(c,ORBIT_CELL[dim],NULL,NULL,NULL,&nb_cofaces,NULL);
	break;
      }
    default:
      {
	cerr << "Stub_gmap: basic_remove: fatal error: dim must be 0, 1 or 2." << endl ;
	cerr.flush();
	exit(0);
      }
    }
    if (nb_cofaces != 2) {
      cerr << "Stub_gmap: basic_remove: fatal error: c must exactly have two cofaces." << endl ;
      cerr.flush();
      exit(0);
    }

    CDart *cf_c = aGmap->alpha(c,dim_coface) ;
    if (!(aGmap->canMerge(c, cf_c, dim_coface))) {
      cerr << "Stub_gmap: basic_remove: fatal error: cannot merge c and cf_c." << endl ;
      cerr.flush();
      exit(0);
    }

    CDart *f = NULL;
    CCoverage *cov = aGmap -> getDynamicCoverage(c,ORBIT_CELL[dim_coface]) ;
    cov->reinit();
    while(cov->cont()) {
      if ( compareNcells(saGmap,**cov,c,dim) != 0) {
	f = **cov;
	break;
      }
      ++*cov;
    }
    delete cov;

    aGmap->merge(c,cf_c,dim_coface,true);

    return f;
  }

  int basic_delete(SGMapVertex* saGmap, CDart* c, int dim, CDart ***nfaces)
  {
    CGMapVertex* aGmap = saGmap->gmap;

    /* On vrifie que la cellule a exactement 2 cofaces */
    int nb_cofaces, nb_faces ;
    int dim_face = dim-1 ;
    switch (dim) {
    case 1:
      {
	aGmap->countCells(c,ORBIT_CELL[dim],&nb_faces,NULL,&nb_cofaces,NULL,NULL);
	break;
      }
    case 2:
      {
	aGmap->countCells(c,ORBIT_CELL[dim],NULL,&nb_faces,NULL,&nb_cofaces,NULL);
	break;
      }
    case 3:
      {
	aGmap->countCells(c,ORBIT_CELL[dim],NULL,NULL,&nb_faces,NULL,NULL);
	nb_cofaces = 1;
	break;
      }
    default:
      {
	cerr << "Stub_gmap: basic_delete: fatal error: dim must be 1, 2 or 3." << endl ;
	cerr.flush();
	exit(0);
      }
    }
    if (nb_cofaces != 1) {
      cerr << "Stub_gmap: basic_delete: fatal error: c must exactly have 1 coface (closed or not !!!)." << endl ;
      cerr.flush();
      exit(0);
    }

    int delete_mark = aGmap->getNewMark();
    int other_mark = aGmap->getNewMark();

    *nfaces = new CDart*[nb_faces] ;
    int cur_faces = 0;

    CCoverage* cov = aGmap->getStaticCoverage(c,ORBIT_CELL[dim]);
    cov->reinit();
    while(cov->cont()) {
      if (!(aGmap->isMarked(**cov,delete_mark)))
	if (!(aGmap->isFree(**cov,dim))) {
	  (*nfaces)[cur_faces] = aGmap->alpha(**cov,dim);
	  cur_faces++;
	  aGmap->unsew(**cov,dim);
	  aGmap->markOrbit(**cov,ORBIT_CELL[dim_face],delete_mark);
	} else {
	  //	  CDart *normal_face = normalForm(saGmap,**cov,dim_face);
	  //	  if (!(aGmap->isMarked(normal_face,other_mark))) {
	  CDart *normal_face = **cov;
	  if (!(aGmap->isWholeCellMarked(normal_face,ORBIT_CELL[dim_face],other_mark))) {
	    aGmap->markOrbit(normal_face,ORBIT_CELL[dim_face],other_mark);
	    (*nfaces)[cur_faces] = normal_face;
	    cur_faces++;
	  }
	}
      ++*cov;
    }

    cov->reinit();
    while(cov->cont()) {
      if (!(aGmap->isFree(**cov,dim_face))) {
	aGmap->unsew(**cov,dim_face);
      }
      ++*cov;
    }
    delete cov;

    aGmap->deleteMarkedDarts(delete_mark);

    /* Trouver un moyen de rcuprer les faces puis les normaliser... */

    for(cur_faces=0;cur_faces<nb_faces;cur_faces++) {
      if (aGmap->isWholeCellMarked((*nfaces)[cur_faces],ORBIT_CELL[dim_face],other_mark))
	aGmap->unmarkOrbit((*nfaces)[cur_faces],ORBIT_CELL[dim_face],other_mark);
    }

    aGmap->freeMark(delete_mark);
    aGmap->freeMark(other_mark);

    return nb_faces;
  }


}
