/*
 *
 *                              MGS
 *                              LaMI
 *                       Universite d'Evry Val d'Essonne
 *                   Olivier MICHEL, Jean-Louis GIAVITTO
 * File: $RCSfile: stub_gmap.old.cc,v $ 
 * Path: %P%
 * Description:
 * Created:
 * Author: Olivier Michel & Jean-Louis Giavitto & Julien Cohen & Antoine Spicher
 * Modified: $Date: 2005-01-13 00:59:54 $ at 
 * Version: $Revision: 1.1 $
 * Status:
 *
 */

#include "g-map-vertex.hh"
#include <list>

#define GMAPPRINTF 	    "Gmap (%10d)\n"\
	    "\tnb darts    = %10d\n"\
	    "\tnb verteces = %10d\n"\
	    "\tnb edges    = %10d\n"\
	    "\tnb faces    = %10d\n"\
	    "\tnb volumes  = %10d\n"

#define NCELLPRINTF "Ncell:(%10d,%10d,%10d,%10d,%s)"

using namespace std;
using namespace GMap3d;


typedef int* Tint;
typedef Tint* TTint;


#define MAX_GMAP 100

typedef struct _SGMapVertex
{
  CGMapVertex *gmap ;
  int DI_Norm[4];
} SGMapVertex;

extern "C" 
{

  /****************************************************************************************/
  /****************************************************************************************/
  /** Normalisation des cellules                                                         **/
  /****************************************************************************************/
  /****************************************************************************************/

  CDart* normalForm(SGMapVertex *saGmap, CDart *dart, int dim)
  {
    return (saGmap->gmap->getDirectInfoAsDart(dart,saGmap->DI_Norm[dim]));
  }

  /* Retourne le brin d'adresse la plus petite dans une cellule */
  CDart *minDart(CGMapVertex* gmap, CDart *dart, int dim)
  {
    CDart* minDart = dart ;
    CCoverage *cov=gmap->getDynamicCoverage(dart,ORBIT_CELL[dim]);
    cov->reinit() ;
    while (cov->cont()) {
      if ( (cov->operator*()) < minDart )
	minDart = (cov->operator*());
      cov->operator++();
    }
    delete cov;

    return (minDart) ;    
  }

  void normDart(SGMapVertex* sgmap, CDart* min, int dim, int mark)
  {
    CCoverage *cov=sgmap->gmap->getDynamicCoverage(min,ORBIT_CELL[dim]);
    cov->reinit() ;
    while (cov->cont()) {
      sgmap->gmap->setDirectInfo(**cov,sgmap->DI_Norm[dim],(void*)min);
      sgmap->gmap->setMark(**cov,mark);
      cov->operator++();
    }
    delete cov;
  }

  void normTabNCell(SGMapVertex* saGmap, int nbDart, CDart** tab)
  {
    CGMapVertex* aGmap = saGmap->gmap;
    int mark0 = aGmap->getNewMark();
    int mark1 = aGmap->getNewMark();
    int mark2 = aGmap->getNewMark();
    int mark3 = aGmap->getNewMark();

    for(int i=0;i<nbDart;i++) {
      if (!(aGmap->isMarked(tab[i],mark0))) normDart(saGmap, (minDart(aGmap,tab[i],0)), 0, mark0);
      if (!(aGmap->isMarked(tab[i],mark1))) normDart(saGmap, (minDart(aGmap,tab[i],1)), 1, mark1);
      if (!(aGmap->isMarked(tab[i],mark2))) normDart(saGmap, (minDart(aGmap,tab[i],2)), 2, mark2);
      if (!(aGmap->isMarked(tab[i],mark3))) normDart(saGmap, (minDart(aGmap,tab[i],3)), 3, mark3);
    }

    aGmap->unmarkAll(mark0);
    aGmap->unmarkAll(mark1);
    aGmap->unmarkAll(mark2);
    aGmap->unmarkAll(mark3);
    aGmap->freeMark(mark0);
    aGmap->freeMark(mark1);
    aGmap->freeMark(mark2);
    aGmap->freeMark(mark3);
  }

  void normOrbit(SGMapVertex* saGmap, CDart* d, TOrbit aOrbit)
  {
    CGMapVertex* aGmap = saGmap->gmap;
    int mark0 = aGmap->getNewMark();
    int mark1 = aGmap->getNewMark();
    int mark2 = aGmap->getNewMark();
    int mark3 = aGmap->getNewMark();

    CCoverage *cov=aGmap->getDynamicCoverage(d,aOrbit);
    cov->reinit() ;
    while (cov->cont()) {
      if (!(aGmap->isMarked(**cov,mark0))) normDart(saGmap, (minDart(aGmap,**cov,0)), 0, mark0);
      if (!(aGmap->isMarked(**cov,mark1))) normDart(saGmap, (minDart(aGmap,**cov,1)), 1, mark1);
      if (!(aGmap->isMarked(**cov,mark2))) normDart(saGmap, (minDart(aGmap,**cov,2)), 2, mark2);
      if (!(aGmap->isMarked(**cov,mark3))) normDart(saGmap, (minDart(aGmap,**cov,3)), 3, mark3);
      cov->operator++();
    }
    delete cov;

    aGmap->unmarkAll(mark0);
    aGmap->unmarkAll(mark1);
    aGmap->unmarkAll(mark2);
    aGmap->unmarkAll(mark3);
    aGmap->freeMark(mark0);
    aGmap->freeMark(mark1);
    aGmap->freeMark(mark2);
    aGmap->freeMark(mark3);
  }

  void normAll(SGMapVertex* saGmap)
  {
    CGMapVertex* aGmap = saGmap->gmap;
    int mark0 = aGmap->getNewMark();
    int mark1 = aGmap->getNewMark();
    int mark2 = aGmap->getNewMark();
    int mark3 = aGmap->getNewMark();

    CDynamicCoverageAll *cov = new CDynamicCoverageAll(aGmap);

    cov->reinit() ;
    while (cov->cont()) {
      if (!(aGmap->isMarked(**cov,mark0))) normDart(saGmap, (minDart(aGmap,**cov,0)), 0, mark0);
      if (!(aGmap->isMarked(**cov,mark1))) normDart(saGmap, (minDart(aGmap,**cov,1)), 1, mark1);
      if (!(aGmap->isMarked(**cov,mark2))) normDart(saGmap, (minDart(aGmap,**cov,2)), 2, mark2);
      if (!(aGmap->isMarked(**cov,mark3))) normDart(saGmap, (minDart(aGmap,**cov,3)), 3, mark3);
      cov->operator++();
    }
    delete cov;

    aGmap->unmarkAll(mark0);
    aGmap->unmarkAll(mark1);
    aGmap->unmarkAll(mark2);
    aGmap->unmarkAll(mark3);
    aGmap->freeMark(mark0);
    aGmap->freeMark(mark1);
    aGmap->freeMark(mark2);
    aGmap->freeMark(mark3);
  }

  void normSeveralOrbits_init(SGMapVertex* saGmap, int *mark0, int *mark1, int *mark2, int *mark3)
  {
    CGMapVertex* aGmap = saGmap->gmap;
    *mark0 = aGmap->getNewMark();
    *mark1 = aGmap->getNewMark();
    *mark2 = aGmap->getNewMark();
    *mark3 = aGmap->getNewMark();
  }

  void normSeveralOrbits(SGMapVertex* saGmap, CDart* d, TOrbit aOrbit, int mark0, int mark1, int mark2, int mark3)
  {
    CGMapVertex* aGmap = saGmap->gmap ;
    CCoverage *cov=aGmap->getDynamicCoverage(d,aOrbit);
    cov->reinit() ;
    while (cov->cont()) {
      if (!(aGmap->isMarked(**cov,mark0))) normDart(saGmap, (minDart(aGmap,**cov,0)), 0, mark0);
      if (!(aGmap->isMarked(**cov,mark1))) normDart(saGmap, (minDart(aGmap,**cov,1)), 1, mark1);
      if (!(aGmap->isMarked(**cov,mark2))) normDart(saGmap, (minDart(aGmap,**cov,2)), 2, mark2);
      if (!(aGmap->isMarked(**cov,mark3))) normDart(saGmap, (minDart(aGmap,**cov,3)), 3, mark3);
      cov->operator++();
    }
    delete cov;
  }

  void normSeveralOrbits_finalize(SGMapVertex* saGmap, int mark0, int mark1, int mark2, int mark3)
  {
    CGMapVertex* aGmap = saGmap->gmap ;
    aGmap->unmarkAll(mark0);
    aGmap->unmarkAll(mark1);
    aGmap->unmarkAll(mark2);
    aGmap->unmarkAll(mark3);
    aGmap->freeMark(mark0);
    aGmap->freeMark(mark1);
    aGmap->freeMark(mark2);
    aGmap->freeMark(mark3);
  }

  void normSeveralOrbits_finalize1(SGMapVertex* saGmap, CDart* d, int mark0, int mark1, int mark2, int mark3)
  {
    CGMapVertex* aGmap = saGmap->gmap ;
    aGmap->unmarkOrbit(d,ORBIT_CC,mark0);
    aGmap->unmarkOrbit(d,ORBIT_CC,mark1);
    aGmap->unmarkOrbit(d,ORBIT_CC,mark2);
    aGmap->unmarkOrbit(d,ORBIT_CC,mark3);
  }

  void normSeveralOrbits_finalize2(SGMapVertex* saGmap, int mark0, int mark1, int mark2, int mark3)
  {
    CGMapVertex* aGmap = saGmap->gmap ;
    aGmap->freeMark(mark0);
    aGmap->freeMark(mark1);
    aGmap->freeMark(mark2);
    aGmap->freeMark(mark3);
  }

  /****************************************************************************************/
  /****************************************************************************************/
  /** Cration d'une gmap                                                                **/
  /****************************************************************************************/
  /****************************************************************************************/
  
  SGMapVertex *createGmap()
  {
    SGMapVertex *sg = (new SGMapVertex) ;
    sg->gmap = (new CGMapVertex) ;
    sg->DI_Norm[0] = sg->gmap->getNewDirectInfo() ;
    sg->DI_Norm[1] = sg->gmap->getNewDirectInfo() ;
    sg->DI_Norm[2] = sg->gmap->getNewDirectInfo() ;
    sg->DI_Norm[3] = sg->gmap->getNewDirectInfo() ;
    return sg ;
  }
  
  void deleteGmap(SGMapVertex * g)
  {
    if (g != NULL) {
      delete g->gmap;
      delete g ;
    }
  }
  
  void deleteDartsTable(CDart** darts)
  {
    delete darts ;
  }


  SGMapVertex *copyGmap(SGMapVertex *ssGmap)
  {

#define DUP(D) (sGmap->getDirectInfoAsDart(D, directInfoIndex))

    SGMapVertex *stGmap = createGmap() ;
    CGMapVertex *tGmap = stGmap->gmap ;
    CGMapVertex *sGmap = ssGmap->gmap ;

    int nbDuplicated = 0;
    int duplicated = sGmap->getNewMark();
    int tduplicated = tGmap->getNewMark();
    //ATTENTION    sGmap->unmarkAll(duplicated) ;
    //ATTENTION    tGmap->unmarkAll(tduplicated) ;
    int directInfoIndex = sGmap->getNewDirectInfo();

    CDynamicCoverageAll it(sGmap);

    // Duplication des brins:
    for (; it.cont(); ++it) {
      CDart * newDart = tGmap->addMapDart();
      
      tGmap->setMark(newDart, tduplicated);

      (*it)->setDirectInfo(directInfoIndex, newDart);
      
      ++nbDuplicated;
    }


    // 0, 1, 2 et 3 coutures:
    for (it.reinit(); it.cont(); ++it) {
      CDart * d = (*it);
      
      if (!(sGmap->isFree0(d)) && tGmap->isFree0(DUP(d)))
	tGmap->linkAlpha0(DUP(d), DUP(sGmap->alpha0(d)));

      if (!(sGmap->isFree1(d)) && tGmap->isFree1(DUP(d)))
	tGmap->linkAlpha1(DUP(d), DUP(sGmap->alpha1(d)));

      if (!(sGmap->isFree2(d)) && tGmap->isFree2(DUP(d)))
	tGmap->linkAlpha2(DUP(d), DUP(sGmap->alpha2(d)));

      if (!(sGmap->isFree3(d)) && tGmap->isFree3(DUP(d)))
	tGmap->linkAlpha3(DUP(d), DUP(sGmap->alpha3(d)));

    }

    normAll(stGmap) ;


    // Duplication des plongements:
    int treated = sGmap->getNewMark();
    int ttreated = tGmap->getNewMark();
    //ATTENTION    sGmap->unmarkAll(treated) ;
    //ATTENTION    tGmap->unmarkAll(ttreated) ;
   
    for (int i=ORBIT_0; i<=ORBIT_0123; ++i)
      if (sGmap->isOrbitUsed(TOrbit(i))) {
	// Duplication pour l'orbite i:
	for (it.reinit(); it.cont(); ++it) {
	  CDart * d = (*it);
	  
	  if (!sGmap->isMarked(d, treated)) {
	    CDart * owner = sGmap->getEmbeddingOwner(d, TOrbit(i));
	    CEmbedding * E = owner->getEmbedding(TOrbit(i));
	    CCoverage * cov = sGmap->getDynamicCoverage(d, TOrbit(i));
	    for (; cov->cont(); sGmap->setMark(**cov, treated), ++(*cov))
	      if (!tGmap->isMarked(DUP(**cov), ttreated)) {
		CEmbedding * dupE = E->copy();
		DUP(**cov)->addEmbedding(dupE);
		tGmap->markOrbit(DUP(**cov), TOrbit(i), ttreated);
		tGmap->unmarkOrbit(DUP(**cov), TOrbit(i), tduplicated);
	      }
	     delete cov;
	  }
	}
	
	// Dmarquage:
	if (it.type() == COVERAGE_ALL) {
	  sGmap->negateMaskMark(treated);
	  tGmap->negateMaskMark(ttreated);
	}
	else
	  for (it.reinit(); it.cont(); ++it)
	    {
	      if (sGmap->isMarked(*it, treated))
		sGmap->unmarkOrbit(*it, TOrbit(i), treated);
	      
	      if (tGmap->isMarked(DUP(*it), ttreated))
		tGmap->unmarkOrbit(DUP(*it), TOrbit(i), ttreated);
	    }
      }
    
    // Librations:
    sGmap->unmarkAll(treated) ;
    tGmap->unmarkAll(ttreated) ;
    sGmap->unmarkAll(duplicated) ;
    tGmap->unmarkAll(tduplicated) ;
    sGmap->freeMark(treated);
    tGmap->freeMark(ttreated);
    sGmap->freeDirectInfo(directInfoIndex);
    sGmap->freeMark(duplicated);
    tGmap->freeMark(tduplicated);
    
    return stGmap;
  }

  /****************************************************************************************/
  /****************************************************************************************/
  /** Infos et manipulation des brins                                                    **/
  /****************************************************************************************/
  /****************************************************************************************/

  CDart* copyDart(SGMapVertex *saGmap, CDart *d)
  {
    CDart* c = saGmap->gmap->addMapDart();
    CVertex *v=saGmap->gmap->findVertex(d) ;
    if (v == NULL)
      return c;
    saGmap->gmap->updateVertex(c,CVertex(v->getX(),v->getY(),v->getZ()));
    return c;
  }

  CDart* newDart(SGMapVertex *saGmap, float x, float y, float z)
  {
    CDart* c = saGmap->gmap->addMapDart();
    saGmap->gmap->updateVertex(c,CVertex(x,y,z));
    return c;
  }
  
  int getNewMark(SGMapVertex *saGmap)
  {
    return (saGmap->gmap->getNewMark());
  }

  void freeMark(SGMapVertex *saGmap, int mark)
  {
    saGmap->gmap->freeMark(mark);
  }

  void markDart(SGMapVertex *saGmap, CDart *d, int mark)
  {
    saGmap->gmap->setMark(d,mark);
  }

  void markNcell(SGMapVertex *saGmap, CDart *d, int n, int mark)
  {
    saGmap->gmap->markOrbit(d,ORBIT_CELL[n],mark);
  }

  void unmarkDart(SGMapVertex *saGmap, CDart *d, int mark)
  {
    saGmap->gmap->unsetMark(d,mark);
  }

  void unmarkNcell(SGMapVertex *saGmap, CDart *d, int n, int mark)
  {
    saGmap->gmap->unmarkOrbit(d,ORBIT_CELL[n],mark);
  }

  CDart* alpha(SGMapVertex *saGmap, CDart *dart, int i)
  {
    return (saGmap->gmap->alpha(dart,i));
  }

  int isFree(SGMapVertex *saGmap, CDart *dart, int i)
  {
    if (saGmap->gmap->isFree(dart,i))
      return 1;
    else
      return 0;
  }

  int isNcellFree(SGMapVertex *saGmap, CDart *ncell, int dim, int i)
  {
    if (saGmap->gmap->isIFreeOrbit(ncell,i,ORBIT_CELL[dim]))
      return 1;
    else
      return 0;
  }

  void sew(SGMapVertex *saGmap, CDart *dart1, CDart *dart2, int i)
  {
    if ((saGmap->gmap->isFree(dart1,i)) && (saGmap->gmap->isFree(dart2,i))) {
      saGmap->gmap->dartSew(dart1,dart2,i);
      normOrbit(saGmap,dart1,ORBIT_CC);
    }
  }

  int get_vertex_coord(SGMapVertex* sg, CDart* s, float* x, float* y, float* z)
  {
    CGMapVertex* g=sg->gmap;
    CVertex *v=g->findVertex(s) ;
    if (v == NULL)
      return -1;
    *x = v->getX() ;
    *y = v->getY() ;
    *z = v->getZ() ;
    return 0;
  }

  void set_vertex_coord(SGMapVertex* sg, CDart* c, double x, double y, double z)
  {
    sg->gmap->updateVertex(c,CVertex(x,y,z));
  }

  int checkGmap(SGMapVertex *saGmap)
  {
    if (saGmap->gmap->checkTopology())
      return 1;
    else
      return 0;
  }

  bool commonPCellOf2NCells(SGMapVertex *saGmap, int n, int p,
			    CDart *c1, CDart *c2,
			    CDart **cp1, CDart **cp2)
  {
    CGMapVertex* aGmap = saGmap->gmap ;

    /* marquage de toutes les cellules de dim p incidente  c1 */
    
    int mark = aGmap->getNewMark();
    CCoverage *cov=aGmap->getDynamicCoverage(c1,ORBIT_CELL[n]);
    cov->reinit() ;
    while (cov->cont()) {
      if (!(aGmap->isMarked(**cov,mark)))
	aGmap->markOrbit(**cov,ORBIT_CELL[p],mark);
      cov->operator++();
    }
    delete cov;

    /* On regarde autour de c2 s'il n y a pas un brin marqu... sinon, on choue */
    cov=aGmap->getDynamicCoverage(c2,ORBIT_CELL[n]);
    cov->reinit() ;
    while (cov->cont()) {
      if (aGmap->isMarked(**cov,mark)) {
	goto found1;
	break;
      }
      cov->operator++();
    }
    delete cov;
    aGmap->unmarkAll(mark);
    aGmap->freeMark(mark);
    return false;

  found1:
    *cp2 = (CDart *)(**cov);
    delete cov;
    
    /* Je marque la cellule de dim p  cote de c2 */
    int mark2 = aGmap->getNewMark();
    aGmap->markOrbit(*cp2,ORBIT_CELL[p], mark2);
    
    /* Je parcours c1 de dim n pour chopper un brin marqu par mark2 */
    /* J'en profite pour tout dmarqu */

    cov=aGmap->getDynamicCoverage(c1,ORBIT_CELL[n]);
    cov->reinit() ;
    while (cov->cont()) {
      if (aGmap->isMarked(**cov,mark)) {
	aGmap->unmarkOrbit(**cov,ORBIT_CELL[p],mark);
      }
      if (aGmap->isMarked(**cov,mark2)) {
	aGmap->unmarkOrbit(**cov,ORBIT_CELL[p],mark2);
	*cp1 = (CDart *)(**cov);
      }
      cov->operator++();
    }
    delete cov;
    aGmap->unmarkAll(mark); // normalement inutile
    aGmap->unmarkAll(mark2); // normalement inutile
    aGmap->freeMark(mark);
    aGmap->freeMark(mark2);

    return true;
  }

  bool commonPCellOf2NCells_opti(SGMapVertex *saGmap, int n, int p,
				 CDart *c1, CDart *c2,
				 CDart **cp1, CDart **cp2)
  {
    CGMapVertex* aGmap = saGmap->gmap ;

    /* marquage de toutes les cellules de dim p incidente  c1 */
    
    int mark = aGmap->getNewMark();
    CCoverage *cov=aGmap->getDynamicCoverage(c1,ORBIT_CELL[n]);
    cov->reinit() ;
    while (cov->cont()) {
      CDart *d = normalForm(saGmap,**cov,p);
      if (!(aGmap->isMarked(d,mark)))
	aGmap->setMark(d,mark);
      cov->operator++();
    }
    delete cov;

    /* On regarde autour de c2 s'il n y a pas un brin marqu... sinon, on choue */
    cov=aGmap->getDynamicCoverage(c2,ORBIT_CELL[n]);
    cov->reinit() ;
    while (cov->cont()) {
      CDart *d = normalForm(saGmap,**cov,p);
      if (aGmap->isMarked(d,mark)) {
	goto found1_opti;
	break;
      }
      cov->operator++();
    }
    delete cov;
    
    /* On enlve les marques autour de c1 */
    cov=aGmap->getDynamicCoverage(c1,ORBIT_CELL[n]);
    cov->reinit() ;
    while (cov->cont()) {
      CDart *d = normalForm(saGmap,**cov,p);
      if (aGmap->isMarked(d,mark))
	aGmap->unsetMark(d,mark);
      cov->operator++();
    }
    delete cov;
    aGmap->freeMark(mark);
    return false;

  found1_opti:
    *cp2 = (CDart *)(**cov);
    delete cov;
    
    /* Je marque la cellule de dim p  cote de c2 */
    int mark2 = aGmap->getNewMark();
    aGmap->setMark(normalForm(saGmap,*cp2,p), mark2);
    
    /* Je parcours c1 de dim n pour chopper un brin marqu par mark2 */
    /* J'en profite pour tout dmarqu */

    cov=aGmap->getDynamicCoverage(c1,ORBIT_CELL[n]);
    cov->reinit() ;
    while (cov->cont()) {
      CDart *d=normalForm(saGmap,**cov,p);
      if (aGmap->isMarked(d,mark)) {
	aGmap->unsetMark(d,mark);
      }
      if (aGmap->isMarked(d,mark2)) {
	aGmap->unsetMark(d,mark2);
	*cp1 = (CDart *)(**cov);
      }
      cov->operator++();
    }
    delete cov;
    aGmap->freeMark(mark);
    aGmap->freeMark(mark2);

    return true;
  }

  CDart* sharedDart(SGMapVertex *saGmap,CDart *vol,int p,CDart *ed,int n)
  {
    CDart *volp = normalForm(saGmap, vol, p);
    CCoverage *cov= saGmap->gmap->getDynamicCoverage(ed,ORBIT_CELL[n]);
    cov->reinit() ;
    while (cov->cont()) {
      if ( normalForm(saGmap,**cov,p) == volp) {
	goto found2;
	break;
      }
      cov->operator++();
    }

    delete cov;
    return NULL;

  found2:
    CDart *shared = **cov;
    delete cov;
    return shared;
  }
  /****************************************************************************************/
  /****************************************************************************************/
  /** Construction et manipulation de cellules                                           **/
  /****************************************************************************************/
  /****************************************************************************************/

  CDart* new_ncell(SGMapVertex *saGmap, int dim)
  {
    CGMapVertex* aGmap = saGmap->gmap ;
    CDart* d1=NULL; CDart* d2=NULL; CDart* d3=NULL; CDart* d4=NULL;
    CDart* d_1=NULL;CDart* d_2=NULL;CDart* d_3=NULL;CDart* d_4=NULL;
    CDart* tab[8] ;

    switch(dim)
      {
      case 0 :
	d1 = aGmap->addMapDart();
	set_vertex_coord(saGmap,d1,0,0,0);
	normOrbit(saGmap,d1,ORBIT_CELL[0]);
	return d1 ;
	break ;
      case 1:
	tab[0] = aGmap->addMapDart();
	set_vertex_coord(saGmap,tab[0],0,0,0);
	tab[1] = aGmap->addMapDart();
	set_vertex_coord(saGmap,tab[1],0,0,0);
	aGmap->sew0(tab[0],tab[1]);
	normTabNCell(saGmap, 2, tab ) ;
	return (tab[0]);
	break;
      case 2:
	d1 = aGmap->createTopoTriangle() ;
	set_vertex_coord(saGmap,d1,0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha01(d1),0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha10(d1),0,0,0);
	normOrbit(saGmap,d1,ORBIT_CELL[2]);
	return (d1);
	break;
      case 3:
	d1 = aGmap->createTopoTriangle(); d_1 = aGmap->alpha1(d1) ;
	set_vertex_coord(saGmap,d1,0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha01(d1),0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha10(d1),0,0,0);
	d2 = aGmap->createTopoTriangle(); d_2 = aGmap->alpha1(d2) ;
	set_vertex_coord(saGmap,d2,0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha01(d2),0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha10(d2),0,0,0);
 	d3 = aGmap->createTopoTriangle(); d_3 = aGmap->alpha1(d3) ;
	set_vertex_coord(saGmap,d3,0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha01(d3),0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha10(d3),0,0,0);
 	d4 = aGmap->createTopoTriangle(); d_4 = aGmap->alpha1(d4) ;
	set_vertex_coord(saGmap,d4,0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha01(d4),0,0,0);
	set_vertex_coord(saGmap,saGmap->gmap->alpha10(d4),0,0,0);

	aGmap->dartSew2(d1,d2);
       	d1=aGmap->alpha0(d1);d2=aGmap->alpha0(d2);
	aGmap->dartSew2(d1,d2);
 	d1=aGmap->alpha1(d1); d2=aGmap->alpha1(d2);
	aGmap->dartSew2(d1,d3); aGmap->dartSew2(d2,d_3);
	d1=aGmap->alpha0(d1);d2=aGmap->alpha0(d2);
	d3=aGmap->alpha0(d3);d_3=aGmap->alpha0(d_3);
	aGmap->dartSew2(d1,d3);aGmap->dartSew2(d2,d_3);

	aGmap->dartSew2(d_1,d4);aGmap->dartSew2(d_2,d_4);
	d_1=aGmap->alpha0(d_1);d_2=aGmap->alpha0(d_2);
	d4=aGmap->alpha0(d4);d_4=aGmap->alpha0(d_4);
	aGmap->dartSew2(d_1,d4);aGmap->dartSew2(d_2,d_4);

	d_3=aGmap->alpha1(d_3);d3=aGmap->alpha1(d3);
	d_4=aGmap->alpha1(d_4);d4=aGmap->alpha1(d4);
	aGmap->dartSew2(d_3,d_4);aGmap->dartSew2(d3,d4);
	d_3=aGmap->alpha0(d_3);d3=aGmap->alpha0(d3);
	d_4=aGmap->alpha0(d_4);d4=aGmap->alpha0(d4);
	aGmap->dartSew2(d_3,d_4);aGmap->dartSew2(d3,d4);

	normOrbit(saGmap,d1,ORBIT_CELL[3]);

	return(aGmap->alpha1(d_1));

	break;
      }

    return NULL;
  }

  CDart* new_ncell_polygon(SGMapVertex *aGmap, int nbside)
  {
    CDart* d = aGmap->gmap->createRegularPolygon(nbside);
    normOrbit(aGmap,d,ORBIT_CELL[2]);
    return (d);
  }

  CDart* new_ncell_square(SGMapVertex *aGmap, int Ax,int Ay, int dim)
  {
    CDart* ASquareCorners[2][2] = {{NULL,NULL},{NULL,NULL}};
    aGmap->gmap->createSquareIMeshed(Ax,Ay,dim,ASquareCorners) ;
    normOrbit(aGmap,ASquareCorners[0][0],ORBIT_CELL[3]);
    normOrbit(aGmap,ASquareCorners[0][1],ORBIT_CELL[3]);
    normOrbit(aGmap,ASquareCorners[1][0],ORBIT_CELL[3]);
    normOrbit(aGmap,ASquareCorners[1][1],ORBIT_CELL[3]);
    return (ASquareCorners[0][0]);
  }

  void new_ncell_sphere(SGMapVertex *aGmap, int nbmeridians, int nbparallels, CDart **southPole, CDart **northPole)
  {
    aGmap->gmap->createSphere(nbmeridians, nbparallels, southPole, northPole) ;
    normOrbit(aGmap,*southPole,ORBIT_CELL[3]);
  }

  CDart* new_ncell_polyline(SGMapVertex *aGmap, int nbverteces)
  {
    CDart *ret = NULL ;
    for (int i = 0;i<nbverteces;i++)
      ret = aGmap->gmap->prolongPolyline(ret,CVertex(0,0,0)) ;
    
    normOrbit(aGmap,ret,ORBIT_CELL[3]);
    return ret;
  }

  void glue2FacesBy4Verteces(SGMapVertex *saGmap, CDart *c1, CDart *c2,
			     CDart *v1c1, CDart *v1c2,
			     CDart *v2c1, CDart *v2c2)
  {
    CGMapVertex* aGmap = saGmap->gmap ;
    CDart *new_v1c1 = NULL;
    CDart *new_v1c2 = NULL;
    CDart *new_v2c1 = NULL;
    CDart *new_v2c2 = NULL;
    int mv1 = aGmap->getNewMark() ;
    int mv2 = aGmap->getNewMark() ;
    aGmap->unmarkAll(mv1) ;
    aGmap->unmarkAll(mv2) ;
    aGmap->markOrbit(v1c1, ORBIT_VERTEX, mv1);
    aGmap->markOrbit(v1c2, ORBIT_VERTEX, mv1);
    aGmap->markOrbit(v2c1, ORBIT_VERTEX, mv2);
    aGmap->markOrbit(v2c2, ORBIT_VERTEX, mv2);

    CCoverage * cov = aGmap->getDynamicCoverage(c1, ORBIT_FACE);
    for (int i=0; cov->cont() && i<2; ++(*cov)) {
      CDart *d = **cov ;
      if (aGmap->isMarked(d,mv1) &&  aGmap->isMarked(aGmap->alpha0(d),mv2) ) {
	new_v1c1 = d ;
	new_v2c1 = aGmap->alpha0(d) ;
	i = 2;
      }
      else if (aGmap->isMarked(d,mv2) &&  aGmap->isMarked(aGmap->alpha0(d),mv1) ) {
	new_v2c1 = d ;
	new_v1c1 = aGmap->alpha0(d) ;
	i = 2;
      }
    }
    delete cov ;

    cov = aGmap->getDynamicCoverage(c2, ORBIT_FACE);
    for (int i=0; cov->cont() && i<2; ++(*cov)) {
      CDart *d = **cov ;
      if (aGmap->isMarked(d,mv1) &&  aGmap->isMarked(aGmap->alpha0(d),mv2) ) {
	new_v1c2 = d ;
	new_v2c2 = aGmap->alpha0(d) ;
	i = 2;
      }
      else if (aGmap->isMarked(d,mv2) &&  aGmap->isMarked(aGmap->alpha0(d),mv1) ) {
	new_v2c2 = d ;
	new_v1c2 = aGmap->alpha0(d) ;
	i = 2;
      }
    }
    delete cov ;

    if (!(aGmap->isFree2(new_v1c1))
	||!(aGmap->isFree2(new_v1c2))
	||!(aGmap->isFree2(new_v2c1))
	||!(aGmap->isFree2(new_v2c2)))
      {
	cerr << "Stub_gmap: glue2FacesBy4Verteces: fatal error: some darts are already sewn" << endl ;
	cerr.flush();
	exit(0);
      }
    //cout << (aGmap->getSewDimension(new_v1c1,new_v1c2)) << endl;
    //cout << (aGmap->getSewDimension(new_v2c1,new_v2c2)) << endl;
    //aGmap->intuitiveSew(new_v1c1,new_v1c2);
    aGmap->dartSew2(new_v1c1,new_v1c2);
    aGmap->dartSew2(new_v2c1,new_v2c2);
    //cout << (aGmap->getSewDimension(new_v2c1,new_v2c2)) << endl;
    //if (aGmap->checkTopology()) cout << "ok" << endl; else {cout << "oups" << endl;exit(0);}
    aGmap->unmarkAll(mv1) ;
    aGmap->unmarkAll(mv2) ;
    aGmap->freeMark(mv1) ;
    aGmap->freeMark(mv2) ;

    normOrbit(saGmap,c1,ORBIT_CC);
    normOrbit(saGmap,c2,ORBIT_CC);
  }

  //on insere un sommet sur un arc
  CDart* insert_vertex(SGMapVertex *saGmap, CDart *e, CDart **re1, CDart **re2)
  {
    CGMapVertex *aGmap = saGmap->gmap ;
    CDart *e1, *e2, *v ;

    if (aGmap->isFree0(e)) {
      cerr << "Stub_gmap: insertVertex: warning: e is 0-free !!! strange... but why not" << endl ;
      cerr.flush();
      e2 = NULL;
    } else {
      e2 = aGmap->alpha0(e) ;
    }

    v  = aGmap->insertVertex(e) ;
    e1 = e ;

    normOrbit(saGmap,v,ORBIT_CELL[0]);
    //    normOrbit(saGmap,e1,ORBIT_CELL[0]);
    //    normOrbit(saGmap,e2,ORBIT_CELL[0]);

    *re1 = e1;
    *re2 = e2;
    return v;
  }

  //on insere un arc dans une face
  CDart* insert_edge(SGMapVertex *saGmap, CDart *vv1, CDart *vv2, CDart **rf1, CDart **rf2)
  {
    CGMapVertex *aGmap = saGmap->gmap ;
    CDart *e, *f1, *f2;
    CDart *v1, *v2;

    //v1=vv1;
    //v2=vv2;
    //if ( (normalForm(saGmap,v1,2)) != (normalForm(saGmap,v2,2)) ) {//(!(aGmap->isSameOrbit(v1,v2,ORBIT_CELL[2]))) {
    if (!(commonPCellOf2NCells_opti(saGmap,0,2,vv1,vv2,&v1,&v2))) {
      cerr << "Stub_gmap: insertEdge: fatal error: vv1 and vv2 must belong to the same face" << endl ;
      cerr.flush();
      exit(0);
    }

    /* On teste pour ne pas faire de face non oriente */
    bool flag=false;
    CDart* tmp = aGmap->alpha01(v1);
    while ( (tmp != v1) && !flag ) {
      if (tmp == v2) {
	flag = true;
      }
      tmp = aGmap->alpha01(tmp) ;
    }
    if (flag)
      v2 = aGmap->alpha1(v2) ;

    f1 = v1;
    f2 = (aGmap->isFree1(f1))?NULL:aGmap->alpha1(f1);

    e = aGmap->insertEdge(v1,v2);
    normOrbit(saGmap,e,ORBIT_EDGE);

    *rf1 = f1;
    *rf2 = f2;
    return e;
  }

  //on insere une face dans un volume
  CDart* insert_face(SGMapVertex *saGmap, CDart *vol, CDart **et, int nb_e, CDart **rv1, CDart **rv2)
  {
    CGMapVertex *aGmap = saGmap->gmap ;
    CDart *f = NULL, *v1, *v2;

    if (nb_e <= 1) {
      cerr << "Stub_gmap: insertFace: fatal error: nb_e must be greater than 1" << endl ;
      cerr.flush();
      exit(0);
    }

    for (int i=1;i<nb_e;i++)
      if (vol != (normalForm(saGmap,et[i],3))) {
	CDart* tmp = sharedDart(saGmap,vol,3,et[i],1);
	if (tmp == NULL) {
	  cerr << "Stub_gmap: insertFace: fatal error: all vt must belong to the same volume" << endl ;
	  cerr.flush();
	  exit(0);
	}
	et[i] = tmp;
      }


    v1 = et[0];

    int aMark = aGmap->getNewMark() ;
    for (int i=0;i<nb_e;i++) {
      aGmap->setMark(et[i],aMark) ; //aGmap->markOrbit(et[i],ORBIT_CELL[1],aMark);
    }

    f = aGmap->insertFace(et[0],aMark,true);
    aGmap->unmarkAll(aMark);
    v1 = f;
    v2 = aGmap->alpha3(v1);

    bool flag = true;
    for (CDynamicCoverage01 cov(aGmap,v2); flag & cov.cont(); ++cov)
      if ( !(aGmap->isFree2(*cov)) )
	flag = false;

    if (flag) {
      aGmap->unsew3(v2);
      aGmap->markOrbit(v2,ORBIT_CELL[2],aMark);
      aGmap->deleteMarkedDarts(aMark);
      v2 = NULL;
    }
    aGmap->freeMark(aMark);

    normOrbit(saGmap,f,ORBIT_CC);
    
    *rv1 = v1;
    *rv2 = v2;
    return f;
  }    

  CDart *create_edge(SGMapVertex *saGmap, CDart *v1, CDart *v2)
  {
    CGMapVertex *aGmap = saGmap->gmap;
    CDart *vv1, *vv2;
    if ((!aGmap->isFree1(v1)) || (!aGmap->isFree2(v1)) || (!aGmap->isFree3(v1)) ||
	(!aGmap->isFree1(v1)) || (!aGmap->isFree2(v2)) || (!aGmap->isFree3(v2))
	) {
      cerr << "Stub_gmap: create_edge: fatal error: the both darts must be alpha 1,2 and 3 free"
	   << endl ;
      cerr.flush();
      exit(0);
    }

    if ((aGmap->isFree0(v1)) && (aGmap->isFree0(v2))) {
      vv1 = v1;
      vv2 = v2;
      aGmap->sew0(vv1,vv2);
    } else if ((!(aGmap->isFree0(v1))) && (!(aGmap->isFree0(v2)))) {
      vv1 = aGmap->addMapDart();
      vv2 = aGmap->addMapDart();
      aGmap->sew1(v1,vv1);
      aGmap->sew1(v2,vv2);
      aGmap->sew0(vv1,vv2);
    } else if ((!(aGmap->isFree0(v1))) && (aGmap->isFree0(v2))) {
      vv2 = aGmap->addMapDart();
      vv1 = v1;
      aGmap->sew0(v2,vv2);
      aGmap->sew1(vv1,vv2);
    } else {
      vv1 = aGmap->addMapDart();
      vv2 = v2;
      aGmap->sew0(v1,vv1);
      aGmap->sew1(vv2,vv1);
    }

    normOrbit(saGmap,vv1,ORBIT_CC);

    return vv1;
  }

  /* On donne un tableau de couple : On cherche  fusionner v_tab[i][0] avec v_tab[i][1]*/
  void can_create_face(SGMapVertex *saGmap, CDart ***v_tab, int i, int j)
  {
    CGMapVertex *aGmap = saGmap->gmap ;
    CDart *tmp1, *tmp2;
    if (!(commonPCellOf2NCells(saGmap,0,1,v_tab[i][1],v_tab[j][0],&tmp1,&tmp2))) {
      cerr << "Stub_gmap: create_face: fatal error: some edge doesn't exist" << endl ;
      cerr.flush();
      exit(0);
    }

    if ((!aGmap->isFree2(tmp1)) || (!aGmap->isFree3(tmp1)) ||
	(!aGmap->isFree2(tmp2)) || (!aGmap->isFree3(tmp2))
	) {
      cerr << "Stub_gmap: create_face: fatal error: the both darts must be alpha 2 and 3 free"
	   << endl ;
      cerr.flush();
      exit(0);
    }
    v_tab[i][1] = tmp1 ;
    v_tab[j][0] = tmp2 ;
  }

  void double_edge(SGMapVertex *saGmap, CDart* d1, CDart* d2, CDart** t1, CDart** t2)
  {
    CGMapVertex *aGmap = saGmap->gmap;
    if ((aGmap->isFree1(d1)) && (aGmap->isFree1(d2))) {
      *t1 = d1;
      *t2 = d2;
      return;
    }

    /* Il faut ddoubler */
    //float x,y,z;
    CDart *n1 = aGmap->addMapDart();
    //get_vertex_coord(saGmap,d1,&x,&y,&z);
    //set_vertex_coord(saGmap,n1,x,y,z);
    CDart *n2 = aGmap->addMapDart();
    //get_vertex_coord(saGmap,d2,&x,&y,&z);
    //set_vertex_coord(saGmap,n2,x,y,z);

    aGmap->dartSew2(n1,d1);
    aGmap->dartSew2(n2,d2);
    aGmap->dartSew0(n1,n2);

    *t1 = n1;
    *t2 = n2;
    return;   
  }

  CDart *create_face(SGMapVertex *saGmap, CDart ***v_tab, int v_nb)
  {
    CGMapVertex *aGmap = saGmap->gmap;
    CDart *t1, *t2;

    /* Trouver les bons brins ; on en profite pour les vrif...*/
    if (v_nb <=1) {
      cerr << "Stub_gmap: create_face: fatal error: v_nb must be greater than 1" << endl ;
      cerr.flush();
      exit(0);
    }
    for(int i=0;i<v_nb-1;i++) {
      can_create_face(saGmap,v_tab,i,i+1);
    }
    can_create_face(saGmap,v_tab,v_nb-1,0);

    /* On cherche  ddoubler les arcs appartenant dj  une face (alpha1 cousu) */
    for(int i=0;i<v_nb-1;i++) {
      double_edge(saGmap,v_tab[i][1],v_tab[i+1][0],&t1,&t2);
      v_tab[i][1] = t1;
      v_tab[i+1][0] = t2;
    }
    double_edge(saGmap,v_tab[v_nb-1][1],v_tab[0][0],&t1,&t2);
    v_tab[v_nb-1][1]=t1;
    v_tab[0][0]=t2;

    /* Reste  tourner une dernire fois pour faire les alpha1 !!!! */

    for(int i=0;i<v_nb;i++) {
      aGmap->dartSew1(v_tab[i][0],v_tab[i][1]);
    }

    /* Normaliser !!! pas oublier pour create_edge... */
    normOrbit(saGmap,v_tab[0][0],ORBIT_CC);
    return (v_tab[0][0]);

  }

  CDart *double_face(SGMapVertex *saGmap, CDart* f)
  {
    CGMapVertex *aGmap = saGmap->gmap;
    CCoverage* cov = NULL;
    int dInfo;

    dInfo = aGmap->getNewDirectInfo();

    cout << "passe1" << endl ; cout.flush() ;
    cov = aGmap->getStaticCoverage(f, ORBIT_FACE);
    cov->reinit();
    while(cov->cont()) {
      CDart* n = aGmap->addMapDart();
      aGmap->setDirectInfo(**cov,dInfo,(void*)n);
      ++*cov;
    }

    cout << "passe2" << endl ; cout.flush() ;
    cov->reinit();
    while(cov->cont()) {
      cout << "passe2.1" << endl ; cout.flush() ;
      CDart* d0 = (CDart*)aGmap->getDirectInfo(aGmap->alpha0(**cov),dInfo);
      cout << "passe2.2" << endl ; cout.flush() ;
      CDart* d1 = (CDart*)aGmap->getDirectInfo(aGmap->alpha1(**cov),dInfo);
//       cout << "passe2.3" << endl ; cout.flush() ;
//       CDart* d2 = (CDart*)aGmap->getDirectInfo(aGmap->alpha2(**cov),dInfo);
      cout << "passe2.4" << endl ; cout.flush() ;
      CDart* d3 = (CDart*)aGmap->getDirectInfo(aGmap->alpha3(**cov),dInfo);

      cout << "passe2.5" << endl ; cout.flush() ;
      CDart* n = (CDart*)aGmap->getDirectInfo(**cov,dInfo);

      cout << "passe2.6" << endl ; cout.flush() ;
      if ((n != d0)&&(aGmap->isFree0(n))) aGmap->dartSew0(n,d0);
      cout << "passe2.7" << endl ; cout.flush() ;
      if ((n != d1)&&(aGmap->isFree1(n))) aGmap->dartSew1(n,d1);
      cout << "passe2.8" << endl ; cout.flush() ;
//       On fait pas puisque ont est dans une face... Il faut pas change de face !!!!
//       if ((n != d2)&&(aGmap->isFree2(n))) aGmap->dartSew2(n,d2);
//       cout << "passe2.9" << endl ; cout.flush() ;
      if ((n != d3)&&(aGmap->isFree3(n))) aGmap->dartSew3(n,d3);
      cout << "passe2.10" << endl ; cout.flush() ;
      
      ++*cov;
    }
    
    cout << "passe3" << endl ; cout.flush() ;
    cov->reinit();
    while(cov->cont()) {
      aGmap->dartSew3(**cov,(CDart*)aGmap->getDirectInfo(**cov,dInfo));
      ++*cov;
    }

    delete cov;
    CDart* df = (CDart*)aGmap->getDirectInfo(f,dInfo);
    aGmap->freeDirectInfo(dInfo);
    return df;
  }

  CDart *create_volume(SGMapVertex *saGmap, CDart **f_tab, int f_nb, CDart ***v_tab, int *l_tab, int v_nb)
  {
    CGMapVertex *aGmap = saGmap->gmap;
    int nb_faces = 0, i, j;
    CCoverage* cov;

    cout << v_nb << endl ; cout.flush() ;
    /* On marque les cellules qui vont tre utilises */

    /* Vrification : chaque face doit tre 3-libre - on marque les cellules  doubler */
    if (v_nb <=1) {
      cerr << "Stub_gmap: volume_face: fatal error: v_nb must be greater than 1" << endl ;
      cerr.flush();
      exit(0);
    }

    int face_mark = aGmap->getNewMark();
    int double_mark = aGmap->getNewMark();
    for(i=0;i<v_nb;i++) {
      for(j=0;j<l_tab[i];j++) {
	cout << i << " " << j << endl ; cout.flush() ;
	CDart* f = v_tab[i][j] ;
	if (!(aGmap->isMarked(f,face_mark))) {
	  bool flag = false ;
	  aGmap->markOrbit(f,ORBIT_FACE,face_mark);
	  nb_faces++;
	  cov = aGmap->getDynamicCoverage(f,ORBIT_FACE) ;
	  while(cov->cont()) {
	    if (!(aGmap->isFree3(**cov))) {
	      delete cov;
	      cerr << "Stub_gmap: volume_face: fatal error: all faces must be 3-free" << endl ;
	      cerr.flush();
	      exit(0);
	    }
	    if (!(aGmap->isFree2(**cov)))
	      flag = true;
	    ++*cov;
	  }
	  delete cov;
	  if (flag)
	    aGmap->markOrbit(f,ORBIT_FACE,double_mark);
	}
      }
    }
    cout << nb_faces << endl ; cout.flush() ;

    /* On ddouble les faces marques par double_mark et on en profite pour dtruire cette marque */

    for(i=0;i<v_nb;i++) {
      for(j=0;j<l_tab[i];j++) {
	cout << i << " - " << j << endl ; cout.flush() ;
	CDart* f = v_tab[i][j] ;
	if (aGmap->isMarked(f,face_mark)) {
	  cout << "Je suis marque par face..." ; cout.flush() ;
	  if (aGmap->isMarked(f,double_mark)) {
	    cout << " et par double" << endl ; cout.flush() ;
	    v_tab[i][j] = double_face(saGmap,f) ;
	  } else {
	    cout << " mais pas par double" << endl ; cout.flush() ;
	  }
	  aGmap->unmarkOrbit(f,ORBIT_FACE,face_mark);
	} else {
	  cout << "Je ne suis pas marque par face..." ; cout.flush() ;
	  if (aGmap->isMarked(f,double_mark)) {
	    cout << " mais par double oui" << endl ; cout.flush() ;
	    v_tab[i][j] = aGmap->alpha3(f) ;
	  } else {
	    cout << " ni par double" << endl ; cout.flush() ;
	  }
	}
	
      }
    }

    return NULL;
  }



  /* Enlever une ncell en fusionnant les exactement 2 (n+1)cells qui la partage */
  CDart* basic_remove(SGMapVertex* saGmap, CDart* c, int dim)
  {
    CGMapVertex* aGmap = saGmap->gmap;

    /* On vrifie que la cellule a exactement 2 cofaces */
    int nb_cofaces ;
    int dim_coface = dim + 1;
    switch (dim) {
    case 0:
      {
	aGmap->countCells(c,ORBIT_CELL[dim],NULL,&nb_cofaces,NULL,NULL,NULL);
	break;
      }
    case 1:
      {
	aGmap->countCells(c,ORBIT_CELL[dim],NULL,NULL,&nb_cofaces,NULL,NULL);
	break;
      }
    case 2:
      {
	aGmap->countCells(c,ORBIT_CELL[dim],NULL,NULL,NULL,&nb_cofaces,NULL);
	break;
      }
    default:
      {
	cerr << "Stub_gmap: basic_remove: fatal error: dim must be 0, 1 or 2." << endl ;
	cerr.flush();
	exit(0);
      }
    }
    if (nb_cofaces != 2) {
      cerr << "Stub_gmap: basic_remove: fatal error: c must exactly have two cofaces." << endl ;
      cerr.flush();
      exit(0);
    }

    CDart *cf_c = aGmap->alpha(c,dim_coface) ;
    if (!(aGmap->canMerge(c, cf_c, dim_coface))) {
      cerr << "Stub_gmap: basic_remove: fatal error: cannot merge c and cf_c." << endl ;
      cerr.flush();
      exit(0);
    }

    CDart *f = NULL;
    CCoverage *cov = aGmap -> getDynamicCoverage(c,ORBIT_CELL[dim_coface]) ;
    cov->reinit();
    while(cov->cont()) {
      if ( (normalForm(saGmap,**cov,dim)) != (normalForm(saGmap,c,dim)) ) {
	f = **cov;
	break;
      }
      ++*cov;
    }
    delete cov;

    aGmap->merge(c,cf_c,dim_coface,true);

    normOrbit(saGmap,f,ORBIT_CC);

    return f;
  }

  int basic_delete(SGMapVertex* saGmap, CDart* c, int dim, CDart ***nfaces)
  {
    CGMapVertex* aGmap = saGmap->gmap;

    /* On vrifie que la cellule a exactement 2 cofaces */
    int nb_cofaces, nb_faces ;
    int dim_face = dim-1 ;
    switch (dim) {
    case 1:
      {
	aGmap->countCells(c,ORBIT_CELL[dim],&nb_faces,NULL,&nb_cofaces,NULL,NULL);
	break;
      }
    case 2:
      {
	aGmap->countCells(c,ORBIT_CELL[dim],NULL,&nb_faces,NULL,&nb_cofaces,NULL);
	break;
      }
    case 3:
      {
	aGmap->countCells(c,ORBIT_CELL[dim],NULL,NULL,&nb_faces,NULL,NULL);
	nb_cofaces = 1;
	break;
      }
    default:
      {
	cerr << "Stub_gmap: basic_delete: fatal error: dim must be 1, 2 or 3." << endl ;
	cerr.flush();
	exit(0);
      }
    }
    if (nb_cofaces != 1) {
      cerr << "Stub_gmap: basic_delete: fatal error: c must exactly have 1 coface (closed or not !!!)." << endl ;
      cerr.flush();
      exit(0);
    }

    int delete_mark = aGmap->getNewMark();
    int other_mark = aGmap->getNewMark();

    *nfaces = new (CDart*)[nb_faces] ;
    int cur_faces = 0;

    CCoverage* cov = aGmap->getStaticCoverage(c,ORBIT_CELL[dim]);
    cov->reinit();
    while(cov->cont()) {
      if (!(aGmap->isMarked(**cov,delete_mark)))
	if (!(aGmap->isFree(**cov,dim))) {
	  (*nfaces)[cur_faces] = aGmap->alpha(**cov,dim);
	  cur_faces++;
	  aGmap->unsew(**cov,dim);
	  aGmap->markOrbit(**cov,ORBIT_CELL[dim_face],delete_mark);
	} else {
	  CDart *normal_face = normalForm(saGmap,**cov,dim_face);
	  if (!(aGmap->isMarked(normal_face,other_mark))) {
	    aGmap->setMark(normal_face,other_mark);
	    (*nfaces)[cur_faces] = normal_face;
	    cur_faces++;
	  }
	}
      ++*cov;
    }

    cov->reinit();
    while(cov->cont()) {
      if (!(aGmap->isFree(**cov,dim_face))) {
	aGmap->unsew(**cov,dim_face);
      }
      ++*cov;
    }
    delete cov;

    aGmap->deleteMarkedDarts(delete_mark);

    /* Trouver un moyen de rcuprer les faces puis les normaliser... */

    for(cur_faces=0;cur_faces<nb_faces;cur_faces++) {
      if (aGmap->isMarked((*nfaces)[cur_faces],other_mark))
	aGmap->unsetMark((*nfaces)[cur_faces],other_mark);
      normOrbit(saGmap,(*nfaces)[cur_faces],ORBIT_CC);
    }

    aGmap->freeMark(delete_mark);
    aGmap->freeMark(other_mark);

    return nb_faces;
  }

  /****************************************************************************************/
  /****************************************************************************************/
  /** Enumration de cellules et des voisins...                                          **/
  /****************************************************************************************/
  /****************************************************************************************/
  
  /* Tableaux de tous les brins d'une gmap */
  
  int allDarts(SGMapVertex * sg, CDart ***darts)
  {
    CGMapVertex* g=sg->gmap;
    int nb_darts = 0;
    g->getGlobalCharacteristics(&nb_darts,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL) ;
    
    if (darts == NULL) return nb_darts ;
    
    *darts = new (CDart*)[nb_darts] ;
    CStaticCoverageAll* cov = new CStaticCoverageAll(g) ;
    cov->reinit() ;
    int i=0 ;
    while (cov->cont()) {      
      (*darts)[i] = (cov->operator*());
      i++;      
      cov->operator++() ;
    }
    delete cov;
    
    return nb_darts ;
  }
  
  /* Tableau de toutes les n-cells d'une gmap */
  
  int allNCells(SGMapVertex *sg, int dim, CDart ***ncells)
  {
    CGMapVertex* g=sg->gmap;
    int nb_ncells = 0;
    TOrbit aOrbit = ORBIT_CC ;
    switch (dim) {
    case 0: // verteces
      g->getGlobalCharacteristics(NULL,&nb_ncells,NULL,NULL,NULL,NULL,NULL,NULL,NULL,NULL) ;
      aOrbit = ORBIT_VERTEX ;
      break;
    case 1: // edges
      g->getGlobalCharacteristics(NULL,NULL,&nb_ncells,NULL,NULL,NULL,NULL,NULL,NULL,NULL) ;
      aOrbit = ORBIT_EDGE ;
      break;
    case 2: // faces
      g->getGlobalCharacteristics(NULL,NULL,NULL,&nb_ncells,NULL,NULL,NULL,NULL,NULL,NULL) ;
      aOrbit = ORBIT_FACE ;
      break;
    case 3: // volumes
      g->getGlobalCharacteristics(NULL,NULL,NULL,NULL,&nb_ncells,NULL,NULL,NULL,NULL,NULL) ;
      aOrbit = ORBIT_VOLUME ;
      break;
    }
    
    if (ncells == NULL) return nb_ncells ;
    
    *ncells = new (CDart*)[nb_ncells] ;
    CStaticCoverageAll* cov = new CStaticCoverageAll(g) ;
    cov->reinit() ;
    int i=0 ;
    
    int aMark = g->getNewMark() ;
    //ATTENTION g->unmarkAll(aMark) ;
    
    while (cov->cont()) {
      CDart *current = (cov->operator*()) ;
      if (!(current->getMark(aMark))) {
	g->markOrbit(current,aOrbit,aMark) ;
	(*ncells)[i] = normalForm(sg,current,dim) ;
	i++;
      }
      cov->operator++() ;
    }
    g->unmarkAll(aMark) ;
    g->freeMark(aMark) ;
    delete cov;
    
    return nb_ncells ;
  }
  
  /* tableau de toutes les (n-1)-faces d'une n-cell */
  int  allFaces(SGMapVertex* sg, CDart* ncell, int dim, CDart ***faces)
  {
    CGMapVertex* g=sg->gmap;
    if (dim == 0) return 0;
    
    int nb_faces = 0;
    TOrbit aOrbit = ORBIT_CC ;
    TOrbit coOrbit = ORBIT_CC ;
    switch (dim) {
    case 1: // edges
      coOrbit = ORBIT_EDGE ;
      aOrbit = ORBIT_VERTEX ;
      g->countCells(ncell,coOrbit,&nb_faces,NULL,NULL,NULL,NULL,NULL) ;
      break;
    case 2: // faces
      coOrbit = ORBIT_FACE ;
      aOrbit = ORBIT_EDGE ;
      g->countCells(ncell,coOrbit,NULL,&nb_faces,NULL,NULL,NULL,NULL) ;
      break;
    case 3: // volumes
      coOrbit = ORBIT_VOLUME ;
      aOrbit = ORBIT_FACE ;
      g->countCells(ncell,coOrbit,NULL,NULL,&nb_faces,NULL,NULL,NULL) ;
      break;
    }
    
    if (faces == NULL) return nb_faces;
    
    *faces = new (CDart*)[nb_faces] ;
    CCoverage *cov=g->getStaticCoverage(ncell,coOrbit);
    cov->reinit() ;
    int i=0 ;
    
    int aMark = g->getNewMark() ;
    //ATTENTION g->unmarkAll(aMark) ;
    
    while (cov->cont()) {
      CDart *current = (cov->operator*()) ;
      if (!(current->getMark(aMark))) {
	g->markOrbit(current,aOrbit,aMark) ;
	(*faces)[i] = normalForm(sg,current,dim-1) ;
	i++;
      }
      cov->operator++() ;
    }
    g->unmarkAll(aMark) ;
    g->freeMark(aMark) ;
    delete cov;
    
    return nb_faces;
  }
  
  /* tableau de toutes les (n+1)-cofaces d'une n-cell */

  int allCofaces(SGMapVertex* sg, CDart* ncell, int dim, CDart ***cofaces)
  {
    CGMapVertex* g = sg->gmap;
    if (dim == 3) return 0;
    
    int nb_cofaces = 0;
    TOrbit aOrbit = ORBIT_CC ;
    TOrbit coOrbit = ORBIT_CC ;
    switch (dim) {
    case 0: // edges
      coOrbit = ORBIT_EDGE ;
      aOrbit = ORBIT_VERTEX ;
      g->countCells(ncell,aOrbit,NULL,&nb_cofaces,NULL,NULL,NULL,NULL) ;
      break;
    case 1: // faces
      coOrbit = ORBIT_FACE ;
      aOrbit = ORBIT_EDGE ;
      g->countCells(ncell,aOrbit,NULL,NULL,&nb_cofaces,NULL,NULL,NULL) ;
      break;
    case 2: // volumes
      coOrbit = ORBIT_VOLUME ;
      aOrbit = ORBIT_FACE ;
      g->countCells(ncell,aOrbit,NULL,NULL,NULL,&nb_cofaces,NULL,NULL) ;
      break;
    }
    
    if (cofaces == NULL) return nb_cofaces;
    
    *cofaces = new (CDart*)[nb_cofaces] ;
    CCoverage *cov=g->getStaticCoverage(ncell,aOrbit);
    cov->reinit() ;
    int i=0 ;
    
    int aMark = g->getNewMark() ;
    //ATTENTION g->unmarkAll(aMark) ;
    
    while (cov->cont()) {
      CDart *current = (cov->operator*()) ;
      if (!(current->getMark(aMark))) {
	g->markOrbit(current,coOrbit,aMark) ;
	(*cofaces)[i] = normalForm(sg,current,dim+1) ;
	i++;
      }
      cov->operator++() ;
    }
    g->unmarkAll(aMark) ;
    g->freeMark(aMark) ;
    delete cov;
    
    return nb_cofaces;
  }

  /* Recherche des i-cells incidentes  une n-cells */
  
  int allIncidentCells(SGMapVertex* sg, CDart* ncell, int n, int i, CDart ***icells)
  {
    CGMapVertex* g=sg->gmap;
    int nb_icells = 0;
    
    TOrbit nOrbit = ORBIT_CC ;
    TOrbit iOrbit = ORBIT_CC ;
    
    switch(n) {
    case 0:
      nOrbit = ORBIT_VERTEX;
      break;
    case 1:
      nOrbit = ORBIT_EDGE;
      break;
    case 2:
      nOrbit = ORBIT_FACE;
      break;
    case 3:
      nOrbit = ORBIT_VOLUME;
      break;
    }
    
    switch(i) {
    case 0:
      iOrbit = ORBIT_VERTEX ;
      g->countCells(ncell,nOrbit,&nb_icells,NULL,NULL,NULL,NULL,NULL) ;
      break;
    case 1:
      iOrbit = ORBIT_EDGE ;
      g->countCells(ncell,nOrbit,NULL,&nb_icells,NULL,NULL,NULL,NULL) ;
      break;
    case 2:
      iOrbit = ORBIT_FACE ;
      g->countCells(ncell,nOrbit,NULL,NULL,&nb_icells,NULL,NULL,NULL) ;
      break;
    case 3:
      iOrbit = ORBIT_VOLUME ;
      g->countCells(ncell,nOrbit,NULL,NULL,NULL,&nb_icells,NULL,NULL) ;
      break;
    }
    
    if (icells == NULL) return nb_icells;
    
    
    *icells = new (CDart*)[nb_icells] ;
    CCoverage *cov=g->getStaticCoverage(ncell,nOrbit);
    cov->reinit() ;
    int j=0 ;
    
    int aMark = g->getNewMark() ;
    //ATTENTION g->unmarkAll(aMark) ;
    
    while (cov->cont()) {
      CDart *current = (cov->operator*()) ;
      if (!(current->getMark(aMark))) {
	g->markOrbit(current,iOrbit,aMark) ;
	(*icells)[j] = normalForm(sg,current,i) ;
	j++;
      }
      cov->operator++() ;
    }
    g->unmarkAll(aMark) ;
    g->freeMark(aMark) ;
    delete cov;
    
    return nb_icells;
  }

  /* Recherche des brins composant une n-cell */

  int allDartsOfNCell(SGMapVertex* sg, CDart* ncell, int dim, CDart ***darts)
  {
    CGMapVertex* g=sg->gmap;
    int nb_darts = 0;

    TOrbit aOrbit = ORBIT_CELL[dim] ;

    g->countCells(ncell,aOrbit,NULL,NULL,NULL,NULL,NULL,&nb_darts) ;
    if (darts == NULL) return nb_darts;
    
    *darts = new (CDart*)[nb_darts] ;
    CCoverage *cov=g->getStaticCoverage(ncell,aOrbit);
    cov->reinit() ;
    int i=0 ;
    
    while (cov->cont()) {
      (*darts)[i] = (cov->operator*()) ;
      i++;
      cov->operator++() ;
    }
    delete cov;
    
    return nb_darts;
  }

  int allCloseNCells(SGMapVertex* saGmap, CDart* ncell, int dim, int vdim, CDart ***darts)
  {
    CGMapVertex* aGmap = saGmap->gmap;
    list<CDart *> closeNcellslist ;
    list<CDart *>::iterator it ;
    int i = 0;

    int rouge = aGmap->getNewMark() ;
    int bleue = aGmap->getNewMark() ;
    //ATTENTION aGmap->unmarkAll(rouge) ;
    //ATTENTION aGmap->unmarkAll(bleue) ;


    aGmap->markOrbit(ncell,ORBIT_CELL[dim], rouge);
    CCoverage * covRed = aGmap->getDynamicCoverage(ncell, ORBIT_CELL[dim]);

    for (; covRed->cont(); ++(*covRed)) {
      CDart *vncell = **covRed ;

      if (!(aGmap->isMarked(vncell,bleue))) {
	aGmap->markOrbit(vncell,ORBIT_CELL[vdim], bleue);
	CCoverage * covBlue = aGmap->getDynamicCoverage(vncell, ORBIT_CELL[vdim]);

	for (; covBlue->cont(); ++(*covBlue)) {
	  CDart *closeNcell = **covBlue ;

	  if (!(aGmap->isMarked(closeNcell,rouge))) {
	    /* mettre closeNcell de ct */
	    closeNcellslist.push_back(normalForm(saGmap,closeNcell,dim));
	    i++;
	    aGmap->markOrbit(closeNcell,ORBIT_CELL[dim], rouge);
	  }

	}
	
	delete covBlue;
      }
      
    }
    
    delete covRed ;
    aGmap->unmarkAll(rouge) ;
    aGmap->unmarkAll(bleue) ;
    aGmap->freeMark(rouge) ;
    aGmap->freeMark(bleue) ;

    *darts = new (CDart*)[i] ;

    int j=0;
    for (it = closeNcellslist.begin() ; it != closeNcellslist.end() ; ++it) {
      CDart* d = *it ;
      (*darts)[j] = d;
      j++;
    }

    if (i != j) {cerr << "allCloseNCells: i!=j\n";cerr.flush();exit(0);}
    return j ;
  }

  /****************************************************************************************/
  /****************************************************************************************/
  /** Affichage d'informations sur la gmap                                               **/
  /****************************************************************************************/
  /****************************************************************************************/
  
  char *stringGmap(SGMapVertex *aGmap)
  {
    char *buffer = NULL ;
    buffer = (char *)malloc((60 + strlen(GMAPPRINTF))*sizeof(char)) ;// a priori - de 187
    sprintf(buffer,GMAPPRINTF, 
	    (int)aGmap,
	    allDarts(aGmap,NULL),
	    allNCells(aGmap,0,NULL),
	    allNCells(aGmap,1,NULL),
	    allNCells(aGmap,2,NULL),
	    allNCells(aGmap,3,NULL)
	    );
    return (buffer) ;
  }

  void printGmap(SGMapVertex *aGmap)
  {
    cout<<stringGmap(aGmap);
  }

  char *stringNcell(SGMapVertex *aGmap, CDart* d, int dim)
  {
    char *buffer = NULL ;
    buffer = (char *)malloc((44 + strlen(NCELLPRINTF))*sizeof(char)) ;// a priori - de 187
    sprintf(buffer,
	    NCELLPRINTF,
	    (int)d,
	    dim,
	    ((dim!=0)?(allIncidentCells(aGmap,d,dim,dim-1,NULL)):0),
	    ((dim!=3)?(allIncidentCells(aGmap,d,dim,dim+1,NULL)):0),
	    ((d==(normalForm(aGmap,d,dim)))?"yes":"no")
	    );
    return buffer;
  }
  
  void printNcell(SGMapVertex *aGmap, CDart* d, int dim)
  {
    cout<<stringNcell(aGmap,d,dim);
  }

  /****************************************************************************************/
  /****************************************************************************************/
  /** Loading objects                                                                    **/
  /****************************************************************************************/
  /****************************************************************************************/
  
  typedef struct {
    CVertex coord;
    list<CDart *> faces;
  } ASEpoint ;

  int exist(SGMapVertex *saGmap,ASEpoint *liste_points,int A,int B,CDart **ABa,CDart **ABb)
  {
    list<CDart *>::iterator itA;
    list<CDart *>::iterator itB;
    CDart *curA = NULL;
    CDart *curB = NULL;

//     cout << "enter in exist..." << endl ;
    for(itA = liste_points[A].faces.begin() ; itA != liste_points[A].faces.end() ; itA++) {
      curA = *itA ;
      for(itB = liste_points[B].faces.begin() ; itB != liste_points[B].faces.end() ; itB++) {
	curB = *itB ;
	if (normalForm(saGmap, curA, 2) == normalForm(saGmap, curB, 2)) {
// 	  cout << "\t" << curA << " " << curB << " match" << endl ;
// 	  cout << "\t" << normalForm(saGmap, curA, 2) << " == " << normalForm(saGmap, curB, 2) << endl ;
	  goto i_find_it;
	}
// 	cout << "\t" << curA << " " << curB << " don't match" << endl ;
      }
    }

    return false;

  i_find_it:
//     remove(itA, ++itA, curA); // faux car plusieurs arcs sur la meme face...
//     remove(itB, ++itB, curB);
//     if (normalForm(saGmap, saGmap->gmap->alpha0(curA), 1) == normalForm(saGmap, curB, 1)) {
//       *ABa = curA;
//       *ABb = saGmap->gmap->alpha0(curA);
//     }
//     else {
//       *ABa = saGmap->gmap->alpha1(curA);
//       *ABb = saGmap->gmap->alpha0(*ABa);  
//     }

    *ABa = curA; *ABb = curB;

    return true;

  }

  int loadASE(SGMapVertex *saGmap, const char *AseFileName, CDart*** ret)
  {
    int nb_points = 0;
    int nb_faces = 0;
    ASEpoint *liste_points = NULL;
    CGMapVertex *aGmap = saGmap->gmap;
    FILE *fichier;
    char buffer[100];
    int index;
    bool moreThanOneObject = false;
    
    fichier=fopen(AseFileName,"r");
    if(fichier == NULL){
      cerr << "loadASE: error: can't open the file " << AseFileName << " for reading." << endl;
      cerr.flush();
      return 0;
    }
    rewind(fichier);

    while(!feof(fichier))
      {
	fscanf(fichier,"%s",buffer);
	if(!strcmp(buffer,"*GEOMOBJECT"))
	  {
	    if (moreThanOneObject) {
	      cout << "warning: there are several objects in this scene. Only the first is returned." << endl;
	      cout.flush();
	      goto suite;
	    }
	    moreThanOneObject = true;
	  }
	if(!strcmp(buffer,"*MESH_NUMVERTEX"))
	  {
	    fscanf(fichier,"%d",&nb_points);
	    liste_points = new ASEpoint[nb_points];
	  }
	if(!strcmp(buffer,"*MESH_NUMFACES"))
	  {
	    fscanf(fichier,"%d",&nb_faces);
	  }
	if(!strcmp(buffer,"*MESH_VERTEX"))
	  {
	    fscanf(fichier,"%d",&index);
	    float x,y,z;
	    fscanf(fichier,"%f %f %f",&x,&y,&z);
	    liste_points[index].coord = CVertex(x,y,z);
 	  }
	if(!strcmp(buffer,"*MESH_FACE"))
	  {
	    int A,B,C;
	    CDart *ABa, *ABb;
	    CDart *ACa, *ACc;
	    CDart *BCb, *BCc;
	    fscanf(fichier,"%d",&index);
	    fscanf(fichier,":    A:    %d B:    %d C:    %d",&A,&B,&C);

	    CDart *a = aGmap->createTopoTriangle() ;
	    normOrbit(saGmap,a,ORBIT_CELL[2]);
	    CDart *b = aGmap->alpha01(a);
	    CDart *c = aGmap->alpha01(b);


	    if (exist(saGmap,liste_points,A,B,&ABa,&ABb))
	      glue2FacesBy4Verteces(saGmap, a, ABa, a, ABa, b, ABb) ;

	    if (exist(saGmap,liste_points,A,C,&ACa,&ACc))
	      glue2FacesBy4Verteces(saGmap, a, ACa, a, ACa, c, ACc) ;

	    if (exist(saGmap,liste_points,B,C,&BCb,&BCc))
	      glue2FacesBy4Verteces(saGmap, b, BCb, b, BCb, c, BCc) ;

	    liste_points[A].faces.push_back(a);
	    liste_points[B].faces.push_back(b);
	    liste_points[C].faces.push_back(c);
	  }
      }
  suite:
    fclose(fichier);

    int i = 0;
    if (ret == NULL) exit(1) ;
    *ret = new (CDart*)[nb_faces] ;
    int al_read = aGmap->getNewMark() ;

    for (index = 0; index < nb_points; index++) {
      list<CDart *>::iterator it;
      for (it=liste_points[index].faces.begin();it!=liste_points[index].faces.end();it++) {
	if (aGmap->findVertex(*it) == NULL) {
	  aGmap->setVertex(*it,liste_points[index].coord);
	}

	if (!(aGmap->isMarked(*it,al_read))) {
	  aGmap->markOrbit(*it,ORBIT_CELL[3], al_read);
	  CDart* d = normalForm(saGmap, *it, 3);
	  (*ret)[i] = d;
	  i++;
	}
      }
    }

    delete liste_points ;
    aGmap->unmarkAll(al_read) ;
    aGmap->freeMark(al_read) ;
    return i;
  }

  void fprint_int(FILE* fich, int i)
  {
    if (i < 10) fprintf(fich,"    %d",i);
    else if (i < 100) fprintf(fich,"   %d",i);
    else if (i < 1000) fprintf(fich,"  %d",i);
    else fprintf(fich," %d",i);
  }

  int saveASE(SGMapVertex *saGmap, const char *AseFileName)
  {
    CGMapVertex *aGmap = saGmap->gmap ;
    FILE *fichier;
    int info, nb_points, nb_faces, mark, ret = 0;

    /* On ouvre le fichier AseFileName */

    fichier=fopen(AseFileName,"w");
    if(fichier == NULL) {
      cerr << "saveASE: error: can't open the file " << AseFileName << " for writing." << endl;
      cerr.flush();
      return 0;
    }
    
    fprintf(fichier,
	    "*3DSMAX_ASCIIEXPORT\t200\n"
	    "*COMMENT \"MGS AsciiExport\"\n"
	    "*SCENE {\n"
	    "\t*SCENE_FILENAME \"\"\n"
	    "\t*SCENE_FIRSTFRAME 0\n"
	    "\t*SCENE_LASTFRAME 30\n"
	    "\t*SCENE_FRAMESPEED 30\n"
	    "\t*SCENE_TICKSPERFRAME 160\n"
	    "\t*SCENE_BACKGROUND_STATIC 0.0000	0.0000	0.0000\n"
	    "\t*SCENE_AMBIENT_STATIC 0.0000	0.0000	0.0000\n"
	    "}\n"
	    "*GEOMOBJECT {\n"
	    "\t*NODE_NAME \"MGS_Gmap\"\n"
	    "\t*NODE_TM {\n"
	    "\t\t*NODE_NAME \"MGS_Gmap\"\n"
	    "\t\t*INHERIT_POS 0 0 0\n"
	    "\t\t*INHERIT_ROT 0 0 0\n"
	    "\t\t*INHERIT_SCL 0 0 0\n"
	    "\t\t*TM_ROW0 1.0000\t0.0000\t0.0000\n"
	    "\t\t*TM_ROW1 0.0000\t1.0000\t0.0000\n"
	    "\t\t*TM_ROW2 0.0000\t0.0000\t1.0000\n"
	    "\t\t*TM_ROW3 0.0000\t0.0000\t0.0000\n"
	    "\t\t*TM_POS 0.0000\t0.0000\t0.0000\n"
	    "\t\t*TM_ROTAXIS 0.0000\t0.0000\t0.0000\n"
	    "\t\t*TM_ROTANGLE 0.0000\n"
	    "\t\t*TM_SCALE 1.0000\t1.0000\t1.0000\n"
	    "\t\t*TM_SCALEAXIS 0.0000\t0.0000\t0.0000\n"
	    "\t\t*TM_SCALEAXISANG 0.0000\n"
	    "\t}\n"
	    "\t*MESH {\n"
	    "\t\t*TIMEVALUE 0\n"
	    );

    /* Je rcupre le nombre de faces et de points */
    nb_points=0;
    nb_faces =0;
    aGmap->getGlobalCharacteristics(NULL,&nb_points,NULL,&nb_faces,NULL,NULL,NULL,NULL,NULL,NULL) ;
    fprintf(fichier,
	    "\t\t*MESH_NUMVERTEX %d\n"
	    "\t\t*MESH_NUMFACES %d\n"
	    "\t\t*MESH_VERTEX_LIST {\n",
	    nb_points,
	    nb_faces);    


    /* parcours des points pour leur affecter un entier.             */
    /* on vrifie galement que tous les points sont plongs dans R3 */

    info = aGmap->getNewDirectInfo() ;
    mark = aGmap->getNewMark();
    int i=0;
    CStaticCoverageAll* cov = new CStaticCoverageAll(aGmap) ;
    while (cov->cont()) {      
      CDart *d = **cov;
      if (!(aGmap->isMarked(d,mark))) {
	CVertex *v = aGmap->findVertex(d);
	if (v == NULL) {
	  cout << "saveASE: error: can't save gmap, all points aren't dived..." << endl;
	  cout.flush();
	  goto error1;
	}

	fprintf(fichier,"\t\t\t*MESH_VERTEX");
	fprint_int(fichier,i);
	fprintf(fichier,"\t%.4f\t%.4f\t%.4f\n",v->getX(),v->getY(),v->getZ());

	aGmap->markOrbit(d,ORBIT_CELL[0], mark);
	aGmap->setDirectInfo(normalForm(saGmap, d, 0),info,(void *)i);
	i++;
      }
      cov->operator++() ;
    }

    if (i != nb_points) {
      cerr << "saveASE: error: problem with the number of vertices" << endl;
      cerr.flush();
      goto error1;
    }

    fprintf(fichier,"\t\t}\n\t\t*MESH_FACE_LIST {\n");

    i = 0;
    cov->reinit();
    while (cov->cont()) {      
      CDart *a = **cov;

      if (aGmap->isMarked(a,mark)) {
	CDart *b = aGmap->alpha01(a);
	CDart *c = aGmap->alpha01(b);

	/* On vrifie que la face est triangulaire */
	if (a != aGmap->alpha01(c)) {
	  cerr << "saveASE: error: all the faces aren't a triangle" << endl;
	  cerr.flush();
	  goto error1;
	}

	int A=(int) (aGmap->getDirectInfo(normalForm(saGmap, a, 0),info));
	int B=(int) (aGmap->getDirectInfo(normalForm(saGmap, b, 0),info));
	int C=(int) (aGmap->getDirectInfo(normalForm(saGmap, c, 0),info));

	fprintf(fichier,"\t\t\t*MESH_FACE"); fprint_int(fichier,i);
	fprintf(fichier,":    A:"); fprint_int(fichier,A);
	fprintf(fichier,"     B:"); fprint_int(fichier,B);
	fprintf(fichier,"     C:"); fprint_int(fichier,C);
	fprintf(fichier," AB:    1 BC:    1 CA:    1\t *MESH_SMOOTHING \t*MESH_MTLID 0\n");
	

	aGmap->unmarkOrbit(a,ORBIT_CELL[2], mark);
	i++;
      }

      cov->operator++() ;
    }

    if (i != nb_faces) {
      cerr << "saveASE: error: problem with the number of faces" << endl;
      cerr.flush();
      goto error1;
    }

    fprintf(fichier,
	    "\t\t}\n\t}\n"
	    "\t*PROP_MOTIONBLUR 0\n"
	    "\t*PROP_CASTSHADOW 1\n"
	    "\t*PROP_RECVSHADOW 1\n"
	    "}"
	    );

    ret = 1;

  error1:
    aGmap->unmarkAll(mark);
    aGmap->freeMark(mark);
    aGmap->freeDirectInfo(info);

    fclose(fichier);

    return 0;
  }


  int loadMoka(SGMapVertex *saGmap, const char *MokaFileName, CDart*** ret) //AsciiFormat 	 BinaryFormat
  {
#define NbMaxVolume 10
    CGMapVertex *aGmap = saGmap->gmap;
    int mark = aGmap->getNewMark() ;

    CStaticCoverageAll* cov = new CStaticCoverageAll(aGmap) ;
    cov->reinit() ;
    while (cov->cont()) {
      CDart *d = **cov ; 
      aGmap->setMark(d,mark);
      cov->operator++() ;
    }
    delete cov;

    if (saGmap->gmap->load(MokaFileName,AsciiFormat)) {

      int i=0 ;
      *ret = new (CDart*)[NbMaxVolume] ;
      cov = new CStaticCoverageAll(aGmap) ;
      cov->reinit() ;
      while ((i<NbMaxVolume) && (cov->cont())) {
	CDart *d = **cov ; 
	if (!(aGmap->isMarked(d,mark))) {
	  normOrbit(saGmap,d,ORBIT_CELL[3]);
	  aGmap->markOrbit(d,ORBIT_CELL[3], mark);
	  (*ret)[i] = normalForm(saGmap, d, 3);
	  i++;
	}
	cov->operator++() ;
      }
      delete cov;

      aGmap->unmarkAll(mark);
      aGmap->freeMark(mark);
      return i;
    }
    else {
      aGmap->unmarkAll(mark);
      aGmap->freeMark(mark);
      return 0;
    }

  }

  int saveMoka(SGMapVertex *saGmap, const char *MokaFileName) //AsciiFormat 	 BinaryFormat
  {
    if (saGmap->gmap->save(MokaFileName,AsciiFormat))
      return 1;
    else
      return 0;
  }


}
