/*
 *
 *                              MGS
 *                              LaMI
 *                       Universite d'Evry Val d'Essonne
 *                   Olivier MICHEL, Jean-Louis GIAVITTO
 * File: $RCSfile: stub_gmap_ml.c,v $ 
 * Path: %P%
 * Description:
 * Created:
 * Author: Olivier Michel & Jean-Louis Giavitto
 * Modified: $Date: 2005-09-07 18:37:43 $ at 
 * Version: $Revision: 1.7 $
 * Status:
 *
 */

#include <stdio.h>
#include <mlvalues.h>
#include <memory.h>
#include <fail.h>
#include <custom.h>

#include "stub_gmap_ml.h"

/* Lecture des blocs CAML */

CAMLprim SGMapVertex* get_gmap(value vgmap)
{
  CAMLparam1(vgmap);
  SGMapVertex *gmap ;
  if (debugAff) {printf("get_gmap\n");fflush(stdout);}
  gmap = * ((SGMapVertex **)Data_custom_val(vgmap));
  CAMLreturn (gmap) ;
}

CAMLprim CDart* get_dart(value vdart)
{
  CAMLparam1(vdart);
  CDart *d;
  if (debugAff) {printf("get_dart\n");fflush(stdout);}
  d = * ((CDart **)Data_custom_val(vdart));
  CAMLreturn (d) ;
}

CAMLprim CDart* get_coverage(value vcoverage)
{
  CAMLparam1(vcoverage);
  CCoverage *c;
  if (debugAff) {printf("get_coverage\n");fflush(stdout);}
  c = * ((CCoverage **)Data_custom_val(vcoverage));
  CAMLreturn (c) ;
}

CAMLprim int get_dim_from_ncell(value ncell)
{
  CAMLparam1(ncell);
  if (debugAff) {printf("get_dim_from_ncell\n");fflush(stdout);}
  CAMLreturn (Long_val(Field(ncell,2))) ;
}

CAMLprim CDart* get_dart_from_ncell(value ncell)
{
  CAMLparam1(ncell);
  CDart* d;
  if (debugAff) {printf("get_dart_from_ncell\n");fflush(stdout);}
  d = get_dart(Field(ncell,1)) ;
  CAMLreturn (d) ;
}

CAMLprim SGMapVertex* get_gmap_from_ncell(value ncell)
{
  CAMLparam1(ncell);
  SGMapVertex* g;
  if (debugAff) {printf("get_gmap_from_ncell\n");fflush(stdout);}
  g = get_gmap(Field(ncell,0));
  CAMLreturn (g) ;
}

/* Customisation des gmaps */

void finalize_gmap(value v)
{
  if (debugAff) {printf("finalize_gmap\n"); fflush(stdout);}
}

int compare_gmap(value v1, value v2)
{
  CAMLparam2(v1,v2);
  SGMapVertex *g1;
  SGMapVertex *g2;
  if (debugAff) {printf("compare_gmap\n"); fflush(stdout);}
  g1 = get_gmap(v1);
  g2 = get_gmap(v2);
  CAMLreturn(((long)g1) - ((long)g2)) ;
}

long hash_gmap(value v)
{
  CAMLparam1(v);
  SGMapVertex *g;
  if (debugAff) {printf("hash_gmap\n"); fflush(stdout);}
  g = get_gmap(v);
  CAMLreturn(((long)g));
}

struct custom_operations ops_gmap = {"ops_gmap",
				     finalize_gmap,
				     compare_gmap,
				     hash_gmap};

void finalize_dart(value v)
{
  if (debugAff) {printf("finalize_dart\n"); fflush(stdout);}
}

int compare_dart(value v1, value v2)
{
  CAMLparam2(v1,v2);
  CDart *d1;
  CDart *d2;
  if (debugAff) {printf("compare_dart\n"); fflush(stdout);}
  d1 = get_dart(v1);
  d2 = get_dart(v2);
  CAMLreturn(((long)d1) - ((long)d2)) ;
}

long hash_dart(value v)
{
  CAMLparam1(v);
  CDart *d;
  if (debugAff) {printf("hash_dart\n"); fflush(stdout);}
  d = get_dart(v);
  CAMLreturn(((long)d));
}

struct custom_operations ops_dart = {"ops_dart",
				     finalize_dart,
				     compare_dart,
				     hash_dart};

void finalize_coverage(value v)
{
  if (debugAff) {printf("finalize_coverage\n"); fflush(stdout);}
}

int compare_coverage(value v1, value v2)
{
  CAMLparam2(v1,v2);
  CCoverage *c1;
  CCoverage *c2;
  if (debugAff) {printf("compare_coverage\n"); fflush(stdout);}
  c1 = get_coverage(v1);
  c2 = get_coverage(v2);
  CAMLreturn(((long)c1) - ((long)c2)) ;
}

long hash_coverage(value v)
{
  CAMLparam1(v);
  CCoverage *c;
  if (debugAff) {printf("hash_coverage\n"); fflush(stdout);}
  c = get_coverage(v);
  CAMLreturn(((long)c));
}

struct custom_operations ops_coverage = {"ops_coverage",
					 finalize_coverage,
					 compare_coverage,
					 hash_coverage};

/* Ecriture des blocs CAML */

CAMLprim value build_gmap(SGMapVertex *gmap)
{
  CAMLparam0();
  CAMLlocal1(gmapBloc);
  SGMapVertex **blocDataPart = NULL ;
  
  if (debugAff) {printf("build_gmap\n");fflush(stdout);}

  gmapBloc = alloc_custom(&ops_gmap, sizeof(SGMapVertex*) ,0, 1) ;
  blocDataPart = (SGMapVertex **)Data_custom_val(gmapBloc);
  *blocDataPart = gmap ;

  CAMLreturn(gmapBloc);
}

CAMLprim value build_dart(CDart *dart)
{
  CAMLparam0();
  CAMLlocal1(dartBloc);
  CDart **blocDataPart = NULL ;
  
  if (debugAff) {printf("build_dart\n");fflush(stdout);}

  dartBloc = alloc_custom(&ops_dart, sizeof(CDart*) ,0, 1) ;
  blocDataPart = (CDart **)Data_custom_val(dartBloc);
  *blocDataPart = dart ;
  
  CAMLreturn(dartBloc);
}

CAMLprim value build_coverage(CCoverage *coverage)
{
  CAMLparam0();
  CAMLlocal1(coverageBloc);
  CCoverage **blocDataPart = NULL ;
  
  if (debugAff) {printf("build_coverage\n");fflush(stdout);}

  coverageBloc = alloc_custom(&ops_coverage, sizeof(CCoverage*) ,0, 1) ;
  blocDataPart = (CCoverage **)Data_custom_val(coverageBloc);
  *blocDataPart = coverage ;
  
  CAMLreturn(coverageBloc);
}

CAMLprim value build_ncell(SGMapVertex *gmap, CDart *dart, int dim)
{
  CAMLparam0();
  CAMLlocal3(gmapBloc,dartBloc,ncell);
  
  if (debugAff) {printf("build_ncell\n");fflush(stdout);}

  gmapBloc = build_gmap(gmap) ;
  dartBloc = build_dart(dart) ;

  ncell = alloc_tuple(3);
  Store_field(ncell,0,gmapBloc) ;
  Store_field(ncell,1,dartBloc) ;
  Store_field(ncell,2,Val_long(dim)) ;

  CAMLreturn(ncell);
}

CAMLprim value build_couple(value v1, value v2)
{
  CAMLparam2(v1,v2);
  CAMLlocal1(coupleBloc);
  
  if (debugAff) {printf("build_couple\n");fflush(stdout);}
  coupleBloc = alloc_tuple(2);
  Store_field(coupleBloc,0,v1) ;
  Store_field(coupleBloc,1,v2) ;

  CAMLreturn(coupleBloc);
}

/* fonctions de l'interface... */

CAMLprim value ml_normal_form(value ncell)
{
  CAMLparam1(ncell);               // une cellule caml (gmap * dart * int)
  CAMLlocal1(normalNcell);         // la cellule normalise
  int dim;                         // dimension de ncell
  SGMapVertex* gmap ;              // la gmap dans laquelle on travaille
  CDart* dart ;                    // le brin  normaliser
  CDart* normalDart ;              // le brin normalis

  if (debugAff) {printf("ml_normal_form\n");fflush(stdout);}
  dim  = get_dim_from_ncell(ncell) ;
  dart = get_dart_from_ncell(ncell) ;
  gmap = get_gmap_from_ncell(ncell) ;

  normalDart = normalForm(gmap, dart, dim) ;

  normalNcell = build_ncell(gmap,normalDart,dim);  
  CAMLreturn(normalNcell);
}

CAMLprim value ml_qmfy(value vgmap)
{
  CAMLparam1 (vgmap);
  SGMapVertex* gmap = get_gmap(vgmap);
  if (debugAff) {printf("ml_qmfy\n");fflush(stdout);}
  qmfy(gmap);
  CAMLreturn (Val_unit);
}

CAMLprim value ml_norm(value vgmap)
{
  CAMLparam1 (vgmap);
  SGMapVertex* gmap = get_gmap(vgmap);
  if (debugAff) {printf("ml_norm\n");fflush(stdout);}
  normAll(gmap);
  CAMLreturn (Val_unit);
}

CAMLprim value ml_is_frozen(value vgmap)
{
  CAMLparam1 (vgmap);
  SGMapVertex* gmap = get_gmap(vgmap);
  if (debugAff) {printf("ml_qmfy\n");fflush(stdout);}
  CAMLreturn (Val_bool ( (isFrozen(gmap)) == 1 ));
}

CAMLprim value ml_is_not_frozen(value vgmap)
{
  CAMLparam1 (vgmap);
  SGMapVertex* gmap = get_gmap(vgmap);
  if (debugAff) {printf("ml_qmfy\n");fflush(stdout);}
  CAMLreturn (Val_bool ( (isFrozen(gmap)) != 1 ));
}

CAMLprim value ml_create_gmap()
{
  CAMLparam0();
  CAMLlocal1(vgmap);
  SGMapVertex *gmap;
  if (debugAff) {printf("ml_create_gmap\n");fflush(stdout);}
  gmap = createGmap(False) ;
  vgmap = build_gmap(gmap);
  CAMLreturn(vgmap);
}

CAMLprim value ml_delete_gmap(value gm)
{
  CAMLparam1(gm);
  SGMapVertex* gmap ;
  if (debugAff) {printf("ml_delete_gmap\n");fflush(stdout);}
  gmap = get_gmap(gm) ;
  deleteGmap(gmap);
  CAMLreturn(Val_unit);
}

CAMLprim value ml_copy_gmap(value vgmap)
{
  CAMLparam1(vgmap);
  CAMLlocal1(vgmap1);
  SGMapVertex *gmap;
  if (debugAff) {printf("ml_copy_gmap\n");fflush(stdout);}
  gmap = get_gmap(vgmap);
  vgmap1 = build_gmap(copyGmap(gmap)) ;
  CAMLreturn(vgmap1);
}

CAMLprim value ml_copy_notmarked_dart(value vgmap_src,value vgmap_dst,value vmark)
{
  CAMLparam3(vgmap_src,vgmap_dst,vmark);
  SGMapVertex *g_src = get_gmap(vgmap_src);
  SGMapVertex *g_dst = get_gmap(vgmap_dst);
  int mark = Int_val(vmark) ;
  int di;
  if (debugAff) {printf("ml_copy_marked_dart\n");fflush(stdout);}
  di = copy_notmarked_dart(g_src,g_dst,mark);
  CAMLreturn (Val_int(di));
}



CAMLprim value ml_dart_null(value unit)
{
  CAMLparam1(unit);
  CAMLreturn (build_dart((CDart*) NULL));
}

CAMLprim value ml_copy_dart(value vg, value vd)
{
  CAMLparam2(vg,vd);
  SGMapVertex *g = get_gmap(vg) ;
  CDart *d = get_dart(vd);
  CDart *c = copyDart(g,d);
  if (debugAff) {printf("ml_copy_dart\n");fflush(stdout);}
  CAMLreturn(build_dart(c));
}

CAMLprim value ml_new_dart(value vg, value vx, value vy, value vz)
{
  CAMLparam4(vg,vx,vy,vz);
  SGMapVertex *g = get_gmap(vg) ;
  double x =  Double_val(vx);
  double y =  Double_val(vy);
  double z =  Double_val(vz);
  CDart *c = newDart(g,x,y,z);
  if (debugAff) {printf("ml_new_dart\n");fflush(stdout);}
  CAMLreturn(build_dart(c));
}

CAMLprim value ml_alpha(value vgmap, value vdart, value vi)
{
  CAMLparam3(vgmap,vdart,vi);
  SGMapVertex *gmap = get_gmap(vgmap);
  CDart *dart = get_dart(vdart);
  int i = Long_val(vi);
  if (debugAff) {printf("ml_alpha\n");fflush(stdout);}
  CAMLreturn (build_dart(alpha(gmap, dart, i)));
}

CAMLprim value ml_sew(value vgmap, value vdart1, value vdart2, value vi)
{
  CAMLparam4(vgmap,vdart1,vdart2,vi);
  SGMapVertex *gmap = get_gmap(vgmap);
  CDart *dart1 = get_dart(vdart1);
  CDart *dart2 = get_dart(vdart2);
  int i = Long_val(vi);
  if (debugAff) {printf("ml_sew\n");fflush(stdout);}
  sew(gmap,dart1,dart2,i);
  CAMLreturn (Val_unit);
}

CAMLprim value ml_is_free(value vgmap, value vdart, value vi)
{
  CAMLparam3(vgmap,vdart,vi);
  SGMapVertex *g = get_gmap(vgmap);
  CDart* d = get_dart(vdart);
  int i = Int_val(vi);
  if (debugAff) {printf("ml_is_free\n");fflush(stdout);}
  CAMLreturn (Val_bool ( (isFree(g,d,i)) == 1)) ;
}

CAMLprim value ml_get_new_mark(value vgmap)
{
  CAMLparam1(vgmap);
  SGMapVertex *g = get_gmap(vgmap);
  int i = getNewMark(g);
  if (debugAff) {printf("ml_get_new_mark\n");fflush(stdout);}
  CAMLreturn(Val_int(i));
}

CAMLprim value ml_free_mark(value vgmap, value vi)
{
  CAMLparam2(vgmap,vi);
  SGMapVertex *g = get_gmap(vgmap);
  int i = Int_val(vi);
  if (debugAff) {printf("ml_free_mark\n");fflush(stdout);}
  freeMark(g,i);
  CAMLreturn(Val_unit);
}

CAMLprim value ml_mark_dart(value vg, value vd, value vi)
{
  CAMLparam3(vg,vd,vi);
  SGMapVertex *g = get_gmap(vg) ;
  CDart *d = get_dart(vd);
  int i = Int_val(vi);
  if (debugAff) {printf("ml_mark_dart\n");fflush(stdout);}
  markDart(g,d,i);
  CAMLreturn(Val_unit);
}

CAMLprim value ml_mark_ncell(value vnc, value vi)
{
  CAMLparam2(vnc,vi);
  SGMapVertex *g = get_gmap_from_ncell(vnc) ;
  CDart *d = get_dart_from_ncell(vnc);
  int n = get_dim_from_ncell(vnc);
  int i = Int_val(vi);
  if (debugAff) {printf("ml_mark_ncell\n");fflush(stdout);}
  markNcell(g,d,n,i);
  CAMLreturn(Val_unit);
}

CAMLprim value ml_unmark_dart(value vg, value vd, value vi)
{
  CAMLparam3(vg,vd,vi);
  SGMapVertex *g = get_gmap(vg) ;
  CDart *d = get_dart(vd);
  int i = Int_val(vi);
  if (debugAff) {printf("ml_unmark_dart\n");fflush(stdout);}
  unmarkDart(g,d,i);
  CAMLreturn(Val_unit);
}

CAMLprim value ml_unmark_ncell(value vnc, value vi)
{
  CAMLparam2(vnc,vi);
  SGMapVertex *g = get_gmap_from_ncell(vnc) ;
  CDart *d = get_dart_from_ncell(vnc);
  int n = get_dim_from_ncell(vnc);
  int i = Int_val(vi);
  if (debugAff) {printf("ml_unmark_ncell\n");fflush(stdout);}
  unmarkNcell(g,d,n,i);
  CAMLreturn(Val_unit);
}

CAMLprim value ml_is_marked_dart(value vg,value vd, value vi)
{
  CAMLparam3(vg,vd,vi);
  SGMapVertex *g = get_gmap(vg) ;
  CDart *d = get_dart(vd);
  int i = Int_val(vi);  
  if (debugAff) {printf("ml_is_marked_dart\n");fflush(stdout);}
  CAMLreturn (Val_bool ( (isMarkedDart(g,d,i)) == 1 ));
}

CAMLprim value ml_is_marked_ncell(value vnc, value vi)
{
  CAMLparam2(vnc,vi);
  SGMapVertex *g = get_gmap_from_ncell(vnc) ;
  CDart *d = get_dart_from_ncell(vnc);
  int dim = get_dim_from_ncell(vnc);
  int i = Int_val(vi);  
  if (debugAff) {printf("ml_is_marked_ncell\n");fflush(stdout);}
  CAMLreturn (Val_bool ( (isMarkedNcell(g,d,dim,i)) == 1 ));
}

CAMLprim value ml_get_new_directinfo(value vg)
{
  CAMLparam1(vg);
  SGMapVertex* g = get_gmap(vg);
  if (debugAff) {printf("ml_get_new_directinfo\n");fflush(stdout);}
  CAMLreturn (Val_int(getNewDirectInfo(g)));
}

CAMLprim value ml_free_directinfo(value vg, value vdi)
{
  CAMLparam2(vg,vdi);
  SGMapVertex* g = get_gmap(vg);
  int di = Int_val(vdi);
  if (debugAff) {printf("ml_free_directinfo\n");fflush(stdout);}
  freeDirectInfo(g,di);
  CAMLreturn (Val_unit);
}

CAMLprim value ml_set_directinfo(value vg, value vd, value vdi, value vinfo)
{
  CAMLparam4(vg,vd,vdi,vinfo);
  SGMapVertex *g = get_gmap(vg);
  CDart *d = get_dart(vd);
  int di = Int_val(vdi);
  CDart *info = get_dart(vinfo);
  if (debugAff) {printf("ml_set_directinfo\n");fflush(stdout);}
  setDirectInfo(g,d,di,(void*)info);
  CAMLreturn (Val_unit);
}

CAMLprim value ml_get_directinfo(value vg, value vd, value vdi)
{
  CAMLparam3(vg,vd,vdi);
  SGMapVertex *g = get_gmap(vg);
  CDart *d = get_dart(vd);
  int di = Int_val(vdi);
  CDart *info;
  if (debugAff) {printf("ml_get_directinfo\n");fflush(stdout);}
  info = (CDart*)getDirectInfo(g,d,di);
  CAMLreturn (build_dart(info));
}

CAMLprim value ml_get_coord_from_vertex(value vncell)
{
  CAMLparam1(vncell);
  CAMLlocal1(bloc);
  SGMapVertex *gmap;
  CDart* s;
  int dim;
  float x=0;
  float y=0;
  float z=0;

  if (debugAff) {printf("ml_get_coord_from_vertex\n");fflush(stdout);}

  gmap = get_gmap_from_ncell(vncell) ;
  s    = get_dart_from_ncell(vncell) ;
  dim  = get_dim_from_ncell(vncell) ;

  if ( (dim != 0) || ((get_vertex_coord(gmap, s, &x, &y, &z)) != 0))
      failwith("Stub_gmap_ml: ml_get_coord_from_vertex: any coordinates associated !!!");

  bloc = alloc_tuple(3);
  Store_field(bloc,0,copy_double(x)) ;
  Store_field(bloc,1,copy_double(y)) ;
  Store_field(bloc,2,copy_double(z)) ;

  CAMLreturn(bloc);
}

CAMLprim value ml_set_coord_of_vertex(value vcell, value vx, value vy, value vz)
{
  CAMLparam4(vcell, vx, vy, vz);
  double x = Double_val(vx);
  double y = Double_val(vy);
  double z = Double_val(vz);
  CDart *cell = get_dart_from_ncell(vcell);
  SGMapVertex *gmap = get_gmap_from_ncell(vcell);

  if (debugAff) {printf("ml_set_coord_of_vertex\n");fflush(stdout);}
  set_vertex_coord(gmap, cell, x, y, z);

  CAMLreturn (Val_unit);
}

CAMLprim value ml_check_gmap(value vgmap)
{
  CAMLparam1(vgmap);
  SGMapVertex* gmap = get_gmap(vgmap);
  if (debugAff) {printf("ml_check_gmap\n");fflush(stdout);}
  CAMLreturn (Val_bool ( (checkGmap(gmap)) == 1 ));
}

CAMLprim value ml_compare_ncells(value vnc1,value vnc2)
{
  CAMLparam2(vnc1,vnc2);
  SGMapVertex* gmap1 = get_gmap_from_ncell(vnc1);
  SGMapVertex* gmap2 = get_gmap_from_ncell(vnc2);
  if (debugAff) {printf("ml_compare_ncells\n");fflush(stdout);}
  if (gmap1==gmap2) {
    int n1 = get_dim_from_ncell(vnc1);
    int n2 = get_dim_from_ncell(vnc2);
    if (n1 == n2) {
      CDart* d1 = get_dart_from_ncell(vnc1);
      CDart* d2 = get_dart_from_ncell(vnc2);

      CAMLreturn (Val_int (compareNcells(gmap1,d1,d2,n1)));

    } else {
      CAMLreturn (Val_int (n1 - n2)) ;
    }
  } else {
    CAMLreturn (Val_int (gmap1 - gmap2)) ;
  }
}



CAMLprim value ml_all_coverage(value vgmap)
{
  CAMLparam1(vgmap);
  SGMapVertex *g = get_gmap(vgmap) ;
  CCoverage *c = getAllCoverage(g);
  if (debugAff) {printf("ml_all_coverage\n");fflush(stdout);}
  CAMLreturn(build_coverage(c));
}

CAMLprim value ml_ncell_coverage(value vnc)
{
  CAMLparam1(vnc);
  SGMapVertex *g = get_gmap_from_ncell(vnc) ;
  CDart *d = get_dart_from_ncell(vnc) ;
  int dim = get_dim_from_ncell(vnc) ;
  CCoverage *c = getNcellCoverage(g,d,dim);
  if (debugAff) {printf("ml_ncell_coverage\n");fflush(stdout);}
  CAMLreturn(build_coverage(c));
}

CAMLprim value ml_delete_coverage(value vcov)
{
  CAMLparam1(vcov);
  CCoverage *cov = get_coverage(vcov);
  if (debugAff) {printf("ml_delete_coverage %i\n");fflush(stdout);}
  deleteCoverage(cov);
  CAMLreturn(Val_unit);
}

CAMLprim value ml_reinit_coverage(value vcov)
{
  CAMLparam1(vcov);
  CCoverage *cov = get_coverage(vcov);
  if (debugAff) {printf("ml_reinit_coverage\n");fflush(stdout);}
  reinitCoverage(cov);
  CAMLreturn(Val_unit);
}

CAMLprim value ml_get_current_dart(value vcov)
{
  CAMLparam1(vcov);
  CCoverage *cov = get_coverage(vcov);
  CDart *d = getCurrentDart(cov);
  if (debugAff) {printf("ml_get_current_dart\n");fflush(stdout);}
  CAMLreturn(build_dart(d));
}

CAMLprim value ml_cont_coverage(value vcov)
{
  CAMLparam1(vcov);
  CCoverage *cov = get_coverage(vcov);
  if (debugAff) {printf("ml_cont_coverage\n");fflush(stdout);}
  CAMLreturn (Val_bool ( (cont(cov)) == 1 ));
}

CAMLprim value ml_next_dart(value vcov)
{
  CAMLparam1(vcov);
  CCoverage *cov = get_coverage(vcov);
  if (debugAff) {printf("ml_next_dart\n");fflush(stdout);}
  next(cov);
  CAMLreturn (Val_unit);
}





CAMLprim value ml_load_moka(value vgmap, value vfname)
{
  CAMLparam2(vgmap,vfname);
  CAMLlocal1(vvol);
  char *fname=String_val(vfname);
  SGMapVertex *gmap = get_gmap(vgmap);
  CDart **vol=NULL;
  int nb_vol, i;

  if (debugAff) {printf("ml_load_moka\n");fflush(stdout);}
  nb_vol = loadMoka(gmap,fname,&vol);
  vvol = alloc_tuple(nb_vol);
  for (i=0;i<nb_vol;i++) {
    Store_field(vvol,i,build_ncell(gmap,vol[i],3)) ;
  }
  deleteDartsTable(vol);
  CAMLreturn(vvol);
}

CAMLprim value ml_save_moka(value vgmap, value vfname)
{
  CAMLparam2(vgmap,vfname);
  char *fname=String_val(vfname);
  SGMapVertex *gmap = get_gmap(vgmap);

  if (debugAff) {printf("ml_save_moka\n");fflush(stdout);}
  saveMoka(gmap,fname);
  CAMLreturn(Val_unit);
}










CAMLprim value ml_new_ncell(value vgmap, value vdim)
{
  CAMLparam2(vgmap,vdim);
  CDart *d;
  int dim ;
  SGMapVertex *gmap ;
  if (debugAff) {printf("ml_new_ncell\n");fflush(stdout);}
  dim = Int_val(vdim) ;
  gmap = get_gmap(vgmap);

  d = new_ncell(gmap,dim) ;
  CAMLreturn (build_ncell(gmap,d,dim)) ;
}

CAMLprim value ml_new_polygon(value vgmap, value vnbside)
{
  CAMLparam2(vgmap,vnbside);
  CDart *d;
  int nbside ;
  SGMapVertex *gmap ;
  if (debugAff) {printf("ml_new_polygon\n");fflush(stdout);}
  gmap = get_gmap(vgmap);
  nbside = Int_val(vnbside) ;
  if (nbside == 0)
    failwith("Stub_gmap_ml: ml_new_poygon: unable to create a polygon with 0 side");

  d = new_ncell_polygon(gmap,nbside) ;
  CAMLreturn (build_ncell(gmap,d,2)) ;
}

CAMLprim value ml_new_sphere(value vgmap, value vnbmeridians, value vnbparallels)
{
  CAMLparam3(vgmap, vnbmeridians, vnbparallels);
  CAMLlocal2(vnorth,vsouth);
  CDart *north = NULL;
  CDart *south = NULL;
  int nbmeridians ;
  int nbparallels ;
  SGMapVertex *gmap ;
  if (debugAff) {printf("ml_new_sphere\n");fflush(stdout);}
  gmap = get_gmap(vgmap);
  nbmeridians = Long_val(vnbmeridians) ;
  nbparallels = Long_val(vnbparallels) ;
  if ((nbmeridians == 0) || (nbparallels == 0))
    failwith("Stub_gmap_ml: ml_new_sphere: unable to create a sphere with 0 meridian or 0 parallel");

  new_ncell_sphere(gmap,nbmeridians,nbparallels,&south,&north) ;
  vnorth = build_ncell(gmap,north,0) ;
  vsouth = build_ncell(gmap,south,0) ;
  CAMLreturn (build_couple(vnorth,vsouth)) ;
}

CAMLprim value ml_new_polyline(value vgmap, value vnbverteces)
{
  CAMLparam2(vgmap, vnbverteces);
  CDart *ret = NULL;
  int nbverteces ;
  SGMapVertex *gmap ;
  if (debugAff) {printf("ml_new_polyline\n");fflush(stdout);}
  gmap = get_gmap(vgmap);
  nbverteces = Long_val(vnbverteces) ;
  if (nbverteces == 0)
    failwith("Stub_gmap_ml: ml_new_polyline: unable to create a polyline without vertex");
  
  ret = new_ncell_polyline(gmap,nbverteces) ;
  CAMLreturn (build_ncell(gmap,ret,0)) ;
}

CAMLprim value ml_new_square(value vgmap, value vax, value vay, value vdim)
{
  CAMLparam4(vgmap,vax,vay,vdim);
  CDart *sq = NULL;
  int ax ;
  int ay ;
  int dim ;
  SGMapVertex *gmap ;
  if (debugAff) {printf("ml_new_square\n");fflush(stdout);}
  gmap = get_gmap(vgmap);
  ax = Long_val(vax) ;
  ay = Long_val(vay) ;
  dim = Long_val(vdim) ;
  if ((ax <= 0) || (ay <= 0))
    failwith("Stub_gmap_ml: ml_new_square: unable to create a square with null parameters");

  sq = new_ncell_square(gmap,ax,ay,dim) ;
  CAMLreturn (build_ncell(gmap,sq,dim)) ;
}

CAMLprim value ml_insert_vertex(value vncell)
{
  CAMLparam1(vncell);
  CAMLlocal1(ans);
  SGMapVertex *gmap = get_gmap_from_ncell(vncell);
  CDart *e = get_dart_from_ncell(vncell);
  int dim = get_dim_from_ncell(vncell);
  CDart *e1 = NULL;
  CDart *e2 = NULL;
  CDart *v = NULL;

  if (debugAff) {printf("insert_vertex\n");fflush(stdout);}

  if (dim != 1)
    failwith("Stub_gmap_ml: ml_insert_vertex: this only works on edge !!!");

  v = insert_vertex(gmap, e, &e1, &e2);

  ans = alloc_tuple(3);
  Store_field(ans,0,build_ncell(gmap,v,0)) ;
  Store_field(ans,1,build_ncell(gmap,e1,1)) ;
  Store_field(ans,2,build_ncell(gmap,e2,1)) ;
  CAMLreturn(ans);
}

CAMLprim value ml_insert_edge(value vncell1, value vncell2)
{
  CAMLparam2(vncell1,vncell2);
  CAMLlocal1(ans);
  SGMapVertex *gmap1 = get_gmap_from_ncell(vncell1);
  SGMapVertex *gmap2 = get_gmap_from_ncell(vncell2);
  CDart *v1 = get_dart_from_ncell(vncell1);
  CDart *v2 = get_dart_from_ncell(vncell2);
  int dim1 = get_dim_from_ncell(vncell1);
  int dim2 = get_dim_from_ncell(vncell2);

  CDart *f1 = NULL;
  CDart *f2 = NULL;
  CDart *e = NULL;

  if (debugAff) {printf("insert_edge\n");fflush(stdout);}

  if ((gmap1 != gmap2) || (dim1 != 0) || (dim2 != 0))
    failwith("Stub_gmap_ml: ml_insert_edge: this only works on 2 vertices of the same gmap !!!");

  e = insert_edge(gmap1, v1, v2, &f1, &f2);

  ans = alloc_tuple(5);
  Store_field(ans,0,build_ncell(gmap1,e,1)) ;
  Store_field(ans,1,build_ncell(gmap1,v1,0)) ;
  Store_field(ans,2,build_ncell(gmap1,v2,0)) ;
  Store_field(ans,3,build_ncell(gmap1,f1,2)) ;
  Store_field(ans,4,build_ncell(gmap1,f2,2)) ;
  CAMLreturn(ans);
}

CAMLprim value ml_insert_face(value vvol, value vncells_array, value vlength)
{
  CAMLparam3(vvol,vncells_array, vlength);
  CAMLlocal2(vncell,ans);
  int length = 0;
  int i = 0;
  SGMapVertex *gmap = NULL;
  CDart **edges_tab = NULL;
  CDart *v1 = NULL;
  CDart *v2 = NULL;
  CDart *f = NULL;

  if (debugAff) {printf("insert_face\n");fflush(stdout);}

  if (get_dim_from_ncell(vvol) != 3)
    failwith("Stub_gmap_ml: ml_insert_face: this function is to be applied on a volume");

  length = Int_val(vlength);
  if (length <= 0)
    failwith("Stub_gmap_ml: ml_insert_face: array size must be greater than 0");

  edges_tab = (CDart**)malloc(length * sizeof(CDart*));
  if (edges_tab == NULL)
    failwith("Stub_gmap_ml: ml_insert_face: memory allocation failed");

  gmap = get_gmap_from_ncell(vvol);
  for(i=0;i<length;i++) {
    vncell = Field(vncells_array,i) ;
    if (gmap != (get_gmap_from_ncell(vncell)) ) {
      free(edges_tab);
      failwith("Stub_gmap_ml: ml_insert_face: all ncells must belong to the same gmap");
    }
    if ((get_dim_from_ncell(vncell)) != 1) {
      free(edges_tab);
      failwith("Stub_gmap_ml: ml_insert_face: all ncells must be edges");
    }

    edges_tab[i] = get_dart_from_ncell(vncell) ;
  }

  f = insert_face(gmap,get_dart_from_ncell(vvol),edges_tab,length,&v1,&v2);

  ans = alloc_tuple(3+length);
  Store_field(ans,0,build_ncell(gmap,f,2)) ;
  Store_field(ans,1,build_ncell(gmap,v1,3)) ;
  Store_field(ans,2,build_ncell(gmap,v2,3)) ;
  for (i=0;i<length;i++) {
    Store_field(ans,i+3,build_ncell(gmap,edges_tab[i],1));
  }

  free(edges_tab);
  CAMLreturn(ans);
}

CAMLprim value ml_create_edge(value vncell1, value vncell2)
{
  CAMLparam2(vncell1,vncell2);
  CAMLlocal1(ans);

  SGMapVertex* gmap1 = get_gmap_from_ncell(vncell1);
  SGMapVertex* gmap2 = get_gmap_from_ncell(vncell2);
  int dim1 = get_dim_from_ncell(vncell1);
  int dim2 = get_dim_from_ncell(vncell2);
  CDart* d1 = get_dart_from_ncell(vncell1);
  CDart* d2 = get_dart_from_ncell(vncell2);
  CDart* e = NULL;

  if (debugAff) {printf("create_edge\n");fflush(stdout);}

  if ((gmap1 != gmap2) || (dim1 != 0) || (dim2 != 0))
    failwith("Stub_gmap_ml: ml_create_edge: this only works on 2 vertices of the same gmap !!!");

  e = create_edge(gmap1,d1,d2);

  ans = alloc_tuple(3);
  Store_field(ans,0,build_ncell(gmap1,e,1)) ;
  Store_field(ans,1,build_ncell(gmap1,d1,0)) ;
  Store_field(ans,2,build_ncell(gmap1,d2,0)) ;
  CAMLreturn(ans);
}

CAMLprim value ml_create_face(value v_ncell_x_ncell_array, value vlength)
{
  CAMLparam2(v_ncell_x_ncell_array,vlength);
  CAMLlocal3(vncell,vblock,ans);
  int length = 0;
  int i = 0;
  int j = 0;
  SGMapVertex *gmap = NULL;
  CDart ***vertex_tab_tab = NULL;
  CDart *f = NULL;

  if (debugAff) {printf("create_face\n");fflush(stdout);}

  length = Int_val(vlength);
  if (length <= 0)
    failwith("Stub_gmap_ml: ml_create_face: array size must be greater than 0");

  vertex_tab_tab = (CDart***)malloc(length * sizeof(CDart**));
  if (vertex_tab_tab == NULL)
    failwith("Stub_gmap_ml: ml_create_face: memory allocation failed");
  for (i=0;i<length;i++) {
    vertex_tab_tab[i] = (CDart**)malloc(2 * sizeof(CDart*));
    if (vertex_tab_tab[i] == NULL) {
      for (j=0;j<i;j++) {
	free(vertex_tab_tab[j]);
      }
      free(vertex_tab_tab);
      failwith("Stub_gmap_ml: ml_create_face: memory allocation failed");
    }
  }

  vblock = Field(v_ncell_x_ncell_array,0) ;
  gmap = get_gmap_from_ncell(Field(vblock,0));
  for(i=0;i<length;i++) {
    vblock = Field(v_ncell_x_ncell_array,i) ;
    vncell = Field(vblock,0) ;
    if (gmap != (get_gmap_from_ncell(vncell)) ) {
      free(vertex_tab_tab);
      failwith("Stub_gmap_ml: ml_create_face: all ncells must belong to the same gmap");
    }
    if ((get_dim_from_ncell(vncell)) != 0) {
      free(vertex_tab_tab);
      failwith("Stub_gmap_ml: ml_create_face: all ncells must be vertices");
    }

    vertex_tab_tab[i][0] = get_dart_from_ncell(vncell) ;

    vncell = Field(vblock,1) ;
    if (gmap != (get_gmap_from_ncell(vncell)) ) {
      free(vertex_tab_tab);
      failwith("Stub_gmap_ml: ml_create_face: all ncells must belong to the same gmap");
    }
    if ((get_dim_from_ncell(vncell)) != 0) {
      free(vertex_tab_tab);
      failwith("Stub_gmap_ml: ml_create_face: all ncells must be vertices");
    }

    vertex_tab_tab[i][1] = get_dart_from_ncell(vncell) ;
  }

  f = create_face(gmap,vertex_tab_tab,length);

  ans = alloc_tuple(1+2*length);
  Store_field(ans,0,build_ncell(gmap,f,2)) ;
  for (i=0;i<length;i++) {
    Store_field(ans,2*i+1,build_ncell(gmap,vertex_tab_tab[i][0],0));
    Store_field(ans,2*i+2,build_ncell(gmap,vertex_tab_tab[i][1],0));
  }

  free(vertex_tab_tab);

  CAMLreturn(ans);
}

CAMLprim value ml_create_volume(value v_ncell_array, value vnbfaces, value v_ncell_array_array, value v_int_array, value vlength)
{
  CAMLparam5(v_ncell_array,vnbfaces,v_ncell_array_array,v_int_array,vlength);
  CAMLlocal3(vncell,vblock,ans);
  int length = 0, nbfaces = 0;
  int i = 0;
  int j = 0;
  int k = 0;
  SGMapVertex *gmap = NULL;
  CDart ***vertex_tab_tab = NULL;
  CDart **faces;
  int *length_tab;
  CDart *v = NULL;

  if (debugAff) {printf("create_volume\n");fflush(stdout);}

  length = Int_val(vlength);
  if (length <= 0)
    failwith("Stub_gmap_ml: ml_create_volume: array size must be greater than 0 (1)");

  nbfaces = Int_val(vnbfaces);
  if (nbfaces <= 0)
    failwith("Stub_gmap_ml: ml_create_volume: array size must be greater than 0 (3)");

  faces = (CDart**)malloc(nbfaces * sizeof(CDart*));
  if (faces == NULL)
    failwith("Stub_gmap_ml: ml_create_volume: memory allocation failed (0)");
  printf("length : %i\n",nbfaces); fflush(stdout);

  vblock = Field(v_ncell_array,0) ;
  gmap = get_gmap_from_ncell(vblock);
  for(i=0;i<nbfaces;i++) {
    vblock = Field(v_ncell_array,i);
    if (gmap != (get_gmap_from_ncell(vblock))) {
      free(faces);
      failwith("Stub_gmap_ml: ml_create_volume: all ncells must belong to the same gmap");
    }
    if (2 != (get_dim_from_ncell(vblock))) {
      free(faces);
      failwith("Stub_gmap_ml: ml_create_volume: i'm waiting for faces (dim = 2)");
    }
    faces[i] = get_dart_from_ncell(vblock);
  }

  length_tab = (int*)malloc(length * sizeof(int));
  if (length_tab == NULL) {
    free(faces);
    failwith("Stub_gmap_ml: ml_create_volume: memory allocation failed (1)");
  }
  printf("length : %i\n",length); fflush(stdout);

  vertex_tab_tab = (CDart***)malloc(length * sizeof(CDart**));
  if (vertex_tab_tab == NULL) {
    free(faces);
    free(length_tab);
    failwith("Stub_gmap_ml: ml_create_volume: memory allocation failed (2)");
  }

  for (i=0;i<length;i++) {
    int tmp;
    vblock = Field(v_int_array,i) ;
    tmp = Int_val(vblock);
    if (tmp <= 0) {
      free(faces);
      free(length_tab);
      free(vertex_tab_tab);
      failwith("Stub_gmap_ml: ml_create_volume: array size must be greater than 0 (2)");
    }
    length_tab[i] = tmp;
    printf("length(%i) : %i\n",i,tmp); fflush(stdout);
  }
  
  for (i=0;i<length;i++) {
    vertex_tab_tab[i] = (CDart**)malloc(length_tab[i] * sizeof(CDart*));
    if (vertex_tab_tab[i] == NULL) {
      for (j=0;j<i;j++) {
	free(vertex_tab_tab[j]);
      }
      free(faces);
      free(vertex_tab_tab);
      free(length_tab);
      failwith("Stub_gmap_ml: ml_create_volume: memory allocation failed (3)");
    }
  }

  for(i=0;i<length;i++) {
    vblock = Field(v_ncell_array_array,i) ;
    for(j=0;j<length_tab[i];j++) {
      vncell = Field(vblock,j) ;
      if (gmap != (get_gmap_from_ncell(vncell)) ) {
	for (k=0;k<length;k++) {
	  free(vertex_tab_tab[k]);
	}
	free(faces);
	free(vertex_tab_tab);
	free(length_tab);
	failwith("Stub_gmap_ml: ml_create_volume: all ncells must belong to the same gmap");
      }

      if ((get_dim_from_ncell(vncell)) != 0) {
	for (k=0;k<length;k++) {
	  free(vertex_tab_tab[k]);
	}
	free(faces);
	free(vertex_tab_tab);
	free(length_tab);
	failwith("Stub_gmap_ml: ml_create_volume: all ncells must be vertices");
      }

      vertex_tab_tab[i][j] = get_dart_from_ncell(vncell) ;
    }
  }
  printf("coucou\n"); fflush(stdout);
  failwith("Stub_gmap_ml: ml_create_volume: not yet implemented in fact  :o ");
  create_volume(gmap,faces,nbfaces,vertex_tab_tab,length_tab,length);

  CAMLreturn (Val_unit);
/*   ans = alloc_tuple(1+2*length); */ 
/*   Store_field(ans,0,build_ncell(gmap,normalForm(gmap,f,2),2)) ; */ 
/*   for (i=0;i<length;i++) { */ 
/*     Store_field(ans,2*i+1,build_ncell(gmap,normalForm(gmap,vertex_tab_tab[i][0],0),0)); */ 
/*     Store_field(ans,2*i+2,build_ncell(gmap,normalForm(gmap,vertex_tab_tab[i][1],0),0)); */ 
/*   } */
/*   free(vertex_tab_tab); */ 

/*   CAMLreturn(ans); */ 
}

CAMLprim value ml_remove_ncell(value vncell)
{
  CAMLparam1(vncell);
  CAMLlocal1(ans);
  SGMapVertex* g = get_gmap_from_ncell(vncell);
  CDart* c = get_dart_from_ncell(vncell);
  int d = get_dim_from_ncell(vncell);
  CDart* f = NULL;

  if (debugAff) {printf("remove_ncell\n");fflush(stdout);}

  f = basic_remove(g,c,d);

  ans = build_ncell(g,f,d+1) ;
  CAMLreturn(ans);
}

CAMLprim value ml_delete_ncell(value vncell)
{
  CAMLparam1(vncell);
  CAMLlocal1(vtab);
  SGMapVertex* g = get_gmap_from_ncell(vncell);
  CDart* c = get_dart_from_ncell(vncell);
  int d = get_dim_from_ncell(vncell);
  int ltab = 0, i = 0;
  CDart** tab = NULL;

  if (debugAff) {printf("remove_ncell\n");fflush(stdout);}

  ltab = basic_delete(g,c,d,&tab);
  vtab = alloc_tuple(ltab);
  for (i=0;i<ltab;i++) {
    Store_field(vtab,i,build_ncell(g,tab[i],d-1)) ;
  }
  deleteDartsTable(tab) ;
  CAMLreturn(vtab);
}

CAMLprim value ml_glue2FacesBy4Verteces(value param)
{
  CAMLparam1(param);
  CAMLlocal5(v_c1,v_c2,v_v1c1,v_v1c2,v_v2c1);
  CAMLlocal1(v_v2c2);
  CDart *c1,*c2,*v1c1,*v1c2,*v2c1,*v2c2;
  SGMapVertex *gmap;
  if (debugAff) {printf("ml_glue2FacesBy4Verteces\n");fflush(stdout);}
  v_c1 = Field(Field(param,0),0);
  v_c2 = Field(Field(param,0),1);
  v_v1c1 = Field(Field(param,1),0);
  v_v1c2 = Field(Field(param,1),1);
  v_v2c1 = Field(Field(param,2),0);
  v_v2c2 = Field(Field(param,2),1);
  
  if ( (get_dim_from_ncell(v_c1) != 2)
       || (get_dim_from_ncell(v_c2) != 2)
       || (get_dim_from_ncell(v_v1c1) != 0)
       || (get_dim_from_ncell(v_v1c2) != 0)
       || (get_dim_from_ncell(v_v2c1) != 0)
       || (get_dim_from_ncell(v_v2c2) != 0) )
    {
      failwith("Stub_gmap_ml: ml_glue2FacesBy4Verteces: unable to glue: verify dimension of ncells");
    }
  
  gmap = get_gmap_from_ncell(v_c1) ;
  if ( (get_gmap_from_ncell(v_c2) != gmap)
       || (get_gmap_from_ncell(v_v1c1) != gmap)
       || (get_gmap_from_ncell(v_v1c2) != gmap)
       || (get_gmap_from_ncell(v_v2c1) != gmap)
       || (get_gmap_from_ncell(v_v2c2) != gmap) )
    {
      failwith("Stub_gmap_ml: ml_glue2FacesBy4Verteces: unable to glue: ncells must be in the same gmap");
    }

  c1 = get_dart_from_ncell(v_c1) ;
  c2 = get_dart_from_ncell(v_c2) ;
  v1c1 = get_dart_from_ncell(v_v1c1) ;
  v2c1 = get_dart_from_ncell(v_v2c1) ;
  v1c2 = get_dart_from_ncell(v_v1c2) ;
  v2c2 = get_dart_from_ncell(v_v2c2) ;
  glue2FacesBy4Verteces(gmap,c1,c2,v1c1,v1c2,v2c1,v2c2) ;
  CAMLreturn(Val_unit);
}

