(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: printer_scalar.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2006-01-13 21:47:06 $
 * Version: $Revision: 1.5 $
 * -------------------------------------------------------------------
 *
 * Les fonctions d'impressions sur les scalaires.
 *
 * Il faut implmenter par type Xxx:
 *
 *  -  _uprbval_'outBuffer_xxx : impression non-formatte de xxx
 *  -  _pfbval_'outBufferF_xxx : impression formatte de xxx
 *
 *
 *=========================================================================
 *)

open Buffer;;
open Printf;;
open Format;;

open Global;;
open Valeur;;

open Dispatch_sig;;
open Environnement;;
open Printer;;


__register_version "$RCSfile: printer_scalar.ml,v $" "$Revision: 1.5 $" "$Date: 2006-01-13 21:47:06 $";;






(*** =======================================================
 *** === Impressions non formatts =========================
 *** ======================================================= 
 ***)



(** uprval b x : est une fonction qui imprime une valeur mgs x dans son buffer b.
 ** Cette fonction est la fonction principale d'affichage. L'impression dans une 
 ** string ou bien sur la sortie standard ou bien dans un fichier derivent de 
 ** cette fonction
 **)
let _uprbval_'outBuffer_undef (b, x) = 
  (if x = "" then Printf.bprintf b "<undef>" else Printf.bprintf b "<undef: %s>" x);
  valundef

and _uprbval_'outBuffer_int (b, x) =  Printf.bprintf b "%d" x; valundef

and _uprbval_'outBuffer_bint (b, x) =  Printf.bprintf b "%s" (Big_int.string_of_big_int x); valundef

and _uprbval_'outBuffer_bool (b, x) = Printf.bprintf b "%s" (if x then "true" else "false"); valundef

and _uprbval_'outBuffer_float (b, x)  = Printf.bprintf b "%f" x; valundef

and _uprbval_'outBuffer_string (b, x) = Printf.bprintf b "%s" x; valundef

and _uprbval_'outBuffer_symbol (b, x) = Printf.bprintf b "%s" x; valundef

and _uprbval_'outBuffer_funct (b, x)  = Printf.bprintf b "%s" (name_of_funct x); valundef
;;

(** _upobval1_'outBuffer_string_string_string_funct_XXX (b,pre,inter,post,f,x)
 ** prend un buffer non-format [b], un prefix [pre], un postfix
 ** [post], un separateur [inter], une fonction [f] et une valeur [x] de type XXX. 
 ** Elle ecrit de faon non-formate l'lment de type mgs [x] dans le buffer [b] 
 ** avec pour prfixe [pre], pour infixe [inter], pour suffixe [post].
 ** De plus, elle applique la fonction f sur chaque lment de [x].
 ** Puis, elle renvoie le type valundef
 **)

let rec _upobval1_'outBuffer_string_string_string_funct_undef (b,pre,inter,post,f,x) =
  (if x = "" then Printf.bprintf b "<undef>" else Printf.bprintf b "<undef: %s>" ( x));
  valundef

and _upobval1_'outBuffer_string_string_string_funct_int (b,pre,inter,post,f,x) =
  Printf.bprintf b "%d" x; valundef

and _upobval1_'outBuffer_string_string_string_funct_bint (b,pre,inter,post,f,x) =
  Printf.bprintf b "%s" (Big_int.string_of_big_int x); valundef

and _upobval1_'outBuffer_string_string_string_funct_bool (b,pre, inter,post,f,x) =
  Printf.bprintf b "%s" (if x then "true" else "false"); valundef

and _upobval1_'outBuffer_string_string_string_funct_float (b,pre, inter,post,f,x) =
  Printf.bprintf b "%f" x; valundef

and _upobval1_'outBuffer_string_string_string_funct_string (b,pre,inter,post,f,x) =
  Printf.bprintf b "%s" x; valundef

and _upobval1_'outBuffer_string_string_string_funct_symbol (b,pre,inter,post,f,x) =
  Printf.bprintf b "%s" x; valundef

and _upobval1_'outBuffer_string_string_string_funct_funct (b,pre,inter,post,f,x) =
  Printf.bprintf b "%s" (name_of_funct x); valundef
;;


(*** =======================================================
 *** === Impressions FORMATTS =============================
 *** ======================================================= 
 ***)



(** _pfbval_'outBufferF_XXX (b, x) prend un buffer format [b] et une valeur [x] de type XXX. 
 ** Elle crit de faon formate [x] dans le buffer format [b], puis renvoie valundef
 **)

let _pfbval_'outBufferF_undef (b, x) =
  (if x = "" then Format.fprintf b "<undef>" else Format.fprintf b "<undef: %s>" x);
  valundef

and _pfbval_'outBufferF_int (b, x) = Format.fprintf b "%d" x; valundef

and _pfbval_'outBufferF_bint (b, x) = Format.fprintf b "%s" (Big_int.string_of_big_int x); valundef

and _pfbval_'outBufferF_bool (b, x) = Format.fprintf b "%s" (if x then "true" else "false"); valundef

and _pfbval_'outBufferF_float (b, x) = Format.fprintf b "%f" x; valundef

and _pfbval_'outBufferF_string (b, x) = Format.fprintf b "%s" x; valundef

and _pfbval_'outBufferF_symbol (b, x) = Format.fprintf b "%s" x; valundef

and _pfbval_'outBufferF_funct (b, x) = Format.fprintf b "%s" (name_of_funct x); valundef
;;

(** _pobval1_'outBufferF_int_string_string_string_funct_XXX (b,t,pre,inter,post,f,x)
 ** prend un buffer format [b], une taille d'indentation [t], une prefix [pre], un postfix
 ** [post], un separateur [inter], une fonction [f] et une valeur [x] de type XXX. 
 ** Elle ecrit de faon formate l'lment de type mgs [x] dans le buffer [b] 
 ** avec pour prfixe [pre], pour infixe [inter], pour suffixe [post], le tout
 ** avec l'indentation [t].
 ** De plus, elle applique la fonction f sur chaque lment de [x].
 ** Puis, elle renvoie le type valundef
 **)

let rec _pobval1_'outBufferF_int_string_string_string_funct_undef (b,t,pre,inter,post,f,x) =
  (if x = "" then Format.fprintf b "<undef>" else Format.fprintf b "<undef: %s>" ( x));
  valundef

and _pobval1_'outBufferF_int_string_string_string_funct_int (b,t,pre,inter,post,f,x) =
  Format.fprintf b "%d" x; valundef

and _pobval1_'outBufferF_int_string_string_string_funct_bint (b,t,pre,inter,post,f,x) =
  Format.fprintf b "%s" (Big_int.string_of_big_int x); valundef

and _pobval1_'outBufferF_int_string_string_string_funct_bool (b,t,pre, inter,post,f,x) =
  Format.fprintf b "%s" (if x then "true" else "false"); valundef

and _pobval1_'outBufferF_int_string_string_string_funct_float (b,t,pre, inter,post,f,x) =
  Format.fprintf b "%f" x; valundef

and _pobval1_'outBufferF_int_string_string_string_funct_string (b,t,pre,inter,post,f,x) =
  Format.fprintf b "%s" x; valundef

and _pobval1_'outBufferF_int_string_string_string_funct_symbol (b,t,pre,inter,post,f,x) =
  Format.fprintf b "%s" x; valundef

and _pobval1_'outBufferF_int_string_string_string_funct_funct (b,t,pre,inter,post,f,x) =
  Format.fprintf b "%s" (name_of_funct x); valundef
;;

(* --- fin du fichier ----------------------------------- *)
