(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: scalar.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2008-06-03 12:59:19 $
 * Version: $Revision: 1.11 $
 * -------------------------------------------------------------------
 *
 * Status:  
 *
 *    Implantation au niveau "representation MGS" des fonctions sur les 
 *    scalaire "non-spcifiques" : undef, int bool, float, string, funct
 *
 *    Attention, les scalaires spcifiques correspondant  la position
 *    dans une topologie (comme par exemple posgbf, splx, ncell, gmap) 
 *    sont traits dans le rprtoire de la collection.
 *
 *=========================================================================
 *)


open Global;;
open Valeur;;
open Dispatch_sig;;
open Auxfct;;
open Big_int;;


__register_version "$RCSfile: scalar.ml,v $" "$Revision: 1.11 $" "$Date: 2008-06-03 12:59:19 $";;




let _size_undef  x = VAL_int(-1)
and _size_int    x = VAL_int(-2)
and _size_bint   x = VAL_int(-11)
and _size_bool   x = VAL_int(-3)
and _size_float  x = VAL_int(-4)
and _size_string x = VAL_int(-5)
and _size_funct  x = VAL_int(-6)
and _size_symbol x = VAL_int(-11)
;;

(* Pour les autres voir dans les fichiers concernes. Dans Gbf/gbf.ml par exemple pour _size_gbf *)
(* ou meme size_posgbf                                                                          *)

let _empty_any(x)  = 
  if is_collection(x)
  then
    begin
      Printf.eprintf "\nInternal error: empty on collection type %s is not defined\n" (constructor_string x);
      exit(1)
    end
  else
    VAL_undef("empty is undefined on a scalar")
;;



let _fold_funct_any_int   (fct, zero, n) = Auxfct.int_fold (fun v suite -> (app_2 fct (VAL_int v) suite)) zero n
and _fold_funct_any_float (fct, zero, n) = Auxfct.int_fold (fun v suite -> (app_2 fct (VAL_int v) suite)) zero (int_of_float n)
;;



(** --- Arithmtique ------------------------------------------------------- **)

let _add_int_int       (x, y) = VAL_int (x+y)
and _add_float_int     (x, y) = VAL_float(x +. (float_of_int y))
and _add_int_float     (x, y) = VAL_float ((float_of_int x) +. y)
and _add_float_float   (x, y) = VAL_float (x +. y)

and _add_any_string    (x, y) = VAL_string((to_string (Printer.unformatted_prsvalue x))^y)
and _add_string_any    (x, y) = VAL_string(x^(to_string (Printer.unformatted_prsvalue y)))
and _add_string_string (x, y) = VAL_string(x^y) (* ce cas n'est pas subsumer par les 2 prcdents car faut enlever les guillemets. *)
and _add_any_any (x, y) = VAL_undef("cannot add a "^(constructor_string x)^" to a "^(constructor_string y))


and _add_bint_bint     (x, y) = VAL_bint (add_big_int x y)
and _add_int_bint      (x, y) = VAL_bint (add_big_int (big_int_of_int x) y)
and _add_bint_int      (x, y) = VAL_bint (add_big_int x (big_int_of_int y))
and _add_float_bint    (x, y) = VAL_float(x +. (float_of_big_int y))
and _add_bint_float    (x, y) = VAL_float ((float_of_big_int x) +. y)

;;

let _sub_int_int (x, y) = VAL_int (x-y)
and _sub_float_int   (x, y) = VAL_float (x -. (float_of_int y))
and _sub_int_float   (x, y) = VAL_float ((float_of_int x) -. y)
and _sub_float_float (x, y) = VAL_float (x -. y)

and _sub_any_any (x, y) = VAL_undef("cannot sub a "^(constructor_string x)^" to a "^(constructor_string y))

and _sub_bint_bint     (x, y) = VAL_bint (sub_big_int x y)
and _sub_int_bint      (x, y) = VAL_bint (sub_big_int (big_int_of_int x) y)
and _sub_bint_int      (x, y) = VAL_bint (sub_big_int x (big_int_of_int y))
and _sub_float_bint    (x, y) = VAL_float(x -. (float_of_big_int y))
and _sub_bint_float    (x, y) = VAL_float ((float_of_big_int x) -. y)

;;

let _mult_int_int     (x, y) = VAL_int (x*y)
and _mult_float_int   (x, y) = VAL_float (x *. (float_of_int y))
and _mult_int_float   (x, y) = VAL_float ((float_of_int x) *. y)
and _mult_float_float (x, y) = VAL_float (x *. y)
and _mult_any_any     (x, y) = VAL_undef("cannot mult a "^(constructor_string x)^" to a "^(constructor_string y))

and _mult_bint_bint     (x, y) = VAL_bint (mult_big_int x y)
and _mult_int_bint      (x, y) = VAL_bint (mult_int_big_int x y)
and _mult_bint_int      (x, y) = VAL_bint (mult_int_big_int y x)
and _mult_float_bint    (x, y) = VAL_float(x *. (float_of_big_int y))
and _mult_bint_float    (x, y) = VAL_float ((float_of_big_int x) *. y)
;;

let _pow_float_int   (x, y) = VAL_float (puissance x  (float_of_int y))
and _pow_int_float   (x, y) = VAL_float (puissance (float_of_int x) y)
and _pow_float_float (x, y) = VAL_float (puissance x y)
and _pow_any_any     (x, y) = VAL_undef("cannot power a "^(constructor_string x)^" to a "^(constructor_string y))

and _pow_int_int     (x, y) =
  if (y >= 0)
  then VAL_bint  (power_int_positive_int x y)
  else VAL_float (puissance (float_of_int x) (float_of_int y))
and _pow_bint_int    (x, y) =
  if (y >= 0)
  then VAL_bint  (power_big_int_positive_int x y)
  else VAL_float (puissance (float_of_big_int x) (float_of_int y))
and _pow_int_bint    (x, y) =
  if (ge_big_int y zero_big_int)
  then VAL_bint  (power_int_positive_big_int x y)
  else VAL_float (puissance (float_of_int x) (float_of_big_int y))
and _pow_bint_bint   (x, y) =
  if (ge_big_int y zero_big_int)
  then VAL_bint  (power_big_int_positive_big_int x y)
  else VAL_float (puissance (float_of_big_int x) (float_of_big_int y))
;;

let _div_int_int (x, y) = VAL_int (x/y)
and _div_float_int   (x, y) = VAL_float (x /. (float_of_int y))
and _div_int_float   (x, y) = VAL_float ((float_of_int x) /. y)
and _div_float_float (x, y) = VAL_float (x /. y)

and _div_any_any (x, y) = VAL_undef("cannot div a "^(constructor_string x)^" to a "^(constructor_string y))

and _div_bint_bint     (x, y) = VAL_bint (div_big_int x y)
and _div_int_bint      (x, y) = VAL_bint (div_big_int (big_int_of_int x) y)
and _div_bint_int      (x, y) = VAL_bint (div_big_int x (big_int_of_int y))
and _div_float_bint    (x, y) = VAL_float(x /. (float_of_big_int y))
and _div_bint_float    (x, y) = VAL_float ((float_of_big_int x) /. y)
;;

let _succ_int   (x)  = VAL_int(x+1)
let _succ_bint   (x) = VAL_bint(succ_big_int x)
and _succ_float (x)  = VAL_float(x +. 1.0)
;;


let _leq_int_int      (x, y) = VAL_bool (x <= y)
and _leq_float_int    (x, y) = VAL_bool (x <= (float_of_int y))
and _leq_int_float    (x, y) = VAL_bool ((float_of_int x) <= y)
and _leq_float_float  (x, y) = VAL_bool (x <= y)
and _leq_string_string(x, y) = VAL_bool (x <= y)
and _leq_bint_bint    (x, y) = VAL_bool (le_big_int x y)
and _leq_int_bint     (x, y) = VAL_bool (le_big_int (big_int_of_int x) y)
and _leq_bint_int     (x, y) = VAL_bool (le_big_int x (big_int_of_int y))
and _leq_float_bint   (x, y) = VAL_bool (x <= (float_of_big_int y))
and _leq_bint_float   (x, y) = VAL_bool ((float_of_big_int x) <= y)
and _leq_any_any      (x, y) = 
  raise_error 
    (Printf.sprintf "numerical comparaison (<=) between: %s and %s does not make sense (use 'compare)" 
       (constructor_string x) 
       (constructor_string y))

and _lt_int_int      (x, y) = VAL_bool (x < y)
and _lt_float_int    (x, y) = VAL_bool (x < (float_of_int y))
and _lt_int_float    (x, y) = VAL_bool ((float_of_int x) < y)
and _lt_float_float  (x, y) = VAL_bool (x < y)
and _lt_bint_bint    (x, y) = VAL_bool (lt_big_int x y)
and _lt_int_bint     (x, y) = VAL_bool (lt_big_int (big_int_of_int x) y)
and _lt_bint_int     (x, y) = VAL_bool (lt_big_int x (big_int_of_int y))
and _lt_float_bint   (x, y) = VAL_bool (x < (float_of_big_int y))
and _lt_bint_float   (x, y) = VAL_bool ((float_of_big_int x) < y)
and _lt_string_string(x, y) = VAL_bool (x < y)
and _lt_any_any      (x, y) = 
  raise_error 
    (Printf.sprintf "numerical comparaison (<) between: %s and %s does not make sense (use 'compare)" 
       (constructor_string x) 
       (constructor_string y))


and _geq_int_int      (x, y) = VAL_bool (x >= y)
and _geq_float_int    (x, y) = VAL_bool (x >= (float_of_int y))
and _geq_int_float    (x, y) = VAL_bool ((float_of_int x) >= y)
and _geq_float_float  (x, y) = VAL_bool (x >= y)
and _geq_bint_bint    (x, y) = VAL_bool (ge_big_int x y)
and _geq_int_bint     (x, y) = VAL_bool (ge_big_int (big_int_of_int x) y)
and _geq_bint_int     (x, y) = VAL_bool (ge_big_int x (big_int_of_int y))
and _geq_float_bint   (x, y) = VAL_bool (x >= (float_of_big_int y))
and _geq_bint_float   (x, y) = VAL_bool ((float_of_big_int x) >= y)
and _geq_string_string(x, y) = VAL_bool (x >= y)
and _geq_any_any      (x, y) = 
  raise_error 
    (Printf.sprintf "numerical comparaison (>=) between: %s and %s does not make sense (use 'compare)" 
       (constructor_string x) 
       (constructor_string y))


and _gt_int_int      (x, y) = VAL_bool (x > y)
and _gt_float_int    (x, y) = VAL_bool (x > (float_of_int y))
and _gt_int_float    (x, y) = VAL_bool ((float_of_int x) > y)
and _gt_float_float  (x, y) = VAL_bool (x > y)
and _gt_bint_bint    (x, y) = VAL_bool (gt_big_int x y)
and _gt_int_bint     (x, y) = VAL_bool (gt_big_int (big_int_of_int x) y)
and _gt_bint_int     (x, y) = VAL_bool (gt_big_int x (big_int_of_int y))
and _gt_float_bint   (x, y) = VAL_bool (x > (float_of_big_int y))
and _gt_bint_float   (x, y) = VAL_bool ((float_of_big_int x) > y)
and _gt_string_string(x, y) = VAL_bool (x > y)
and _gt_any_any      (x, y) = 
  raise_error 
    (Printf.sprintf "numerical comparaison (>) between: %s and %s does not make sense (use 'compare)" 
       (constructor_string x) 
       (constructor_string y))
;;


let _maximum_int_int      (x, y) = VAL_int(Pervasives.max x y)
and _maximum_float_int    (x, y) = VAL_float(Pervasives.max x (float_of_int y))
and _maximum_int_float    (x, y) = VAL_float(Pervasives.max (float_of_int x) y)
and _maximum_float_float  (x, y) = VAL_float(Pervasives.max x  y)
and _maximum_bint_bint    (x, y) = VAL_bint(max_big_int x y)
and _maximum_int_bint     (x, y) = VAL_bint(max_big_int (big_int_of_int x) y)
and _maximum_bint_int     (x, y) = VAL_bint(max_big_int x (big_int_of_int y))
and _maximum_float_bint   (x, y) = VAL_float(Pervasives.max x (float_of_big_int y))
and _maximum_bint_float   (x, y) = VAL_float(Pervasives.max (float_of_big_int x) y)
and _maximum_string_string(x, y) = VAL_string(Pervasives.max x y)
and _maximum_any_any      (x, y) =  if 0 < to_int(compareMgsVal x y) then x else y



and _minimum_int_int      (x, y) = VAL_int(Pervasives.min x y)
and _minimum_float_int    (x, y) = VAL_float(Pervasives.min x (float_of_int y))
and _minimum_int_float    (x, y) = VAL_float(Pervasives.min (float_of_int x) y)
and _minimum_float_float  (x, y) = VAL_float(Pervasives.min x  y)
and _minimum_bint_bint    (x, y) = VAL_bint(min_big_int x y)
and _minimum_int_bint     (x, y) = VAL_bint(min_big_int (big_int_of_int x) y)
and _minimum_bint_int     (x, y) = VAL_bint(min_big_int x (big_int_of_int y))
and _minimum_float_bint   (x, y) = VAL_float(Pervasives.min x (float_of_big_int y))
and _minimum_bint_float   (x, y) = VAL_float(Pervasives.min (float_of_big_int x) y)
and _minimum_string_string(x, y) = VAL_string(Pervasives.min x y)
and _minimum_any_any      (x, y) =  if 0 > to_int(compareMgsVal x y) then x else y

and _compareMgsVal_bint_bint  (x, y) = VAL_int(compare_big_int x y)
and _compareMgsVal_int_bint   (x, y) = VAL_int(compare_big_int (big_int_of_int x) y)
and _compareMgsVal_bint_int   (x, y) = VAL_int(compare_big_int x (big_int_of_int y))
and _compareMgsVal_bint_float (x, y) = VAL_int(Pervasives.compare (float_of_big_int x) y)
and _compareMgsVal_float_bint (x, y) = VAL_int(Pervasives.compare x (float_of_big_int y))
;;


let _random_int    (x) = VAL_int(Random.int x)
and _random_float  (x) = VAL_float(Random.float x)
and _random_bool   (x) = VAL_bool(0 <> (Random.int 2))
and _random_string (x) = 
  let sz = String.length x
  and n = Random.int (1 + !lrandom_string) in
  let s = String.create n in
    for i = 0 to n-1 do
      String.set s i (String.get x (Random.int sz))
    done;
    VAL_string(s)
;;

(** Les fonctions quasi_random issues de Gsl ------ **)

let _qrngmake_string_int (s, i) = (try 
				     VAL_int(Gsl_random.make s i)
				   with Not_found -> raise_error "illegal dimension in qrngmake")
and _qrnginit_int i             = try VAL_undef (let _ = Gsl_random.init i in "qrng_init") with Not_found -> raise_error "wrong qrng index"
and _qrngget_int i              = (try
				     VAL_seq (Array.map (fun x -> VAL_float x) (Gsl_random.get i), Type_seq.seq_desc)
				   with Not_found -> raise_error "wrong qrng index")
and _qrngsample_int i           = (try
				     VAL_seq (Array.map (fun x -> VAL_float x) (Gsl_random.sample i), Type_seq.seq_desc)
				   with Not_found -> raise_error "wrong qrng index")
and _qrngname_int i             = try VAL_string (Gsl_random.name i) with Not_found -> raise_error "wrong qrng index"
and _qrngdimension_int i        = try VAL_int (Gsl_random.dimension i) with Not_found -> raise_error "wrong qrng index"
;;


(** Les fonctions rng issues de Gsl ------ **)

let _rngname_any _         = VAL_string (Gsl_random.rng_name())
and _rngdefaultseed_any _  = VAL_int    (Nativeint.to_int(Gsl_random.rng_default_seed()))
and _rngsetseed_int i      = (let _ = (Gsl_random.rng_set_default_seed(Nativeint.of_int i))
                              in
				VAL_int (Nativeint.to_int(Gsl_random.rng_default_seed())))
  
and _rngmin_any _          = VAL_int    (Nativeint.to_int(Gsl_random.rng_min()))
and _rngmax_any _          = VAL_int    (Nativeint.to_int(Gsl_random.rng_max()))
and _rngget_any _          = VAL_float  (let r = (Gsl_random.rng_get())
					 in Printf.printf "%nu\n" r;
					   flush Pervasives.stdout;
					   Nativeint.to_float r


)         (* un entier dans [min, max] *)
and _rnguniform_any _      = VAL_float  (Gsl_random.rng_uniform())     (* un flottant dans [0, 1 [  *)
and _rnguniformpos_any _   = VAL_float  (Gsl_random.rng_uniformpos())  (* un flottant dans ]0, 1[   *)
and _rnguniformint_int i   = VAL_int    (Gsl_random.rng_uniformint i)  (* un entier dans [0, i-1]   *)
;;


(** Les prdicats de type ------------------------- **)

let _intPredicate_int _ = mgstrue
and _intPredicate_any _ = mgsfalse

and _bintPredicate_bint _ = mgstrue
and _bintPredicate_any _  = mgsfalse

and _floatPredicate_float _ = mgstrue
and _floatPredicate_any _   = mgsfalse

and _boolPredicate_bool _ = mgstrue
and _boolPredicate_any _  = mgsfalse

and _stringPredicate_string _ = mgstrue
and _stringPredicate_any _    = mgsfalse

and _functPredicate_funct _ = mgstrue
and _functPredicate_any _   = mgsfalse

and _undefPredicate_undef _ = mgstrue
and _undefPredicate_any _   = mgsfalse

and _symbolPredicate_symbol _ = mgstrue
and _symbolPredicate_any _    = mgsfalse

;;


let _val2bool_undef x = mgsfalse
and _val2bool_int x = VAL_bool (x <> 0)
and _val2bool_bint x = VAL_bool (not (eq_big_int x zero_big_int))
and _val2bool_bool x = VAL_bool x
and _val2bool_float x = VAL_bool (x <> 0.0)
and _val2bool_string x = VAL_bool (0 <> String.length(x))
and _val2bool_funct x = mgsfalse
and _val2bool_symbol x = mgsfalse
;;




(** --- Math IEEE ------------------------------------------------------- **)

let rec auxpower_int x = function
  | 0 -> 1
  | 1 -> x
  | y -> if 0 = y mod 2 then auxpower_int (x*x) (y/2) else x*(auxpower_int x (y - 1))
;;

let _power_float_float (x, y) = VAL_float(x**y)
and _power_float_int   (x, y) = VAL_float(x**(float_of_int y))
and _power_int_float   (x, y) = VAL_float((float_of_int x)**y)
and _power_int_int     (x, y) = 
  if y >= 0 then VAL_int(auxpower_int x y) 
  else VAL_float(1.0 /. (float_of_int (auxpower_int x (0-y))))
and _power_bint_bint   (x, y) =
  if (ge_big_int y zero_big_int)
  then VAL_bint  (power_big_int_positive_big_int x y)
  else VAL_float ((float_of_big_int x)**(float_of_big_int y))
and _power_int_bint    (x, y) =
  if (ge_big_int y zero_big_int)
  then VAL_bint  (power_int_positive_big_int x y)
  else VAL_float ((float_of_int x)**(float_of_big_int y))
and _power_bint_int    (x, y) =
  if (y >= 0)
  then VAL_bint  (power_big_int_positive_int x y)
  else VAL_float ((float_of_big_int x)**(float_of_int y))
and _power_float_bint  (x, y) = VAL_float(x**(float_of_big_int y))
and _power_bint_float  (x, y) = VAL_float((float_of_big_int x)**y)
;;


let _atan2_float_float (x, y) = VAL_float(Pervasives.atan2 x y)
and _atan2_float_int   (x, y) = VAL_float(Pervasives.atan2 x (float_of_int y))
and _atan2_int_float   (x, y) = VAL_float(Pervasives.atan2 (float_of_int x) y)
and _atan2_int_int     (x, y) = VAL_float(Pervasives.atan2 (float_of_int x) (float_of_int y))
and _atan2_float_bint  (x, y) = VAL_float(Pervasives.atan2 x (float_of_big_int y))
and _atan2_bint_float  (x, y) = VAL_float(Pervasives.atan2 (float_of_big_int x) y)
and _atan2_bint_bint   (x, y) = VAL_float(Pervasives.atan2 (float_of_big_int x) (float_of_big_int y))
;;


let _modulo_int_int     (x, y) = try VAL_int(x mod y) with Division_by_zero -> VAL_undef("mod: div by zero")
and _modulo_int_float   (x, y) = VAL_float(Pervasives.mod_float (float_of_int x) y)
and _modulo_float_int   (x, y) = VAL_float(Pervasives.mod_float x (float_of_int y))
and _modulo_float_float (x, y) = VAL_float(Pervasives.mod_float x y)
and _modulo_bint_float  (x, y) = VAL_float(Pervasives.mod_float (float_of_big_int x) y)
and _modulo_float_bint  (x, y) = VAL_float(Pervasives.mod_float x (float_of_big_int y))
and _modulo_bint_bint   (x, y) =
  if (eq_big_int zero_big_int y)
  then VAL_undef("mod: div by zero")
  else VAL_bint(mod_big_int x y)
and _modulo_int_bint    (x, y) =
  if (eq_big_int zero_big_int y)
  then VAL_undef("mod: div by zero")
  else VAL_bint(mod_big_int (big_int_of_int x) y)
and _modulo_bint_int    (x, y) =
  if (y = 0)
  then VAL_undef("mod: div by zero")
  else VAL_bint(mod_big_int x (big_int_of_int y))
;;

let _exp_int   x = VAL_float(Pervasives.exp(float_of_int x))
and _exp_bint  x = VAL_float(Pervasives.exp(float_of_big_int x))
and _exp_float x = VAL_float(Pervasives.exp x)

and _acos_int   x = VAL_float(Pervasives.acos (float_of_int x))
and _acos_float x = VAL_float(Pervasives.acos x)

and _asin_int   x = VAL_float(Pervasives.asin (float_of_int x))
and _asin_bint  x = VAL_float(Pervasives.asin (float_of_big_int x))
and _asin_float x = VAL_float(Pervasives.asin x)

and _atan_int   x = VAL_float(Pervasives.atan (float_of_int x))
and _atan_bint  x = VAL_float(Pervasives.atan (float_of_big_int x))
and _atan_float x = VAL_float(Pervasives.atan x)


and _cos_int   x = VAL_float(Pervasives.cos (float_of_int x))
and _cos_bint  x = VAL_float(Pervasives.cos (float_of_big_int x))
and _cos_float x = VAL_float(Pervasives.cos x)


and _cosh_int   x = VAL_float(Pervasives.cosh (float_of_int x))
and _cosh_bint  x = VAL_float(Pervasives.cosh (float_of_big_int x))
and _cosh_float x = VAL_float(Pervasives.cosh x)


and _log_int   x = VAL_float(Pervasives.log (float_of_int x))
and _log_bint  x = VAL_float(Pervasives.log (float_of_big_int x))
and _log_float x = VAL_float(Pervasives.log x)


and _log10_int   x = VAL_float(Pervasives.log10 (float_of_int x))
and _log10_bint  x = VAL_float(Pervasives.log10 (float_of_big_int x))
and _log10_float x = VAL_float(Pervasives.log10 x)


and _sin_int   x = VAL_float(Pervasives.sin (float_of_int x))
and _sin_bint  x = VAL_float(Pervasives.sin (float_of_big_int x))
and _sin_float x = VAL_float(Pervasives.sin x)


and _sinh_int   x = VAL_float(Pervasives.sinh (float_of_int x))
and _sinh_bint  x = VAL_float(Pervasives.sinh (float_of_big_int x))
and _sinh_float x = VAL_float(Pervasives.sinh x)


and _sqrt_int   x = VAL_float(Pervasives.sqrt (float_of_int x))
and _sqrt_bint  x = VAL_float(Pervasives.sqrt (float_of_big_int x))
and _sqrt_float x = VAL_float(Pervasives.sqrt x)


and _tan_int   x = VAL_float(Pervasives.tan (float_of_int x))
and _tan_bint  x = VAL_float(Pervasives.tan (float_of_big_int x))
and _tan_float x = VAL_float(Pervasives.tan x)


and _tanh_int   x = VAL_float(Pervasives.tanh (float_of_int x))
and _tanh_bint  x = VAL_float(Pervasives.tanh (float_of_big_int x))
and _tanh_float x = VAL_float(Pervasives.tanh x)


and _ceil_float x = VAL_int(int_of_float (Pervasives.ceil x))

and _floor_float x = VAL_int(int_of_float (Pervasives.floor x))

and _abs_int   x = VAL_int(if x < 0 then (0-x) else x)
and _abs_bint  x = VAL_bint(abs_big_int x)
and _abs_float x = VAL_float(Pervasives.abs_float x)

and _cnp_bint_bint (n, p) = VAL_bint(good_cnp_big_int n p)
and _cnp_int_bint  (n, p) = VAL_bint(good_cnp_big_int (big_int_of_int n) p)
and _cnp_bint_int  (n, p) = VAL_bint(good_cnp_big_int n (big_int_of_int p))
and _cnp_int_int   (n, p) = VAL_bint(good_cnp_big_int (big_int_of_int n) (big_int_of_int p))

and _anp_bint_bint (n, p) = VAL_bint(good_anp_big_int n p)
and _anp_int_bint  (n, p) = VAL_bint(good_anp_big_int (big_int_of_int n) p)
and _anp_bint_int  (n, p) = VAL_bint(good_anp_big_int n (big_int_of_int p))
and _anp_int_int   (n, p) = VAL_bint(good_anp_big_int (big_int_of_int n) (big_int_of_int p))

and _floatof_int x = VAL_float(Pervasives.float_of_int x)
and _floatof_bint x = VAL_float(float_of_big_int x)
and _floatof_float x = VAL_float(x)
and _floatof_string x = try VAL_float(Pervasives.float_of_string x) with Failure s -> VAL_undef(s^" failed")

and _intof_int x = VAL_int(x)
and _intof_bint x =
  if (is_int_big_int x)
  then VAL_int(int_of_big_int x)
  else VAL_undef("Scalar: _intof_bint: too big to be cast as an int")
and _intof_float x = VAL_int(Pervasives.int_of_float x)
and _intof_string x = try VAL_int(Pervasives.int_of_string x) with Failure s -> VAL_undef(s^" failed")

and _bintof_int x = VAL_bint(big_int_of_int x)
and _bintof_bint x = VAL_bint(x)
and _bintof_float x = VAL_bint(big_int_of_int (int_of_float x))
and _bintof_string x = try VAL_bint(big_int_of_string x) with e -> VAL_undef(x^" failed")

;;


(** de OCAML/Pervasives.mli
  type fpclass =
  | FP_normal
  | FP_subnormal
  | FP_zero
  | FP_infinite
  | FP_nan
  external classify_float: float -> fpclass = "classify_float"
**)
let _whatfloat_float x = match classify_float x with
  | FP_zero      -> VAL_symbol("ieee_zero") (** Number is 0.0 or -0.0 *)
  | FP_normal    -> VAL_symbol("ieee_normal") (** Normal number, none of the other case *)
  | FP_subnormal -> VAL_symbol("ieee_subnormal") (** Number very close to 0.0, has reduced precision *)
  | FP_infinite  -> VAL_symbol("ieee_infinity") (** Number is positive or negative infinity *)
  | FP_nan       -> VAL_symbol("nan") (** Not a number: result of an undefined operation *)
;;




(** Fonctions sur les strings ------------------------------------------------------- **)


let _stringlength_string s = VAL_int (String.length s)

and _stringget_string_int (s, i) = 
  let ns = String.create 1 in
    begin
      try
	String.set ns 0 (String.get s i);
	VAL_string(ns)
      with Invalid_argument m -> VAL_undef("string_get: "^m)
    end

and _stringmake_int_string (i, s) = 
  try		   
    VAL_string(String.make i (String.get s 0))
  with Invalid_argument m -> VAL_undef("string_make: "^m)


and _stringsub_string_int_int (s, i, j) =
  try
    VAL_string(String.sub s i j)
  with Invalid_argument m -> VAL_undef("string_sub: "^m)


and _stringindex_string_string (s, sc) = 
  let c = String.get sc 0 in 
    try 
      VAL_int(String.index s c)
    with Not_found -> VAL_undef("string_index: char "^(String.sub sc 0 1)^" does not occurs in string")

and _stringrindex_string_string (s, sc) = 
  let c = String.get sc 0 in 
    try 
      VAL_int(String.rindex s c)
    with Not_found -> VAL_undef("string_rindex: char "^(String.sub sc 0 1)^" does not occurs in string")


and _stringindexfrom_string_int_string (s, i, sc) = 
  let c = String.get sc 0 in 
    try 
      VAL_int(String.index_from s i c)
    with Not_found -> VAL_undef("string_index_from: char "^(String.sub sc 0 1)^" does not occurs in string")


and _stringrindexfrom_string_int_string (s, i, sc) = 
  let c = String.get sc 0 in 
    try 
      VAL_int(String.rindex_from s i c)
    with Not_found -> VAL_undef("string_rindex_from: char "^(String.sub sc 0 1)^" does not occurs in string")


and _stringcontains_string_string (s, sc) = VAL_bool(String.contains s (String.get sc 0))

and _stringcontainsfrom_string_int_string (s, i, sc) = 
  VAL_bool(String.contains_from s i (String.get sc 0))

and _stringrcontainsfrom_string_int_string (s, i, sc) =
  VAL_bool(String.rcontains_from s i (String.get sc 0))


and _undef2string_undef s = VAL_string(s)

and _string2undef_string s = VAL_undef(s)
;;




(** Fonctions sur les symbols ------------------------------------------------------- **)

let gensymbcpt = ref 0 ;;

let _gensym_string s =
  incr gensymbcpt ;
  if s.[0] == '`'
  then VAL_symbol(s^string_of_int(!gensymbcpt))
  else VAL_symbol("`"^s^string_of_int(!gensymbcpt))

and _gensym_symbol s =
  incr gensymbcpt ;
  VAL_symbol(s^string_of_int(!gensymbcpt))

;;

let _add_symbol_symbol (s1,s2) =
  let s2' = String.copy s2 in
    s2'.[0] <- '_' ;
    VAL_symbol(s1^s2')
;;

let _split_symbol(s) =
  if (s = "")
  then VAL_seq([||], Type_seq.seq_desc)
  else
    let ls = ref [] in
    let len = ref 0 in
      for i =  (String.length s) - (if (s.[(String.length s) - 1] = '_') then 2 else 1) downto 0 do
	if (s.[i] = '_') || (s.[i] = '`')
	then (
	  ls  := VAL_symbol("`"^(String.sub s (i+1) !len)) :: !ls ;
	  len := 0
	)
	else
	  len := !len + 1
      done;
      (*if (List.length !ls = 1) then List.hd !ls else*)
	VAL_seq(Array.of_list !ls, Type_seq.seq_desc)
;;


