(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: kernel_seq.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2004-11-18 16:21:24 $
 * Version: $Revision: 1.3 $
 * -------------------------------------------------------------------
 *
 * Status: 
 *   voir aussi Common/kernel_collection.ml
 *   qui contient les fcts gnriques sur les collections et les monoidaux.
 *
 *=========================================================================
 *)

open Global;;
open Valeur;;
open Type;;
open Dispatch_sig;;
open Dispatch;;
open Environnement;;
open Seq;;


__register_version "$RCSfile: kernel_seq.ml,v $" "$Revision: 1.3 $" "$Date: 2004-11-18 16:21:24 $";;


print_string "(* loading kernel functions on SEQ into user-space ... ";
flush Pervasives.stdout;;



(*!*)

(**
  @module monoidal
  @name 'left
  @type seq_int_any

  Returns the value at the left of an element in a sequence.

  ['left(l, p)] returns the element at the left of position [p] in the
  sequence [l]. Position in a sequence are integers.

  This function is also a special form in a transformation.

  @voir 'right      'right
  @voir Collections.'collection 'collection
  @voir Functional chapter functional
*) 
addenv_impure "'left"         (fun_2 __left);;


(**
  @module monoidal
  @name 'right
  @type seq_int_any

  Returns the value at the right of an element in a sequence.

  ['right(l, p)] returns the element at the right of position [p] in the
  sequence [l]. Position in a sequence are integers.

  This function is also a special form in a transformation.

  @voir 'left       'left
  @voir Collections.'collection 'collection
  @voir Functional chapter functional
*) 
addenv_impure "'right"        (fun_2 __right);;



(**
  @module collections
  @name 'sequify
  @type collection_seq
  @type posgbf_seq
  @type scalar_seq

  Transform its argument into a sequence.

  ['sequify(c)] returns the sequence of the elements of the collection
  [c]. If [c] is already a sequence, [sequify] is the identity.

  ['sequify(s)] where [s] is a string returns a sequence where the element [i] in the
  sequence is a string with only one character, the [i]th character of[s]

  ['sequify(p)] where [p] is a [posgbf] returns the sequence of the
  coefficient of the smith normal form of [p].

  ['sequify(s)] where [s] is a scalar that is not a [posgbf] or a [string] 
  returns the singleton sequence with [s] as the unique element.

  @example 
  [string_to_seq("abc")] returns ["a", "b", "c", seq:()]
*)
(* FIXME: a generer automatiquement *)
addenv_impure "'sequify" (fun_1 __sequify);;





(**
  @module monoidal
  @name 'reverse
  @type seq_seq

  Sequence reversal.
*)
addenv_impure "'reverse" (fun_1 __reverseseq);;


(**
  @module monoidal
  @name 'rotate_left
  @type seq_seq

  [rotate_left]  cycles the elements in the sequence argument one position to the left.

  @voir rotate_right
*)
addenv_impure "'rotate_left" (fun_1 __rotateleft);;


(**
  @module monoidal
  @name 'rotate_right
  @type seq_seq

  [rotate_right]  cycles the elements in the sequence argument one position to the right.

  @voir rotate_left
*)
addenv_impure "'rotate_right" (fun_1 __rotateright);;


(**
  @module monoidal
  @name 'pad_right
  @type seq_int_any_seq

  [pad_right(s, n, v0)]  makes a sequence of length [n] by truncating 
  or padding [s] with [v0] on the right.

  @voir pad_left
*)
addenv_impure "'pad_right" (fun_3 __padRight);;


(**
  @module monoidal
  @name 'pad_left
  @type seq_int_any_seq

  [pad_left(s, n, v0)]  makes a sequence of length [n] by truncating 
  or padding [s] with [v0] on the left.

  @voir pad_left
*)
addenv_impure "'pad_left" (fun_3 __padLeft);;



(**
  @module monoidal
  @name 'matrix_dim
  @type seq_int

  [matrix_dim(s)]  returns the maximal nesting of sequences in [s].

  @voir 'rank 'rank
*)
addenv_impure "'matrix_dim" (fun_1 __matrixDim);;


(**
  @module monoidal
  @name 'rank
  @type seq_seq

  [matrix_rank(s)] returns the rank of the argument. The rank is a sequence of integers
  where element [i] represents the maximal number of elements in dimension [i] when [s] 
  is viewed as a matrix. The size of the rank (i.e. the length of the returned sequence) 
  is the dimension of the matrix as computed by ['matrix_dim].

  Note that if one will interpret the first element in the rank as 
  the number of rows in the matrix, and the second number as the number of columns, then a
  nested sequence like [(1, 2)::(3, 4)::(5,6)::(7,8)::seq:()] must be visualized as:
  {[  1  2]}
  {[  3  4]}
  {[  5  6]}
  {[  7  8]}

  @example
  {[rank((1, 2)::(3, 4)::(5,6)::(7,8)::seq:());;]}
  returns [(4, 2):'seq]. 

  @voir 'matrix_dim 'matrix_dim
*)
addenv_impure "'rank" (fun_1 __rank);;


(**
  @module monoidal
  @name 'matrixify
  @type seq_seq

  [matrixify(s)]  returns a proper matrix build on [s]. The rank of the returned sequence [m] is
  the rank of [s] and [m] is a proper matrix. In the process of building the proper matrix, if a 
  non-sequence element is provided where a sequence is required, then this element is replicated 
  to form a sequence as many times as necessary. If a provided sequence is shorter than a required
  sequence, then the shorter sequence is padded with [<undef>] values.

  @example
  {[matrixify(1::2::(3::4::5::():seq)::6::():seq);;]}
  returns a [4x3] matrix:
  {[((1, 1, 1):'seq, ]}
  {[ (2, 2, 2):'seq,  ]}
  {[ (3, 4, 5):'seq,  ]}
  {[ (6, 6, 6):'seq):'seq ]}
  and
  {[matrixify((1,2)::(3,4,5)::():seq);;]}
  build the [2x3] matrix:
  {[((1, 2, <undef: matrixify>):'seq,]}
  {[ (3, 4, 5):'seq):'seq]}

  
  @voir 'rank 'rank
*)
addenv_impure "'matrixify" (fun_1 matrixify);;


(**
  @module monoidal
  @name 'matrix_check
  @type any_any

  [matrix_check(s)] checks if [s] is a matrix of dimension 2.
  A matrix is a sequence of sequences of equal length. [matrix_check(s)] returns [false]
  if [s] cannot be considered as a 2 dimensionnal matrix, else it returns a sequence of two integers
  [(n, m)] where [n] is the number of elements at the top level sequence and [m] the length of the
  nested sequences. The elements of a 2 dimensionnal matrix are arbitrary and can be themselves sequences. 
  So a 3 dimensional matrix is also 2 dimensional matrix valued by sequences. In consequence, the value
  [(n, m)] returned by [matrix_check(s)] does not coincide with [rank(s)].
  
  @voir 'rank 'rank
*)
addenv_impure "'matrix_check" (fun_1 matrixCheck);;



(**
  @module monoidal
  @name 'transpose
  @type seq_seq

  [transpose(m)] returns the transposition of the 2-dimensionnal matrix [m].
  
  @voir 'matrix_check  'matrix_check
*)
addenv_impure "'transpose" (fun_1 transpose);;




(**
  @module monoidal
  @name 'outerproduct
  @type funct_funct_seq_seq_seq

  [outerproduct(f1, f2, s1, s2)] computes the outer product of matrix or vector [s1] and [s2].
  
*)
addenv_impure "'outerproduct" (fun_4 outerproduct);;


(**
  @module monoidal
  @name 'scalarproduct
  @type seq_seq_seq

  [scalarproduct(s1, s2)] computes the scalar product of two vectors [s1] and [s2].
  [scalarproduct(s1, s2)] is equivalent to  
  {[outerproduct((\x, y.x+y), (\x, y.x*y), s1, s2)]}
  
*)
addenv_impure "'scalarproduct" (fun_2 scalarproduct);;


(**
  @module monoidal
  @name 'matrix_vector_product
  @type seq_seq_seq

  [matrix_vector_product(m, v)] computes the product of matrix [m] with vector [v].
  [matrix_vector_product(m, v)] is equivalent to  
  {[outerproduct((\x, y.x+y), (\x, y.x*y), m, v)]}

  @example
  {['matrix_vector_product((1, 2)::(3, 4)::(5,6)::(7,8)::seq:(), (10, 11));;]}            
  returns 
  [(32, 74, 116, 158):'seq]
  
*)
addenv_impure "'matrix_vector_product" (fun_2 prodMatrixVector);;


(**
  @module monoidal
  @name 'matrix_matrix_product
  @type seq_seq_seq

  [matrix_matrix_product(m1, m2)] computes the product of two matrices.
  [matrix_matrix_product(m1, m2)] is equivalent to  
  {[outerproduct((\x, y.x+y), (\x, y.x*y), m1, m2)]}

  @example
  {[matrix_matrix_product((1, 2)::(3, 4)::(5,6)::seq:(), (10, 11, 12)::(20, 21, 22)::seq:());;]}
  returns 
  [((50, 53, 56):'seq, (110, 117, 124):'seq, (170, 181, 192):'seq):'seq]. 

  In tabular form, we have computed:
  {[   | 1 2 |   | 10  11  12 |   | 50   53  56 |  ]}
  {[   | 3 4 | . | 20  21  22 | = | 110 117 124 |  ]}
  {[   | 5 6 |                    | 170 181 192 |  ]}
  
*)
addenv_impure "'matrix_matrix_product" (fun_2 prodMatrixMatrix);;





(* --- fin de l'installation des fonctions de base --- *)

print_string "done *)";
print_newline();;

(* ======================================================= *)
