(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: ptopo_seq.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG, AS
 * Modified: $Date: 2006-04-16 16:27:45 $
 * Version: $Revision: 1.1 $
 * -------------------------------------------------------------------
 *
 * Status: voir le commentaire general dans topology.ml
 *
 *=========================================================================
 *)

open Global;;
open Valeur;;
open Ptopology;;
open Def_seq;;
open Core_seq;;
open Printer_seq;;

__register_version "$RCSfile: ptopo_seq.ml,v $" "$Revision: 1.1 $" "$Date: 2006-04-16 16:27:45 $";;

type 'position extendedPos =
  | EP_vertex of 'position
  | EP_edge   of 'position * 'position
;;

let string_of_EP strOfPos = function
  | EP_vertex(p)     -> strOfPos p
  | EP_edge  (p1,p2) -> "("^(strOfPos p1)^", "^(strOfPos p2)^")"
;;

class virtual ptopo_seq ((c : 'collection), couleur) mgsval =
object(self)
  inherit [int extendedPos] generic_ptopo 
    (string_of_EP string_of_int)      (** afficheur de position            **)
    (c, couleur)                      (** collection a transformer         **)
    mgsval                            (** l'argument complet               **)
    (EP_vertex 0)                     (** position de dpart: A randomizer **)
(*
  val _avl  = c
  val _avl_usage = (let tbl = Core_achain.create_achain (Core_achain.size c)
		    in
		      Core_achain.iter (fun k v -> Core_achain.add tbl k PosFree) c;
		      tbl)
  val mutable _batch = ([] : (path_type * 'position path * mgs_value) list)
  val mutable _last_index = Global.crandom (Core_achain.size c)


  method common_position fl cfl tfl tcfl =
    match (fl,cfl) with
      | [],[] -> [||]
      | fe::flq,[]  -> (
	  let fs = List.fold_right 
		     (fun f acc -> SAcell.inter (Core_achain.faces_sacell f) acc )
		     flq
		     (Core_achain.faces_sacell fe)
	  in
	  let common_list  = SAcell.fold
			       (fun acell acc -> if (true) then acell::acc else acc)
			       fs
			       []
	  in
	    Array.of_list common_list)
      | [],cfe::cflq  -> (
	  let cfs = List.fold_right 
		      (fun cf acc -> SAcell.union (Core_achain.cofaces_sacell cf) acc )
		      cflq
		      (Core_achain.cofaces_sacell cfe)
	  in
	  let common_list  = SAcell.fold
			       (fun acell acc -> if (true) then acell::acc else acc)
			       cfs
			       []
	  in
	    Array.of_list common_list)
      | fe::flq,cfe::cflq  -> (
	  let fs = List.fold_right 
		     (fun f acc -> SAcell.inter (Core_achain.faces_sacell f) acc )
		     flq
		     (Core_achain.faces_sacell fe)
	  and cfs = List.fold_right 
		      (fun cf acc -> SAcell.union (Core_achain.cofaces_sacell cf) acc )
		      cflq
		      (Core_achain.cofaces_sacell cfe)
	  in
	  let common_sacell = SAcell.inter fs cfs in
	  let common_list  = SAcell.fold
			       (fun acell acc -> if (true) then acell::acc else acc)
			       common_sacell
			       []
	  in
	    Array.of_list common_list)

  method iter_all fct = Core_achain.iter (fun k _ -> (fct k)) c
  method iter_faces fct pos = List.iter
				(fun k -> if (Core_achain.mem c k) then (fct k))
				(Core_achain.cofaces pos)
  method iter_cofaces fct pos = List.iter
				  (fun k -> if (Core_achain.mem c k) then (fct k))
				  (Core_achain.faces pos)

  method isFaceOf p1 p2 = List.mem p1 (Core_achain.faces p2)
  method isCoFaceOf p1 p2 = List.mem p1 (Core_achain.cofaces p2)
	  
  method use_transitive_faces_cofaces s =
    let tf  = Core_achain.transitive_faces s in
    let tcf = Core_achain.transitive_cofaces s in
    let ret = ref [] in
      List.iter (fun f -> try if (self#usable f) then self#use s f ; ret := f::!ret with Not_found -> ()) tf;
      List.iter (fun cf -> try if (self#usable cf) then self#use s cf ; ret := cf::!ret with Not_found -> ()) tcf;
      !ret

  method dimension s = (acelldim s)

  method state s = Core_achain.find _avl_usage s
  method update_state f s = Core_achain.replace _avl_usage s (f (Core_achain.find _avl_usage s))
  method next_usable =
    try
      Hash.iter_from
	_last_index 
	(fun _ (k,v) -> if ((pos_usable v) && (pos_visitable v)) then raise (FoundPos k))
	_avl_usage;
      raise NoMoreUsable
    with FoundPos p -> (assert
			  begin 
			    if !trace_opt then
			      (Printf.printf "       ptopoACHAIN: next_usable: found %s (%s)\n" (string_of_acell p) (string_of_pos (self#state p))); 
			    true
			  end;
			_last_index <- Hash.last_iter_index();
			p)
  method next_filterable =
    try
      Hash.iter_from
	_last_index 
	(fun _ (k,v) -> if ((pos_filterable v) && (pos_visitable v)) then raise (FoundPos k))
	_avl_usage;
      raise NoMoreFilterable
    with FoundPos p -> (assert
			  begin 
			    if !trace_opt then
			      (Printf.printf "       ptopoACHAIN: next_filterable: found %s\n" (string_of_acell p)); 
			    true
			  end;
			_last_index <- Hash.last_iter_index();
			p)
  method unvisit_all =
    Core_achain.iter (fun s v -> if (pos_visited v) then (Core_achain.replace _avl_usage s (pos_unvisit v))) _avl_usage

  method get_value_from_pos p = Core_achain.find _avl p
  method mgs_value_from_pos p = VAL_acell(p)
  method pos_from_mgs_value m = match m with | VAL_acell(p) -> p | _ -> raise (Erreur "ptopoACHAIN: pos_from_mgs_value: wait for a acell")


  method add_match rt p v = _batch <- (rt, p, v)::_batch

  val _copy = Core_achain.create_achain (Core_achain.size c)

  method copie_force pos =
    let d = acelldim pos in
    let npos = Core_achain.new_acell_force
		 d
		 (List.fold_right
		    (fun f acc -> try match Core_achain.find _copy f with None -> acc | Some p -> p::acc with Not_found -> acc)
		    (Core_achain.faces pos)
		    []
		 )
		 (List.fold_right
		    (fun cf acc -> try match Core_achain.find _copy cf with None -> acc | Some p -> p::acc with Not_found -> acc)
		    (Core_achain.cofaces pos)
		    []
		 )
    in
      Core_achain.add _copy pos (Some npos) ;
      Core_achain.set_modifiable_of_acell npos ;
      npos

  method mark_to_copy pos =
    if (not (Core_achain.mem _copy pos)) then (
      Core_achain.add _copy pos None ;
      List.iter self#mark_to_copy (Core_achain.faces pos) ;
      List.iter self#mark_to_copy (Core_achain.cofaces pos))

  method copie pos =
    try
      match Core_achain.find _copy pos with
	| None -> self#copie_force pos
	| Some p -> p
    with Not_found ->
      self#mark_to_copy pos ;
      self#copie_force pos

  method copied pos =
    try
      match Core_achain.find _copy pos with
	| None -> false
	| Some p -> true
    with Not_found -> false

  method marked_to_copy pos = Core_achain.mem _copy pos

  method add_one_instance ret (rt, (p_f,p_u), v) =
    match v with
      | VAL_seq(a,t) when (t == Type_seq.pRhsSpec_desc) -> (
	  let rhs = self#trad a in
	  (* Pour chaque utilis, vrifier qu il n est pas rcrit *)

	    assert (
	      List.for_all
		      (fun x -> match x with
			   (Filt (c,_)) -> not (List.mem c p_u)
			 | (Symb _)     -> true)
		      rhs
	    ) ;

	  (* Pour chaque filtr abs en partie droite, copie_tfcf, assert tfcf dans rhs, suppr l elt de _copy *)

	  List.iter
	    (
	      fun ac ->
		if (List.for_all
		      (fun x -> match x with
			   (Filt (c,_)) -> (c <> ac)
			 | (Symb _)     -> true)
		      rhs)
		then begin
		  let tf  = Core_achain.transitive_faces ac in
		  let tcf = Core_achain.transitive_cofaces ac in
		    List.map self#copie tf ;
		    List.map self#copie tcf ;
		    assert (true (* compliqu  vrifier, mais  faire... *)
		    );
		    Core_achain.remove _copy ac
		end
	    )
	    p_f;


	  (* rajouter des copies des lments filtrs apparaissant dans la partie droite *)

	  List.iter
	    (
	      fun x -> match x with
		| Filt (c,_) -> ignore(self#copie c)
		| Symb (n,(d,lf,lcf),_) -> ()
	    )
	    rhs;

	  (* pour chaque symbole, copie des bords dont on a besoin pour le construire *)
	  (* construire les nouveaux lments depuis la dim la plus petite... ce qui est faux *)

	  let dico_symb = Hash.create 1223 in
	    List.iter
	      (
		fun x -> match x with
		  | Filt (c,v) -> ignore(Core_achain.replace ret (self#copie c) v)
		  | Symb (n,(d,lf,lcf),v) -> (
		      let lf' = List.fold_right
				  (fun x acc -> match x with
				       VAL_symbol s -> (try (Hash.find dico_symb s)::acc with Not_found -> acc)
				     | VAL_acell  c -> (self#copie c)::acc
				     | _ -> raise (Erreur "Imossible !!!! (1)")
				  )
				  lf
				  []
		      in let lcf' = List.fold_right
				      (fun x acc -> match x with
					   VAL_symbol s -> (try (Hash.find dico_symb s)::acc with Not_found -> acc)
					 | VAL_acell  c -> (self#copie c)::acc
					 | _ -> raise (Erreur "Imossible !!!! (2)")
				      )
				      lcf 
				      []
		      in
		      let npos = Core_achain.new_acell_force d lf' lcf' in
			Core_achain.set_modifiable_of_acell npos ;
			Hash.add dico_symb n npos ;
			ignore(Core_achain.replace ret npos v)
		    )
	    )
	    rhs;

	)
      | _            ->
	  if ((List.length p_f) = 1) then (
	    (*List.map self#copie p_u ;*)
	    let npos = self#copie (List.hd p_f) in
	      ignore(Core_achain.add_acell ret npos v)
	  ) else
	    raise (Erreur "mouai, c bof ca !!!")
	      
  method build_result =
    let c' = Core_achain.create_achain (Core_achain.size c) in
      List.iter (self#add_one_instance c') _batch ;
      Core_achain.iter (fun ac v -> if (not (self#filtered ac)) then Core_achain.replace c' (self#copie ac) v) c ;
      VAL_achain (c', couleur)

*)
end

;;
