(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: instancie_pat.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2007-09-12 13:36:14 $
 * Version: $Revision: 1.8 $
 * -------------------------------------------------------------------
 *
 * status :
 *
 *=========================================================================
 *)

open Global;;
open Valeur;;
open Terme;;
open Printer;;
open Dispatch_sig;;

open Topology;;
open Pattern;;


__register_version "$RCSfile: instancie_pat.ml,v $" "$Revision: 1.8 $" "$Date: 2007-09-12 13:36:14 $";;


let predefined_rule_vars = [
  ("'self", 0);
  ("self", 0)
];;


let env_argname = "$rule_env"
;;

let rec prepare_rhs varindex exp = Abs(env_argname, bind_pvar varindex exp) 
 
and indexation pvar =  
  let index = ref (0) 
  in List.fold_left (fun varindex var -> incr index; (var, !index)::varindex) predefined_rule_vars  pvar 
 
and bind_pvar varindex e = 
  let cond s = List.mem_assoc s varindex 
  in
    match e
    with

      (** variable du pattern **)
      | (Id s) when cond s ->
	  op_apply1 env_argname (Cte (VAL_int (List.assoc s varindex)))
	    
      (** fonctions speciales: pos, left, neighbor, etc. qui n'apparaissent que dans une transformation **)
	    
      | App(Id ("pos" | "'pos"), [Id s])  when cond s -> 
	  op_apply1 env_argname (Cte(VAL_int(-(List.assoc s varindex))))
	    
      | App(Id ("neighbors" | "'neighbors"), [Id s]) when cond s ->
	  let pos  = bind_pvar varindex (op_apply1 "pos" (Id s))
	  and self = bind_pvar varindex (Id "'self")
	  in op_apply2 "'neighbors" self pos

      | App(Id ("neighborspos" | "'neighborspos"), [Id s]) when cond s ->
	  let pos  = bind_pvar varindex (op_apply1 "pos" (Id s))
	  and self = bind_pvar varindex (Id "'self")
	  in op_apply2 "'neighborspos" self pos
	       
      | App(Id ("neighbor" | "'neighbor"), [Id s; fin]) when cond s ->
	  let pos  = bind_pvar varindex (op_apply1 "pos" (Id s))
	  and self = bind_pvar varindex (Id "'self")
	  and fin  = bind_pvar varindex fin
	  in op_apply3 "'neighbor" self pos fin
	       
      | App(Id ("neighborsfold" | "'neighborsfold"), [f; z; Id s]) when cond s ->
	  let neighbors = bind_pvar varindex (op_apply1 "'neighbors" (Id s))
	  and f         = bind_pvar varindex f
	  and z         = bind_pvar varindex z
	  in op_apply3 "'fold" f z neighbors
	       
      | App(Id ("facesfold" | "'facesfold"), [f; z; Id s]) when cond s ->
	  let faces = bind_pvar varindex (op_apply1 "'faces" (op_apply1 "pos" (Id s)))
	  and f     = bind_pvar varindex f
	  and z     = bind_pvar varindex z
	  in op_apply3 "'fold" f z faces
	       
      | App(Id ("cofacesfold" | "'cofacesfold"), [f; z; Id s]) when cond s ->
	  let cofaces = bind_pvar varindex (op_apply1 "'cofaces" (op_apply1 "pos" (Id s)))
	  and f       = bind_pvar varindex f
	  and z       = bind_pvar varindex z
	  in op_apply3 "'fold" f z cofaces
	       
      | App(Id ("left" | "'left"), [Id s]) when cond s ->
	  let pos  = bind_pvar varindex (op_apply1 "pos" (Id s))
	  and self = bind_pvar varindex (Id "'self")
	  in op_apply2 "'left" self pos
	       
      | App(Id ("right" | "'right"), [Id s]) when cond s -> 
	  let pos  = bind_pvar varindex (op_apply1 "pos" (Id s))
	  and self = bind_pvar varindex (Id "'self")
	  in op_apply2 "'right" self pos
	       
      | App(Id ("rightq" | "'rightq"), [Id s]) when cond s ->
	  let pos  = bind_pvar varindex (op_apply1 "pos" (Id s))
	  and self = bind_pvar varindex (Id "'self")
	  in op_apply2 "'!=" (op_apply2 "'right" self pos) (Cte (VAL_undef ""))
	       
      | App(Id ("leftq" | "'leftq"), [Id s]) when cond s ->
	  let pos  = bind_pvar varindex (op_apply1 "pos" (Id s))
	  and self = bind_pvar varindex (Id "'self")
	  in op_apply2 "'!=" (op_apply2 "'left" self pos) (Cte (VAL_undef ""))
	       
      (** Propagation du traitement **)
	       
      | App(f, a)   -> App(bind_pvar varindex f, List.map (bind_pvar varindex) a) 
      | Abs (s, t)  -> Abs(s, bind_pvar (List.remove_assoc s varindex) t) 
      | If(a, b, c) -> If(bind_pvar varindex a, bind_pvar varindex b, bind_pvar varindex c) 
      | Fby(a, b)   -> Fby(bind_pvar varindex a, bind_pvar varindex b) 
      | Set(s, t)   -> Set(s, bind_pvar varindex t) 
      | x -> x 
;;

let compile_rhs_rule x = (if !Global.trace_input then Compile.traduit_evalue else Compile.eval)(x) 
;;


let rec print_string_list = function
    [] -> ""
  | t::q -> t^" "^(print_string_list q)

(* O. : Et pourquoi le fold_named_pattern n'a-t-il pas ete utilise ? *)

and string_of_named_pattern = function
  | PA_BlankPat(ls)               -> "_ as [ "^(print_string_list ls)^"]"
  | PA_CtePat(ls, _, 1, _)        -> "cte as [ "^(print_string_list ls)^"]"
  | PA_CtePat(ls, _, _, pp)       -> "cteP (== "^(string_of_named_pattern pp)^") as [ "^(print_string_list ls)^"]"
  | PA_CommaPat(ls, _, _, p1, p2) -> "("^(string_of_named_pattern p1)^", "^(string_of_named_pattern p2)^") as [ "^(print_string_list ls)^"]"
  | PA_DownPat(ls, p1, p2)        -> "("^(string_of_named_pattern p1)^"\\/ "^(string_of_named_pattern p2)^") as [ "^(print_string_list ls)^"]"
  | PA_GuardedPat(ls, p, _)       -> "("^(string_of_named_pattern p)^") / guard as [ "^(print_string_list ls)^"]"
  | PA_OrPat(ls, p1, p2)          -> "(("^(string_of_named_pattern p1)^") | ("^(string_of_named_pattern p2)^")) as [ "^(print_string_list ls)^"]"
  | PA_RecordPat(ls, _)           -> "record as [ "^(print_string_list ls)^"]"
  | PA_IterPat(ls, _, p)          -> "("^(string_of_named_pattern p)^")* as [ "^(print_string_list ls)^"]"
  | PA_NotPat(ls, p)              -> "~ ("^(string_of_named_pattern p)^") as [ "^(print_string_list ls)^"]"

let rec namepattern pat = namepat false false [] pat

(** namepat traduit un pattern_term en pattern_term_aux qui est un pattern-term nomm.
 **    explicit: un booleen qui dit que je dois donner explicitement un nom
 **    ls      : une liste de noms pour le pattern
 **
 ** Un GuardedPat n'a pas de nom: il est propag au fils. L'ide est de donner un nom 
 ** qu'aux patterns qui filtrent vraiment quelque chose.
 **)
and namepat tilded explicit ls = 
  let newname () = if explicit && (ls = []) then [new_name "$patvar"] else ls
  in function 
    | NotYetPat               -> raise (Erreur "Error: Sorry: pattern not yet implemented")
    | AsPat(s, p)             -> namepat tilded true (s::ls) p
    | BlankPat                -> (
	if tilded then
	  PA_NotPat([],PA_BlankPat(newname ()))
	else
	  PA_BlankPat(newname ()) 
      )
    | CtePat (v, 1)           -> (
	if tilded then
	  PA_NotPat([],PA_CtePat(newname(), v, 1, PA_BlankPat(newname())))
	else
	  PA_CtePat(newname(), v, 1, PA_BlankPat(newname())) (* PA_BlankPat(newname()) est un dummy *)
      )
    | CtePat (v, n)           -> (
	let s = newname() in
	let pat = Auxfct.int_fold
	  (fun _ p -> CommaPat(AllDirections, None, (CtePat(v, 1)), p))
	  (CtePat(v, 1))
	  (n-1)
	in
	  if tilded then
	    namepat tilded explicit s pat
	  else
	    PA_CtePat(s, v, n, namepat tilded explicit s pat)
      )
    | RecordPat l             -> (
	if tilded then
	  PA_NotPat([], PA_RecordPat(newname(), l))
	else
	  PA_RecordPat(newname(), l)
      )
    | GuardedPat(p, t)        -> PA_GuardedPat([], namepat tilded explicit ls p, t)
    | OrPat(p1, p2)           -> PA_OrPat(newname(), namepat tilded explicit [] p1, namepat tilded explicit [] p2)
    | IterPat (ki, p)         -> PA_IterPat(newname(), ki, namepat tilded explicit [] p)
    | CommaPat(d, dn, p1, p2) -> PA_CommaPat(newname(), d, dn, namepat tilded explicit [] p1, namepat tilded explicit [] p2)
    | DownPat(p1, p2)         -> PA_DownPat(newname(), namepat tilded true [] p1, namepat false explicit [] p2)
    | NotPat(p)               -> namepat true true ls p
;;


let rec pattern_var = function
  | PA_BlankPat s          -> s
  | PA_CtePat(s, _, _, p)  -> without_double (s@(pattern_var p))
  | PA_CommaPat (s, _, sd, p1, p2) ->  (if sd <> None then (from_option sd)::s else s)@(pattern_var p1)@(pattern_var p2)
  | PA_DownPat (s, p1, p2) -> s@(pattern_var p1)@(pattern_var p2)
  | PA_GuardedPat(s, p, _) -> s@(pattern_var p)
  | PA_OrPat(s, p1, p2)    -> s@(pattern_var p1)@(pattern_var p2)
  | PA_RecordPat(s, l)     -> List.fold_left (fun acc -> function (None, _) -> acc | (Some id, _) -> id::acc) s l
  | PA_IterPat(s, _, p)    -> s@(pattern_var p)
  | PA_NotPat(s, p)        -> s@(pattern_var p)
;;


let rec one_pat_var = function
  | PA_BlankPat(s::_)             -> s
  | PA_CtePat(s::_, _, _, _)      -> s
  | PA_CommaPat(s::_, _, _, _, _) -> s
  | PA_DownPat(s::_, _, _)        -> s
  | PA_GuardedPat(_, p, _)        -> one_pat_var p
  | PA_OrPat(s::_, _, _)          -> s
  | PA_RecordPat(s::_, _)         -> s
  | PA_IterPat(s::_, _, _)        -> s
  | PA_NotPat(_, p)               -> one_pat_var p
  | x -> assert(false); 
      raise (Erreur "Error: internal: pattren with no name is asked for a name (one_pat_var)")
;;

(* XXX check a faire avant/apres instantiation des patterns:
 *
 * - pas de variable definie 2 fois : x, x  ou alors transformer en : x, (tmp123 / x == tmp123)
 *
 * - verifier que les variables d'un or pattern sont bien definies:
 *   par exemple: (1 as x) | (2 as y) => x+y est errone. On peut demander que seules 
 *   les variables qui apparaissent dans les 2 branches du or soient utilisables a droite
 *   de la fleche. Mais attention, considerez le pattern: (x/ x > 2) | (x, y / y > 3) => x
 *
 * - motif du type (x / y > 3), y est interdit a cause de la ref. en avant. Ou bien la 
 *   premiere occurence de y refere a un y global, mais pas celui de la regle.
 *
 * - motif x* as X / x>4 interdit car le x sous le * est local.
 *
 *)

(** Transforme un named-pattern-term (cf. term.ml) en pattern-objet (cf. matching.ml) 
 **)

let rec instancie_pattern_gen nbvar varindex =
  let indexvar l = List.map (fun x -> List.assoc x varindex) l 
  and patvar = fst (List.split varindex)
  in function (cont : 'a Topology.pattern) -> function pat ->  match pat with
    | PA_BlankPat s -> new blank_pattern (indexvar s) cont
    | PA_CtePat(s, v, 1, _) -> new cte_pattern1 (indexvar s) v cont

    | PA_CtePat(s, v, n, altpat) -> (let alternatif = instancie_pattern_gen nbvar varindex cont altpat
				     in
				       new cte_patternN (indexvar s) v n alternatif cont)

    | PA_RecordPat(s, l) -> new record_pattern (indexvar s) l varindex cont
	
    | PA_OrPat(s, p1, p2) as p ->
	(new or_pattern (indexvar s) (is_elementary_path p) 
	   p1 p2 
	   (instancie_pattern_gen nbvar varindex (new end_pattern) p1) 
	   (instancie_pattern_gen nbvar varindex (new end_pattern) p2) 
	   cont)
	
    | PA_CommaPat(s, dir, dirname, p1, p2) ->
	let suite = (if s = [] 
		     then cont
		     else (new update_pattern 
			     (indexvar s)
			     (List.assoc (one_pat_var p1) varindex)
			     (List.assoc (one_pat_var p2) varindex)
			     cont))
	and index = if (dirname = None) then -1 else List.assoc (from_option dirname) varindex in
	let fin   = instancie_pattern_gen nbvar varindex suite p2 in
	let comma = new comma_pattern (indexvar s) dir dirname index fin 
	in instancie_pattern_gen nbvar varindex comma p1

    | PA_DownPat(s, p1, p2) -> (
        (*Si p1 n'est pas un elementary Path, et pour que le down         *)
	(*soit possible, ce ne peut-etre que un (x* as l) ; il n'y a alors*)
	(*aucune chance de pouvoir filtrer quoi que ce soit dans l car le *)
	(*filtrage de l aura consomme *tous* les elements de l.           *)
	
	let pvarD     = pattern_var p2 in        (* On obtient les variables dfinies dans le sous-filtre      *)
	let patvarD   = indexvar pvarD in        (* On obtient les index de ces variables dans l'environnement *)

	let pat1 = instancie_pattern_gen nbvar varindex (new end_pattern) p1
	and pat2 = instancie_pattern_gen nbvar varindex (new end_pattern) p2 in

	  new down_pattern

	    (indexvar s)
	    (pat1#get_patvar)
	    (List.hd (pat1#get_patvar))

	    (is_elementary_path p1)
	   p1 p2                                                      
	   pat1 pat2
	   ({cm_motif = pat2; cm_nbvar = nbvar; cm_patvar = patvarD})

	   (function
	      | (VAL_rec c) as mgsval -> (
		  let (topoD : < apply_one_matching : ((('a pattern) compiled_motif)  -> ('a path) option); .. >)
		    = Obj.magic (new Topo_rec.topo_record c mgsval)
		  in
		    topoD
		)

	      | (VAL_seq c) as mgsval -> (
		  let (topoD : < apply_one_matching : ((('a pattern) compiled_motif)  -> ('a path) option); .. >)
		    = Obj.magic (new Topo_seq.topo_sequence c mgsval)
		  in
		    topoD
		)

	      | (VAL_gbf c) as mgsval -> (
		  let (topoD : < apply_one_matching : ((('a pattern) compiled_motif)  -> ('a path) option); .. >)
		    = Obj.magic (new Topo_gbf.topo_gbf c mgsval)
		  in
		    topoD
		)		  

	      | (VAL_set c) as mgsval -> (
		  let (topoD : < apply_one_matching : ((('a pattern) compiled_motif)  -> ('a path) option); .. >)
		    = Obj.magic (new Topo_set.topo_set c mgsval (Core_set.cardinal (fst c)))
		  in
		    topoD
		)		  

	      | (VAL_bag c) as mgsval -> (
		  let n = (Core_bag.cardinal_b (fst c)) in
		  let (topoD : < apply_one_matching : ((('a pattern) compiled_motif)  -> ('a path) option); .. >)
		    = Obj.magic (new Topo_bag.topo_bag c mgsval n)
		  in
		    topoD
		)		  
(*
	      | (VAL_gbf c) as mgsval -> (
		  let (topoD : < apply_one_matching : ((('a pattern) compiled_motif)  -> ('a path) option); .. >)
		    = Obj.magic (new Topo_gbf.topo_gbf c mgsval)  in
		  let pvarD     = pattern_var p2 in 
		  let varindexD = indexation pvarD in
		  let nbvarD    = List.length pvarD in
		  let motifD    = instancie_pattern_gen nbvarD varindexD (new end_pattern) p2
		  in
		    topoD, {cm_motif = motifD; cm_nbvar = nbvarD}
		)
		  
	      | (VAL_grf c) as mgsval -> (
		  let (topoD : < apply_one_matching : ((('a pattern) compiled_motif)  -> ('a path) option); .. >)
		    = Obj.magic (new Topo_graph.topo_grf c mgsval) in
		  let pvarD     = pattern_var p2 in 
		  let varindexD = indexation pvarD in
		  let nbvarD    = List.length pvarD in
		  let motifD    = instancie_pattern_gen nbvarD varindexD (new end_pattern) p2
		  in
		    topoD, {cm_motif = motifD; cm_nbvar = nbvarD}
		)
		  
	      | (VAL_prox c) as mgsval -> (
		  let (topoD : < apply_one_matching : ((('a pattern) compiled_motif)  -> ('a path) option); .. >)
		    = Obj.magic (new Topo_prox.topo_prox c mgsval) in
		  let pvarD     = pattern_var p2 in 
		  let varindexD = indexation pvarD in
		  let nbvarD    = List.length pvarD in
		  let motifD    = instancie_pattern_gen nbvarD varindexD (new end_pattern) p2
		  in
		    topoD, {cm_motif = motifD; cm_nbvar = nbvarD}
		)

	      | (VAL_del c) as mgsval -> (
		  let (topoD : < apply_one_matching : ((('a pattern) compiled_motif)  -> ('a path) option); .. >)
		    = Obj.magic (new Topo_del.topo_del c mgsval) in
		  let pvarD     = pattern_var p2 in 
		  let varindexD = indexation pvarD in
		  let nbvarD    = List.length pvarD in
		  let motifD    = instancie_pattern_gen nbvarD varindexD (new end_pattern) p2
		  in	  
		    topoD, {cm_motif = motifD; cm_nbvar = nbvarD}
		)
		  
	      | (VAL_qmf c) as mgsval -> (
		  let (topoD : < apply_one_matching : ((('a pattern) compiled_motif)  -> ('a path) option); .. >)
		    = Obj.magic (new Topo_qmf.topo_qmf c mgsval 0 1) in
		  let pvarD     = pattern_var p2 in 
		  let varindexD = indexation pvarD in
		  let nbvarD    = List.length pvarD in
		  let motifD    = instancie_pattern_gen nbvarD varindexD (new end_pattern) p2
		  in
		    topoD, {cm_motif = motifD; cm_nbvar = nbvarD}
		)
		  
	      | (VAL_achain c) as mgsval -> (
		  let (topoD : < apply_one_matching : ((('a pattern) compiled_motif)  -> ('a path) option); .. >)
		    = Obj.magic (new Topo_achain.topo_achain c mgsval 0 1) in
		  let pvarD     = pattern_var p2 in 
		  let varindexD = indexation pvarD in
		  let nbvarD    = List.length pvarD in
		  let motifD    = instancie_pattern_gen nbvarD varindexD (new end_pattern) p2
		  in
		    topoD, {cm_motif = motifD; cm_nbvar = nbvarD}
		)
*)	  
	      | _       -> raise (Erreur "Instancie_pat: instancie_pattern_gen: not yet implemented downpat on other collection than aggregate types "))
	   cont
	)

    | PA_GuardedPat(s, p, exp) ->
	let fin = new guarded_pattern 
		    (indexvar s) 
		    (indexvar (Terme.appear_list patvar exp))
		    exp 
		    (compile_rhs_rule(prepare_rhs varindex exp)) 
		    cont
	in instancie_pattern_gen nbvar varindex fin p
	     
    | PA_IterPat(s, kiter, p) ->
	let iterpat = instancie_pattern_gen nbvar varindex (new end_pattern) p
	in new iter_pattern (indexvar s) kiter iterpat cont

    | PA_NotPat(s, p) ->
	new not_pattern (indexvar s) (instancie_pattern_gen nbvar varindex cont p)
;;
