(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: instancie_pat_ac.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2008-06-03 12:59:19 $
 * Version: $Revision: 1.6 $
 * -------------------------------------------------------------------
 *
 * status :
 *
 *=========================================================================
 *)

open Global;;
open Valeur;;
open Type_rec;;
open Terme;;
open Printer;;
open Printer_term;;
open Pattern;;
open Pattern_ac;;
open Instancie_pat;;

__register_version "$RCSfile: instancie_pat_ac.ml,v $" "$Revision: 1.6 $" "$Date: 2008-06-03 12:59:19 $";;

(*** pour savoir si on peut faire l optimisation... ***)

let rec is_optimizable_ac = 
  function
    | PA_BlankPat s         -> true
    | PA_CtePat(s, v, 1, _) -> true
    | PA_CtePat(s, v, _, _) -> true
    | PA_OrPat(s, _, _)     -> true
    | PA_RecordPat(s, l)    -> false (* Peut-etre a prendre en compte un jour... *)
    | PA_CommaPat(s, dir, dirname, p1, p2) -> (if dir <> AllDirections
					       then false
					       else ((is_optimizable_ac p1)&&(is_optimizable_ac p2)))
    | PA_GuardedPat(s, p, exp) -> is_optimizable_ac p
    | PA_IterPat(s, kiter, p) -> true
    | _ -> false
;;

(*** applatissement des pattern ***)

type flat_pattern =
  | FP_BlankPat
  | FP_CtePat of mgs_value
  | FP_CtePatN of mgs_value * int * ((int list * flat_pattern * int ref) list)
  | FP_RecordPat of fields_pattern_specification * ((string * int) list)
  | FP_GuardedPat of (int list) * term * mgs_value
  | FP_UpdatePat of int * int
  | FP_IterPat of (int list) * pat_iteration_kind * ((int list * flat_pattern * int ref) list)
;;

(*** fonctions pour l'affichage des flat_pattern ***)

let string_of_list f l =
  let rec _string_of_list l = match l with
    | [] -> ""
    | [a] -> f a
    | t::q -> (f t)^";"^(_string_of_list q)
  in
    "["^(_string_of_list l)^"]"
;;

let rec string_of_flatPat fp = match fp with
  | FP_BlankPat -> "_ "
  | FP_CtePat v -> (to_string(Printer.prsvalue v))^" "
  | FP_CtePatN (v,n,p) -> "#"^(string_of_int n)^" "^(to_string(Printer.prsvalue v))^" alt:"^(string_of_list string_of_trip p)^" "
  | FP_RecordPat(l,_) -> (to_string (Printer_term.prspat (RecordPat l)))^" "
  | FP_GuardedPat (dep,exp,guard) -> "//"^(to_string(Printer_term.prsterm exp))^"// "
  | FP_UpdatePat(v1,v2) -> "["^(string_of_int v1)^","^(string_of_int v2)^"] "
  | FP_IterPat(dep,kiter,iterpat) ->
    (
      let kind =
        match kiter with
          | PatStar _ -> "*"
	  | PatPlus _ -> "+"
      in
        "("^(string_of_list string_of_trip iterpat)^")"^"||"^(string_of_list string_of_int dep)^"||"^kind
    )

and string_of_trip (varDef, fp, pred) =
	"< "^(string_of_flatPat fp)^(string_of_list string_of_int varDef)^" "^(string_of_int !pred)^" >"
;;

(*** calcul de la liste de dpendance d'un iterpat ***)

let rec no_double acc = function
	| [] -> acc
	| t::q -> if List.mem t acc then
	            no_double acc q
		  else
		    no_double (t::acc) q
;;

let dep_iterpat iterpat =
  let rec dep__iterpat (idxvars, dep) = function
    | [] ->  (idxvars, dep) 
    | (indexvar,t,ref)::q ->
    (
      let new_idxvars = idxvars@indexvar
      in
        match t with
          | FP_BlankPat  -> dep__iterpat (new_idxvars,dep) q
	  | FP_CtePat _  -> dep__iterpat (new_idxvars,dep) q
	  | FP_CtePatN _ -> dep__iterpat (new_idxvars,dep) q
	  | FP_RecordPat _ -> dep__iterpat (new_idxvars,dep) q
	  | FP_GuardedPat(gdep,_,_) -> dep__iterpat (new_idxvars, dep@(List.filter (fun x -> not (List.mem x new_idxvars)) gdep)) q
	  | FP_UpdatePat(v1,v2) -> dep__iterpat (new_idxvars, dep@(List.filter (fun x -> not (List.mem x new_idxvars)) ([v1;v2]))) q
	  | FP_IterPat(idep,_,_) -> dep__iterpat (new_idxvars, dep@(List.filter (fun x -> not (List.mem x new_idxvars)) idep)) q
    )
  in
  let (_,dep) = dep__iterpat ([],[]) iterpat
  in
    no_double [] dep
;;

(*** suppression des or_pattern ***)
let rec remove_OrPat orvar = function
  | PA_BlankPat s                -> [PA_BlankPat (s@orvar)]
  | PA_CtePat (s, v, n, altpat)  -> [PA_CtePat (s@orvar, v, n, altpat)]
  | PA_RecordPat (s, l)          -> [PA_RecordPat (s@orvar, l)]

  | PA_NotPat (s, p)             -> (List.map
				       (function p -> PA_NotPat (s@orvar, p))
				       (remove_OrPat [] p)
				    )
  | PA_GuardedPat (s, p, exp)    -> (List.map
				       (function p -> PA_GuardedPat (s@orvar, p, exp))
				       (remove_OrPat [] p)
				    )
  | PA_IterPat (s, kiter, p)     -> (List.map
				       (function p -> PA_IterPat (s@orvar, kiter, p))
				       (remove_OrPat [] p)
				    )
  | PA_DownPat (s, p1, p2)       -> (let pp1 = remove_OrPat [] p1
				     and pp2 = remove_OrPat [] p2
				     in
				       List.flatten (List.map
						       (function ppp1 -> (List.map
									    (function ppp2 ->
									       PA_DownPat (s@orvar, ppp1, ppp2) 
									    )
									    pp2
									 )
						       )
						       pp1
						    )
				    )
  | PA_CommaPat (s, dir, dirname, p1, p2) -> (let pp1 = remove_OrPat [] p1
					      and pp2 = remove_OrPat [] p2
					      in
						List.flatten (List.map
								(function p1 -> (List.map
										   (function p2 ->
										      PA_CommaPat (s@orvar, dir, dirname, p1, p2)
										   )
										   pp2
										)
								)
								pp1
							     )
					     )


  | PA_OrPat (s, p1, p2) -> (remove_OrPat (s@orvar) p1)@(remove_OrPat (s@orvar) p2)

;;


(*** applatissement des pattern ***)

let rec flat_named_pat varindex =
  let indexvar l = List.map (fun x -> List.assoc x varindex) l
  and patvar     = fst (List.split varindex)
  in function cont -> function
    | PA_BlankPat s         -> ((indexvar s), FP_BlankPat, ref 0)::cont
    | PA_CtePat(s, v, 1, _) -> ((indexvar s), FP_CtePat v, ref 0)::cont
    | PA_CtePat(s, v, n, altpat) -> (
	let alternatif = flat_named_pat varindex [] altpat in
	  ((indexvar s), FP_CtePatN (v,n,alternatif), ref 0)::cont
      )
    | PA_RecordPat(s, l)    -> ((indexvar s), FP_RecordPat (l, varindex), ref 0)::cont
	
    | PA_GuardedPat(s, p, exp) ->
	let fin = ((indexvar s), FP_GuardedPat(
		     indexvar (Terme.appear_list patvar exp),
		     exp,
		     (compile_rhs_rule(prepare_rhs varindex exp))),
	           ref 0
                  )::cont
	in flat_named_pat varindex fin p
	     
    | PA_CommaPat(s, dir, dirname, p1, p2) ->
        let suite = (if s = []
		     then cont
		     else 
		       (
		         ((indexvar s),FP_UpdatePat( 
			    (List.assoc (one_pat_var p1) varindex),
			    (List.assoc (one_pat_var p2) varindex)),
			  ref 0
			 )::cont
		       )
	            ) in
	let fin   = flat_named_pat varindex suite p2 
	in
	  flat_named_pat varindex fin p1

    | PA_IterPat(s, kiter, p) ->
    	let iterpat = flat_named_pat varindex [] p in
	((indexvar s),FP_IterPat(
	   dep_iterpat iterpat,
	   kiter,
	   iterpat),
	 ref 0
	)::cont
    | _ -> []
;;

(*** recherche de qui definit quoi ? ***)

let rec defines pat x = match pat with
	| []                  -> raise Not_found
	| ((l, _, _) as p)::q -> if (List.mem x l) then p else defines q x
;;


(*** tri topologique ***)

(* Liste des contraintes... *)
let selfDep pat p = match p with
  | defList, FP_GuardedPat(varUsed, exp, guard), pred -> (
      List.concat (List.map ( fun x ->
				try (
				  let def = defines pat x in
				    pred := (!pred + 1); [def, p]
				) with Not_found -> []
			    ) varUsed)
    )
  | defList, FP_UpdatePat(v1, v2), pred -> (
      List.concat (List.map ( fun x ->
				try (
				  let def = defines pat x in
				    pred := (!pred + 1); [def, p]
				) with Not_found -> []
			    ) [v1;v2])
    )
  | defList, FP_RecordPat _, pred -> []
      (*
	List.concat (List.map ( fun x ->
	try
	(
	let def = pat#defines x in
	p#incPred; [pat#defines x, p]
	)
	with Not_found -> []
	) p#varUsed)
	*)
  | defList, FP_IterPat(varUsed, _, _), pred ->	(
      List.concat (List.map ( fun x ->
				try (
				  let def = defines pat x in
				    pred := (!pred + 1); [def, p]
				) with Not_found -> []
			    ) varUsed)
    )
  | _ -> []
;;

let cmpPat pat1 pat2 = match (pat1, pat2) with
  | (_,FP_RecordPat _,_), _           -> []
  | _, (_,FP_RecordPat _,_)           -> []
  | (_,FP_UpdatePat _,_), _              -> []
  | _, (_,FP_UpdatePat _,_)              -> []

  | (_,FP_GuardedPat(varUsed1,_,_),pred1),(_,FP_GuardedPat(varUsed2,_,_),pred2) ->
      if ( (List.length varUsed1) < (List.length varUsed2) )
      then ( pred2 := !pred2 + 1; [(pat1,pat2)] )
      else
	if ( (List.length varUsed1) > (List.length varUsed2) )
	then ( pred1 := !pred1 + 1; [(pat2,pat1)] )
	else []

  | (_,FP_GuardedPat _,_), _         -> []
  | _, (_,FP_GuardedPat _,_)         -> []

  | (_,FP_CtePatN(v1,n1,_),pred1),(_,FP_CtePatN(v2,n2,_),pred2) ->
      if (n1 < n2)
      then ( pred1 := !pred1 + 1; [(pat2,pat1)] )
      else
	if (n2 < n1)
	then ( pred2 := !pred2 + 1; [(pat1,pat2)] )
	else []
  | (_,FP_CtePatN _,_), (_,_,pred)      -> pred := !pred + 1; [(pat1,pat2)]
  | (_,_,pred), (_,FP_CtePatN _,_)      -> pred := !pred + 1; [(pat2,pat1)]

  | (_,FP_CtePat _,_),(_,FP_CtePat _,_) -> []
  | (_,FP_CtePat _,_), (_,_,pred)       -> pred := !pred + 1; [(pat1,pat2)]
  | (_,_,pred), (_,FP_CtePat _,_)       -> pred := !pred + 1; [(pat2,pat1)]




(*	| PO_RecordPat, PO_RecordPat -> 
	if ( (List.length pat1#varUsed) < (List.length pat2#varUsed) )
	                                  then ( pat2#incPred; [(pat1,pat2)] )
					  else if ( (List.length pat1#varUsed) > (List.length pat2#varUsed) )
					       then ( pat1#incPred; [(pat2,pat1)] )
					       else []
	| PO_RecordPat, _           -> pat2#incPred; [(pat1,pat2)]
	| _, PO_RecordPat           -> pat1#incPred; [(pat2,pat1)]
*)

  | (varDef1,FP_BlankPat,pred1), (varDef2,FP_BlankPat,pred2) ->
      (
	match ( (List.length varDef1), (List.length varDef2) ) with
	  | 0,0 -> []
	  | 0,_ -> pred1 := !pred1 + 1; [pat2,pat1]
	  | _,0 -> pred2 := !pred2 + 1; [pat1,pat2]
	  | _ -> []
      )
  | (_,FP_BlankPat,_), (_,_,pred) -> pred := !pred + 1; [(pat1,pat2)]
  | (_,_,pred), (_,FP_BlankPat,_) -> pred := !pred + 1; [(pat2,pat1)]

  | (_,FP_IterPat(varUsed1,_,_),pred1), (_,FP_IterPat(varUsed2,_,_),pred2) ->
      if ( (List.length varUsed1) < (List.length varUsed2) )
      then ( pred2 := !pred2 + 1; [(pat1,pat2)] )
      else
	if ( (List.length varUsed1) > (List.length varUsed2) )
	then ( pred1 := !pred1 + 1; [(pat2,pat1)] )
	else []
(*	| PO_IterPat, PO_IterPat   -> []
	| PO_IterPat, _            -> pat1#incPred; [(pat2,pat1)]
	| _, PO_IterPat            -> pat2#incPred; [(pat1,pat2)]*)
;;

let rec cmp p pat = match pat with
	| [] -> []
	| t::q -> (cmpPat p t)@(cmp p q)
;;

let allDep pat =
  let rec allDep_ p pat = match p with
  	| []   -> []
	| t::q -> (selfDep pat t)@(cmp t q)@(allDep_ q pat)
    in
    allDep_ pat pat
;;

(*********** tri topologique ****************)

let rec withoutPred pat =
	match pat with
	| [] -> []
	| t::q -> let (_,_,pred) = t in if (!pred = 0) then t::(withoutPred q) else (withoutPred q)
;;

let rec chooseZero zeros = match zeros with
	| [] -> raise (Erreur "chooseZero on empty list")
	| [e] -> e,[]
	| ((_, tType, _) as t)::q ->
	(
	  let (((_, eType, _) as e), n) = chooseZero q 
	  in
	    match (tType,eType) with
		
	      | FP_GuardedPat _, _     -> t,q
	      | _, FP_GuardedPat _     -> e,t::n
		  
	      | FP_UpdatePat _, _      -> t,q
	      | _, FP_UpdatePat _      -> e,t::n
		  
	      | FP_CtePatN _, _        -> t,q
	      | _, FP_CtePatN _        -> e,t::n
		  
	      | FP_CtePat _, _         -> t,q
	      | _, FP_CtePat _         -> e,t::n
		  
	      | FP_RecordPat _, _      -> t,q
	      | _, FP_RecordPat _      -> e,t::n
		  
	      | FP_BlankPat, _         -> t,q
	      | _, FP_BlankPat         -> e,t::n
		  
	      | FP_IterPat _, _        -> t,q
	      (*| _, FP_IterPat _        -> e,t::n  (*match case unused !!! *)*)
	)
;;

let topoSort pat l =
	let decPred (_,_,pred) = (pred := !pred - 1) in
	let getNbPred (_,_,pred) = !pred in
	let rec topoSort_ listeCouples acc zeros = match zeros with
	| [] -> acc
	| _  ->
	(
	  	let (zero, q)    = chooseZero zeros in
	  	let (dec, new_l) = List.partition (fun (x,y) -> (Pervasives.compare x zero) = 0) listeCouples in
	  	let new_dec      = List.fold_right (fun (x,y) acc -> decPred y; if ((getNbPred y)=0) then (y::acc) else acc) dec [] 
		in
		  topoSort_ new_l (acc@[zero]) (new_dec@q)
	)
	in
	topoSort_ l [] (withoutPred pat)
;;

let rec sort pat =
  let rec searchIterPat = function
    | [] -> []
    | (a,t,b)::q ->
    (
      match t with
        | FP_IterPat(dep,kiter,iterpat) ->
          (a, FP_IterPat(dep,kiter,sort iterpat), b)::(searchIterPat q)
        | _ -> (a,t,b)::(searchIterPat q)
    )
  in
    let newPat = searchIterPat pat in
    let dep = allDep newPat in
      assert begin if !trace_opt then
	(print_string ("liste des contraintes :\n"^
			 (string_of_list ( fun (x,y) ->
					     "\n" ^ (string_of_trip x) ^ " est avant " ^ (string_of_trip y) ^ "\n"
					 )
			    dep
			 )
		      ); print_newline(); flush stdout);
	true
      end;
      topoSort newPat dep
;;


(*** transformation du flat_pattern en 'position ???_pattern_ac ***)

let rec object_of_flatPat fin fpl =
	let rec object_of_flatPat_ fpl = match fpl with
	  | [] -> fin, false
	  | (patvar,flatPat,_)::q ->
	  (
	  	let (next,need) = object_of_flatPat_ q in
		match flatPat with
		  | FP_BlankPat                   -> new blank_pattern_ac patvar next need, true
		  | FP_CtePat v                   -> new cte_pattern_ac patvar v next need, false
		  | FP_CtePatN (v,n,altpat)       -> (
		      let alternatif, need2 = object_of_flatPat next altpat
		      in
			assert(not need2);
			new cteN_pattern_ac patvar v n alternatif next need, false
		    )
		  | FP_RecordPat(l,varindex)      -> new record_pattern_ac patvar l varindex next need, true
		  | FP_GuardedPat (dep,exp,guard) -> new guarded_pattern_ac patvar dep exp guard next need, false
		  | FP_UpdatePat(v1,v2)           -> new update_pattern_ac patvar v1 v2 next need, false
                  | FP_IterPat(dep,kiter,iterpat) ->
                  (
                    let iterpat_obj, need2 = object_of_flatPat (new end_pattern) iterpat
                    in
                    new iter_pattern_ac patvar kiter iterpat_obj next need need2, false
                  )
	  )
	in (object_of_flatPat_ fpl)
;;

(*** instanciation des motifs pour les ensembles et multiensembles... ***)

let opt_ac = ref true;;

let instancie_pattern_ac nbvar varindex cont pat =
	if (is_optimizable_ac pat)
	then (
	  let newres = flat_named_pat varindex [] pat in

	    assert begin if !trace_opt then 
	      (print_string ("flat pat : "^(string_of_list string_of_trip newres)); print_newline(); flush stdout);
	      true
	    end;
	   
	    let listeTrie = sort newres in

	      assert begin if !trace_opt then 
		(print_string ("flat pat  tri : "^(string_of_list string_of_trip listeTrie)); print_newline(); flush stdout);
		true
	      end;

	      let ans = fst(object_of_flatPat cont listeTrie) in

		assert begin if !trace_opt then 
		  (print_string ("'position ???_pattern_ac : "^ans#to_full_string); print_newline(); flush stdout);
		  true
		end;
	   
		ans
	)
	else instancie_pattern_gen nbvar varindex cont pat
;;















(*	   let dep = allDep newres in

	   assert begin if !trace_opt then 
	     (print_string ("liste des contraintes :\n"^
		(string_of_list ( fun (x,y) ->
					"\n" ^ (string_of_trip x) ^ " est avant " ^ (string_of_trip y) ^ "\n"
				)
				dep
		)
	      ); print_newline(); flush stdout);
	     true
	   end;
*)
