(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: options.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2007-09-10 17:08:40 $
 * Version: $Revision: 1.13 $
 * -------------------------------------------------------------------
 *
 *   Dans ce fichier,
 *      - on traite les options
 *      - on traite les itrations.
 * Les fonctions de ce fichier ne sont pas  appeler par le parseur:
 * c'est un fichier "auxilliaire" de macro.ml qui contient les fonctions
 *  appeler. pour le traitement des itrations, la fct de macro.ml est
 * un alias de make_application qui est dfinie ici et qui est le point
 * d'entre.
 *
 * Traitement des options d'une fonction
 * -------------------------------------
 *
 * Seule les fonctions (definie avec fun) nommes peuvent avoir des
 * options. Ces options correspondent  des variables impratives, vues 
 * uniquement dans le corp de la fonction et qui persistent  travers
 * les appels. Ca correspond un peut  une variable locale dclare static
 * en C.
 *
 * Si on oublie la visibilite locale des options, alors la syntaxe de 
 * definition: 
 *     fun f[a=U](x) = ... 
 * correspond a 
 *      a := U ; fun f(x) = ... 
 * Remarquez en particulier que la valeur de a est initialise au moment 
 * de la definition de f. 
 * Et la syntaxe de l'appel
 *      f[a=W](V) 
 * correspond a:   
 *      a := W; f(V) 
 * En particulier, l'appel suivant  f s'evaluera avec un a qui a pour 
 * valeur un W et pas la valeur par defaut V. 
 *
 * Ce comportement est le seul compatible avec l'idee de garder de 
 * l'information d'une iteration a l'autre, ou d'un appel a l'autre. En 
 * effet, nous *voulons* que  
 *     f[iter=n+1](...)   ===   f(f[iter=n](...)) 
 *
 *
 * Une table de hachage (option_tbl) est utilise pour associer 
 *  chaque nom de fonction-avec-option un doublet: 
 *     - l'environnement d'option
 *     - la liste d'association qui indique l'index d'un arg optionnel dans 
 *       un environnement 
 * 
 * Un environnement d'option est tout simplement un array de mgs value. 
 * Chaque option d'une fonction est associe  un index dans ce tableau.
 * Cet environnement est partag par tous les appels de la fonction. Fournir
 * de nouvelle valeur lors de l'application est simplement une maniere d'affecter
 * cet environnement.
 *
 * Lors de la dfinition d'une fonction-avec-option, la fonction make_application 
 * est le point d'entre (au parsing) qui le bon terme.
 * La fonction make_application appelle make_real_apply avec les bon arguments et 
 * cette dernire gnre la forme macro-expanse
 *
 * On cree l'environnement 
 * (make_option_env). Le corp de la fonction est modifi (make_body) afin que les 
 * rferences  une option se transforment en acces  l'environnement courant des
 * options (get_option). 
 * Ces accs ne doivent pas tre effectu au moment de la dfinition, mais au 
 * moment de l'excution d'un appel, on force donc le retard grce  la fonction 
 * impure delay (qui retourne son argument, cf. Environnement.wrap_delay).
 *
 * Lors d'une application de fonction, si la fonction applique est une 
 * fonction-avec-option, on transforme l'appel 
 *     f[a0=v0, ..., ap=vp](x1, ..., xq))
 * en un nouveau terme:
 *     set_option 0 v0; 
 *     ...;  
 *     set_option p vp;
 *     f(x1, ..., xq)
 * La gnration des set_options avant l'appel se fait par update_options.
 *
 *
 *
 *
 * Traitement des itrations d'une fonction
 * ----------------------------------------
 *
 * Une fonction peut tre applique:
 *  - une fois
 *  - n fois
 *  - jusqu' un point fixe
 * et :
 *  - la dtection du point fixe peut se faire par un prdicat fourni 
 *    par le programmeur
 *  - les itrations (n fois o par point-fixe) peuvent s'entrelacer 
 *    avec des fonctions de prlude, interlude et postlude.
 *  - la variable spciale 'iteration est utilisable dans le corp d'une 
 *    fonction et prend pour valeur le rang d el'itration courante.
 * Voir la doc en ligne pour plus de dtails, entre: 'iter et 'fixpoint.
 * 
 * Les diffrents paramtres des itrations utilisent la mme syntaxe que les 
 * options dfinies par l'utilisateur. Mais aussi des syntaxes particulire.
 * Par exemple f[*] est un alias pour f['fixpoint].
 * Cela rend la dtection des options un peu pnible et il faut donc les chercher 
 * dans la liste des options si celel-ci existe et vrifier qu'il n'y a pas de contradiction. 
 * Cette recherche est faite  l'aide des fonctions auxilliaires : 
 * search_lude, search_iteration, wrap_iterfct, iteration_name, check_no_option, etc.
 *
 * L'ide est de "macro-expanser" un appel indiquant une itration, en un
 * appel  une fonction d'ordre sup F qui prend la fonction  itrer et les
 * diffrents xxx_lude et effectue le calcul ncessaire.
 * Il y a 2 fonctions F : 
 *   - itere_fct_n_with_lude        (pour itrer n fois)
 *   - itere_fct_fix_pred_with_lude (pour iterer jusqu'au point fixe)
 *
 * Ces fonctions sont appeles avec les arguments optionnels ou bien des 
 * valeurs par dfaut de ces arguments :
 * - standard_comparaison pour le prdicat d'galit du point-fixe
 * - donothing qui est l'identit pour les xxx_ludes.
 *
 * Le point d'entre est le mme que le prcdent: make_application fait la 
 * macro-expansion ncessaire.
 *
 *=========================================================================
 *)

open Global;;
open Valeur;;
open Dispatch_sig;;
open Terme;;
open Printer;;
open Printer_term;;
open Environnement;;

__register_version "$RCSfile: options.ml,v $" "$Revision: 1.13 $" "$Date: 2007-09-10 17:08:40 $";;





(** === Traitement de l'iteration ======================================================= **)


(** Wrapper iteration et point-fixe **)

type possible_option = 
  | NoIteration
  | OptionFixPoint
  | OptionFixRule 
  | OptionFixPred of term
  | OptionItere of term 
  | OptionStrategy of strategy
  | OptionList of (string*term) list


and strategy = 
    StDefault             (* Parallele maximale *)
  | StStochastic          (* Stochastique       *)
  | StGillespie of g_type (* Gillespie          *)

and g_type =              
  | GillDirect            (* Methode directe                 *)
  | GillFirst             (* Methode de la premiere reaction *)
;;


let iteration_name = "'iteration"
and strategy_name = "'strategy"
and tau_name      = "'tau"
let index_var_iter = 
  (Environnement.addenv_impure iteration_name (VAL_int 1);
   index_var iteration_name)
and index_var_strategy =
  (Environnement.addenv_impure strategy_name (VAL_int 0);
   index_var strategy_name)
and index_var_tau =
  (Environnement.addenv_impure tau_name (VAL_float 0.0);
   index_var tau_name)
;;

let check_lude = function
  | VAL_funct f -> f
  | x ->
      let err = "Error: bad argument. A function is expected for pre/inter/post-lude."
      in Printf.fprintf stderr "%s\nGot:\t%a\n" err prfvalue2 x; raise (Erreur err)

and check_predicate = function
  | VAL_funct f -> f
  | x ->
      let err = "Error: bad argument. A function is expected for the predicate in guarded fixpoint."
      in Printf.fprintf stderr "%s\nGot:\t%a\n" err prfvalue2 x; raise (Erreur err)

;;

let push_X x v = 
  let save = Environnement.get_varidx(x)
  in
    Environnement.set_varidx x v;
    save

and pop_X x saved =
  Environnement.set_varidx x saved
;;

let push_strategy = push_X index_var_strategy
and pop_strategy  = pop_X index_var_strategy

and get_tau () = Environnement.get_varidx(index_var_tau)
and set_tau v  = Environnement.set_varidx index_var_tau v

;;


let fct_generique = 
  VAL_funct (fun fct ->
	       (VAL_funct (fun st -> 
			     (VAL_funct (fun arg ->
					   let save = push_strategy st 
					   and stau = get_tau() in
					     try
					       let r = (pf1 fct) arg 
					       in
						 pop_strategy save;
						 set_tau stau;
						 r
					     with MGSExn(a1,a2) ->
					       pop_strategy save;
					       set_tau stau;
					       raise (MGSExn(a1,a2))
					)))))
	       

let itere_fct_n_with_lude = 
  VAL_funct(function prelude' -> 
	      VAL_funct(function interlude' -> 
			  VAL_funct(function postlude' -> 
				      let prelude = check_lude prelude' 
				      and interlude = check_lude interlude'
				      and postlude = check_lude postlude' in
					VAL_funct(function i -> 
   						    VAL_funct(function fct -> VAL_funct (fun st ->
								VAL_funct(function arg ->
									    match i with
									      | VAL_int(0) -> arg  
									      | VAL_int(n) when n < 0 -> VAL_undef("negative iteration")
									      | VAL_int(n) -> (
										  let nn = Pervasives.max 0 n 
										  and f = pf1 fct 
										  and v = ref arg 
										  and flag = ref false
										  and save_st   = push_strategy st
										  and stau      = get_tau()
										  and save_iter = Environnement.get_varidx(index_var_iter) in 
										    try
										      begin
											Environnement.set_varidx index_var_iter (VAL_int 0);
											v := prelude(!v);
											for i = 1 to nn do 
											  (if !flag then v := interlude(!v) else flag := true);
											  Environnement.set_varidx index_var_iter (VAL_int i);
											  v := f(!v) 
											done; 
											let resultat = postlude(!v) in
 											  Environnement.set_varidx index_var_iter (save_iter);
											  pop_strategy save_st;
											  set_tau stau;
											  resultat
										      end
										    with MGSExn(a1,a2) -> 
										      Environnement.set_varidx index_var_iter (save_iter);
										      pop_strategy save_st;
										      set_tau stau;
										      raise (MGSExn(a1,a2))
										)
									      | x ->  
										  let err = "Error: iteration argument must be an integer" in
										    Printf.fprintf stderr "%s. Got\n\t" err; 
										    ignore(prfvalue stderr x);
										    raise (Erreur err) )))))))
    
    
and itere_fct_fix_pred_with_lude = 
  VAL_funct(function prelude' -> 
	      VAL_funct(function interlude' -> 
			  VAL_funct(function postlude' -> 
				      let prelude = check_lude prelude' 
				      and interlude = check_lude interlude'
				      and postlude = check_lude postlude' in
					VAL_funct(function predicate -> let pequal = check_predicate predicate in
						    VAL_funct(function fct -> VAL_funct (fun st ->
								VAL_funct(function arg -> 
									    let f = pf1 fct 
									    and save_iter = Environnement.get_varidx(index_var_iter)
									    and iter      = ref 0
									    and save_st   = push_strategy st
									    and save_tau  = get_tau()
									    and v0        = ref arg
									    and v1        = ref (VAL_undef(new_name "dummy_initial_value_for_fixpoint_iteration"))
									    in
									      try
										Environnement.set_varidx index_var_iter (VAL_int 0);
										v0 := prelude(!v0);
										while true do
										  incr(iter);
										  Environnement.set_varidx index_var_iter (VAL_int !iter);
										  v1 := f(!v0);
										  if Auxfct.bool_from_val((pf1 (pequal !v0)) !v1) then raise BreakLoop
										  else v0 := interlude(!v1)
										done;
										VAL_undef "Internal Error: impossible case in itere_fct FixPointFct"
									      with
										| BreakLoop ->
										    let resultat = postlude(!v1) in
 										      Environnement.set_varidx index_var_iter (save_iter);
										      pop_strategy save_st;
										      set_tau save_tau;
										      resultat
										| MGSExn(a1,a2) -> 
										    Environnement.set_varidx index_var_iter (save_iter);
										    pop_strategy save_st;
										    set_tau save_tau;
										    raise (MGSExn(a1,a2))
									 )))))))
    
and standard_comparaison = 
  Cte(VAL_funct(function x -> VAL_funct(function y -> equalMgsVal x y)))
    
and user_comparaison pred =  
  Cte(VAL_funct(function x -> VAL_funct(function y -> (pf1 (pred x) y))))

and donothing = Cte(VAL_funct(function x -> x))
;;



let err_too_much_arg_for_iteration fct = 
  let err = "Error: iteration of function "
	    ^(to_string (Printer_term.prsterm fct))
	    ^" takes only one argument"
  in prerr_endline err; raise (Erreur err)
;;




(** === Traitement des options ======================================================= **)

(** A chaque nom de fonction-avec-argument-optionnel, on associe
 **     - l'environnement
 **     - la liste d'association qui indique l'index d'un arg optionnel dans un environnement 
 **)
type option_env = mgs_value array
and  dico_variable_index = (string*int) list
and  option_env_dico = (string, option_env * dico_variable_index) Hashtbl.t
;;

let option_tbl = ((Hashtbl.create 101) : option_env_dico)
;;

let fct_with_option s = Hashtbl.mem option_tbl s
;;
  


let get_option env = function 
  | VAL_int n -> assert(n >= 0);  assert(n < Array.length env); 
      env.(n)
  | x -> raise (Erreur "Error: internal: acessing option needs an integer")

and set_option env = function 
  | VAL_int n -> assert(n >= 0); assert(n < Array.length env);
      VAL_funct(function v -> 
		  env.(n) <- v; v)
  | x -> raise (Erreur "Error: internal: acessing option needs an integer")
;;


(* Creer une fonction avec option :
 *	process_option  nom-de-la-fonction  liste-argument  body
 * avec 
 *   nom-de-la-fonction = l'identificateur d'un terme fun ...
 *   liste-argument : une liste de couples ("nom arg", la-valeur-par-defaut)
 * et ca retourne un body modifie
 *
 * make_body: transforme le body pour changer les acces aux options en un appel 
 *  stackaccess.
 *)
let rec make_option_env fname argl =
  let sz = List.length argl in
  let idx = ref(-1)
  and env = Array.make sz (VAL_undef "initial value for an option") in
  let varindex = List.map (function (id, v) -> incr idx; (id, !idx)) argl in
  let access = wrap_delay(Cte(VAL_funct(get_option env)))
  and setaccess = wrap_delay(Cte(VAL_funct(set_option env))) in
    begin
      (** declaration d'une fonction avec option **)
      Hashtbl.replace option_tbl fname (env, varindex);
       (** calcul de l'environnement par defaut **)
      List.iter (function (id, t) -> env.(List.assoc id varindex) <- Compile.eval t) argl; 
      (** on renvoit a pour permettre l'ajustement des rfrences aux options dans le corp de la fct **)
      (varindex, access, setaccess)
    end
    

and process_option_fct fname argl body = 
  let (varindex, access, setaccess) = make_option_env fname argl
  in make_body_fct access setaccess varindex body

and process_option_trans tname argl rules =
  let (varindex, access, setaccess) = make_option_env tname argl
  in make_body_trans access setaccess varindex rules

and process_option_patch tname argl rules =
  let (varindex, access, setaccess) = make_option_env tname argl
  in make_body_patch access setaccess varindex rules

and make_body_fct access setaccess idl = function
  | Id(s) when List.mem_assoc s idl -> App(access, [Cte(VAL_int (List.assoc s idl))])

  | Set (s, t) when List.mem_assoc s idl ->
      App(setaccess, [Cte(VAL_int (List.assoc s idl)); make_body_fct access setaccess idl t])

  | App(t1, t2) -> App(make_body_fct access setaccess idl t1, List.map (make_body_fct access setaccess idl) t2)

  | Abs(s, _) as t  when List.mem_assoc s idl -> make_body_fct access setaccess (List.remove_assoc s idl) t

  | Abs(s, t) -> Abs(s, make_body_fct access setaccess idl t)

  | If(t1, t2, t3) -> If(make_body_fct access setaccess idl t1, 
			 make_body_fct access setaccess idl t2, 
			 make_body_fct access setaccess idl t3)

  | Fby(t1, t2) -> Fby(make_body_fct access setaccess idl t1, make_body_fct access setaccess idl t2)

  | Set (s, t) -> Set(s, make_body_fct access setaccess idl t)

  | t -> t

and make_body_trans access setaccess idl rules = List.map (make_body_rule access setaccess idl) rules

and make_body_patch access setaccess idl rules = List.map (make_body_prule access setaccess idl) rules

and make_body_rule access setaccess idl (id, pat, kind, exp) =
  (id,
   make_body_pat access setaccess idl pat,
   make_body_kind access setaccess idl kind,
   make_body_fct access setaccess idl exp
  )

and make_body_prule access setaccess idl (id, pat, kind, exp) =
  (id, make_body_ppat access setaccess idl pat, kind, make_body_fct access setaccess idl exp)

and make_body_pat access setaccess idl pat = 
  let recur = make_body_pat access setaccess idl
  in match pat with
    | AsPat(s, p)             -> AsPat(s, recur p)
    | OrPat(p1, p2)           -> OrPat(recur p1, recur p2)
    | (RecordPat _) as r      -> r
    | GuardedPat(p, t)        -> GuardedPat(recur p, make_body_fct access setaccess idl t)
    | IterPat (ki, p)         -> IterPat(ki, recur p)
    | CommaPat(d, dn, p1, p2) -> CommaPat(d, dn, recur p1, recur p2)
	(* FIXME: mettre les cas exhaustifs *)
    | x -> x

and make_body_ppat access setaccess idl ppat =
  let recur = make_body_ppat access setaccess idl in
  let treat_guard = function
    | Test t -> Test (make_body_fct access setaccess idl t)
    | Pred (s,t) -> Pred (s,make_body_fct access setaccess idl t)
  in match ppat with
    | [] -> []
    | (s,(b,gl,sl1,a',sl2,b'))::q -> (s,(b,(List.map treat_guard gl),sl1,a',sl2,b'))::(recur q)

and make_body_kind access setaccess idl kind =
  (match kind.a_gillespie_c with
     | GAiExp | GCiExp -> kind.a_gillespie_t <- (make_body_fct access setaccess idl (kind.a_gillespie_t))
     | _               -> ()
  );
  (match kind.a_stochastic with
     | Some (StochExp t) -> kind.a_stochastic <- Some (StochExp (make_body_fct access setaccess idl t))
     | _                 -> ()
  );
  kind
;;



(* Fonctions qui cree un terme qui met a jour les valeurs des options
 *)
let rec update_options fid env varindex = function
  | [] -> []

  | (s, t)::l when (not(iteration_name s) & (not(strategy_name s))) ->
      let index = 
	try List.assoc s varindex 
	with Not_found -> raise (Erreur ("Error: function "^fid^" does not have optional argument: "^s))
      in App(wrap_delay(Cte(VAL_funct(set_option env))), [Cte(VAL_int index); t])::(update_options fid env varindex l)      

  | _::l -> update_options fid env varindex l
  

and wrap_option fid option args =
  let apply = make_real_apply (Id fid) args option in
    match option with
      | OptionList [] -> apply
      | OptionList opts ->
	  begin
	    try
	      let (env, varindex) = 
		try Hashtbl.find option_tbl fid 
		with Not_found -> raise (Erreur ("Error: function "^fid^" has no option")) in
	      let befor = update_options fid env varindex opts
	      in 
		List.fold_right (fun t acc -> Fby(t, acc)) befor apply
	    with Not_found ->
	      raise (Erreur ("Error: function "^fid^" has no optionnal argument: "
			     ^(List.fold_left (fun s (id, _) -> id^" "^s) "" opts)))
	  end
      | _ -> apply
	

and make_real_apply fct args opts =
  let prelude   = search_prelude opts
  and interlude = search_interlude opts
  and postlude  = search_postlude opts
  and strategy  = Cte(search_strategy opts)
  in match (search_iteration opts) with
    | NoIteration -> (** application normale **)
	if (strategy = Cte(VAL_int 0)) then
	  if args = []
	  then App(fct, [Cte(VAL_undef("dummy unit"))])
	  else App(fct, (List.rev args))
	else
	  if args = []
	  then App(Cte(fct_generique), fct::strategy::[Cte(VAL_undef("dummy unit"))])
	  else App(Cte(fct_generique), fct::strategy::(List.rev args))

    | OptionItere nexp -> 
	if 1 <> List.length args then err_too_much_arg_for_iteration fct
	else App(Cte(itere_fct_n_with_lude), prelude::interlude::postlude::nexp::fct::strategy::args)

    | OptionFixRule -> 
	prerr_endline "Warning: fixrule not yet implemented. Use of 'fixpoint instead";
	if 1 <> List.length args then err_too_much_arg_for_iteration fct
	else App(Cte(itere_fct_fix_pred_with_lude),
		 prelude::interlude::postlude::(standard_comparaison)::fct::strategy::args)

    | OptionFixPoint ->  
	if 1 <> List.length args then err_too_much_arg_for_iteration fct
	else App(Cte(itere_fct_fix_pred_with_lude), 
		 prelude::interlude::postlude::(standard_comparaison)::fct::strategy::args)

    | OptionFixPred pred ->  
	if 1 <> List.length args then err_too_much_arg_for_iteration fct
	else App(Cte(itere_fct_fix_pred_with_lude), 
		 prelude::interlude::postlude::pred::fct::strategy::args)

    | OptionList l -> 
	raise (Erreur "Error: internal error: unexpected OptionList in make_real_apply")

    | OptionStrategy _ -> raise (Erreur "OPTION: internal error: this case cannot happend")
	

and search_iteration = function
  | OptionList l -> search_iteration2 l
  | x            -> x

and search_iteration2 = function
  | [] ->  NoIteration

  | ("'iter", nexp)::l 
  | ("iter", nexp)::l
    -> no_iter l; OptionItere nexp

  | ("'fixpoint", Cte(VAL_undef _))::l
  | ("fixpoint", Cte(VAL_undef _))::l
      ->  no_iter l; OptionFixPoint

  | ("'fixpoint", upred)::l
  | ("fixpoint", upred)::l
    ->  no_iter l; OptionFixPred upred

  | ("'fixrule", exp)::l
  | ("fixrule", exp)::l 
    -> prerr_endline "Warning: fixrule not yet implemented. Use of 'fixpoint instead";
      search_iteration2 (("'fixpoint", exp)::l)

  | _::l -> search_iteration2 l

and search_strategy = function
    OptionList l -> search_strategy2 l
  | x            -> VAL_int 0                                                    (* OptionStrategy StDefault *)


(* Ca, c'est moche :( Coder par des entiers alors que l'on a des types sommes... *)

and search_strategy2 = function
    []                                                              -> VAL_int 0  (* OptionStrategy StDefault *)
  | ( (("'strategy", (Cte (VAL_symbol "`default"))) |
       ("strategy",  (Cte (VAL_symbol "`default")))) :: _)          -> VAL_int 6  (* On force ici la strategy par defaut. Le vrai default prend la courante *)

  | ( (("'strategy", (Cte (VAL_symbol "`gillespie"))) |
       ("strategy",  (Cte (VAL_symbol "`gillespie")))) :: _)        -> VAL_int 1  (* OptionStrategy (StGillespie GillDirect) *)

  | ( (("'strategy", (Cte (VAL_symbol "`multiStochastic"))) |
       ("strategy",  (Cte (VAL_symbol "`multiStochastic")))) :: _)  -> VAL_int 2

  | ( (("'strategy", (Cte (VAL_symbol "`singleStochastic"))) |
       ("strategy",  (Cte (VAL_symbol "`singleStochastic")))) :: _) -> VAL_int 3

  | ( (("'strategy", (Cte (VAL_symbol "`stochastic"))) |
       ("strategy",  (Cte (VAL_symbol "`stochastic")))) :: _)       -> VAL_int 4

  | ( (("'strategy", (Cte (VAL_symbol "`asynchronous"))) |
       ("strategy",  (Cte (VAL_symbol "`asynchronous")))) :: _)     -> VAL_int 5

  | _::t                                                            -> search_strategy2 t



and search_lude s = function 
  | OptionList opts ->
      begin
	try
	  List.assoc s opts
	with Not_found -> 
	  try
	    List.assoc (String.sub s 1 ((String.length s) - 1)) opts
	  with Not_found -> donothing
      end
  | _ -> donothing

and search_prelude opts   = search_lude "'prelude" opts
and search_interlude opts = search_lude "'interlude" opts
and search_postlude opts  = search_lude "'postlude" opts
	
and iteration_name = function
  | "'fixpoint"
  | "fixpoint"
  | "'fixrule"
  | "fixrule"
  | "'iter"
  | "iter"
    -> true
  | s when wrap_iterfct s -> true
  | _ -> false

and strategy_name = function
    "'strategy"
  | "strategy"  -> true
  | _           -> false

and wrap_iterfct =  function
  | "prelude"
  | "'prelude"
  | "postlude"
  | "'postlude"
  | "'interlude"
  | "interlude"
    -> true
  | _ -> false

and no_iter = function
  | [] -> ()
  | (s,_)::l when (iteration_name s) & not(wrap_iterfct s)
      -> Printf.fprintf stderr "Warning: contradictory iteration specification in function application\n"; ()
  | _::l-> no_iter l

and check_no_option fct = function
  | OptionList l -> check_no_option2 fct l
  | _ -> ()

and check_no_option2 fct = function
  | [] -> ()
  | (a,_)::l when iteration_name a -> check_no_option2 fct l
  | (a,_)::l when strategy_name a  -> ()
  | _ as l ->
      let err = "Error: cannot use optionnal argument with an expression as a function head"
      in 
	Printf.fprintf stderr "%s\nGot\thead:%a\n" err prfterm2 fct;
	List.iter (function (s, t) -> Printf.fprintf stderr "\targ: %s = %s\n" s (to_string (prsterm t))) l;
	raise (Erreur err)
;;



(** Fonction "interface" qui gere l'application: on appelle cette
 ** fonction et c'est elle qui cree "le bon terme"
 **)
let rec make_application fct option args =
  match fct with
    | Id id when fct_with_option id -> wrap_option id option args
    | Id id -> (* On applique une fonction qui n'a pas d'option ou bien qui est inconnue.
		* Si la variable est deja defini, tout va bien et on traite comme le cas ci-dessous.
		* Si la variable n'est as defini, c'est peut tre une recursion croise, et dans ce cas, 
		* il ne faut pas dclarer une variable inconnue. On cree donc cette variable.
		*)
	Environnement.find_or_create_var id;
	check_no_option fct option;
	make_real_apply fct args option
    | _ -> (** seules options possibles : vide, iter ou bien fixpoint **)
	begin
	  check_no_option fct option;
	  make_real_apply fct args option
	end
;;



let check_new_fct_name_before_definition s =
  if is_defined_var s then
    begin
      if not(is_undef(!var_env.(index_var(s))))
      then  Printf.fprintf stderr "Warning: redefining %s as a function\n" s;
      (** je dois annuler les options prcdentes **)
      if (fct_with_option s) then
	begin
	  Printf.fprintf stderr "Warning: redefining '%s' a function with option may cause troubles\n" s;
	  Printf.fprintf stderr "in the handling of the optional argument of the previous definition\n";
	  Hashtbl.remove option_tbl s
	end
    end
;;

