(*=======================================================-*-tuareg-*-======
 *
 * MGS : a general model of systems
 *
 *    Olivier Michel      : michel@lami.univ-evry.fr
 *    Jean-Louis Giavitto : giavitto@lami.univ-evry.fr
 *    Antoine Spicher     : aspicher@lami.univ-evry.fr
 *    Julien Cohen        : jcohen@lami.univ-evry.fr
 *    Mailing-list        : mgs@lami.univ-evry.fr
 *    WWW                 : http://mgs.lami.univ-evry.fr
 *
 * LaMI umr 8042 du CNRS, Genopole (C), Universite d'Evry Val d'Essonne
 *
 *
 * -------------------------------------------------------------------
 * File: $RCSfile: printer_code.ml,v $
 * Symbolic name : $Name:  $
 * Author: OM, JLG
 * Modified: $Date: 2004-09-23 21:24:23 $
 * Version: $Revision: 1.1 $
 * -------------------------------------------------------------------
 *
 * Les fonctions d'impressions du code
 *
 *
 *=========================================================================
 *)

open Global;;
open Valeur;;
open Byte_code;;
open Printer;;
open Environnement;;
open Buffer;;


__register_version "$RCSfile: printer_code.ml,v $" "$Revision: 1.1 $" "$Date: 2004-09-23 21:24:23 $";;


(** Attention: les noms collisionnent dans Printf et Format.
 ** Aussi on n'ouvre pas ces modules mais on dfinit ici 2 alias
 **)
let uprintf = Printf.bprintf (** impression standard dans un buffer **)
and fprintf = Format.bprintf (** impression formatt dans un buffer **)
;;


let rec prbcode b = function

  | C_I  -> uprintf b "I"
  | C_K  -> uprintf b "K"
  | C_B  -> uprintf b "B"
  | C_C  -> uprintf b "C"
  | C_S  -> uprintf b "S"
  | C_S4 -> uprintf b "S4"

  | Cp s -> uprintf b"#%s" s
  | Cd s -> uprintf b "%s" s
  | Cv v -> uprbvalue b v; ()

  | Cabs(s, p) ->  uprintf b("(%s)[%a]") s prbcode p

  | Capp(p, q) ->  uprintf b "(%a %a)" prbcode p prbcode q

  | Cupl(Oset(idx), j, [a]) -> uprintf b "%s<%d> := %a"  (name_of_var idx) j prbcode a

  | Cupl(Ofct(idx), j, []) -> uprintf b "%s<%d>" (name_of_var idx) j

  | Cupl(Ofct(idx), j, l) -> uprintf b "%s<%d>@%a" (name_of_var idx) j (prlcode "(" ", " ")") l

  | Cupl(Ofby, level, l) -> prlcode ("FBY<"^(string_of_int level)^"> ")   " ; "   " YBF"  b  l

  | Cupl(Oif, level, [p;q;r]) -> 
      uprintf b "IF<%d> %a Then %a Else %a Fi" level prbcode p prbcode q prbcode r

  | Cupl(_, _, l) as t -> uprintf b "\nERROR: Cupl syntax: %a\n" (prlcode "[" "/" "]\n") l
      

and prlcode deb sep fin b l = 
  uprintf b "%s" deb;
  prlcode2 sep b l;
  uprintf b "%s" fin;

and prlcode2 sep b = function
  | [] -> ()
  | [a] -> uprintf b "%a" prbcode a
  | a::l -> uprintf b "%a%s%a" prbcode a sep (prlcode2 sep) l
;;


let prcode x = 
  clear out_buffer;
  prbcode out_buffer x; 
  output_buffer stdout out_buffer; 
  x

(** prfcode out x: sortie de x sur le fichier out **)
and prfcode out x = 
  clear out_buffer;
  prbcode out_buffer x; 
  output_buffer out out_buffer; 
  x

(** prscode x : retourne une string qui represente la valeur x **)
and prscode x =
  clear out_buffer;
  prbcode out_buffer x; 
  VAL_string(contents out_buffer)
;;

(** idem que prfcode, mais renvoie l'argument plutot que unit **)
let prfcode2 out x = (prfcode out x; ())
and error_c x = (print_string "\nError: "; prcode x; print_string "\n");;

