/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Ce fichier fait partie de "g-map-generic.hh".
//******************************************************************************

public:

/**
 * @param ADart Un pointeur sur un brin
 * @param ADim Une dimension
 * @return Un boolen indiquant si la cellule peut tre contracte.
 */
bool canContract(CDart* ADart, int ADim);

/**
 * Contracte la cellule de dimension ADim dsignes par le brin ADart.
 *
 * @param ADart Un pointeur sur brin
 * @param ADim Une dimension
 * @param ADeleteDarts Un boolen indiquant s'il faut supprimer les brins
 *   correspondant  la cellule contracte.
 *
 * @precondition canContract(ADart, ADim)
 */
void contract(CDart* ADart, int ADim, bool ADeleteDarts = false);

/**
 * Contracte lorsque cela est possible les cellules dont un brin au moins
 * est marqu.
 *
 * @param AMarkNumber Un numro de marque
 * @param ADim Une dimension (1, 2 ou 3)
 * @param ADeleteDarts Un boolen indiquant si la contraction doit se faire
 *   en supprimant des brins
 * @return Le nombre de contractions effectues
 */
int contractMarkedCells(int AMarkNumber, int ADim,
			bool ADeleteDarts = false);

//******************************************************************************
