/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Ce fichier fait partie de "g-map-generic.hh".
//******************************************************************************

public:

/**
 * Duplique les brins marqus et respecte dans la mesure du possible les
 * coutures.
 * Si ADirectInfoIndex est positif ou nul, chaque brin dupliqu est point
 * par le brin  partir duquel il alpha t cr (par le biais du champ
 * directInfo correspondant).
 *
 * Lorsqu'un champ A-Sew vaut 'faux', aucune couture de dimension A-Sew
 * n'est effectue entre les brins crs.
 *
 * @param AMarkNumber Un numro de marque
 * @param ADirectInfoIndex Un indice de champ directInfo
 * @param ADuplicateEmbeddings Un boolen indiquant s'il faut ou non
 *   dupliquer les plongements des brins
 * @param A-Sew Des boolens indiquant quelles coutures doivent tre
 *   prserves
 * @return Le nombre de brins dupliqus
 */
int duplicateMarkedDarts(int AMarkNumber, int ADirectInfoIndex = -1,
			 bool ADuplicateEmbeddings = true,
			 bool A0Sew = true, bool A1Sew = true,
			 bool A2Sew = true, bool A3Sew = true);

/**
 * Duplique les brins correspondant  l'orbite donne en dupliquant les
 * plongements et prservant les coutures qui peuvent l'tre.
 *
 * Si ADirectInfoIndex est positif ou nul, chaque brin dupliqu est point
 * par le brin  partir duquel il alpha t cr (par le biais du champ
 * directInfo correspondant).
 *
 * @param ADart Un brin de la carte
 * @param AOrbit Une orbite quelconque
 * @param ADirectInfoIndex Un indice de champ directInfo
 * @param ADuplicateEmbeddings Un boolen indiquant s'il faut ou non
 *   dupliquer les plongements des brins
 * @return Le brin dupliqu correspondant  ADart
 */
CDart* duplicateDarts(CDart* ADart, TOrbit AOrbit,
		      int ADirectInfoIndex = -1,
		      bool ADuplicateEmbeddings = true);

protected:

/**
 * Coud les brins dupliqus entre eux.
 * Les brins traits sont ceux parcourus qui sont marqus avec la marque
 * AMarkNumber.
 *
 * @param ACoverage Un itrateur quelconque
 * @param AMarkNumber Un numro de marque
 * @param ADirectInfoIndex Un indice de champ directInfo
 * @param A-Sew Des boolens indiquant quelles coutures doivent tre
 *   traites
 */
void sewDuplicatedDarts(CCoverage* ACoverage,
			int AMarkNumber, int ADirectInfoIndex,
			bool A0Sew = true, bool A1Sew = true,
			bool A2Sew = true, bool A3Sew = true);

/**
 * Duplique les plongements des brins dupliqus.
 *
 * @param ACoverage Un itrateur quelconque
 * @param AMarkInitial Un numro de marque
 * @param AMarkDuplicated Un numro de marque
 * @param ADirectInfoIndex Un indice de champ directInfo
 */
void duplicateEmbeddingsOfDuplicatedDarts(CCoverage* ACoverage,
					  int AMarkInitial,
					  int AMarkDuplicated,
					  int ADirectInfoIndex);

//******************************************************************************
