/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Ce fichier fait partie de "g-map-generic.hh".
//******************************************************************************

public:

/**
 * Insre un sommet topologique sur l'arte incidente au brin ADart.
 *
 * @param ADart Un pointeur sur un brin
 * @return Un brin du sommet topologique insr
 *
 * @precondition canInsertVertex(ADart)
 */
virtual CDart* insertVertex(CDart* ADart);

/**
 * Insre un sommet sur chaque arte dont un brin au moins est marqu,
 * lorsque cela est possible.
 *
 * @param AMarkNumber Un numro de marque
 * @return Le nombre d'insertions effectues
 */
int insertVertexOnMarkedEdges(int AMarkNumber);

/**
 * Insre une arte topologique entre les sommets incidents aux brins
 * ADart1 et ADart2.
 *
 * @param ADart1 Un pointeur sur un brin
 * @param ADart2 Un pointeur sur un brin
 * @return Un brin de l'arte topologique insre
 *
 * @precondition canInsertEdge(ADart1, ADart2)
 */
virtual CDart* insertEdge(CDart* ADart1, CDart* ADart2);

/**
 * Insre une arte sur chaque face dont deux sommets sont slectionns,
 * lorsque cela est possible.
 * Si ANoCrossedFace est faux, une arte n'est insre sur une face que si
 * exactement 2 brins de cette face sont marqus.
 * Si ANoCrossedFace est vrai, une arte n'est insre sur une face que si
 * 2 sommets de cette face sont marqus (2, 3 ou 4 brins marqus).
 *
 * @param AMarkNumber Un numro de marque
 * @param ANoCrossedFace Un boolen indiquant s'il faut vrifier qu'on ne
 *   va pas crer une face 'croise', c'est--dire une face en forme
 *   de 8, avec deux artes 2-cousues qui se croisent
 * @param ANoTwoEdgesFace Un boolen indiquant s'il faut vrifier qu'on ne
 *   va pas crer de face ferme compose de seulement deux artes
 * @return Le nombre d'artes insres
 */
int insertEdgeOnMarkedFaces(int AMarkNumber,
			    bool ANoCrossedFace = false,
			    bool ANoTwoEdgesFace = false);

/**
 * Teste s'il est possible d'insrer une face topologique dans un volume.
 * Si ANoCrossedVolume est vrai, au moins un brin de chaque demi-arte du
 * volume (orbite 0) correspondant  la face  crer doit tre slectionn.
 * Si ANoCrossedVolume est faux, au moins un brin de chaque arte du volume
 * (orbite 02) correspondant  la face  crer doit tre slectionn.
 *
 * @param ADart Un brin marqu du volume dans lequel insrer la face
 * @param AMarkNumber Un numro de marque
 * @param ANoCrossedVolume Un boolen indiquant s'il faut vrifier qu'on ne
 *   va pas crer un volume 'crois', c'est--dire un volume en forme
 *   de 8, avec deux faces 3-cousues qui se croisent
 * @param ANoTwoEdgesFace Un boolen indiquant s'il faut vrifier qu'on ne
 *   va pas crer de face ferme compose de seulement deux artes
 * @param ANoTwoFacesVolume Un boolen indiquant s'il faut vrifier qu'on
 *   ne va pas crer de volume ferm compos de seulement deux faces
 * @return Un boolen
 */
bool canInsertFace(CDart* ADart, int AMarkNumber,
		   bool ANoCrossedVolume = false,
		   bool ANoTwoEdgesFace = false,
		   bool ANoTwoFacesVolume = false);

/**
 * Insre une face le long des artes slectionnes du volume donn.
 *
 * @param ADart Un brin marqu du volume dans lequel insrer la face
 * @param AMarkNumber Un numro de marque
 * @param ANoCrossedVolume Un boolen indiquant s'il faut vrifier qu'on ne
 *   va pas crer un volume 'crois', c'est--dire un volume en forme
 *   de 8, avec deux faces 3-cousues qui se croisent
 * @return Un brin de la face cre
 *
 * @precondition canInsertFace(ADart, AMarkNumber)
 */
virtual CDart* insertFace(CDart* ADart, int AMarkNumber,
			  bool ANoCrossedVolume = false);

/**
 * Insre une face sur chaque volume dont plusieurs artes sont
 * slectionnes, lorsque cela est possible.
 *
 * @param AMarkNumber Un numro de marque
 * @param ANoCrossedVolume Un boolen indiquant s'il faut vrifier qu'on ne
 *   va pas crer un volume 'crois', c'est--dire un volume en forme
 *   de 8, avec deux faces 3-cousues qui se croisent
 * @param ANoTwoEdgesFace Un boolen indiquant s'il faut vrifier qu'on ne
 *   va pas crer de face ferme compose de seulement deux artes
 * @param ANoTwoFacesVolume Un boolen indiquant s'il faut vrifier qu'on
 *   ne va pas crer de volume ferm compos de seulement deux faces
 * @return Le nombre de faces insres
 */
int insertFaceOnMarkedVolumes(int AMarkNumber,
			      bool ANoCrossedVolume = false,
			      bool ANoTwoEdgesFace = false,
			      bool ANoTwoFacesVolume = false);


protected:

/**
 * Cette mthode est utilise par les mthodes 'canInsertFace' et 'insertFace'.
 * Elle permet de se dplacer le long des artes marques d'une volume.
 * Partant d'un brin donn elle tourne autour d'un sommet (orbite 12)
 * jusqu' trouver s'il existe un brin marqu d'une autre arte que celle
 * sur laquelle se trouve ADart.
 * Les deux derniers paramtres sont des paramtres en sortie.
 *
 * @param ADart Un brin de la carte, par lequel on est arriv sur ce sommet
 * @param ANoCrossedVolume Un boolen indiquant si l'insertion de face doit
 *   ne pas entraner l'apparition d'un volume crois (voir la mthode
 *   'insertFace')
 * @param AMarkNumber Un numro de marque permettant de reprer les brins
 *   slectionns
 * @param ANext Un pointeur sur un brin (paramtre en sortie uniquement)
 * @param ACrossed Un boolen indiquant si le brin ANext est tel que la
 *   face  insrer est croise (paramtre en sortie uniquement)
 * @return Un boolen indiquant si le sommet est valide, c'est--dire si au
 *   plus de 2 des artes qui lui sont incidentes sont slectionnes
 */
bool turnAroundVertex(CDart* ADart,
		      bool ANoCrossedVolume, int AMarkNumber,
		      CDart* & ANext, bool & ACrossed);

//******************************************************************************
