/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Ce fichier fait partie de "g-map-generic.hh".
//******************************************************************************

public:

/**
 * Retourne la plus petite dimension dans laquelle ADart1 et ADart2 peuvent
 * tre cousus.
 *
 * @param ADart1 Un brin quelconque
 * @param ADart2 Un brin quelconque
 * @return La plus petite dimension dans laquelle ADart1 et ADart2 peuvent
 *   tre cousus
 *
 * @postcondition La valeur retourne vaut 0, 1, 2, 3 ou 4.
 */
int getSewDimension(CDart* ADart1, CDart* ADart2);

/**
 * Essaie de coudre ADart1 et ADart2 en dterminant la dimension de la
 * couture  l'aide de la mthode 'getSewDimension'.
 *
 * @param ADart1 Un brin quelconque
 * @param ADart2 Un brin quelconque
 * @return Un boolen indiquant si la couture alpha pu se faire
 */
bool intuitiveSew(CDart* ADart1, CDart* ADart2);

/**
 * Essaie de coudre par alpha-ADimension des cellules de dimension
 * ADimension,  partir des brins ADart1 et ADart2. Le bord de dimension
 * ADimension incident  chaque brin de dpart est parcouru en parallle en
 * se limitant du premier ct  la rgion dlimite par la marque.
 *
 * @param ADart1 Un brin de la carte
 * @param AMarkNumber1 Un numro de marque
 * @param ADart2 Un brin de la carte
 * @param ADimension Une dimension de couture (2 ou 3)
 * @return Le nombre de coutures qui ont t effectues
 */
int multiSew(CDart* ADart1, int AMarkNumber1,
	     CDart* ADart2, int ADimension);

//******************************************************************************
