/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Ce fichier fait partie de "g-map-generic.hh".
//******************************************************************************

public:

/**
 * Bouche le bord de dimension ADim incident au brin ADart avec
 * une cellule de dimension ADart.
 * La fonction est virtuelle afin de pouvoir tre surcharge pour grer les
 * problmes de plongement. Par exemple, dans le cas de sommets plongs le
 * bouchage d'un 0-bord entrane la cration d'un nouveau sommet
 * topologique.
 *
 * @param ADart Un brin quelconque d'un bord de dimension ADim
 * @param ADim Une dimension
 */
virtual void stopUp(CDart* ADart, int ADim);

/**
 * Bouche les bords de dimension ADim incident aux brins slectionns.
 *
 * @param AMarkNumber Un numro de marque
 * @param ADim Une dimension
 * @return Le nombre de bords qui ont t bouchs
 */
int stopUpMarkedBorders(int AMarkNumber, int ADim);

/**
 * Bouche les bords incidents aux brins slectionns, en dterminant de
 * manire automatique la dimension des bords.
 *
 * @param AMarkNumber Un numro de marque
 * @return Le nombre de bords qui ont t bouchs
 */
int intuitiveStopUpMarkedBorders(int AMarkNumber);

//******************************************************************************
