/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
#ifndef ATTRIBUTE_VERTEX_HH
#define ATTRIBUTE_VERTEX_HH
//******************************************************************************
#include "attribute.hh"
#include "vertex.hh"
#include "inline-macro.hh"

#include <iostream>
//******************************************************************************

/**
 * Les diffrents formats de fichiers supports pour la sauvegarde et le
 * chargement.
 */
typedef enum { UnknownFormat=-1, AsciiFormat, BinaryFormat } TSaveFormat;

/**
 * Dfinit la classe dfinissant un attribut sommet, pouvant tre utilis par
 * une carte plonge sommets.
 *
 * @author MOKA TEAM
 */
class CAttributeVertex: public CAttribute, public CVertex
{
public:
  // @name Constructeurs et destructeur
  // @{

  /**
   * Constructeur par dfaut
   *
   * Construction d'une nouvelle instance de la classe, initialise
   * au sommet origine (0,0,0).
   *
   * @return Une nouvelle instance de la classe
   */
  CAttributeVertex();

  /**
   * Constructeur avec initialisation
   *
   * Construction d'une nouvelle instance de la classe, initialise au
   * sommet de coordonnes (Ax,Ay,Az).
   *
   * @param Ax La premire  coordonne du sommet
   * @param Ay La deuxime  coordonne du sommet
   * @param Az La troisime coordonne du sommet
   * @return Une nouvelle instance de la classe
   */
  CAttributeVertex(TCoordinate Ax, TCoordinate Ay, TCoordinate Az);

  /**
   * Constructeur par copie (CVertex)
   *
   * Construction d'une nouvelle instance de la classe, en copiant le sommet
   * pass en paramtre.
   *
   * @param AVertex le sommet  copier
   * @return Une nouvelle instance de la classe
   */
  CAttributeVertex(const CVertex& AVertex);

  /**
   * Constructeur par copie (CAttributeVertex)
   *
   * Construction d'une nouvelle instance de la classe, en copiant le sommet
   * pass en paramtre.
   *
   * @param AAttributeVertex le sommet  copier
   * @return Une nouvelle instance de la classe
   */
  CAttributeVertex(const CAttributeVertex& AAttributeVertex);

  /// Destructeur
  ~CAttributeVertex();

  // @}
  /**
   * @name Implmentation des mthodes virtuelles pures hrites de la classe
   *   CAttribute.
   */
  // @{

  /**
   * Retourne le type de l'attribut, c'est--dire la valeur de la variable de
   * classe 'TYPE'.
   *
   * @return Le type de l'attribut
   */
  TAttributeId getType() const;

  /**
   * Cre une nouvelle instance de CAttributeVertex, clone de l'instance
   * courante.
   *
   * @return Une nouvelle instance de l'attribut
   */
  CAttribute* copy() const;

  /// Dtruit l'objet (appel au destructeur).
  void destroy();

  /**
   * crit l'attribut dans le flot donn, en vue d'une sauvegarde.
   *
   * @param AStream Un flot dans lequel crire l'attribut
   */
  void save(std::ostream& AStream) const;

  /**
   * Lit l'attribut  partir du flot donn.
   *
   * @param AStream Un flot dans lequel lire l'attribut
   */
  void load(std::istream& AStream);

  // @}
};

// @name Affichage
// @{

/**
 * @param AStream Le flot dans lequel afficher le sommet
 * @param AVertex Le sommet  afficher
 @ @return Le flot utilis pour l'affichage
 */
std::ostream& operator<<(std::ostream& AStream,
			 const CAttributeVertex& AVertex);

// @}
//******************************************************************************
#include INCLUDE_INLINE("attribute-vertex.icc")
//******************************************************************************
#endif // ATTRIBUTE_VERTEX_HH
//******************************************************************************
