/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
#include "g-map-vertex.hh"
using namespace GMap3d;
//******************************************************************************
void CGMapVertex::updateAllBurstDarts()
{
  int selected = getNewMark();
  negateMaskMark(selected);
  updateMarkedBurstDarts(selected, false);
  negateMaskMark(selected);
  freeMark(selected);
}
//******************************************************************************
#define GET_VERTEX(DART,INDEX) (getDirectInfoAsAttributeVertex(DART, INDEX))
//******************************************************************************
void CGMapVertex::updateMarkedBurstDarts(int AMarkNumber,
					 bool AUpdateIncidentVolumes)
{
  if (AUpdateIncidentVolumes)
    {
      int selected = getNewMark();
      markIncidentCells(ORBIT_VERTEX, AMarkNumber, selected);
      markIncidentCells(ORBIT_VOLUME, selected);
      updateMarkedBurstDarts(selected, false);
      unmarkAll(selected);
      freeMark(selected);
      return;
    }

  TCoordinate k1 = getBurstCoef(1);
  TCoordinate k2 = getBurstCoef(2);
  TCoordinate k3 = getBurstCoef(3);

  int directIndex = getNewDirectInfo();

  int markV = getNewMark(); // volume mark

  int mark3 = getNewMark(); // volume mark
  int mark2 = getNewMark(); // face mark
  int mark1 = getNewMark(); // edge mark

  pointDirectInfoToAttributeVertex(directIndex);

  // CARTE ENTIRE:
  for (CDynamicCoverageAll all(this); all.cont(); ++all)
    if (!isMarked(*all, mark3))
      {
	if (!isMarked(*all, AMarkNumber))
	  {
	    setMark(*all, mark3);
	    setMark(*all, mark2);
	    setMark(*all, mark1);
	  }
	else
	  {
	    // VOLUME:
	    CVertex bary3 =
	      basicBarycenter(*all, ORBIT_012, markV, mark3, directIndex);

	    negateMaskMark(markV);
	
	    for (CBasicDynamicCoverage012 vol(this, *all, markV);
		 vol.cont(); ++vol)
	      if (!isMarked(*vol, mark2))
		{
		  if (!isMarked(*vol, AMarkNumber))
		    {
		      setMark(*vol, mark2);
		      setMark(*vol, mark1);
		    }
		  else
		    {
		      // FACE:
		      CVertex bary2 =
			barycenter(*vol, ORBIT_01, directIndex);
		
		      markOrbit(*vol, ORBIT_01, mark2);
		
		      CVertex delta2 = bary2 - bary3;
		
		      for (CDynamicCoverage01 fac(this, *vol);
			   fac.cont(); ++fac)
			if (!isMarked(*fac, mark1))
			  {
			    if (!isMarked(*fac, AMarkNumber))
			      setMark(*fac, mark1);
			    else
			      {
				// ARTE:
				CVertex bary1 =
				  barycenter(*fac, ORBIT_0, directIndex);
				
				markOrbit(*fac, ORBIT_0, mark1);
				
				CVertex delta1 = bary1 - bary2;
				
				CVertex & b1 = getBurstVertex(   *fac );
				CVertex & b2 = getBurstVertex(alpha0(*fac));

				CAttributeVertex & v1 =
				  * GET_VERTEX(   *fac , directIndex);

				CAttributeVertex & v2 =
				  * GET_VERTEX(alpha0(*fac), directIndex);
			
				/* Les lignes suivantes sont quivalentes :
				 *
				 * b1 = bary3 +
				 *      k3*(delta2 +
				 *          k2*(delta1 +
				 *              k1*(v1-bary1)));
				 *
				 * b2 = bary3 +
				 *      k3*(delta2 +
				 *          k2*(delta1 +
				 *              k1*(v2-bary1)));
				 *
				 * Pour minimiser le nombre d'allocations et de
				 * dsallocations, elles sont remplaces par:
				 */
				
				b1.setX(bary3.getX() +
					k3*(delta2.getX() +
					    k2*(delta1.getX() +
						k1*(v1.getX()-bary1.getX()))));
				
				b1.setY(bary3.getY() +
					k3*(delta2.getY() +
					    k2*(delta1.getY() +
						k1*(v1.getY()-bary1.getY()))));
				
				b1.setZ(bary3.getZ() +
					k3*(delta2.getZ() +
					    k2*(delta1.getZ() +
						k1*(v1.getZ()-bary1.getZ()))));
				
				b2.setX(bary3.getX() +
					k3*(delta2.getX() +
					    k2*(delta1.getX() +
						k1*(v2.getX()-bary1.getX()))));
				
				b2.setY(bary3.getY() +
					k3*(delta2.getY() +
					    k2*(delta1.getY() +
						k1*(v2.getY()-bary1.getY()))));
				
				b2.setZ(bary3.getZ() +
					k3*(delta2.getZ() +
					    k2*(delta1.getZ() +
						k1*(v2.getZ()-bary1.getZ()))));
			      }
			  }
		    }
		}

	    negateMaskMark(markV);
	  }
      }

  freeMark(markV);
  negateMaskMark(mark3); freeMark(mark3);
  negateMaskMark(mark2); freeMark(mark2);
  negateMaskMark(mark1); freeMark(mark1);

  freeDirectInfo(directIndex);
}
//******************************************************************************
#undef GET_VERTEX
//******************************************************************************
