/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Ce fichier fait partie de "g-map-vertex.hh".
//******************************************************************************

public:

/**
 * Alloue et si ncessaire initialise (dans le cas o la librairie GMP est
 * utilise) l'espace mmoire ncessaire pour stocker une valeur de type
 * TCoordinate, ceci pour le champ directInfo[ADirectInfoIndex] des brins
 * marqus avec la marque AMarkNumber.
 *
 * Si AOrbitUsed est diffrent de ORBIT_NONE, l'allocation n'est faite que
 * pour le brin porteur du plongement AOrbitUsed de chaque cellule d'orbite
 * AOrbitUsed dont un brin au moins est marqu.
 *
 * @param AMarkNumber Un numro de marque
 * @param ADirectInfoIndex L'indice du champ directInfo  utiliser
 * @param AOrbitUsed Une orbite quelconque
 *
 */
void allocCoordToDirectInfo(int AMarkNumber, int ADirectInfoIndex,
			    TOrbit AOrbitUsed = ORBIT_NONE);

/**
 * Alloue et si ncessaire initialise (dans le cas o la librairie GMP est
 * utilise) l'espace mmoire ncessaire pour stocker une valeur de type
 * TCoordinate, ceci pour le champ directInfo[ADirectInfoIndex] des brins
 * correspondant  l'orbite donne par le couple (ADart, AOrbit).
 *
 * Si AOrbitUsed est diffrent de ORBIT_NONE, l'allocation n'est faite que
 * pour le brin porteur du plongement AOrbitUsed de chaque cellule d'orbite
 * AOrbitUsed dont un brin au moins est sur l'orbite (ADart, AOrbit).
 *
 * @param ADart Un brin quelconque de la carte
 * @param AOrbit Une orbite
 * @param ADirectInfoIndex L'indice du champ directInfo  utiliser
 * @param AOrbitUsed Une orbite quelconque
 */
void allocCoordToDirectInfo(CDart* ADart, TOrbit AOrbit, int ADirectInfoIndex,
			    TOrbit AOrbitUsed = ORBIT_NONE);

/**
 * Duplique le plongement sommet des sommets slectionns.
 * Le sommet dupliqu est rattach  un brin de l'orbite sommet, qui est le
 * mme que celui qui porte l'attribut original.
 *
 * Un sommet est slectionn si au moins un de ses brins est marqu avec la
 * marque AMarkNumber.
 *
 * L'espace allou par cette mthode peut tre libr  l'aide de la mthode
 * 'deleteDuplicatedVertex'.
 *
 * @param AMarkNumber Un numro de marque
 * @param ADirectInfoNumber Un indice indiquant quel directInfo utiliser
 */
void duplicateVertexToDirectInfo(int AMarkNumber, int ADirectInfoNumber);

/**
 * Libre l'espace mmoire allou par la mthode 'duplicateVertexToDirectInfo'.
 * Les champs directInfo librs ne sont pas positionns  NULL.
 *
 * Attention: les brins slectionns devraient tre les mmes qu'au moment de
 * l'appel  duplicateVertexToDirectInfo!
 *
 * @param ADirectInfoNumber Un indice indiquant quel directInfo utiliser
 */
void deleteDuplicatedVertex(int ADirectInfoNumber);

/**
 * Copie les attributs sommets vers le champ directInfo spcifi, pour les
 * brins dont ce champ est non NULL.
 *
 * Attention: ces brins slectionns devraient tre les mmes qu'au moment de
 * l'appel  duplicateVertexToDirectInfo!
 *
 * @param ADirectInfoNumber Un indice indiquant quel directInfo utiliser
 */
void updateDirectInfoWithVertex(int ADirectInfoNumber);

/**
 * Pour chaque brin b de la carte, fait pointer le champ directInfo numro
 * ADirectInfoNumber vers le plongement sommet du sommet auquel b appartient.
 *
 * @param ADirectInfoIndex Un indice indiquant quel directInfo utiliser
 */
void pointDirectInfoToAttributeVertex(int ADirectInfoIndex);

/**
 * Pour chaque brin b de l'orbite sommet correspondant  ADart, fait pointer
 * le champ directInfo numro ADirectInfoNumber vers le plongement sommet du
 * sommet auquel b appartient.
 *
 * @param ADirectInfoIndex Un indice indiquant quel directInfo utiliser
 * @param ADart Un brin de la carte
 */
void pointDirectInfoToAttributeVertex(int ADirectInfoIndex, CDart* ADart);

/**
 * Pour chaque brin b de l'orbite AOrbit correspondant  ADart, fait pointer
 * le champ directInfo numro ADirectInfoNumber de tous les brins du sommet
 * vers le plongement de ce sommet.
 *
 * @param ADirectInfoIndex Un indice indiquant quel directInfo utiliser
 * @param ADart  Un brin de la carte
 * @param AOrbit Une orbite
 */
void pointDirectInfoToAttributeVertex(int ADirectInfoIndex, CDart* ADart,
				      TOrbit AOrbit);

//******************************************************************************
