/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
#include "g-map-vertex.hh"
using namespace GMap3d;
//******************************************************************************
void CGMapVertex::geoSew(CDart * ADart1, CDart * ADart2, int ADim,
			 bool ARotateCells, bool AScaleCells,
			 bool ATranslateCells)
{
  assert(ADart1!=NULL && ADart2!=NULL);
  assert(ADim>=0 && ADim<=3);
  assert(canSew(ADart1, ADart2, ADim));

  if (ADim!=0 && canPlate(ADart1, ADart2))
    plate(ADart1, ADart2, ADim-1, ARotateCells, AScaleCells, ATranslateCells);

  sew(ADart1, ADart2, ADim);
}
//******************************************************************************
bool CGMapVertex::intuitiveGeoSew(CDart * ADart1, CDart * ADart2,
				  bool ARotateCells, bool AScaleCells,
				  bool ATranslateCells)
{
  assert(ADart1!=NULL && ADart2!=NULL);

  int dim = getSewDimension(ADart1, ADart2);

  bool possible = dim!=4;

  if (possible)
    geoSew(ADart1, ADart2, dim, ARotateCells, AScaleCells, ATranslateCells);

  return possible;
}
//******************************************************************************
void CGMapVertex::geoBorderSew(CDart * ADart1, CDart * ADart2, int ADim,
			       bool ARotateCells, bool AScaleCells,
			       bool ATranslateCells)
{
  assert(ADart1!=NULL && ADart2!=NULL);
  assert(ADim>=0 && ADim<=3);
  assert(canSewBorders(ADart1, ADart2, ADim));

  if (ADim!=0 && canPlate(ADart1, ADart2))
    borderPlate(ADart1, ADart2, ADim, ARotateCells, AScaleCells, ATranslateCells);

  borderSew(ADart1, ADart2, ADim);
}
//******************************************************************************
bool CGMapVertex::intuitiveGeoBorderSew(CDart * ADart1, CDart * ADart2,
					bool ARotateCells, bool AScaleCells,
					bool ATranslateCells)
{
  assert(ADart1!=NULL && ADart2!=NULL);

  int dim= getBorderDimension(ADart1);

  bool possible = dim!=4 && canSewBorders(ADart1, ADart2, dim);

  if (possible)
    geoBorderSew(ADart1, ADart2, dim, ARotateCells,AScaleCells,ATranslateCells);

  return possible;
}
//******************************************************************************
