/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Ce fichier fait partie de "g-map-vertex.hh".
//******************************************************************************

public:

/**
 * Destruction de tous les brins de la carte.
 * Cette mthode appelle la mthode 'empty' de la classe CGMapBasic, puis
 * slectionne l'orbite ORBIT_VERTEX comme orbite utilise.
 */
void empty();

/**
 * Cre une nouvelle instance de la classe CDartVertex.
 *
 * @return Un nouveau brin, de type CDartVertex
 */
virtual CDart* newDart();

/**
 *  Supprime le brin pass en paramtre (de type CDartVertex)
 *
 *  @param ADart Le brin  supprimer
 */
virtual void delDart(CDart* ADart);

/**
 * Cre une nouvelle instance de CDartVertex.
 * Le brin cr n'est pas plong.
 *
 * @return Un nouveau brin, de type CDartVertex
 */
CDart* addMapDart();

/**
 * Cre une nouvelle instance de CDartVertex, dont le plongement est donn par
 * AVertex.
 *
 * @param AVertex Un sommet
 * @return Un nouveau brin de type CDartVertex
 */
CDartVertex* addMapDart(const CVertex & AVertex);

/**
 * Retourne le coefficient d'clatement de la dimension ADim.
 *
 * @param ADim Une dimension
 * @return Une valeur comprise entre 0 et 1
 */
TCoordinate getBurstCoef(int ADim) const;

/**
 * Positionne le coefficient d'clatement de la dimension ADim  la valeur
 * ACoef.
 *
 * @param ADim Une dimension
 * @param ACoef Une valeur comprise entre 0 et 1.
 */
void setBurstCoef(int ADim, TCoordinate ACoef);

/**
 * Renvoie le sommet associ  un brin, en parcourant si ncessaire tous les
 * brins du sommet.
 *
 * @param ADart Un brin de la carte
 * @return Le plongement sommet du brin
 */
CAttributeVertex* findVertex(CDart* ADart);

/**
 * Renvoie le sommet associ  un brin.
 * La mthode renvoie NULL si le plongement sommet n'est pas port par ADart.
 *
 * @param ADart Un brin de la carte
 * @return Le plongement sommet du brin
 */
CAttributeVertex* getVertex(CDart* ADart) const;

/**
 * Affecte le plongement AVertex au brin ADart.
 *
 * @param ADart Un brin de la carte
 * @param AVertex Un sommet
 *
 * @precondition findVertex(ADart)==NULL
 */
void setVertex(CDart* ADart, const CVertex & AVertex);
void setVertex(CDart* ADart, CAttributeVertex* AVertex);

/**
 * Ecrase s'il existe le plongement sommet du brin ADart avec le sommet AVertex.
 * Si le sommet topologique incident  ADart n'est pas plong, la mthode
 * est quivalente  un appel  'setVertex'.
 *
 * @param ADart Un brin de la carte
 * @param AVertex Un sommet
 */
void updateVertex(CDart* ADart, const CVertex & AVertex);

/**
 * Supprime s'il existe le plongement sommet du brin ADart.
 *
 * @param ADart Un brin de la carte
 */
void delVertex(CDart* ADart);

/**
 * Dtache le plongement sommet du sommet topologique incident  ADart.
 * Si le sommet n'est pas plong, la mthode retourne NULL.
 *
 * @param ADart Un brin de la carte
 */
CAttributeVertex* removeVertex(CDart* ADart);

/**
 * Retourne le plongement clat de la premire extrmit du brin ADart donn.
 *
 * @param ADart Un brin de la carte
 * @return Le plongement clat de ADart
 */
CVertex & getBurstVertex(CDart* ADart) const;

/**
 * Retourne le plongement clat de la deuxime extrmit du brin ADart donn.
 *
 * @param ADart Un brin de la carte
 * @return Le plongement clat de l'extrmit du brin ADart donn
 */
CVertex computeBurstExtremity(CDart* ADart) const;

/**
 * Retourne le champ directInfo demand du brin ADart aprs l'avoir
 * retyp en pointeur sur un attribut sommet.
 *
 * @param ADart Un brin de la carte
 * @param ADirectInfoIndex Un indice de champ directInfo
 * @return Un pointeur sur un attribut sommet
 */
CAttributeVertex* getDirectInfoAsAttributeVertex(CDart* ADart,
						 int ADirectInfoIndex) const;

/**
 * Retourne le champ directInfo demand du brin ADart aprs l'avoir
 * retyp en pointeur sur un sommet.
 *
 * @param ADart Un brin de la carte
 * @param ADirectInfoIndex Un indice de champ directInfo
 * @return Un pointeur sur un sommet
 */
CVertex* getDirectInfoAsVertex(CDart* ADart, int ADirectInfoIndex) const;

/**
 * Retourne le champ directInfo demand du brin ADart aprs l'avoir
 * retyp en pointeur sur une coordonne (TCoordinate).
 
 * @param ADart Un brin de la carte
 * @param ADirectInfoIndex Un indice de champ directInfo
 * @return Un pointeur sur une coordonne
 */
TCoordinate* getDirectInfoAsCoord(CDart* ADart, int ADirectInfoIndex) const;

//******************************************************************************
