/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Ce fichier fait partie de "g-map-vertex.hh".
//******************************************************************************

public:

/**
 * Teste si les deux cellules incidentes aux brins passs en paramtres
 * peuvent tre plaques (c'est le cas si elles ne sont pas dans la mme
 * composante connexe).
 *
 * @param ADart1 Un brin de la carte
 * @param ADart2 Un brin de la carte
 * @param Un boolen indiquant si les deux cellules incidentes aux brins
 *   donns peuvent tre plaques
 */
bool canPlate(CDart* ADart1, CDart* ADart2);

/**
 * Calcule la dimension maximum des cellules incidentes aux brins donns qui
 * peuvent tre plaques, en tenant compte de l'tat des coutures de chacun
 * de ces brins (par exemple, pour faire un placage de faces, il faut qu'au
 * moins un des brins soit 3-cousu).
 * Si le placage est impossible, la valeur -1 est retourne, sinon la valeur
 * retourne vaut 0, 1 ou 2.
 *
 * @param ADart1 Un brin de la carte
 * @param ADart2 Un brin de la carte
 * @return Un entier indiquant la dimension des cellules incidentes aux brins
 *   donns pouvant tre plaques
 */
int getPlateDimension(CDart* ADart1, CDart* ADart2);

/**
 * Plaque les cellules de dimension ADim incidentes aux brins ADart1 et ADart2.
 * La cellule incidentes  ADart1 voit son plongement modifi pour se plaquer
 *  la cellule incidente  ADart2. Le plongement de la cellule incidente 
 * ADart2 n'est pas modifi.
 * Le placage tient compte de l'tat des boolen 'orientation', 'scale' et
 * 'position'.
 *
 * @param ADart1 Un brin de la carte
 * @param ADart2 Un brin de la carte
 * @param ADim Une dimension
 * @param ARotateCells Un boolen pour modifier ou non l'orientation
 * @param AScaleCells Un boolen pour modifier ou non la taille
 * @param ATranslateCells Un boolen pour modifier ou non la position
 */
void plate(CDart* ADart1, CDart* ADart2, int ADimbool,
	   bool ARotateCells = true, bool AScaleCells = true,
	   bool ATranslateCells = true);

/**
 * Effectue un placage comme la mthode 'plate'.
 * La dimension du placage est dtermine par la mthode 'getPlateDimension'.
 *
 * @param ADart1 Un brin de la carte
 * @param ADart2 Un brin de la carte
 * @param ARotateCells Un boolen pour modifier ou non l'orientation
 * @param AScaleCells Un boolen pour modifier ou non la taille
 * @param ATranslateCells Un boolen pour modifier ou non la position
 * @return true si le placage alpha t possible
 */
bool intuitivePlate(CDart* ADart1, CDart* ADart2,
		    bool ARotateCells = true, bool AScaleCells = true,
		    bool ATranslateCells = true);

/**
 * Effectue un placage des bords de dimension ADim incidents aux brins ADart1
 * et ADart2. Tout comme pour la mthode 'plate', c'est le bord incident 
 * ADart1 qui est dplac vers le bord incident  ADart2.
 *
 * @param ADart1 Un brin de la carte
 * @param ADart2 Un brin de la carte
 * @param ADim Une dimension
 * @param ARotateCells Un boolen pour modifier ou non l'orientation
 * @param AScaleCells Un boolen pour modifier ou non la taille
 * @param ATranslateCells Un boolen pour modifier ou non la position
 */
void borderPlate(CDart* ADart1, CDart* ADart2, int ADim,
		 bool ARotateCells = true, bool AScaleCells = true,
		 bool ATranslateCells = true);

//******************************************************************************
