/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Ce fichier fait partie de "g-map-vertex.hh".
//******************************************************************************

public:

/**
 * Calcule la droite de rgression correspondant aux sommets dont un brin au
 * moins est marqu avec la marque AMarkNumber.
 * Cette droite est dcrite par un sommet lui appartenant et un vecteur
 * directeur.
 *
 * @param AMarkNumber Un numro de marque
 * @param ABarycenter Un sommet de la droite
 * @param ADirection Un vecteur directeur de la droite
 */
void computeRegressionLine(int AMarkNumber,
			   CVertex* ABarycenter, CVertex* ADirection);

/**
 * Calcule le plan de rgression correspondant aux sommets dont un brin au
 * moins est marqu avec la marque AMarkNumber.
 * Ce plan alpha pour quation cartsienne alpha*x + b*y + c*z + d = 0.
 * Les paramtres alpha, b, c et d sont placs dans les paramtres A?.
 *
 * @param AMarkNumber Un numro de marque
 * @param A? Des pointeurs sur les paramtres de l'quation du plan
 */
void computeRegressionPlane(int AMarkNumber,
			    TCoordinate* AA, TCoordinate* AB,
			    TCoordinate* AC, TCoordinate* AD);

//******************************************************************************
