/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
// Generic coverage class following different orbits in gmaps.
//******************************************************************************
#ifndef COVERAGE_HH
#define COVERAGE_HH
//******************************************************************************
#include "kernel-types.hh"
#include "inline-macro.hh"
//******************************************************************************
#define COVERAGE0       (0)
#define COVERAGE1       (1)
#define COVERAGE2A      (2)
#define COVERAGE2B      (3)
#define COVERAGE3A      (4)
#define COVERAGE3B      (5)
#define COVERAGE4       (6)

#define COVERAGE_BORDER (7)
#define COVERAGE_ALL    (8)
//******************************************************************************
namespace GMap3d
{
  class CDart;
  class CGMapBasic;

  /**
   * Classe abstraite pour la dfinition des itrateurs permettant de parcourir
   * les brins d'une carte.
   *
   * @author MOKA TEAM
   */
  class CCoverage
  {
  public:
    // @name Constructeur et destructeur.
    // @{

    /// Constructeur.
    CCoverage(CGMapBasic* AMap);

    /// Destructeur virtuel.
    virtual ~CCoverage();

    // @}

    /// Accesseur  l'objet G-Carte.
    CGMapBasic* getMap() const;

    // @name Mthodes virtuelles pures implmentes dans les classes filles
    // @{

    /// Rinitialisation du parcours.
    virtual void reinit() = 0;

    /// Itrateur prfixe.
    virtual void operator++() = 0;

    /// Itrateur postfixe
    virtual CDart* operator++(int) = 0;

    /// Drapeau d'arrt du parcours.
    virtual bool cont() = 0;

    /// Drefrencement du pointeur de parcours en pointeur de brin.
    virtual CDart* operator*() = 0;

    /// Type du dernier appel  ++.
    virtual TOperationState prevOperationType();

    /// Identificateur du type de parcours.
    virtual int type() const = 0;

    // @}

  protected:
    // @name Champs privs
    // @{

    /// La G-Carte sur laquelle va s'effectuer les parcours.
    CGMapBasic* FMap;

    // @}
  };

} // namespace GMap3d
//******************************************************************************
#include INCLUDE_INLINE("coverage.icc")
//******************************************************************************
#endif // COVERAGE_HH
//******************************************************************************
