/*
 * lib-gmapkernel : Un noyau de 3-G-cartes et des oprations.
 * Copyright (C) Moka Team, moka@sic.univ-poitiers.fr
 *               http://www.sic.univ-poitiers.fr/moka/
 *
 * This file is part of lib-gmapkernel
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

//******************************************************************************
#ifndef VECTOR_HH
#define VECTOR_HH
//******************************************************************************
#include "math-extension.hh"
#include "inline-macro.hh"

#include <iostream>
//******************************************************************************
class CTransformationMatrix;

/**
 * Dfinit la classe permettant de manipuler des vecteurs de dimension 4.
 *
 * @author MOKA TEAM
 */
class CVector
{
private:
  TCoordinate* FElements;

public:
  // @name Constantes
  // @{

  /**
   * Types de vecteurs particuliers.
   */
  typedef enum { NullVector, AnyVector } TVectorType;

  // @}

  // @name Constructeurs et destructeur
  // @{

  /**
   * Constructeur par dfaut
   * Construction d'une nouvelle instance de la classe, NON INITIALISE.
   *
   * @param AType Le type d'initialisation  effectuer (AnyVector par dfaut)
   * @return Une nouvelle instance de la classe
   */
  CVector(TVectorType AType = AnyVector);

  /**
   * Constructeur par copie
   * Construction d'une nouvelle instance de la classe, en copiant le vecteur
   * pass en paramtre.
   *
   * @param AVector le vecteur  copier
   * @return Une nouvelle instance de la classe
   */
  CVector(const CVector& AVector);

  /**
   * Constructeur avec initialisation
   * Construction d'une nouvelle instance de la classe, avec initialisation
   * explicite.
   *
   * @param A-- Les valeurs des cellules du vecteur.
   * @return Une nouvelle instance de la classe
   */
  CVector(TCoordinate A0, TCoordinate A1, TCoordinate A2, TCoordinate A3);

  /**
   * Destructeur
   */
  ~CVector();

  // @}
  // @name Initialisations
  // @{

  /**
   * Rinitialise l'instance au vecteur nul.
   */
  void setToNull();

  /**
   * @param AVector Le vecteur initial
   * @return La matrice courante obtenue aprs affectation
   */
  CVector& operator=(const CVector& AVector);

  // @}
  // @name Accesseur aux lments du vecteur
  // @{

  /**
   * Permet d'accder (en lecture et en criture)  un lment du vecteur.
   *
   * @param AIndex L'indice de l'lment  atteindre
   *
   * @precondition 0 <= AIndex <= 3
   */
  TCoordinate& operator[](int AIndex) const;

  // @}
  // @name Produit de matrices
  // @{

  /**
   * Calcule le produit du vecteur courant par la matrice AMatrix passe
   * en paramtre.
   *
   * @param AMatrix Une matrice quelconque
   * @return Le rsultat du produit du vecteur courant par la matrice
   *   AMatrix
   */
  CVector operator*(const CTransformationMatrix& AMatrix) const;

  /**
   * Calcule le produit du vecteur courant par le vecteur AVector pass
   * en paramtre.
   *
   * @param AVector Un vecteur quelconque
   * @return Le rsultat du produit du vecteur courant par le vecteur
   *   AVector
   */
  TCoordinate operator*(const CVector& AVector) const;

  // @}
};

// @name Affichage
// @{

/**
 * @param AStream Le flot dans lequel afficher le vecteur
 * @param AVector Le vecteur  afficher
 * @return Le flot utilis pour l'affichage
 */
std::ostream& operator<<(std::ostream& AStream, const CVector& AVector);

// @}
//******************************************************************************
#include INCLUDE_INLINE("vector.icc")
//******************************************************************************
#endif // VECTOR_HH
//******************************************************************************
