(**
  Minimum Degree algorithm
  
  Based on the article:
  The Minimum Degree Heuristic and the Minimal
  Triangulation Process
  by A. Berry, Pinar Heggernes & Genevive Simonet.
  
  @author Matthieu Sozeau
  @author Pierre-Loic Garoche
  
  $Id: md.mli,v 1.1 2004/06/28 07:32:55 filliatr Exp $
*)

module MD :
  sig
    module P :
      functor (G : Sig.P) ->
        sig
          module VertexSet : Set.S with type elt = G.V.t
          type edgeset = (G.V.t * G.V.t) list

	  (** [md g] return a tuple [(g', e, o)] where [g'] is 
	    a triangulated graph, [e] is the triangulation of [g] and
	    [o] is a perfect elimination order of [g']
	  *)
          val md : G.t -> G.t * edgeset * G.V.t list

	  (** [triangulate g] return the graph [g'] produced by applying 
	    miminum degree to [g].
	  *)
          val triangulate : G.t -> G.t
        end

    module I :
      functor (G : Sig.I) ->
        sig
          module VertexSet : Set.S with type elt = G.V.t
          type edgeset = (G.V.t * G.V.t) list

	  (** [md g] return a tuple [(g', e, o)] where [g'] is 
	    a triangulated graph, [e] is the triangulation of [g] and
	    [o] is a perfect elimination order of [g']
	  *)
          val md : G.t -> G.t * edgeset * G.V.t list

	  (** [triangulate g] return the graph [g'] produced by applying 
	    miminum degree to [g].
	  *)
          val triangulate : G.t -> G.t
        end
  end
